package org.modelgen.importexport.Export;

import org.modelgen.importexport.oo.model.Field;
import org.modelgen.importexport.oo.model.OOClass;
import org.modelgen.importexport.oo.model.OOObject;
import org.modelgen.importexport.oo.model.OOSchema;
import org.modelgen.importexport.oo.model.ObjectGeneralization;
import org.modelgen.importexport.oo.model.Reference;
import org.modelgen.importexport.oo.model.ReferenceComponent;
import org.modelgen.importexport.oo.model.ValueComponent;
import org.modelgen.importexport.persistence.ConnectionFactory;
import org.modelgen.importexport.persistence.DbConnection;
import org.modelgen.importexport.persistence.ImportExportPersistence;
import org.modelgen.importexport.utilities.ImportExportException;
import org.modelgen.importexport.utilities.PersistenceException;
import org.modelgen.importexport.utilities.datatype.DB2ObjectOrientedDataType;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

/**Exports an OOSchema to a Connection towards a DB2 Database.
 * An Exporter is a class which has the rule to translate the given schema of type T
 * in a correct form for the Destination E.
 * For example we can export a XML schema both to a File and a DBMS with XML capabilities;
 * we can threfore define two class: one which has the rule to export to a File and one to a DBMS 
 * 
 * @version 0.9
 *
 */

public class DB2OOExporter implements Exporter<OOSchema,Destination<Connection>> {
	//suffisso per il nome dei tipi
	private static final String TYPE_SUFFIX = "T";
	//suffisso per creare un tipo (normale oppure parent di una gerarchia)
	private static final String MODE_DB2SQL_TYPE = "NOT FINAL" +
			" MODE DB2SQL WITH FUNCTION ACCESS REF USING INTEGER";
	//suffisso per creare un tipo figlio in una gerarchia
	private static final String MODE_DB2="NOT FINAL MODE DB2SQL";
	/*mappa che mi serve nel momento che vado a creare le istanze delle typed table*/
	private Map<Integer, Integer> objectGeneralization=new TreeMap<Integer, Integer>();

	/**Check that the model Given can be export to the Destination.
	 * When we define a class that implements this interface we can export not only
	 * the model it was written for, but also other model with less costructs
	 * For example a class that export a RelationalSchema can also export a RelationalSchema
	 * in which model does not exist FK.
	 * 
	 *  Per maggiori info vedere modello massimale e minimale nella tesi
	 * 
	 * @param model OID of the model as stored in ModelGen db
	 * @return true if the model is between maximal e minimal model
	 * @throws ImportExportException
	 */
	public boolean checkModel(int model) throws ImportExportException {
		
		return true;
	}
	
	/**Export the given schema to the given destination.
	 * Based on the type of the Destination this class must export the information contained in the Schema.
	 * For example in case of a OOSchema to a DBMS Destination for each typed table this metod must 
	 * create a SQL command CREATE TYPE .. adn CREATE TABLE .. OF ..
	 * 
	 * 
	 * @param schema the schema to be exported
	 * @param destination the destination to which export the schema
	 * @throws ImportExportException any error during the export
	 */
	public void exportSchema(OOSchema schema, Destination<Connection> destination)	throws ImportExportException {	
		//la lista dei comandi da eseguire,  semplicemente una lista di stringhe
		//chiedere a paolo se usare il batch oppure no
		Vector<String> comandi=new Vector<String>();
		//prima genero lo schema
		StringBuilder sb =  new StringBuilder();
		sb.append("CREATE SCHEMA ");
		sb.append(schema.getName());
		comandi.add(sb.toString());
		
		List<OOClass> ordered=schema.generateHierachy();
		for(OOClass ooClass : ordered) {
			OOClass temp=schema.getParent(ooClass);	//null se non ha padre
			//genero il tipo
			comandi.add(generateSQLOOType(ooClass, temp, schema.getName()));
			//genero la typed table
			comandi.add(generateSQLOOTable(ooClass, temp, schema.getName()));
		}
		
		Connection con=destination.getDestination();
		try {
			Statement st = con.createStatement();
			for(String s:comandi) {
				System.out.println(s);
				System.out.println();
				st.executeUpdate(s);
			}
			st.close();				
		} catch (SQLException e) {
			throw new ImportExportException("Errore nell'Export dell schema");
		}		
	}

	/**Export the instances of the given shcema to the given location.
	 * Based on the type of the Destination this class must export the information contained in the Schema.
	 * For example in case of an OOSchema to a DBMS Destination for each object this metod must 
	 * create a SQL command INSERT INTO ... VALUES (...)
	 * 
	 * @param schema the schema to be exported
	 * @param destination the destination to which export the schema
	 * @throws ImportExportException any error during the export
	 */
	public void exportInstance(OOSchema schema, Destination<Connection> destination) throws ImportExportException {
		/*per prima cosa vado a popolare la mappa dei collegamenti tra oggetti
		 * serve perch cos la ricerca dei collegamenti tra oid degli oggetti cinclusi in una generalizzazione
		 *  molto pi veloce*/
		for(ObjectGeneralization objGen:schema.getObjectGeneralization()) {
			int objParent=objGen.getParentObject().getObjectOID();
			int objChild=objGen.getChildObject().getObjectOID();
			/*li posso inserire in questo ordine perch sono certo che OGNI oggetto abbia un OID
			 * diverso da quello degli altri
			 * Mi serve perch io scendo nella gerarchia fino a trovare l'oggetto che non ha figli in
			 * nessuna della classi sottostanti e lo creo risalendo la gerarchia fino all'antenato
			 * vedi tesi*/
			this.objectGeneralization.put(objChild,objParent);
		}
		
		//lista dei comandi che devo eseguire
		Vector<String> comandi=new Vector<String>();
				
		/*genero i vari insert in base alla gerarchia*/
		List<OOClass> ordered=schema.generateHierachy();	
		for(OOClass ooClass : ordered) {
			OOClass parent=schema.getParent(ooClass);

			for(OOObject oob: ooClass.getOOObjects()) {
				StringBuilder temp=new StringBuilder();
				temp.append("INSERT INTO ");
				temp.append(schema.getName());
				temp.append(".");
				temp.append(ooClass.getName());
				String values;
				if(parent!=null) {
					values=generateValues(oob,schema,parent);
					temp.append(values);
				}
				else {
					temp.append(" VALUES ");
					values=generateSQLValues(oob,schema);
					temp.append(values);
				}					
				if(values.length()>0)
					comandi.add(temp.toString());
				}			
		}

		Connection con=destination.getDestination();
		try {
			Statement st = con.createStatement();
			for(String s:comandi) {
				System.out.println(s);
				System.out.println();
				st.executeUpdate(s);
			}
			st.close();		
		} catch (SQLException e) {
			throw new ImportExportException("Errore nell'Export delle istanze");
		}		
	}
	
	/**Genera il comando SQL per creare un tipo
	 * @throws ImportExportException
	 */
	private String generateSQLOOType(OOClass child,OOClass parent,String schema_name) throws ImportExportException {
		
		String typeName = child.getName() + TYPE_SUFFIX;
		//per creare il tipo
		StringBuilder type = new StringBuilder();
		String newline = System.getProperty("line.separator");
		
		//preamble
		type.append("CREATE TYPE ");
		type.append(schema_name);
		type.append(".");
		
		//create type
		type.append(typeName);
		if(parent!=null){
			type.append(" UNDER ");
			type.append(schema_name);
			type.append(".");
			type.append(parent.getName()+TYPE_SUFFIX);
			type.append(newline);
		}
		/*controllo che il tipo abbia almeno un field oppure una reference*/
		if(child.getFields().size()==0 && child.getReferences().size()==0)
			throw new ImportExportException("Non  possibile gestire tipi senza attributi o reference " +
					"in DB2("+child.getName()+")");

		if( child.getFields().size()>0 || child.getReferences().size()>0) {
			type.append(newline);
			type.append("AS (");
		}
		//fields
		for (Field ooField : child.getFields()) {
			type.append(ooField.getName());
			type.append(" ");
			String tipo=DB2ObjectOrientedDataType.getOOType(ooField.getType());
			type.append(tipo);
			type.append(", ");
		}
		//references
		for (Reference ooReference : child.getReferences()) {
			type.append(ooReference.getName());
			type.append("  REF(");
			type.append(schema_name);
			type.append(".");
			type.append(ooReference.getClassTo().getName()+TYPE_SUFFIX);
			type.append(")");
			type.append(", ");
		}
		/*cancello l'ultima virgola e ci metto una parentesi per chiudere*/
		if( child.getFields().size()>0 || child.getReferences().size()>0) 
			type.replace(type.length() - 2, type.length(), ") ");
		
		/*Se sono sottotipi NON devono avere ne with function access ne Ref using*/
		if(parent==null){
			type.append(newline + MODE_DB2SQL_TYPE);
		}
		else {
			type.append(newline + MODE_DB2);
		}	
		
		return type.toString();
	}
	
	/**Genera il codice SQL per create una typed table
	 */
	private String generateSQLOOTable(OOClass child,OOClass parent,String schema_name) {
		//il nome del tipo corrispondente
		String typeName = child.getName() + TYPE_SUFFIX;
		String newline = System.getProperty("line.separator");
		
		StringBuilder typedTable = new StringBuilder();
		typedTable.append("CREATE TABLE ");
		
		//crete typed table on the type definition
		typedTable.append(schema_name);
		typedTable.append(".");
		typedTable.append(child.getName());
		typedTable.append(" OF ");
		typedTable.append(schema_name);
		typedTable.append(".");
		typedTable.append(typeName);
		typedTable.append(newline);
		/*se sottotabella in una generalizzazione ci va...*/
		if(parent!=null) {
			typedTable.append("UNDER ");
			typedTable.append(schema_name);
			typedTable.append(".");
			typedTable.append(parent.getName());
			typedTable.append(" INHERIT SELECT PRIVILEGES");
			typedTable.append(newline);
		}
		/*a questo punto devo aprire la parentesi per specificare qualcosa solo se..*/
		boolean elements=false;
		/*1) qualche Field  chiave o non pu essere NULL*/
		for (Field ooField : child.getFields()) {
			if (ooField.isIdentifier()){
				elements=true;
				break;
			}
			if (!ooField.isNullable()){
				elements=true;
				break;
			}
		}
		/*2) devo specificare lo scope di una Ref*/
		if(child.getReferences().size()>0)
			elements=true;
		/*3) se la typed table non ha padre devo specificare quale sar l'OID e di che tipo*/
		if(parent==null)
			elements=true;
		
		if(elements)
			typedTable.append("(");

		Vector<String> primaryKey = new Vector<String>();
		
		for (Field ooField : child.getFields()) {
			if (!ooField.isNullable())  {
				typedTable.append(ooField.getName());
				typedTable.append(" WITH OPTIONS NOT NULL, ");
				//primary key
				if (ooField.isIdentifier())
					primaryKey.add(ooField.getName());
			}
		}
		
		for (Reference ooReference : child.getReferences()) {
			typedTable.append(ooReference.getName());
			
			if(!ooReference.isNullable())
				typedTable.append(" WITH OPTIONS NOT NULL SCOPE ");
			else
				typedTable.append(" WITH OPTIONS SCOPE ");
			
			typedTable.append(schema_name);
			typedTable.append(".");
			typedTable.append(ooReference.getClassTo().getName());
			typedTable.append(", ");			
		}
		/*ora ho l'eventuale chiave primaria*/		
		if(primaryKey.size()>0) {
			typedTable.append("CONSTRAINT pk PRIMARY KEY (");
			for (String field : primaryKey) {
				typedTable.append(field);
				typedTable.append(", ");
			}
			typedTable.replace(typedTable.length() - 2, typedTable.length(), "), ");
		}		
		/*se il parent  null devo specificare quale sar il tipo di ref per gli oggetti
		 * di questa typed table(se ha un parent eredita l'identificativo del parent)*/
		if(parent==null) {
			typedTable.append("REF IS OID" + child.getName() + " USER GENERATED, ");
		}		
		/*se avevo aperto le parentesi adesso le devo chiudere togliendo l'ultima virgola*/
		if(elements)
			typedTable.replace(typedTable.length() - 2, typedTable.length(), ") ");		
		
		return typedTable.toString();
	}
	
	/**Questo metodo genera i valori per un oggetto che NON  inserito in una gerarchia e che quindi non 
	 * necessita di cercare altri oggetti per deinirne il valore
	 * 
	 * @param object
	 * @param schema
	 */
	private String generateSQLValues(OOObject object,OOSchema schema) {
		/*se non devo andare ad inserire l'oggetto corrente, ma uno che star pi sotto nella 
		 * gerarchia (uqesto cotnrollo  solo per sicurezza)*/
		if(!this.checkObjectGeneralization(object.getObjectOID())) {
			StringBuilder row = new StringBuilder("(");
			row.append(schema.getName()+"."+object.getOoClass().getName() + TYPE_SUFFIX + "(");
			row.append(object.getObjectOID() + "), ");
			for (ValueComponent valueComponent : object.getValueComponents()) {
				row.append(this.getSQLValue(valueComponent));
				row.append(", ");
			}
			for (ReferenceComponent referenceComponent : object.getReferenceComponents()) {
				row.append(generateSQLDatabase(referenceComponent,schema.getName()));
				row.append(", ");
			}
			row.replace(row.length()-2, row.length(), ")");
			
			return row.toString();
		}
		else
			return "";	
	}
	
	/**Genera il codice SQL per un oggetto che  inserito in una gerarchia.
	 * 
	 * A livello di oggetti ho una situazione del genere:
	 * oggetto della classe Persona: OID 1 Nome "Leonardo" ..
	 * oggetto della classe Manager: OID:13 Mult 10 ...
	 * 
	 * corrispondenza tra oggetti (13,1); sta a significare che l'oggetto 13 e l'oggetto 1 sono
	 * la stessa cosa quindi non devo costruire l'oggetto 1, ma solo il Manager di OID 13 prenderdo 
	 * gli altri valori dall'oggetto Persona 1
	 * 
	 * vedi tesi per maggiori dettagli
	 * 
	 * @param object
	 * @param schema
	 * @param parent
	 * @throws ImportExportException
	 */
	private String generateValues(OOObject object,OOSchema schema,OOClass parent) throws ImportExportException{
		/*verifico che l'oggetto non sia l'oggetto padre di nessuno,
		 * quindi pu essere inserito...*/
		if(!this.checkObjectGeneralization(object.getObjectOID())) {
			//i nomi delle colonne
			StringBuilder col=new StringBuilder("(");
			//i valori
			StringBuilder values=new StringBuilder("(");
			
			/*inserisco gli elementi dell'ultimo oggetto della gerarchia*/
			for (ValueComponent valueComponent : object.getValueComponents()) {
				col.append(valueComponent.getOoField().getName());
				col.append(", ");
				values.append(this.getSQLValue(valueComponent));
				values.append(", ");
			}
			for(ReferenceComponent ooref:object.getReferenceComponents()) {
				col.append(ooref.getOoReference().getName());
				col.append(", ");
				values.append(generateSQLDatabase(ooref,schema.getName()));
				values.append(", ");
			}
			/*adesso devo risalire nella lista degli oggetti fino a che non ho raggiunto 
			 * la vetta della gerarchia*/
			int objectOID=object.getObjectOID();
			OOClass last=object.getOoClass();
			
			while(parent!=null) {
				/*prendo il prossimo oggetto*/
				int objectToOID=getObjectGeneralization(objectOID);
				/*prendo l'oggeto corrispondente al prossimo OID*/
				OOObject o=parent.getOOObjectByOID(objectToOID);	
				
				/*anche per questo prendo gli elementi*/
				for (ValueComponent valueComponent : o.getValueComponents()) {
					col.append(valueComponent.getOoField().getName());
					col.append(", ");
					values.append(this.getSQLValue(valueComponent));
					values.append(", ");
				}
				for(ReferenceComponent ooref:o.getReferenceComponents()) {
					col.append(ooref.getOoReference().getName());
					col.append(", ");
					values.append(generateSQLDatabase(ooref,schema.getName()));
					values.append(", ");
				}
				
				/*mi preparo per il ciclo successivo*/
				last=parent;
				parent=schema.getParent(parent);
				objectOID=objectToOID;
			}
			/*devo inserire ancora il nome della colonna di self-referencing
			 * che ha il nome dell'ultima clase della gerarchia*/
			col.append("OID" + last.getName());
			col.append(") ");
			/*aggiungo ai valori il corrispondente valore di identificazione, ma questo va preso
			 * dal primo oggetto, cio quello pi basso nella gerarchia*/
			String leafClass=object.getOoClass().getName();
			values.append(schema.getName()+"."+leafClass+TYPE_SUFFIX+"(");
			values.append(object.getObjectOID() + ")) ");
			
			/*fondo le due stringhe insieme, in maniera corretta*/
			col.append(" VALUES ");
			col.append(values);
			return col.toString();
		}
		/*..oppure devo rimandarlo a pi tardi*/
		else 
			return "";		
	}
	
	/**Return true if the given object OID is not present as a values in a couple 
	 * of object generalizations
	 * @param objectOID
	 */
	private boolean checkObjectGeneralization(int objectOID) {
		if(this.objectGeneralization.containsValue(objectOID))
			return true;
		else 
			return false;
	}
	
	/**Return an OID corrisponding to the object OID of the parent of the given object OID
	 * @param objectOID
	 * @throws ImportExportException if this Object has'n a parent object
	 */
	private int getObjectGeneralization(int objectOID) throws ImportExportException {
		if(this.objectGeneralization.containsKey(objectOID))
			return this.objectGeneralization.get(objectOID);
		else
			throw new ImportExportException("Non  stato possibile caricare l'oggetto padre di "+objectOID);
	}
	
	/**Returns a valid SQL value for the 
	 * @param type
	 */
	private String getSQLValue(ValueComponent value) {
		if(value.getValue()==null)
			return "NULL";
		/*devo verificare che tipo sia in modo da aggiungere o no gli apici*/
		else {
			return DB2ObjectOrientedDataType.getOOValue(value.getOoField().getType(), value.getValue());
		}
	}
	
	/**This metod generates the SQL code for a Reference
	 * */
	private String generateSQLDatabase(ReferenceComponent ref,String schema_name) {
		if(ref.getOoObjectTo()==null) {
			return "NULL";
		}
		else {
			OOClass ooClass =ref.getOoObjectTo().getOoClass(); 
			String typeName = ooClass.getName() + TYPE_SUFFIX;
			StringBuilder sql = new StringBuilder(schema_name);
			sql.append(".");
			sql.append(typeName + "(");
			sql.append(ref.getOoObjectTo().getObjectOID());
			sql.append(")");
			return sql.toString();
		}		
	}	
	
	/**Main di prova
	 * */
	public static void main(String[] args) {
		try {
			OOSchema schema=ImportExportPersistence.getInstance().loadOOSchema(1);
			ImportExportPersistence.getInstance().loadOOInstance(schema,1);
			//System.out.println(schema.toString());
			Exporter<OOSchema, Destination<Connection>> exporter=new DB2OOExporter();
			DbConnection con=ConnectionFactory.getConnection(ConnectionFactory.DB2, "Temp1","localhost", 
					50000, "db2admin","db2admin","");
			Destination<Connection> destination=new Destination<Connection>(con.getConnection());
			destination.getDestination().setAutoCommit(false);
			schema.generateHierachy();
			exporter.exportSchema(schema, destination);
			exporter.exportInstance(schema, destination);
			destination.getDestination().rollback();
			
		} catch (PersistenceException e) {
			e.printStackTrace();
		} catch (SQLException e) {
			e.printStackTrace();
		} 
		catch (ImportExportException e) {
			e.printStackTrace();
		}
	}
}
