package org.modelgen.importexport.Export;

import org.modelgen.importexport.oo.model.Field;
import org.modelgen.importexport.oo.model.OOClass;
import org.modelgen.importexport.oo.model.OOObject;
import org.modelgen.importexport.oo.model.ObjectGeneralization;
import org.modelgen.importexport.oo.model.Reference;
import org.modelgen.importexport.oo.model.ReferenceComponent;
import org.modelgen.importexport.oo.model.ValueComponent;
import org.modelgen.importexport.or.model.ComplexType;
import org.modelgen.importexport.or.model.ComplexTypeComponent;
import org.modelgen.importexport.or.model.Insertable;
import org.modelgen.importexport.or.model.Nest;
import org.modelgen.importexport.or.model.NestComponent;
import org.modelgen.importexport.or.model.ORClass;
import org.modelgen.importexport.or.model.ORObject;
import org.modelgen.importexport.or.model.ORReference;
import org.modelgen.importexport.or.model.ORReferenceComponent;
import org.modelgen.importexport.or.model.ORSchema;
import org.modelgen.importexport.or.model.ORTable;
import org.modelgen.importexport.or.model.ORTuple;
import org.modelgen.importexport.persistence.ConnectionFactory;
import org.modelgen.importexport.persistence.DbConnection;
import org.modelgen.importexport.persistence.ImportExportPersistence;
import org.modelgen.importexport.relational.model.Column;
import org.modelgen.importexport.relational.model.ComponentOfForeignKey;
import org.modelgen.importexport.relational.model.ForeignKey;
import org.modelgen.importexport.relational.model.Table;
import org.modelgen.importexport.relational.model.TupleComponent;
import org.modelgen.importexport.utilities.ImportExportException;
import org.modelgen.importexport.utilities.PersistenceException;
import org.modelgen.importexport.utilities.datatype.DB2ObjectOrientedDataType;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

/**Export a ORSchema to a DB2 database.
 * An Exporter is a class which has the rule to translate the given schema of type T
 * in a correct form for the Destination E.
 * For example we can export a XML schema both to a File and a DBMS with XML capabilities;
 * we can threfore define two class: one which has the rule to export to a File and one to a DBMS 
 * 
 * @version 0.9
 */
public class DB2ORExporter implements Exporter<ORSchema,Destination<Connection>> {
	//suffisso usato per distinguere il nome di un tipo 
	private static final String TYPE_SUFFIX = "T";
	//suffisso da aggiungere alla definizione di un tipo(radice della gerarchia)
	private static final String MODE_DB2SQL_TYPE = "NOT FINAL" +
			" MODE DB2SQL WITH FUNCTION ACCESS REF USING INTEGER";
	//suffisso da aggiungere alla definizione di un tipo(foglie)
	private static final String MODE_DB2="NOT FINAL MODE DB2SQL";
	//mappa di collegamento tra oggetti
	private Map<Integer, Integer> objectGeneralization;
	
	public DB2ORExporter(){
		this.objectGeneralization=new HashMap<Integer, Integer>();
	}

	/**Check that the model Given can be export to the Destination.
	 * When we define a class that implements this interface we can export not only
	 * the model it was written for, but also other model with less costructs
	 * For example a class that export a RelationalSchema can also export a RelationalSchema
	 * in which model does not exist FK.
	 * 
	 *  Per maggiori info vedere modello massimale e minimale nella tesi
	 * 
	 * @param model OID of the model as stored in ModelGen db
	 * @return true if the model is between maximal e minimal model
	 * @throws ImportExportException
	 */
	public boolean checkModel(int model) throws ImportExportException {
		return true;
	}
	
	/**Export the given schema to the given destination.
	 * Based on the type of the Destination this class must export the information contained in the Schema.
	 * For example in case of a ORSchema to a DBMS Destination for each table this metod must 
	 * create a SQL command CREATE TABLE .. and for each ORClass it must create a SQL command CREATE TYPE..
	 * and another CREATE TABLE .. OF ..
	 * 
	 * 
	 * @param schema the schema to be exported
	 * @param destination the destination to which export the schema
	 * @throws ImportExportException any error during the export
	 */
	public void exportSchema(ORSchema schema, Destination<Connection> destination) throws ImportExportException {
		/*la destinazione  una connessione verso il DBMS*/
		Connection con=destination.getDestination();
		
		/*questa  la lista dei comandi da eseguire
		 * Decidere se usare batch ecc. chiedere a paolo*/
		Vector<String> comandi=new Vector<String>(); 
		comandi.add("CREATE SCHEMA "+schema.getName());
		
		/*devo generare i comandi SQL, ma rispettando necessariamente la gerarchia*/
		List<Insertable> ordered=schema.generateHierachy();
		Iterator<Insertable> i=ordered.iterator();
		
		while(i.hasNext()) {
			Object obj=i.next();
			if (obj instanceof ORTable) {
				ORTable table=(ORTable)obj;
				for(ComplexType complex:table.getComplexes()) {
					String complexSQL=exportComplexType(complex, schema );
					/*se il complex non  gi stato definito, lo aggiungo*/
					if(!controllaEsistenza(complex, comandi, schema.getName()))
						comandi.add(complexSQL);
				}
				comandi.add(exportTable(table, schema.getName()));
			}
			else if (obj instanceof ORClass) {
				ORClass ooClass=(ORClass)obj;
				OOClass parent=schema.getParent(ooClass);
				for(ComplexType complex:ooClass.getComplexes()) {
					String complexSQL=exportComplexType(complex, schema );
					/*se il complex non  gi stato definito, lo aggiungo*/
					if(!controllaEsistenza(complex, comandi, schema.getName()))
						comandi.add(complexSQL);
				}
				/*se il complex non  gi stato definito, lo aggiungo*/
				if(!controllaEsistenza(ooClass,comandi,schema.getName())){
					comandi.add(exportORType(ooClass,parent,schema.getName()));
				}
				comandi.add(exportORClass(ooClass,parent,schema.getName()));
			}
			if (obj instanceof ComplexType) {
				ComplexType complex=(ComplexType)obj;
				String complexSQL=exportComplexType(complex, schema );
				if(!comandi.contains(complexSQL))
					comandi.add(complexSQL);			
			}
		}
		
		boolean istanceLoaded=false;
		for(Table table:schema.getTables()) {
			if(table.getTuples().size()>0) {
				istanceLoaded=true;
			}
		}
		for(ORClass orClass:schema.getORClasses()) {
			if(orClass.getORObjects().size()>0)
				istanceLoaded=true;
		}
		
		if(!istanceLoaded){
			comandi.addAll(generateForeignKey(schema));
		}

		try {
			Statement st = con.createStatement();
			for(String s:comandi) {
				System.out.println(s);
				System.out.println();
				st.executeUpdate(s);
			}
			st.close();					
		} catch (SQLException e) {
			e.printStackTrace();
			throw new ImportExportException("Errore nell'Export dell schema");
		}			
	}
	
	/**Metodo che controlla se nella lista di comandi gi creati  presente un complex relativo ad una typed table
	 * con lo stesso nome
	 * @return true if already existsa complex with the same name, false otherwise
	 */
	private boolean controllaEsistenza(ORClass orClass,List<String> comandi,String schema) {
		String sql=new String("CREATE TYPE "+schema+"."+orClass.getName()+TYPE_SUFFIX);
		for(String s:comandi) {
			if(s.startsWith(sql))
				return true;
		}
		return false;
	}
	
	/**Metodo che controlla se nella lista di comandi gi creati  presente un complex con lo stesso nome
	 * @return true if already existsa complex with the same name, false otherwise
	 */
	private boolean controllaEsistenza(ComplexType complex,List<String> comandi,String schema) {
		String sql=new String("CREATE TYPE "+schema+"."+complex.getName()+TYPE_SUFFIX);
		for(String s:comandi) {
			if(s.startsWith(sql))
				return true;
		}
		return false;
	}
	
	/**Metodo che crea un comando SQL per ogni tabella
	 * Genera cio il comando SQL CREATE TABLE .. ([COLUMN],[REF],[COMPLEX TYPE])
	 * @throws ImportExportException 
	 * 
	 * */
	private String exportTable(ORTable table,String schema) throws ImportExportException {
		StringBuilder sql = new StringBuilder();
		String newline = System.getProperty("line.separator");
		
		sql.append("CREATE TABLE ");
		sql.append(schema);
		sql.append(".");
		sql.append(table.getName());
		sql.append(" (");
		sql.append(newline);
		
		Vector<Column> primaryKey = new Vector<Column>();
		//1) le colonne
		for (Column column : table.getColumns()) {
			sql.append(column.getName());
			sql.append(" ");
			String type=DB2ObjectOrientedDataType.getOOType(column.getType());
			sql.append(type);
			if(!column.isNull()) {
				sql.append(" NOT NULL");
			}
			sql.append(", ");

			if (column.isKey()) 
				primaryKey.add(column);
		}
		//2) le eventuali REF verso una typed table
		for(ORReference ref:table.getReferences()){
			sql.append(ref.getName());
			sql.append(" REF(");
			sql.append(schema);
			sql.append(".");
			sql.append(ref.getTo().getName()+TYPE_SUFFIX);			
			sql.append(") SCOPE ");
			sql.append(schema);
			sql.append(".");
			sql.append(ref.getTo().getName());
			sql.append(" , ");
		}
		//3)tutti gli eventuali complex type(nota che non si deve verificare che stiano in un NEST)
		for(ComplexType complex:table.getComplexes()) {
			sql.append(complex.getName());
			sql.append(" ");
			sql.append(schema);
			sql.append(".");
			sql.append(complex.getName()+TYPE_SUFFIX);
			sql.append(", ");
		}
		
		if(primaryKey.size()>=1){
			sql.append("primary key (");
			for (Column column : primaryKey) {
				sql.append(column.getName());
				sql.append(", ");
			}
		}	
		/*chiudo le parentesi*/
		if(primaryKey.size()>=1)
			sql.replace(sql.length()-2, sql.length(), "))");
		else
			sql.replace(sql.length()-2, sql.length(), ")");
		
		return sql.toString();
	}
	
	/**Genera il comando SQL CREATE TYPE per una ORClass.
	 * 
	 * @param child
	 * @param parent
	 * @param schema
	 * @throws ImportExportException nel caso la ORClass non definisca ne un Field ne una REF(non consentito in DB2)
	 */
	private String exportORType(ORClass child,OOClass parent,String schema) throws ImportExportException {
		if(child.getFields().size()==0 && child.getReferences().size()==0)
			throw new ImportExportException("Non  possibile gestire tipi senza attributi/reference " +
					"in DB2("+child.getName()+")");
		
		//preamble
		StringBuilder type = new StringBuilder();
		String newline = System.getProperty("line.separator");
		
		type.append("CREATE TYPE ");
		type.append(schema);
		type.append(".");
		type.append(child.getName() + TYPE_SUFFIX);
		if(parent!=null){
			type.append(" UNDER ");
			type.append(schema);
			type.append(".");
			type.append(parent.getName()+TYPE_SUFFIX);
			type.append(newline);
		}
		if( child.getFields().size()>0 || child.getReferences().size()>0 || child.getComplexes().size()>0) {
			type.append(newline);
			type.append("AS (");
			//1)fields
			for (Field ooField : child.getFields()) {
				type.append(ooField.getName());
				type.append(" ");
				String tipo=DB2ObjectOrientedDataType.getOOType(ooField.getType());
				type.append(tipo);
				type.append(", ");
			}
		}		
		//2) Reference		
		for (Reference ooReference : child.getReferences()) {
			type.append(ooReference.getName());
			type.append("  REF(");
			type.append(schema);
			type.append(".");
			type.append(ooReference.getClassTo().getName()+TYPE_SUFFIX);
			type.append(")");
			type.append(", ");
		}
		//3)eventuali complex type
		for(ComplexType complex:child.getComplexes()) {
			type.append(complex.getName());
			type.append("  ");
			type.append(schema);
			type.append(".");
			type.append(complex.getName()+TYPE_SUFFIX);
			type.append(", ");
		}
				
		if(child.getReferences().size()!=0 || child.getFields().size()!=0 || child.getComplexes().size()>0)
			type.replace(type.length() - 2, type.length(), ") ");
		
		/*Se foglie di una gerarchia devono usare il suffisso senza identificatore della typed table*/
		if(parent==null){
			type.append(newline + MODE_DB2SQL_TYPE);
		}
		else {
			type.append(newline + MODE_DB2);
		}	
		
		return type.toString();
	}

	/**Metodo che restituisce una stringa per ogni classe.
	 * Genera cio il comando SQL CREATE TABLE .. OF [TIPO] ([OPZIONI])
	 * 
	 * @throws ImportExportException 
	 */
	private String exportORClass(ORClass child,OOClass parent,String schema) throws ImportExportException {
		String newline = System.getProperty("line.separator");
		StringBuilder typedTable = new StringBuilder();

		if(child.getFields().size()==0 && child.getReferences().size()==0)
			throw new ImportExportException("Non  possibile gestire tipi senza attributi/reference " +
					"in DB2("+child.getName()+")");

		//crete typed table on the type definition
		typedTable.append("CREATE TABLE ");
		typedTable.append(schema);
		typedTable.append(".");
		typedTable.append(child.getName());
		typedTable.append(" OF ");
		typedTable.append(schema);
		typedTable.append(".");
		typedTable.append(child.getName() + TYPE_SUFFIX);
		typedTable.append(newline);
		/*se questa typed table  figlia di un'altra va indicato il padre*/
		if(parent!=null) {
			typedTable.append("UNDER ");
			typedTable.append(schema);
			typedTable.append(".");
			typedTable.append(parent.getName());
			typedTable.append(" INHERIT SELECT PRIVILEGES");
			typedTable.append(newline);
		}
		
		/*boolean TRUE=>ci saranno elementi di cui specificare qualcosa(in pratica se devo aprire 
		 * le parentesi per specificare qualche opzione tipo riferimento, non nullit di un Field o Ref ecc)*/
		boolean elements=false;
		for (Field ooField : child.getFields()) {
			if (ooField.isIdentifier()){
				elements=true;
				break;
			}
			if (!ooField.isNullable()){
				elements=true;
				break;
			}
		}
		if(child.getReferences().size()>0)
			elements=true;
		if(parent==null)
			elements=true;

		/*come detto se ci sono apro le parentesi*/
		if(elements)
			typedTable.append("(");

		Vector<String> primaryKey = new Vector<String>();
		//1)tutti i Field
		for (Field ooField : child.getFields()) {
			if (ooField.isIdentifier())
				primaryKey.add(ooField.getName());
			
			if (!ooField.isNullable()) {
				typedTable.append(ooField.getName());
				typedTable.append(" WITH OPTIONS NOT NULL, ");
			}
		}
		//2) le eventuali Ref
		for (Reference ooReference : child.getReferences()) {
			typedTable.append(ooReference.getName());
			
			if(!ooReference.isNullable())
				typedTable.append(" WITH OPTIONS NOT NULL SCOPE ");
			else
				typedTable.append(" WITH OPTIONS SCOPE ");
			
			typedTable.append(schema);
			typedTable.append(".");
			typedTable.append(ooReference.getClassTo().getName());
			typedTable.append(", ");			
		}

		//3)chiave primaria
		if(primaryKey.size()>0) {
			typedTable.append("CONSTRAINT pk PRIMARY KEY (");
			for (String field : primaryKey) {
				typedTable.append(field);
				typedTable.append(", ");
			}
			typedTable.replace(typedTable.length() - 2, typedTable.length(), "), ");
		}		
		
		/*se questa typed table NON  una foglia in una gerarchia allora devo indicare il tipo di riferimento*/
		if(parent==null) {
			typedTable.append("REF IS OID" + child.getName() + " USER GENERATED, ");
		}			
		//chiudo le parentesi
		if(elements)
			typedTable.replace(typedTable.length() - 2, typedTable.length(), ") ");

		return typedTable.toString();
	}
	
	/**Genera il comando SQL per creare un tipo complesso
	 * Genera cio CREATE TYPE .. AS ()
	 * 
	 * @param complex
	 * @param schema
	 * @throws ImportExportException 
	 */
	private String exportComplexType(ComplexType complex,ORSchema schema) throws ImportExportException{
		// preamble
		StringBuilder type = new StringBuilder();
		String newline = System.getProperty("line.separator");
		
		type.append("CREATE TYPE ");
		type.append(schema.getName());
		type.append(".");
		type.append(complex.getName() + TYPE_SUFFIX);
		
		type.append(newline);
		type.append("AS (");
		//fields
		for (Field ooField : complex.getFields()) {
			type.append(ooField.getName());
			type.append(" ");
			String tipo=DB2ObjectOrientedDataType.getOOType(ooField.getType());
			type.append(tipo);
			type.append(", ");
		}
		//REF
		for (Reference ooReference : complex.getReferences()) {
			type.append(ooReference.getName());
			type.append("  REF(");
			type.append(schema.getName());
			type.append(".");
			type.append(ooReference.getClassTo().getName()+TYPE_SUFFIX);
			type.append(")");
			type.append(", ");
		}
		/*c' inoltre la possibilit che questo complex type appartenga ad un Nest, quindi*/		
		for(Nest nest:schema.getNestes()) {
			if(nest.getTop()==complex){
				ComplexType nested=nest.getNested();
				type.append(nested.getName());
				type.append(" ");
				type.append(schema.getName());
				type.append(".");
				type.append(nested.getName()+TYPE_SUFFIX);
				type.append(", ");
			}
		}
		type.replace(type.length() - 2, type.length(), ") ");
		/*aggiungo il suffisso perch questo non facendo parte di una typed table 
		 * non pu far parte di una gerarchia*/
		type.append(newline + MODE_DB2SQL_TYPE);

		return type.toString();
	}
	
	/**Export the instances of the given schema to the given location.
	 * Based on the type of the Destination this class must export the information contained in the Schema.
	 * For example in case of a ORSchema to a DBMS Destination for each Tuple this metod must 
	 * create a SQL command INSERT INTO ... VALUES (...) and the same for each ORClass
	 * 
	 * @param schema the schema to be exported
	 * @param destination the destination to which export the schema
	 * @throws ImportExportException any error during the export
	 */
	public void exportInstance(ORSchema schema, Destination<Connection> destination) throws ImportExportException {
		/*per prima cosa vado a popolare la mappa dei collegamenti tra oggetti
		 * serve perch cos la ricerca dei collegamenti tra oid degli oggetti cinclusi in una generalizzazione
		 *  molto pi veloce*/
		for(ObjectGeneralization objGen:schema.getObjectGeneralization()) {
			int objParent=objGen.getParentObject().getObjectOID();
			int objChild=objGen.getChildObject().getObjectOID();
			/*li posso inserire in questo ordine perch sono certo che OGNI oggetto abbia un OID
			 * diverso da quello degli altri
			 * Mi serve perch io scendo nella gerarchia fino a trovare l'oggetto che non ha figli in
			 * nessuna della classi sottostanti e lo creo risalendo la gerarchia fino all'antenato
			 * vedi tesi*/
			this.objectGeneralization.put(objChild,objParent);
		}
				
		//lista dei comandi che devo eseguire
		Vector<String> comandi=new Vector<String>();
		/*devo generare i comandi SQL, ma rispettando necessariamente la gerarchia*/
		List<Insertable> ordered=schema.generateHierachy();
		Iterator<Insertable> i=ordered.iterator();
		
		while(i.hasNext()) {
			Object obj=i.next();
			if (obj instanceof ORTable) {
				ORTable table=(ORTable)obj;
				for(ORTuple tuple:table.getORTuples()) {
					comandi.add(generateValuesORTable(tuple, table, schema));
				}				
			}
			else if (obj instanceof ORClass) {
				ORClass orClass=(ORClass)obj;
				OOClass parent=schema.getParent(orClass);
				for(ORObject object:orClass.getORObjects()) {
					comandi.add(generateValuesORObject(object, schema, parent));
				}				
			}
			if (obj instanceof ComplexType) {
				/*non devo generare valori, li genero insieme alla tabelle o alla classi*/
			}
		}
		
		comandi.addAll(generateForeignKey(schema));
		
		try {
			Statement st = destination.getDestination().createStatement();
			for(String s:comandi) {
				if(s.length()>1) {
					System.out.println(s);
					System.out.println();
					st.executeUpdate(s);
				}
			}
			st.close();					
		} catch (SQLException e) {
			e.printStackTrace();
			throw new ImportExportException("Errore nell'Export dell schema:");
		}	
	}
	
	/**Genera il codice SQL per inserie una tupla di una tabella.
	 * In particolare crea il comando INSERT INTO [TABELLA]([COLONNE],[REF],[COMPLEX]) VALUES (..)
	 * 
	 */
	private String generateValuesORTable(ORTuple tuple,ORTable table,ORSchema schema) {
		StringBuilder sb = new StringBuilder();
		sb.append("INSERT INTO ");
		sb.append(schema.getName());
		sb.append(".");
		sb.append(table.getName());
		sb.append(" (");
		//1)tutte le colonne
		for(TupleComponent tupleComponent:tuple.getTupleComponents()) {
			sb.append(tupleComponent.getColumn().getName());
			sb.append(", ");
		}
		//2)tutte le Ref
		for(ORReferenceComponent ref:tuple.getReferenceComponents()) {
			sb.append(ref.getRef().getName());
			sb.append(", ");
		}
		//3)tutti i complex
		for(ComplexTypeComponent complexComponent:tuple.getComplexComponents()) {
			ComplexType complex=complexComponent.getComplex();
			sb.append(complex.getName());
			sb.append(", ");		
		}
		sb.replace(sb.length()-2, sb.length(), ")");
		/*adesso genero tutti valori*/
		sb.append(" VALUES( ");
		//1)i valori delle colonne
		for(TupleComponent tupleComponent:tuple.getTupleComponents()) {
			/*verifico se le istanze di una certa colonna hanno bisogno degli apici 
			 * per le loro istanze, come ad esempio varchar ecc. vedi sotto*/
			if(tupleComponent.getValue()==null)
				sb.append("null");
			else {
				String value=DB2ObjectOrientedDataType.getOOValue(tupleComponent.getColumn().getType(), tupleComponent.getValue());
				sb.append(value);
			}
			sb.append(", ");
		}
		//2)i valori delle Ref
		for(ORReferenceComponent ref:tuple.getReferenceComponents()) {
			if(ref.getObject()==null) {
				sb.append("NULL");
			}
			else {
				sb.append(schema.getName());
				sb.append(".");
				sb.append(ref.getRef().getTo().getName()+TYPE_SUFFIX);
				sb.append("(");
				sb.append(ref.getObject().getObjectOID());
				sb.append(")");
			}	
			sb.append(", ");
		}
		//3)i valori dei complex collegati
		for(ComplexTypeComponent complexComponent:tuple.getComplexComponents()) {
			/*genero i valori per il complexComponent attuale*/
			sb.append(generateSQLComplexComponent(complexComponent,schema.getName()));
			/*verifico se questi  inserito in un nest*/
			for(Nest nest:schema.getNestes()) {
				for(NestComponent nestComponent:nest.getNestedComponent()) {
					if(nestComponent.getTop().getOID()==complexComponent.getOID()) {
						ComplexTypeComponent nested=nestComponent.getNested();
						sb.append("..");
						sb.append(nested.getComplex().getName());
						sb.append("(");
						sb.append(generateSQLComplexComponent(nested,schema.getName()));
						sb.append(generetaSQLNestedComponent(schema,nested));
						sb.append(")");
					}
				}
			}
			sb.append(", ");
		}
		sb.replace(sb.length()-2, sb.length(), ")");
		
		return sb.toString();
	}
	
	private String generetaSQLNestedComponent(ORSchema schema,ComplexTypeComponent complexTop) {
		StringBuilder sb=new StringBuilder();
		for(Nest nest:schema.getNestes()) {
			for(NestComponent nestComponent:nest.getNestedComponent()) {
				if(nestComponent.getTop().getOID()==complexTop.getOID()) {
					ComplexTypeComponent nested=nestComponent.getNested();
					sb.append("..");
					sb.append(nested.getComplex().getName());
					sb.append("(");
					sb.append(generateSQLComplexComponent(nested,schema.getName()));
					sb.append(generetaSQLNestedComponent(schema,nested));
					sb.append(")");
				}
			}
		}		
		return sb.toString();
	}
	
	/**Genera la stringa per l'inserimento di un ComplexTypeComponent.
	 * In particolare genera la string ain questo modo :
	 * 		nomeTipo()..[nomeField(valore)]..[nomeRef(valore)]
	 * 
	 * @param component
	 */
	private String generateSQLComplexComponent(ComplexTypeComponent component,String schema){
		StringBuilder sb=new StringBuilder();
		sb.append(schema);
		sb.append(".");
		sb.append(component.getComplex().getName()+TYPE_SUFFIX);
		sb.append("()..");
		for(ValueComponent value:component.getValues()) {
			if(value.getValue()!=null) {
				sb.append(value.getOoField().getName());
				sb.append("(");
				String valore=DB2ObjectOrientedDataType.getOOValue(value.getOoField().getType(), value.getValue());
				sb.append(valore);
				sb.append(")..");
			}			
		}
		for(ReferenceComponent ref:component.getReferencesComponets()) {
			if(ref.getOoObjectTo()!=null) {
				sb.append(ref.getOoReference().getName());
				sb.append("(");
				sb.append(schema);
				sb.append(".");
				sb.append(ref.getOoObjectTo().getOoClass().getName()+TYPE_SUFFIX);
				sb.append("(");
				sb.append(ref.getOoObjectTo().getObjectOID());
				sb.append(")");
				sb.append(")..");
			}			
		}
		sb.replace(sb.length()-2, sb.length(), "");
		return sb.toString();
	}
	
	/**Genera il comando SQL per inserire i valori in una typed table.
	 * In particolare il comando che va a generare :
	 * INSERT INTO .. ([OID],[FIELD],[REF],[COMPLEX]) VALUES (..) 
	 * 
	 * @throws ImportExportException
	 */
	private String generateValuesORObject(ORObject object,ORSchema schema,OOClass parent) throws ImportExportException {
		if(!this.checkObjectGeneralization(object.getObjectOID())) {
			//i nomi delle colonne
			StringBuilder col=new StringBuilder("INSERT INTO ");
			col.append(schema.getName());
			col.append(".");
			col.append(object.getOoClass().getName());
			col.append(" (");
			//i valori
			StringBuilder values=new StringBuilder("(");
			
			/*inserisco gli elementi dell'ultimo oggetto della gerarchia*/
			//1)i valori
			for (ValueComponent valueComponent : object.getValueComponents()) {
				col.append(valueComponent.getOoField().getName());
				col.append(", ");
				values.append(getSQLValue(valueComponent));
				values.append(", ");
			}
			//2)le Ref
			for(ReferenceComponent ooref:object.getReferenceComponents()) {
				col.append(ooref.getOoReference().getName());
				col.append(", ");
				values.append(generateSQLDatabase(ooref,schema.getName()));
				values.append(", ");
			}
			//3)tutti i complex
			for(ComplexTypeComponent complexComponent:object.getComplexComponets()) {
				ComplexType complex=complexComponent.getComplex();
				col.append(complex.getName());
				col.append(", ");
				
				/*genero i valori per il complexComponent attuale*/
				values.append(generateSQLComplexComponent(complexComponent,schema.getName()));
				/*verifico se questi  inserito in un nest*/
				for(Nest nest:schema.getNestes()) {
					for(NestComponent nestComponent:nest.getNestedComponent()) {
						if(nestComponent.getTop().getOID()==complexComponent.getOID()) {
							ComplexTypeComponent nested=nestComponent.getNested();
							values.append("..");
							values.append(nested.getComplex().getName());
							values.append("(");
							values.append(generateSQLComplexComponent(nested,schema.getName()));
							values.append(generetaSQLNestedComponent(schema,nested));
							values.append(")");
						}
					}
				}
				values.append(", ");
			}
			/*adesso devo risalire nella lista degli oggetti fino a che non ho raggiunto 
			 * la vetta della gerarchia*/
			int objectOID=object.getObjectOID();
			OOClass last=object.getOoClass();
			
			while(parent!=null) {
				/*prendo il prossimo oggetto*/
				int objectToOID=getObjectGeneralization(objectOID);
				/*prendo l'oggeto corrispondente al prossimo OID*/
				OOObject o=parent.getOOObjectByOID(objectToOID);	
				
				/*anche per questo prendo gli elementi*/
				for (ValueComponent valueComponent : o.getValueComponents()) {
					col.append(valueComponent.getOoField().getName());
					col.append(", ");
					values.append(this.getSQLValue(valueComponent));
					values.append(", ");
				}
				for(ReferenceComponent ooref:o.getReferenceComponents()) {
					col.append(ooref.getOoReference().getName());
					col.append(", ");
					values.append(generateSQLDatabase(ooref,schema.getName()));
					values.append(", ");
				}
				
				/*mi preparo per il ciclo successivo*/
				last=parent;
				parent=schema.getParent(parent);
				objectOID=objectToOID;
			}
			/*devo inserire ancora il nome della colonna di self-referencing
			 * che ha il nome della classe pi in alto nella gerarchia*/
			col.append("OID" + last.getName());
			col.append(") ");
			/*aggiungo ai valori il corrispondente valore di identificazione, ma questo va preso
			 * dal primo oggetto, cio quello pi basso nella gerarchia*/
			String leafClass=object.getOoClass().getName();
			values.append(schema.getName()+"."+leafClass+TYPE_SUFFIX+"(");
			values.append(object.getObjectOID() + ")) ");
			
			/*fondo le due stringhe insieme, in maniera corretta*/
			col.append(" VALUES ");
			col.append(values);
			return col.toString();
		}
		else 
			return "";
	}
	
	/**Return an OID corrisponding to the object OID of the parent of the given object OID
	 * @param objectOID
	 * @throws ImportExportException if this Object has'n a parent object
	 */
	private int getObjectGeneralization(int objectOID) throws ImportExportException {
		if(this.objectGeneralization.containsKey(objectOID))
			return this.objectGeneralization.get(objectOID);
		else
			throw new ImportExportException("Non  stato possibile caricare l'oggetto padre di "+objectOID);
	}
	
	private boolean checkObjectGeneralization(int objectOID) {
		if(this.objectGeneralization.containsValue(objectOID))
			return true;
		else 
			return false;
	}
	
	/**Returns a valid SQL value for the 
	 * @param type
	 */
	private String getSQLValue(ValueComponent value) {
		if(value.getValue()==null)
			return "NULL";
		/*devo verificare che tipo sia in modo da aggiungere o no gli apici*/
		else {
			return DB2ObjectOrientedDataType.getOOValue(value.getOoField().getType(), value.getValue());
		}
	}
	
	/**This metod generates the SQL code for a Reference
	 * */
	private String generateSQLDatabase(ReferenceComponent ref,String schema_name) {
		if(ref.getOoObjectTo()==null) {
			return "NULL";
		}
		else {
			OOClass ooClass =ref.getOoObjectTo().getOoClass(); 
			String typeName = ooClass.getName() + TYPE_SUFFIX;
			StringBuilder sql = new StringBuilder(schema_name);
			sql.append(".");
			sql.append(typeName + "(");
			sql.append(ref.getOoObjectTo().getObjectOID());
			sql.append(")");
			return sql.toString();
		}		
	}	
	
	/**Generate the SQL commands to add the Foreign Keys.
	 * It generates:
	 * 
	 * ALTER TABLE .. ADD FOREIGN KEY ..
	 * 
	 * @throws PersistenceException 
	 */
	private List<String> generateForeignKey(ORSchema schema) throws ImportExportException {
		List<String> comandi=new ArrayList<String>();
		
		for(ForeignKey fk:schema.getForeignKeys()) {
			/*1)alter table*/
			StringBuilder sb=new StringBuilder("ALTER TABLE ");
			sb.append(schema.getName());
			sb.append(".");
			
			/*2)nome della tabella di origine*/
			Table tableFrom=null;
			ORClass classFrom=null;
			
			if(fk.getType()==ForeignKey.TableToTable || fk.getType()==ForeignKey.TableToAbstract) {
				tableFrom=schema.getTableByOID(fk.getFromOID());
				if(tableFrom!=null)
					sb.append(tableFrom.getName());
				else 
					throw new ImportExportException("La FK "+fk.getOID()+" definisce una tabella From" +
							" non trovata");
			}
			/*2')nome della classe di origine*/
			else if(fk.getType()==ForeignKey.AbstractToAbstract || fk.getType()==ForeignKey.AbstractToTable) {
				classFrom=schema.getClassByOID(fk.getFromOID());
				if(classFrom!=null)
					sb.append(classFrom.getName());
				else 
					throw new ImportExportException("La FK "+fk.getOID()+" definisce una classe From" +
					" non trovata");
			}
			/*3)add foreign key*/
			sb.append(" ADD FOREIGN KEY ( ");
			
			/*4)i componenti*/
			for(ComponentOfForeignKey cFk:fk.getComponents()) {
				if(tableFrom!=null){
					Column from=schema.getColumnByOID(cFk.getFromOID());
					if(from!=null) {
						sb.append(from.getName());
						sb.append(", ");
					}
					else
						throw new ImportExportException("Il componente "+cFk.getFromOID()+" Non  stato trovato");
						
				}
				else if(classFrom!=null){
					Field from=schema.getFieldByOID(cFk.getFromOID());
					if(from!=null) {
						sb.append(from.getName());
						sb.append(", ");
					}
					else
						throw new ImportExportException("Il componente "+cFk.getFromOID()+" Non  stato trovato");
				}
			}
			
			/*5)References*/
			sb.replace(sb.length()-2, sb.length(), " ) ");
			sb.append("REFERENCES ");
			sb.append(schema.getName());
			sb.append(".");
			
			/*6)tabella o class di destinazione*/
			Table tableTo=null;
			ORClass orClassTo=null;
			
			if(fk.getType()==ForeignKey.AbstractToTable || fk.getType()==ForeignKey.TableToTable) {
				tableTo=schema.getTableByOID(fk.getToOID());
				if(tableTo!=null) {
					sb.append(tableTo.getName());
					sb.append(" ( ");
				}
				else 
					throw new ImportExportException("La FK "+fk.getOID()+" definisce una tabella To" +
						" non trovata");
			}
			else if(fk.getType()==ForeignKey.AbstractToAbstract || fk.getType()==ForeignKey.TableToAbstract) {
				orClassTo=schema.getClassByOID(fk.getToOID());
				if(orClassTo!=null) {
					sb.append(orClassTo.getName());
					sb.append(" ( ");
				}
				else 
					throw new ImportExportException("La FK "+fk.getOID()+" definisce una classe To" +
						" non trovata");
			}
			
			/*7)componenti destinazione*/
			for(ComponentOfForeignKey cFk:fk.getComponents()) {
				if(tableTo!=null) {
					Column to=schema.getColumnByOID(cFk.getToOID());
					if(to!=null) {
						sb.append(to.getName());
						sb.append(", ");
					}
					else
						throw new ImportExportException("Il componente "+cFk.getToOID()+" Non  stato trovato");
				}
				else if(orClassTo!=null) {
					Field to=schema.getFieldByOID(cFk.getToOID());
					if(to!=null) {
						sb.append(to.getName());
						sb.append(", ");
					}
					else
						throw new ImportExportException("Il componente "+cFk.getToOID()+" Non  stato trovato");
				}
			}
			
			/*8)fine*/
			sb.replace(sb.length()-2, sb.length(), " )");
			//System.out.println(sb.toString()+"\n");
			comandi.add(sb.toString());
		}		
		return comandi;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			ORSchema schema = ImportExportPersistence.getInstance().loadORSchema(1);
			ImportExportPersistence.getInstance().loadORInstance(schema, 1);
			//System.out.println(schema.toString());
			Exporter<ORSchema, Destination<Connection>> exporter=new DB2ORExporter();
			DbConnection con=ConnectionFactory.getConnection(ConnectionFactory.DB2, "Temp1","localhost", 
					50000, "db2admin","db2admin","");
			Destination<Connection> destination=new Destination<Connection>(con.getConnection());
			destination.getDestination().setAutoCommit(false);
			
			exporter.exportSchema(schema, destination);
			exporter.exportInstance(schema, destination);
			
			destination.getDestination().rollback();
			
		} catch (PersistenceException e) {
			e.printStackTrace();
		} catch (SQLException e) {
			e.printStackTrace();
		} 
		catch (ImportExportException e) {
			e.printStackTrace();
		}
	}
}
