package org.modelgen.importexport.Export;

import org.modelgen.importexport.utilities.ImportExportException;
import org.modelgen.importexport.utilities.XmlSchemaSourceDestination;
import org.modelgen.importexport.xml.model.ComplexElement;
import org.modelgen.importexport.xml.model.ComplexInstance;
import org.modelgen.importexport.xml.model.ComplexType;
import org.modelgen.importexport.xml.model.ComponentOfForeignKey;
import org.modelgen.importexport.xml.model.Element;
import org.modelgen.importexport.xml.model.ForeignKey;
import org.modelgen.importexport.xml.model.InstanceElement;
import org.modelgen.importexport.xml.model.SimpleElement;
import org.modelgen.importexport.xml.model.SimpleInstance;
import org.modelgen.importexport.xml.model.XmlSchema;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

/**Export a XMLSchema to a File.
 * An Exporter is a class which has the rule to translate the given schema of type T
 * in a correct form for the Destination E.
 * For example we can export a XML schema both to a File and a DBMS with XML capabilities;
 * we can threfore define two class: one which has the rule to export to a File and one to a DBMS 
 * 
 * @version 0.9
 */
public class XMLSchemaExporter implements Exporter<XmlSchema, Destination<XmlSchemaSourceDestination>> {
	/*namaspace di default*/
	private static final String NAMESPACE="xsd";
	/*URI del namespace di default*/
	private static final String URI="http://www.w3.org/2001/XMLSchema";
	/*base per il nome della chiave*/
	private static final String KEY_BASE="key_";
	private String fkElementTo="";
	/*namespace delle istanze*/
	private static final String INSTANCE_NAMESPACE="xsi";
	/*URI del namespace per le istanze*/
	private static final String INSTANCEURI="http://www.w3.org/2001/XMLSchema-instance";
	
	/**Check that the model Given can be export to the Destination.
	 * When we define a class that implements this interface we can export not only
	 * the model it was written for, but also other model with less costructs
	 * For example a class that export a RelationalSchema can also export a RelationalSchema
	 * in which model does not exist FK.
	 * 
	 *  Per maggiori info vedere modello massimale e minimale nella tesi
	 * 
	 * @param modelOID of the model as stored in ModelGen db
	 * @return true if the model is between maximal e minimal model
	 * @throws ImportExportException
	 */
	public boolean checkModel(int modelOID) throws ImportExportException {
		return true;
	}
	
	/**Export the given schema to the given destination.
	 * Based on the type of the Destination this class must export the information contained in the Schema.
	 * For example in the case of a XMLSchema to a File it must create the correct elements
	 * 
	 * <xsd:element name="" ...>
	 * 		<simplexType/complexType>
	 * 		...
	 * 		</simplexType/complexType>
	 * </xsd:element>
	 * 
	 * @param schema the schema to be exported
	 * @param destination the destination to which export the schema
	 * @throws ImportExportException any error during the export
	 */
	public void exportSchema(XmlSchema schema, Destination<XmlSchemaSourceDestination> destination) throws ImportExportException {
		try {
			/*dalla destination tiro fuori il file*/
			FileWriter output=new FileWriter(destination.getDestination().getSchema());

			/*creo un writer XML sul FileWriter appena definito*/
			XMLStreamWriter writer=XMLOutputFactory.newInstance().createXMLStreamWriter(output);
			
			/*setto il prefisso di default*/
			writer.setPrefix(NAMESPACE, URI);
			/*specifico il namespace e la relativa URI*/
			writer.writeStartElement(NAMESPACE,"schema",URI);
			writer.writeNamespace(NAMESPACE,URI);
			/*va accapo :) */
			writer.writeCharacters("\n");
			
			/*creo un complexElement fittizzio che racchiuder tutti gli altri elementi*/
			ComplexType complexType=new ComplexType("");
			ComplexElement complexElement=new ComplexElement(schema.getName(),false,false,false,complexType);
			
			for(Element e:schema.getTopElements()) {
				complexType.addElement(e);
			}			
			/*esporto questo unico "grande elemento" che ricorsivamente conterr tutti gli altri*/
			exportComplexElement(writer, complexElement,1,schema);
			
			/*chiudo l'elemento schema, faccio il flush e chiudo il writer*/
			writer.writeEndElement();
			writer.flush();
			writer.close();
			output.close();
			
		} catch (XMLStreamException e) {
			e.printStackTrace();
			throw new ImportExportException("Errore nella scrittura degli elementi nel file XSD");
		} catch (IOException e) {
			throw new ImportExportException("Errore di I/O nella scrittura del file XSD");
		}
	}
	
	/**Exports a SimpleElement to the given writer.
	 * For more information about global/local element declaration see W3C or thesis
	 *  
	 * @param depth int used to express the indentation this element should have
	 * @throws XMLStreamException
	 */
	private void exportSimpleElement(XMLStreamWriter writer,SimpleElement e,int depth) throws XMLStreamException {
		//1)indentation
		indentation(writer, depth);
		
		//2)element e suoi attributi
		writer.writeStartElement(URI,"element");
		writer.writeAttribute("name",e.getName());
		if(e.isNullable())
			writer.writeAttribute("nillable","true");
		if(e.isOptional())
			writer.writeAttribute("minOccurs","0");
		if(e.isUnbounded())
			writer.writeAttribute("maxOccurs","unbounded");
		writer.writeAttribute("type",NAMESPACE+":"+e.getType());
		writer.writeCharacters("\n");
		
		//3)chiudo l'element
		indentation(writer, depth);
		writer.writeEndElement();
		writer.writeCharacters("\n");
	}
	
	/**Exports a ComplexElement to the given writer.
	 * For more information about global/local element declaration see W3C or thesis
	 *  
	 * @param depth int used to express the indentation this element should have
	 * @throws XMLStreamException
	 */
	private void exportComplexElement(XMLStreamWriter writer,ComplexElement e,int depth,XmlSchema schema) throws XMLStreamException {
		//1)indentation
		indentation(writer, depth);
		
		//2)element e suoi attrbuti
		writer.writeStartElement(URI,"element");
		writer.writeAttribute("name",e.getName());
		if(e.isNullable())
			writer.writeAttribute("nillable","true");
		if(e.isOptional())
			writer.writeAttribute("minOccurs","0");
		if(e.isUnbounded())
			writer.writeAttribute("maxOccurs","unbounded");
		writer.writeCharacters("\n");
		
		//3)specifico che  un complexType
		indentation(writer, depth+1);
		writer.writeStartElement(URI,"complexType");
		writer.writeCharacters("\n");
		
		//4)il complexType  una sequence
		indentation(writer, depth+2);
		writer.writeStartElement(URI,"sequence");
		writer.writeCharacters("\n");
		
		//5)per ogni elemento sfrutto ricorsivamente i metodi per i simpleElement/complexElement
		for(Element element:e.getComplexType().getElements()) {
			if (element instanceof SimpleElement) {
				SimpleElement simple = (SimpleElement) element;
				exportSimpleElement(writer,simple,depth+3);
			}
			else if (element instanceof ComplexElement) {
				ComplexElement complex = (ComplexElement) element;
				exportComplexElement(writer,complex,depth+3,schema);
			}
		}
		
		//6)chiudo tutti gli elementi aperti facendo attenzione all'indentazione
		indentation(writer, depth+2);
		writer.writeEndElement();
		writer.writeCharacters("\n");
		indentation(writer, depth+1);
		writer.writeEndElement();
		writer.writeCharacters("\n");
		
		//7)aggiungo l'eventuale chiave e le eventuali FK
		if(depth==1) {
			for(Element child:e.getChilds()){
				if(child instanceof ComplexElement)
					generateKeys(writer,(ComplexElement)child,"./");
			}				
			generateFk(writer, e, "", schema);
		}
		
		//8)chiudo l'element
		indentation(writer, depth);
		writer.writeEndElement();
		writer.writeCharacters("\n");
	}

	/**Generates the space needed for the indentation of the next Element.
	 * 
	 * @param writer
	 * @param depth the depth to which the next Element will be
	 * @throws XMLStreamException
	 */
	private void indentation(XMLStreamWriter writer,int depth) throws XMLStreamException {
		for(int i=0;i<depth;i++) {
			writer.writeCharacters("\t");
		}
	}
	
	/**Generates the Key element for the given Element e.
	 * This metod has the rule to search the tree of child of the given element looking for
	 * a child of type SimpleElement which has the property isIdentifier=true; the it generates 
	 * the corrispondending key element for the root of the tree
	 * 
	 * @param writer the output to which writes the key element
	 * @param e the actual element to check
	 * @param base String which represents the elements already seen
	 * @throws XMLStreamException
	 */
	private void generateKeys(XMLStreamWriter writer,ComplexElement e,String base) throws XMLStreamException  {
		/*qui devo andare in tutti i figli di questo nodo e se per qualcuno di questi ho isId=true
		 * devo generare la chiave*/
		boolean key=false;
		for(Element child:e.getComplexType().getElements()) {
			if(child instanceof SimpleElement && ((SimpleElement)child).isIdentifier())
				key=true;
		}	
		if(key) {
			//1)elemento key			
			indentation(writer, 1);
			writer.writeStartElement(URI,"key");
			writer.writeAttribute("name",KEY_BASE+e.getName());
			writer.writeCharacters("\n");
			
			//2)selector
			indentation(writer, 2);
			writer.writeStartElement(URI,"selector");
			writer.writeAttribute("xpath",base+e.getName());
			writer.writeEndElement();	
			writer.writeCharacters("\n");

			for(Element child:e.getComplexType().getElements()) {
				if(child instanceof SimpleElement && ((SimpleElement)child).isIdentifier()) {
					//3)i field della chiave				
					indentation(writer, 2);
					writer.writeStartElement(URI,"field");
					writer.writeAttribute("xpath",child.getName());
					writer.writeEndElement();
					writer.writeCharacters("\n");
				}
			}
			//4chiudo la chiave
			indentation(writer, 1);
			writer.writeEndElement();
			writer.writeCharacters("\n");			
		}
		/*per ultima cosa faccio la ricorsione di questo metodo su tutti gli elementi complessi, figli 
		 * dell'elemento attuale */
		for(Element child:e.getComplexType().getElements()) {
			if(child instanceof ComplexElement)
				generateKeys(writer,(ComplexElement)child,base+e.getName()+"/");
		}
	}
	
	/**Generates the elements keyref corrisponding to the Foreign Key in ModelGen.
	 * A keyref in Xml Schema has a name which is unique and it refers to a key (like in relational database).
	 * This element contains a selector which specifies the element this fk refer and a number of field corrisponding to
	 * the component of the foreign key.
	 * 
	 * @param writer
	 * @param e
	 * @param base
	 * @param schema
	 * @throws XMLStreamException
	 */
	private void generateFk(XMLStreamWriter writer, ComplexElement e, String base,XmlSchema schema) throws XMLStreamException {
		Vector<ForeignKey> fks=schema.getFk(e);
		if(fks.size()>0) {
			//System.out.println("per il complex ho fk "+e.getName());
			
			for(ForeignKey fk:fks) {
				//1)elemento keyref		
				indentation(writer, 1);
				writer.writeStartElement(URI,"keyref");
				writer.writeAttribute("name",fk.getName());
				writer.writeAttribute("refer",KEY_BASE+fk.getTo().getName());
				writer.writeCharacters("\n");
				
				//2)selector
				for(Element child:schema.getTopElements()) {
					findElement(child,fk.getFrom(),".");
				}
				indentation(writer, 2);
				writer.writeStartElement(URI,"selector");
				writer.writeAttribute("xpath",fkElementTo);
				writer.writeEndElement();	
				writer.writeCharacters("\n");
				
				for(ComponentOfForeignKey cFk:fk.getComponents()) {
					//3)i field della chiave				
					indentation(writer, 2);
					writer.writeStartElement(URI,"field");
					writer.writeAttribute("xpath",cFk.getFrom().getName());
					writer.writeEndElement();
					writer.writeCharacters("\n");
				}
				fkElementTo="";
				//4chiudo la chiave
				indentation(writer, 1);
				writer.writeEndElement();
				writer.writeCharacters("\n");				
			}
		}
		/*per ultima cosa faccio la ricorsione di questo metodo su tutti gli elementi complessi, figli 
		 * dell'elemento attuale */
		for(Element child:e.getComplexType().getElements()) {
			if(child instanceof ComplexElement)
				generateFk(writer,(ComplexElement)child, base, schema);
		}
	}
	

	/**Search the element to from the root of the tree saving the path used to get there.
	 * 
	 * @param child
	 * @param to
	 * @param base
	 */
	private void findElement(Element child, ComplexElement to,String base) {
		if(child==to) {
			//System.out.println("ho trovato l'elemento "+to.getName()+" e il selector vale "+base+"/"+child.getName());
			fkElementTo=base+"/"+child.getName();
		}
		else{
			//System.out.println("L'element "+child.getName()+" non  quello cercato, nei sui figli con base "+base+"/"+child.getName());
			for(Element e:child.getChilds())
				findElement(e, to, base+"/"+child.getName());
		}		
	}
	
	/**Exports the instance of a XmlSchema to a Xml File. 
	 * 
	 * 
	 *  @see org.modelgen.importexport.Export.Exporter#exportInstance(java.lang.Object, java.lang.Object)
	 */
	public void exportInstance(XmlSchema schema, Destination<XmlSchemaSourceDestination> destination) throws ImportExportException {
		try {
			/*dalla destination tiro fuori il file*/
			FileWriter output=new FileWriter(destination.getDestination().getInstance().get(0));

			/*creo un writer XML sul FileWriter appena definito*/
			XMLStreamWriter writer=XMLOutputFactory.newInstance().createXMLStreamWriter(output);
			writer.writeStartDocument();
			writer.writeCharacters("\n");
			
			/*creo primo elemento, ovvero la radice di tutti*/
			writer.writeStartElement(schema.getName());
			writer.writeNamespace(INSTANCE_NAMESPACE, INSTANCEURI);
			writer.writeAttribute(INSTANCEURI,"noNamespaceSchemaLocation", destination.getDestination().getSchema().getName());
			/*va accapo :) */
			writer.writeCharacters("\n");
			
			for(Element child:schema.getTopElements()) {
				for(InstanceElement childInstance:schema.getInstances()) {
					if(childInstance.getSchemaElement()==child && childInstance instanceof ComplexInstance) {
						/*scrivo il nome dell'elemento complesso*/
						indentation(writer, 1);
						writer.writeStartElement(child.getName());
						writer.writeCharacters("\n");
						/*scrivo ricorsivamente gli elementi che ho dentro*/
						exportIstance(writer,(ComplexInstance)childInstance,2);
						/*chiudo l'elemento*/
						indentation(writer, 1);
						writer.writeEndElement();
						writer.writeCharacters("\n");
					}
				}
			}
			/*chiudo l'elemento radice e il documento stesso*/
			writer.writeEndElement();
			writer.writeEndDocument();
			writer.flush();
			writer.close();
			output.close();
			
		} catch (XMLStreamException e) {
			e.printStackTrace();
			throw new ImportExportException("Errore nella scrittura degli elementi nel file XSD");
		} catch (IOException e) {
			throw new ImportExportException("Errore di I/O nella scrittura del file XSD");
		}
	}
	
	/**Metodo ricorsivo privato che scrive gli elementi delle istanze.
	 * 
	 * @param writer
	 * @param instance
	 * @param depth
	 * @throws XMLStreamException
	 */
	private void exportIstance(XMLStreamWriter writer,ComplexInstance instance, int depth) throws XMLStreamException {
		Element schema=instance.getSchemaElement();
		if(schema!=null && schema instanceof ComplexElement) {
			ComplexElement complex=(ComplexElement)schema;
			
			 for(Element child:complex.getChilds()) {
				for(InstanceElement childInstance:instance.getInstances()) {
					if(childInstance.getSchemaElement()==child && childInstance instanceof SimpleInstance) {
						indentation(writer, depth);
						writer.writeStartElement(child.getName());
						/*qui aggiungere controllo sui tipi*/
						String value=((SimpleInstance)childInstance).getValue();
						if(value!=null)
							writer.writeCharacters(value);
						else {
							writer.writeAttribute("xsi:nil", "true");
							writer.writeCharacters("");
						}
						writer.writeEndElement();
						writer.writeCharacters("\n");
					}
					else if(childInstance.getSchemaElement()==child && childInstance instanceof ComplexInstance) {
						indentation(writer, depth);
						writer.writeStartElement(child.getName());
						writer.writeCharacters("\n");						
						exportIstance(writer,(ComplexInstance)childInstance,depth+1);
						indentation(writer, depth);
						writer.writeEndElement();
						writer.writeCharacters("\n");
					}
				}
			 }			
		}
	}
}
