package org.modelgen.importexport.or.model;

import org.modelgen.importexport.oo.model.Field;
import org.modelgen.importexport.oo.model.Reference;

import java.util.List;
import java.util.Vector;

/**Represents a ComplexType.
 * A ComplexType is the representations of a complex attribute.It can have a number of 
 * Field(as for an OOClass) and Reference (also has a OOClass).
 * 
 * For example:
 * <code>
 * create type uomo(nome varchar(20),et integer,domicilio REF(indirizzo))
 * </code>
 * 
 * @version 0.9
 * 
 */
public class ComplexType implements Insertable{
	//OID
	private int OID;
	//name
	private String name;
	//type, it different from the name, during import it is the original name of the type
	private String type;
	private boolean isOptional;
	private boolean isNullable;
	//list of the fields of this complex
	private Vector<Field> fields;
	//list of the References of this complex
	private Vector<Reference> references;
	
	//list of the instances
	private Vector<ComplexTypeComponent> components;
	
	
	/**Create a new Complex with the given parameters.
	 * @param oid
	 * @param name
	 * @param isOptional
	 * @param isNullable
	 */
	public ComplexType(int oid, String name,String type, boolean isOptional, boolean isNullable) {
		OID = oid;
		this.name = name;
		this.type=type;
		this.isOptional = isOptional;
		this.isNullable = isNullable;
		this.fields=new Vector<Field>();
		this.references=new Vector<Reference>();
		this.components=new Vector<ComplexTypeComponent>();
	}
	
	/**Create a new Complex with the given parameters(without the type).
	 * @param oid
	 * @param name
	 * @param isOptional
	 * @param isNullable
	 */
	public ComplexType(int oid, String name,boolean isOptional, boolean isNullable) {
		OID = oid;
		this.name = name;
		this.type="";
		this.isOptional = isOptional;
		this.isNullable = isNullable;
		this.fields=new Vector<Field>();
		this.references=new Vector<Reference>();
		this.components=new Vector<ComplexTypeComponent>();
	}
	
	public void addField(Field field) {
		this.fields.add(field);
	}
	
	public void addReference(Reference ref) {
		this.references.add(ref);
	}
	
	public boolean add(ComplexTypeComponent o) {
		return components.add(o);
	}

	public Vector<Field> getFields() {
		return fields;
	}

	public boolean isNullable() {
		return isNullable;
	}

	public boolean isOptional() {
		return isOptional;
	}

	public String getName() {
		return name;
	}

	public int getOID() {
		return OID;
	}

	public Vector<Reference> getReferences() {
		return references;
	}

	public String getType() {
		return type;
	}	

	public Vector<ComplexTypeComponent> getComponents() {
		return components;
	}

	/** Returns true if this ComplexType is insertable based on the item already ordered.
	 * 
	 * An ComplexType is insertable if:
	 * 1)it has No Reference
	 * 2)it has References but the OOClass each Reference refers to is already inserted
	 * 2')this complex belongs to a Nest but the other elements belonging to this nest are 
	 * 	  already in the list
	 * 
	 * 
	 * @see org.modelgen.importexport.or.model.Insertable#isInsertable(List, ORSchema)
	 */
	public boolean isInsertable(List<Insertable> alreadyOrdered,ORSchema schema) {
		for(Reference ref:this.references) {
			if(!alreadyOrdered.contains(ref.getClassTo())) {
				//System.out.println(this.getName()+" NON  inseribile perch "+ref.getClassTo().getName()+" Non  ancora inserito");
				return false;
			}
		}
		/*o non ha reference o sono soddisfatte, ora per devo vedere se questo  l'elemento 
		 * pi piccolo*/
		/*devo pure verificare che se questo  il top di qualcuno sia stato 
		 * inserito in lista anche il nested*/
		Vector<ComplexType> nested=schema.getNested(this);
		if(nested.size()>0) {
			for(ComplexType complexType:nested) {
				if(!alreadyOrdered.contains(complexType))
					return false;
			}
		}		
		return true;
	}
	
	/**Returns a Strin representation of this object
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuilder sb=new StringBuilder("\tOID: ");
		sb.append(OID);
		sb.append(" name: ");
		sb.append(name);
		sb.append(" type: ");
		sb.append(type);
		sb.append(" isOpt: ");
		sb.append(isOptional);
		sb.append(" isNullable: ");
		sb.append(isNullable);
		
		sb.append("\n\t\tFields :( ");
		for(Field field:this.fields) {
			sb.append(field.getOid());
			sb.append("/");
			sb.append(field.getName());
			sb.append(",");
		}
		sb.replace(sb.length()-1, sb.length(), ")");
		sb.append(" REF :( ");
		for(Reference ref:this.references) {
			sb.append(ref.getOid());
			sb.append("/");
			sb.append(ref.getName());
			sb.append(",");
		}
		sb.replace(sb.length()-1, sb.length(), ")");
		
		sb.append("\n");
		for(ComplexTypeComponent component:components) {
			sb.append("\t");
			sb.append(component.toString());
		}
		
		return sb.toString();
	}
	
}
