package org.modelgen.importexport.or.model;

import org.modelgen.importexport.oo.model.OOClass;
import org.modelgen.importexport.oo.model.OOObject;
import org.modelgen.importexport.oo.model.Reference;

import java.util.List;
import java.util.Vector;

/**Represents an ORClass.
 * An ORClass is a OOClass(see..) but it can also contains a number of ComplexType.
 * 
 * For example
 * <code>
 * create type indirizzo (..)
 * create type uomo (nome varchar(20),cognome varchar(20),domicilio indirizzo,residenza indirizzo)
 * create type table uomini ..
 * </code>
 * 
 * @version 0.9
 * 
 */
public class ORClass extends OOClass implements Insertable {
	//lista dei complex
	private Vector<ComplexType> complexes;
	//lista delle istanze
	private Vector<ORObject> objects;

	/**Create a new ORClass with the given parameters
	 * @param oid
	 * @param name
	 */
	public ORClass(int oid, String name) {
		super(oid, name);
		this.complexes=new Vector<ComplexType>();
		this.objects=new Vector<ORObject>();
	}
	
	public void addComplex(ComplexType t) {
		this.complexes.add(t);
	}

	public Vector<ComplexType> getComplexes() {
		return complexes;
	}

	public Vector<ORObject> getORObjects() {
		return objects;
	}
	
	public Vector<OOObject> getOOObjects() {
		Vector<OOObject> temp=new Vector<OOObject>();
		for(ORObject obj:objects)
			temp.add(obj);
		return temp;
	}
	
	public boolean addORObject(ORObject o) {
		return objects.add(o);
	}
	
	/**Returns the OOObject identified by its OID(truly is a ORObject which "is a" OOObject) 
	 * @see org.modelgen.importexport.oo.model.OOClass#getOOObjectByOID(int)
	 */
	public OOObject getOOObjectByOID(int OID) {
		for(OOObject o:this.objects) {
			if(o.getObjectOID()==OID)
				return o;
		}
		return null;
	}
	
	public ORObject getORObjectByOID(int OID) {
		for(ORObject obj:objects) {
			if(obj.getObjectOID()==OID)
				return obj;
		}
		return null;
	}
	
	/**Returns true if this ORClass can be inserted in the ordered List
	 * 
	 * Each OOClass to be inserted in the ordere list needs that both all of its References and OOgeneralizations 
	 * are satisfied; each Reference points to another OOClass, which means that this 
	 * OOClass must be already in the ordered list.
	 * A OOGeneralization means tha each child can be inserted only after all its parent
	 * 
	 * @see org.modelgen.importexport.or.model.Insertable#isInsertable(List, ORSchema)
	 * 
	 * ATTENZIONE qui manca anckora la generalizzazione
	 */
	public boolean isInsertable(List<Insertable> alreadyOrdered,ORSchema schema) {
		/*se questa classe  inserita in una gerarchia suo padre deve gi essere stato inserito*/
		OOClass parent=schema.getParent(this);
		if(parent!=null && !alreadyOrdered.contains(parent))
			return false;
		/*caso banale: se NON ha ORRef e NON ha complex  inseribile*/
		if(this.references.size()==0 && this.complexes.size()==0){
			return true;
		}
		/*caso complesso: devo verificare che tutte le REF che ha siano soddisfatte*/
		else {
			for(Reference ref:this.references) {
				if(!alreadyOrdered.contains(ref.getClassTo())) {
					//System.out.println(this.getName()+" NON  inseribile perch "+ref.getClassTo().getName()+" Non  ancora inserito");
					return false;
				}
			}
			
			for(ComplexType complex:this.complexes) {
				for(Reference ref:complex.getReferences()) {
					if(!alreadyOrdered.contains(ref.getClassTo())) {
						//System.out.println(this.getName()+" NON  inseribile perch "+ref.getClassTo().getName()+" Non  ancora inserito");
						return false;
					}						
				}
				/*devo pure verificare che se questo  il top di qualcuno sia stato 
				 * inserito in lista anche il nested*/
				Vector<ComplexType> nested=schema.getNested(complex);
				if(nested.size()>0) {
					for(ComplexType complexType:nested) {
						if(!alreadyOrdered.contains(complexType))
							return false;
					}
				}
			}
			/*ho controllato tutte le Ref e sono soddisfatte quindi  inseribile*/
			return true;			
		}
	}
	
	/**Returns a String representations of this object
	 * @see org.modelgen.importexport.oo.model.OOClass#toString()
	 */
	public String toString() {
		StringBuilder sb=new StringBuilder(super.schemaToString());
		String newline = System.getProperty("line.separator");
		
		if(complexes.size()>0) {
			sb.append(newline);
			sb.append("\tComplex Type:");
			sb.append(newline);
			for(ComplexType t:complexes) {
				sb.append(t.toString());
				sb.append(newline);
			}
			sb.append(newline);
			
		}
		else {
			sb.append(newline);
			sb.append("\tNO Complex Type");
			sb.append(newline);
		}	
		
		if(this.objects.size()>0) {
			sb.append(newline);
			sb.append("\tInstances" + newline);
			for(ORObject obj:this.objects){
				sb.append(obj.toString());
			}
		}
		else{
			sb.append(newline);
			sb.append("\tNO Instance");
			sb.append(newline);
		} 
		
		return sb.toString();
	}
}
