package org.modelgen.importexport.or.model;

import org.modelgen.importexport.oo.model.Field;
import org.modelgen.importexport.oo.model.OOClass;
import org.modelgen.importexport.oo.model.OOGeneralization;
import org.modelgen.importexport.oo.model.ObjectGeneralization;
import org.modelgen.importexport.oo.model.Reference;
import org.modelgen.importexport.relational.model.Column;
import org.modelgen.importexport.relational.model.ForeignKey;
import org.modelgen.importexport.utilities.Schema;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;


/**Represents an Object Relational schema.
 * An OR schema is composed of table (ORTable) and typed table(ORClass) and a number of
 * Foreign Key between them.
 * 
 * The differences between ORTable and Table are the the first could have Reference to an OOClass
 * and a number of ComplexType(see ORTable)
 * 
 * The differences between ORClass and OOClass is that the first could have a number of ComplexType
 * (see ORClass)
 * 
 * @version 0.9
 * 
 */

public class ORSchema extends Schema{
	//tables belonging to the schema.
	private Vector<ORTable> tables;
	//foreign keys
	private Vector<ForeignKey> foreignKeys;
	//orClasses
	private Vector<ORClass> orClasses;
	//the generalizations of the schema
	private Vector<OOGeneralization> generalizations;
	//list of objectGeneralization
	private Vector<ObjectGeneralization> objectGeneralization;
	//questa li sta contiene i nest
	private Vector<Nest> nestes;
	//questa lista serve solo per essere pi rapido a ritrovarli
	private Vector<ComplexType> nested;

	/**Create a new ORSchema with the given parameters
	 * @param oid OID as stored in ModelGen
	 * @param name name as stored in ModelGen 
	 */
	public ORSchema(int oid, String name) {
		super(oid,name);
		this.tables=new Vector<ORTable>();
		this.foreignKeys=new Vector<ForeignKey>();
		this.orClasses=new Vector<ORClass>();
		this.generalizations=new Vector<OOGeneralization>();
		this.objectGeneralization=new Vector<ObjectGeneralization>();
		this.nestes=new Vector<Nest>();
		this.nested=new Vector<ComplexType>();
	}
	
	public void addORTable(ORTable table){
		this.tables.add(table);
	}
	
	public void addForeignKey(ForeignKey fk) {
		this.foreignKeys.add(fk);
	}
	
	public void addNest(Nest nest) {
		this.nestes.add(nest);
		this.nested.add(nest.getNested());
	}
	
	public Vector<ComplexType> getAllNested() {
		return nested;
	}

	public void addORClass(ORClass oc){
		this.orClasses.add(oc);
	}

	public Vector<ORClass> getORClasses() {
		return orClasses;
	}

	public Vector<ORTable> getTables() {
		return tables;
	}

	public String getName() {
		return name;
	}

	public int getOID() {
		return OID;
	}

	public Vector<ForeignKey> getForeignKeys() {
		return foreignKeys;
	}	
	
	public Vector<Nest> getNestes() {
		return nestes;
	}

	public void setForeignKeys(Vector<ForeignKey> foreignKeys) {
		this.foreignKeys = foreignKeys;
	}

	public void setGeneralizations(Vector<OOGeneralization> generalizations) {
		this.generalizations = generalizations;
	}

	public void setOrClasses(Vector<ORClass> orClasses) {
		this.orClasses = orClasses;
	}

	/**Returns a ORClass identified by its type.
	 * As we know the correspondence between a type and OOClass is one-to-one, so it posssible to
	 * identified a ORClass by its type
	 * 
	 * @param type the string that represent the type
	 * @return the ORClass is if it exists null otherwise
	 */
	public ORClass getORClassByType(String type) {
		for(ORClass o:orClasses) {
			if(o.getType().equalsIgnoreCase(type))
				return o;
		}
		return null;		
	}
	
	/**Return true if the given OOClass is a parent in any generalization
	 * @param o the OOCLass to check
	 * @return	true if it is a parent
	 */
	public boolean checkGeneralization(OOClass o) {
		for(OOGeneralization gen:generalizations) {
			if(gen.getParent()==o)
				return true;
		}
		return false;
	}
	
	/**Return the OOGeneralization to which the given OOClass is the parent
	 * @param parent the parent of the OOGeneralization
	 * @return the OOgenerlization if exists null otherwise
	 */
	public OOGeneralization getGeneralization(OOClass parent) {
		for(OOGeneralization gen:generalizations) {
			if(gen.getParent()==parent) {
				return gen;
			}
		}
		return null;
	}
	
	public Vector<OOGeneralization> getGeneralizations() {
		return this.generalizations;
	}
	
	/**Return the parent of the given OOCLass
	 * @param o the OOCLass to check
	 * @return	its parent if it exist null otherwise
	 */
	public OOClass getParent(OOClass o) {
		for(OOGeneralization gen : generalizations) {
			if(gen.getOOChilds().contains(o))
				return gen.getParent();
		}		
		return null;
	}
	
	public void addOOGeneralization(OOGeneralization o) {
		this.generalizations.add(o);
	}
	

	public boolean addObjectGeneralization(ObjectGeneralization objGen) {
		return objectGeneralization.add(objGen);
	}

	
	public Vector<ObjectGeneralization> getObjectGeneralization() {
		return objectGeneralization;
	}

	public ORTable getTableByOID(int tableOID) {
		for(ORTable table:tables) {
			if(table.getOID()==tableOID)
				return table;
		}
		return null;
	}
	
	public ORClass getClassByOID(int classOID) {
		for(ORClass o:orClasses) {
			if(o.getOid()==classOID)
				return o;
		}
		return null;
	}
	
	public ORClass getClassByName(String name) {
		for(ORClass o:orClasses) {
			if(o.getName().equals(name))
				return o;
		}
		return null;
	}
	
	public Column getColumnByOID(int columnOID) {
		for(ORTable table:tables) {
			for(Column c:table.getColumns()) {
				if(c.getOID()==columnOID)
					return c;
			}
		}
		return null;
	}
	
	public Field getFieldByOID(int fieldOID) {
		for(ORClass o:orClasses) {
			for(Field f:o.getFields()) {
				if(f.getOid()==fieldOID)
					return f;
			}
			for(ComplexType complex:o.getComplexes()) {
				for(Field field:complex.getFields()) {
					if(field.getOid()==fieldOID)
						return field;
				}
			}
		}
		for(ORTable table:tables) {
			for(ComplexType complex:table.getComplexes()) {
				for(Field field:complex.getFields()) {
					if(field.getOid()==fieldOID)
						return field;
				}
			}
		}
		for(ComplexType complex:nested) {
			for(Field field:complex.getFields()) {
				if(field.getOid()==fieldOID)
					return field;
			}
		}
		return null;
	}
	
	public Reference getReferenceByOID(int refOID) {
		for(ORClass orClass:orClasses) {
			for(Reference ref:orClass.getReferences()) {
				if(ref.getOid()==refOID)
					return ref;
			}
			for(ComplexType complex:orClass.getComplexes()) {
				for(Reference ref:complex.getReferences()) {
					if(ref.getOid()==refOID)
						return ref;
				}
			}
		}
		for(ORTable table:tables) {
			for(ComplexType complex:table.getComplexes()) {
				for(Reference ref:complex.getReferences()) {
					if(ref.getOid()==refOID)
						return ref;
				}
			}
		}
		for(ComplexType complex:nested) {
			for(Reference ref:complex.getReferences()) {
				if(ref.getOid()==refOID)
					return ref;
			}
		}
		return null;
	}
	
	public ForeignKey getForeignKeyByOID(int fkOID) {
		for(ForeignKey fk:foreignKeys) {
			if(fk.getOID()==fkOID)
				return fk;
		}
		return null;
	}
	
	public ORReference getORReferenceByOID(int OID) {
		for(ORTable table:tables) {
			for(ORReference ref:table.getReferences()) {
				if(ref.getOID()==OID)
					return ref;
			}
		}
		return null;
	}
	
	public ORTuple getORTuple(int OID) {
		for(ORTable table:tables) {
			for(ORTuple tuple:table.getORTuples()) {
				if(tuple.getTupleOID()==OID)
					return tuple;
			}
		}
		return null;
	}
	
	public ORTable getORTableByName(String name) {
		for(ORTable table:tables) {
			if(table.getName().equals(name))
				return table;
		}
		return null;
	}
	
	public ComplexType getComplexTypeByOID(int OID) {
		for(ORTable table:tables) {
			for(ComplexType complex:table.getComplexes()) {
				if(complex.getOID()==OID)
					return complex;
			}
		}
		for(ORClass orClass:orClasses) {
			for(ComplexType complex:orClass.getComplexes()) {
				if(complex.getOID()==OID)
					return complex;
			}
		}
		for(ComplexType complex:nested) {
			if(complex.getOID()==OID)
				return complex;
		}
		return null;
	}
	
	public ORObject getORObjectByOID(int OID) {
		for(ORClass orClass:orClasses) {
			ORObject obj=orClass.getORObjectByOID(OID);
			if(obj!=null)
				return obj;
		}
		return null;
	}
	
	public Nest getNestByOID(int OID) {
		for(Nest nest:nestes) {
			if(nest.getOID()==OID)
				return nest;
		}
		return null;
	}
	
	public ComplexTypeComponent getNestComponentByOID(int nestOID) {
		for(ComplexType complex:nested) {
			for(ComplexTypeComponent component:complex.getComponents()) {
				if(component.getOID()==nestOID)
					return component;
			}
		}
		return null;
	}
	
	/**Check if the given ORClass has already in its parent a ComplexType of the name
	 * given
	 * 
	 * @param orClass
	 * @param name
	 * @return true if already exist a ComplexType of the given name, false otherwise
	 */
	public boolean existComplexType(ORClass orClass,String name){
		System.out.println("qui ci sono passato..");
		OOClass temp=getParent(orClass);
		while(temp!=null) {
			ORClass parent=getClassByOID(temp.getOid());			
			for(ComplexType complex:parent.getComplexes()) {
				if(complex.getName().equalsIgnoreCase(name))
					return true;
			}
			temp=getParent(temp);
			//parent=getClassByOID(temp.getOid());
		}
		return false;
	}
	
	/**Generate the Hierachy of this schema.
	 * To do this we create a List with all the ORTable,all the ComplexType nested and ORClasses.
	 * The we search in this List until all element are insertable in the ordered list.
	 * 
	 * For each condition see isInsertable or thesis
	 * 
	 * 
	 * @return an ordered List
	 */
	public List<Insertable> generateHierachy() {
		List<Insertable> result=new ArrayList<Insertable>();

		Vector<Insertable> temp=new Vector<Insertable>();
		temp.addAll(this.nested);
		temp.addAll(this.tables);
		temp.addAll(this.orClasses);
		
		while(temp.size()>0) {
			Iterator<Insertable> i=temp.iterator();
			
			while(i.hasNext()) {
				Insertable ins=i.next();
				if(ins.isInsertable(result,this)){
					result.add(ins);
					i.remove();
				}
			}
		}		
		return result;
	}
	
	
	/**Returns a List of all Nested of the given ComplexType
	 * @param complex
	 */
	public Vector<ComplexType> getNested(ComplexType complex) {
		Vector<ComplexType> result=new Vector<ComplexType>();
		for(Nest nest:nestes) {
			if(nest.getTop()==complex) {
				result.add(nest.getNested());
			}
		}		
		return result;
	}
	
	/**Return the Nest which has the given ComplexType as top and nested
	 * @param top ComplexType top
	 * @param nested ComplexType nested
	 * @return Nest if it exists null otherwise
	 */
	public Nest getNest(ComplexType top,ComplexType nested) {
		for(Nest nest:this.nestes) {
			if(top==nest.getTop() && nested==nest.getNested()) {
				return nest;
			}
		}
		return null;
	}
	
	/**Check if a OOclass belongs to a OOGeneralization both as a parent or child
	 * @param o the OOCLass to check
	 * @return true if belongs false otherwise
	 */
	public boolean isImplicated(ORClass o) {
		for(OOGeneralization gen : generalizations){
			if(gen.getParent()==o)
				return true;
			else {
				for(OOClass child:gen.getOOChilds())
					if(child==o)
						return true;
			}
		}
		return false;
	}
	
	/**Returns a String representation of this schema
	 * @see java.lang.Object#toString()
	 */
	public String toString(){
		StringBuilder sb=new StringBuilder();
		sb.append("OID: ");
		sb.append(this.OID);
		sb.append(" nome: ");
		sb.append(name);
		sb.append("\n");
		
		for (ORTable table : tables) {
			sb.append(table.toString());
			sb.append("\n");
		}
		
		for (ORClass orClass : orClasses) {
			sb.append(orClass.toString());
			sb.append("\n");
		}
		
		for(ForeignKey fk: foreignKeys){
			sb.append(fk.toString());
			sb.append("\n");
		}
		
		sb.append("\nGeneralizzazioni:\n\n");
		for(OOGeneralization gen:generalizations) {
			sb.append(gen.toString());	
			sb.append("\n");
		}
		sb.append("\nCorrispondenze tra oggetti:\n");
		for(ObjectGeneralization objGen:this.objectGeneralization) {
			sb.append(objGen.toString());
			sb.append("\n");
		}
		
		sb.append("\nComplexNested:\n");
		for(ComplexType complex:this.nested) {
			sb.append("\t");
			sb.append(complex.toString());
			sb.append("\n\n");
		}
		
		sb.append("\nNest:\n");
		for(Nest nest:nestes) {
			sb.append("\t");
			sb.append(nest.toString());
			sb.append("\n");
		}
		
		return sb.toString();	
	}
}
