package org.modelgen.importexport.or.model;

import org.modelgen.importexport.oo.model.Reference;
import org.modelgen.importexport.relational.model.Column;
import org.modelgen.importexport.relational.model.Table;
import org.modelgen.importexport.relational.model.Tuple;
import org.modelgen.importexport.relational.model.TupleComponent;

import java.util.List;
import java.util.Vector;

/**Represent an ORTable which is a Table plus ComplexType and ORReference.
 * 
 * A Table can have only lexical Column; an ORTable can have number of ORReference, which are 
 * normal Reference but between a Table and a OOClass.
 * A complexType is a type used as an attribute of a Column(see ComplexType)
 * 
 * For example:
 * <code>
 * create type indirizzo(..)
 * create typed table indirizzi of indirizzo(..)
 * create table impiegati (nome varchar(20),et int,..,domicilio REF(indirizzo) scope indirizzi,residenza indirizzo)
 * </code>
 * 
 * @version 0.9
 * 
 */
public class ORTable extends Table implements Insertable {
	//le reference
	private Vector<ORReference> references;
	//gli attributi complessi
	private Vector<ComplexType> complexes;
	//le tuple
	private Vector<ORTuple> tuples;
	
	/**Creates a new ORTable with the given parameters
	 * @param oid OID as stored in ModelGen
	 * @param name name as stored in ModelGen
	 */
	public ORTable(int oid, String name) {
		super(oid,name);
		this.references=new Vector<ORReference>();
		this.tuples=new Vector<ORTuple>();
		this.complexes=new Vector<ComplexType>();
	}
	
	public void addORReference(ORReference ref){
		this.references.add(ref);
	}

	public Vector<ORReference> getReferences() {
		return references;
	}
	
	public void addComplex(ComplexType complex) {
		this.complexes.add(complex);
	}

	public Vector<ComplexType> getComplexes() {
		return complexes;
	}
	
	public void addORTuple(ORTuple tuple) {
		this.tuples.add(tuple);
	}

	public Vector<ORTuple> getORTuples() {
		return tuples;
	}
	
	public Vector<Tuple> getTuples() {
		Vector<Tuple> temp=new Vector<Tuple>();
		for(ORTuple tuple:tuples) {
			temp.add(tuple);
		}
		return temp;
	}

	/** Returns true if this ORTable is insertable based on the item already ordered.
	 * 
	 * An ORTable is insertable if:
	 * 1)it has No Reference and no ComplexTypes
	 * 2)it has References but the OOClass each Reference refers to is already inserted
	 * 2')is has ComplexTypes but ALL the nested ComplexType of each ComplexType and 
	 *    the Reference of each ComplexType are already inserted
	 *    
	 *    For more detail see thesis
	 * 
	 * 
	 * @see org.modelgen.importexport.or.model.Insertable#isInsertable(List, ORSchema)
	 */
	public boolean isInsertable(List<Insertable> alreadyOrdered,ORSchema schema) {

		/*caso banale: se NON ha ORRef e NON ha complex  inseribile*/
		if(this.references.size()==0 && this.complexes.size()==0) {
			return true;
		}
		/*caso complesso: devo verificare che tutte le REF che ha, sia di tipo OR che dei suoi complex
		 * siano soddisfatte*/
		else {
			for(ORReference ref:this.references) {
				if(!alreadyOrdered.contains(ref.getTo())) {
					//System.out.println(this.getName()+" NON  inseribile perch "+ref.getTo().getName()+" Non  ancora inserito");
					return false;				
				}
			}
			
			for(ComplexType complex:this.complexes) {
				for(Reference ref:complex.getReferences()) {
					if(!alreadyOrdered.contains(ref.getClassTo())) {
						//System.out.println(this.getName()+" NON  inseribile perch "+ref.getClassTo().getName()+" Non  ancora inserito");
						return false;
					}
				}
				/*devo pure verificare che se questo  il top di qualcuno sia stato 
				 * inserito in lista anche il nested*/
				Vector<ComplexType> nested=schema.getNested(complex);
				if(nested.size()>0) {
					for(ComplexType complexType:nested) {
						if(!alreadyOrdered.contains(complexType))
							return false;
					}
				}
			}
			/*ho controllato tutte le Ref e sono soddisfatte quindi  inseribile*/
			return true;
		}
	}
	
	/**Returns a String representation of this object
	 * @see org.modelgen.importexport.relational.model.Table#toString()
	 */
	public String toString() {
		StringBuffer table = new StringBuffer("\nTable: " + name + " oid: "+this.OID+" (");
		String newline = System.getProperty("line.separator");
		table.append(newline);
		table.append("Columns :\n");
		for (Column column : columns) {
			table.append(column);
			table.append(newline);
		}
		table.append("References :\n");
		for(ORReference ref: references){
			table.append(ref.toString());
			table.append(newline);
		}
		table.append("ComplexTypes :\n");
		for(ComplexType complex: complexes){
			table.append(complex.toString());
			table.append(newline);
		}
		
		table.append(")" + newline);
		
		//now let's print the instances
		table.append(newline);
		table.append(newline);
		table.append("INSTANCES" + newline);
		if(tuples.size()==0)
			table.append("No instances");
		else {
			for (ORTuple tuple : tuples) {
				table.append("Tuple OID: ");
				table.append(tuple.getTupleOID());
				table.append(newline);
				table.append("\t");
				for (TupleComponent cTuple : tuple.getTupleComponents()) {
					table.append(cTuple.toString());
				}
				for(ORReferenceComponent ref:tuple.getReferenceComponents()) {
					table.append(ref.toString());
				}
				table.append(newline);
				for(ComplexTypeComponent component:tuple.getComplexComponents()) {
					table.append(component.toString());
					table.append(newline);
				}				
				table.append(newline);
			}
		}		
		
		return table.toString();
	}
}
