package org.modelgen.jpf.automaticSignatureApplication.facade;

import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.scheme.Scheme;
import org.modelgen.signature.MacroRuleSignature;
import org.modelgen.signature.ModelSignature;
import org.modelgen.signature.application.loader.RulesLoader;
import org.modelgen.signature.application.macroruleapplication.NaiveCacheMacroRuleSignature;
import org.modelgen.signature.application.macroruleapplication.SignatureApplicationUtility;
import org.modelgen.signature.application.ruleapplication.ApplicateRuleSignature;
import org.modelgen.signature.application.treeSearcher.OrderedReductionsReader;
import org.modelgen.signature.application.treeSearcher.TreeSearcher;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

public class FindTranslationFacade {

	//segnature di tutte le riduzioni definite nel sistema
	private List<MacroRuleSignature> reductions = null;

	//singleton
	private static FindTranslationFacade instance = null;

	public static FindTranslationFacade getInstance(){
		if(instance==null)
			instance = new FindTranslationFacade();
		return instance;
	}

	/*
	 * Inizializza la cache delle macroregole
	 */
	private FindTranslationFacade(){
		if(NaiveCacheMacroRuleSignature.getInstance().getRegole()==null){
			SignatureApplicationUtility.getInstance().loadCache();
		}
		try {
			this.initReductions();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/*
	 * Inizializza la lista delle segnature di tutte le riduzioni esistenti
	 */
	private void initReductions() throws IOException{
		if(this.reductions==null){
			this.reductions = new LinkedList<MacroRuleSignature>();
			List<String> redNames = RulesLoader.getInstance().loadReductions();
			for(String reduction: redNames){
				MacroRuleSignature mrd = NaiveCacheMacroRuleSignature.getInstance().getRegola(reduction);
				reductions.add(mrd);
			}
		}
	}

	/**
	 * Implementa l'algoritmo di traduzione da uno schema sorgente verso un modello
	 * target.
	 * Ritorna una lista di nomi di file di macroregole che implementano la trasformazione.
	 * Ritorna null se la traduzione non  automatizzabile, cio se lo schema o il modello 
	 * non sono associati ad alcuna famiglia.
	 * Ritorna la lista vuota se non sono state trovate riduzioni
	 * @param source lo Scheme sorgente
	 * @param target il Model a cui si vuole giungere
	 * @return una lista di nomi di file di macroregole che implementano la trasformazione,
	 * null se la traduzione non  automatizzabile, cio se lo schema o il modello non sono
	 * associati ad alcuna famiglia.
	 * @throws IOException
	 */
	public List<String> translate(Scheme source, Model target) throws IOException{
		//determino la famiglia dello schema sorgente
		Model sourceFamily = this.getSchemeFamily(source);
		//determino la famiglia per il modello target
		Model targetFamily = this.getModelFamily(target);
		//se almeno una delle due famiglie non esiste, la traduzione non 
		//automatizzabile
		if(sourceFamily==null || targetFamily==null)
			return null;
		//se lo schema e il modello appartengono alla stessa famiglia, facciamo
		//una semplice ricerca delle riduzioni
		if(sourceFamily.getFamilyOid()==targetFamily.getFamilyOid())
			return this.findReductions(sourceFamily.getModelSignature(), target.getModelSignature());
		//altrimenti calcola la trafsormazione
		return this.findTransformation(sourceFamily.getModelSignature(), source.getModelSignature(), targetFamily.getModelSignature(), target.getModelSignature(), target.isAncestor());
	}

	/*
	 * Restituisce il capostipite della famiglia di uno schema
	 */
	private Model getSchemeFamily(Scheme schema){
		return schema.getMetaInfo().getFamily();
	}

	/*
	 * Restituisce il capostipite della famiglia di un modello
	 */
	private Model getModelFamily(Model model){
		return model.getFamily();
	}

	/**
	 * Data la segnatura del capostipite di una famiglia e quella di un modello
	 * della stessa famiglia, restituisce la lista delle riduzioni (nomi di files)
	 * @param sourceFamily la segnatura del capostipite di una famiglia
	 * @param modelTarget la segnatura di un modello della stessa famiglia
	 * @return la lista delle riduzioni (nomi di files)
	 */
	private List<String> findReductions(ModelSignature sourceFamily, ModelSignature modelTarget){
		List<String> reduction = new LinkedList<String>();
		List<MacroRuleSignature> listReduction = (new TreeSearcher()).getReductionFromFamilyToModel(sourceFamily, modelTarget, this.reductions);
		for(MacroRuleSignature sign: listReduction)
			reduction.add(sign.getFileName());
		return reduction;
	}

	/*
	 * Algoritmo di ricerca delle macroregole per la traduzione automatica.
	 * Se non vengono trovate, riporta la lista vuota
	 */
	private List<String> findTransformation(ModelSignature sourceFamily, ModelSignature schemeSource, ModelSignature targetFamily, ModelSignature modelTarget, boolean mode) throws IOException{
		List<String> result = new LinkedList<String>();
		//carico le informazioni sulle trasformazioni associabili alla famiglia sorgente
		OrderedReductionsReader in = new OrderedReductionsReader();
		List<List<MacroRuleSignature>> sourceAssociatedTrans = in.readFamilyReportingSignature(sourceFamily.getName());
		if(sourceAssociatedTrans!=null){
			//controllo tutte le varie possibilit
			for(List<MacroRuleSignature> transformation: sourceAssociatedTrans){
				//applico tutte le macroregole
				ApplicateRuleSignature applicator = new ApplicateRuleSignature();	
				ModelSignature reachedModel = sourceFamily;
				for(MacroRuleSignature macro: transformation){
					//non dar mai null, perch le regole sono state selezionate sulla famiglia
					reachedModel = applicator.createModelSignature(macro, reachedModel);
				}
				//se il modello ottenuto  compatibile con la famiglia target, seleziono le
				//macroregole utili, che possono essere un sottoinsieme di quelle definite per
				//la sua famiglia
				if(SignatureApplicationUtility.getInstance().isCompatibile(targetFamily, reachedModel)){
					ModelSignature model = schemeSource;
					ModelSignature tmpModel;
					for(MacroRuleSignature macro: transformation){
						if(SignatureApplicationUtility.getInstance().isApplicableMacroRule(macro, model)){
							tmpModel = applicator.createModelSignature(macro, model);
							//se la macroregola non  compatibile con lo schema sorgente.
							//Pu accadere finch non vengono gestite le reference nelle signature
							if(tmpModel==null){
								//evita la ricerca finale
								mode = true;
								break;
							}
							result.add(macro.getFileName());
							model = tmpModel;
						}
					}
					//se il modello target non  un capostipite, occorre calcolare le riduzioni dalla sua famiglia
					if(!mode && !SignatureApplicationUtility.getInstance().isCompatibile(modelTarget, reachedModel))
						result.addAll(this.findReductions(reachedModel, modelTarget));
					break;
				}
			}
		}
		return result;
	}

}
