package org.modelgen.jpf.opgen;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import javax.swing.JButton;
import javax.swing.JOptionPane;

import org.modelgen.jpf.opgen.ormHandling.MacroSaver;
import org.modelgen.regole.MacroRegola;

/**
 * represents a listener for button events
 * @author luigi, francesca
 *
 */
public class ButtonActionListener implements ActionListener {
	
	/* reference to the main frame */
	private OpGenTool ogt;		
	
	public ButtonActionListener(OpGenTool ogt) {
		this.ogt = ogt;
	}
	
	
	/**
	 * delegates action execution according to the
	 * type of button hit
	 */
	public void actionPerformed(ActionEvent e) {
		
		/* cast is safe since we build this instance only */
		/* when we pass it to addActionListener of */
		/* a JButton */
		String nomePulsante = ((JButton)e.getSource()).getName();
		
		if(nomePulsante.equals("Diff"))
			diffCommand();
		else if(nomePulsante.equals("Merge"))
			mergeCommand();
		else if(nomePulsante.equals("Save"))
			saveCommand();
			
	}
	
	
	/**
	 * performs a diff command
	 *
	 */
	private void diffCommand() {
		/* "please wait.." message */
		ogt.showHourglass(true);
		
		/* resets the current stored rule (to allow garbage collection) */
		ogt.resetRule();	

		/* invokes the domain operation */
		invokeCommandWorker(new DiffCommandWorker());
		
		/* task completed */
		ogt.showHourglass(false);
		
		displayRuleQuestion("DiffRule");
		
		
	}


	/**
	 * invokes command workers given the concrete instance of the worker.
	 * Note: Swing is (and must be) updated only by the event-dispatcher thread.
	 * @param cw
	 */
	private void invokeCommandWorker(CommandWorker cw) {
		
		Future<MacroRegola> mrF;
		ExecutorService es = Executors.newFixedThreadPool(1);
		
		
		/* invokes the domain operation */
		mrF = es.submit(cw);
		try {
			ogt.setMacroRule(mrF.get());
		} catch(Exception ex) {
			System.out.println("OPGEN: Execution error");
		}
	}

	/**
	 * asks and eventually displays the generated rule
	 *
	 */
	private void displayRuleQuestion(String ruleName) {
		int resp = ogt.showQuestionMessage("Show rule?", "Rule correctly generated. \nDo you want to display the generated rule? (Note it might take several seconds)");
		
		/* updates the gui */
		if(resp==JOptionPane.YES_OPTION) 
			ogt.updateView();		
		else
			ogt.setBodyText(ruleName + " rule correctly generated. You chose not to display it. You can now save the rule.");
	}
	
	/**
	 * performs a merge command
	 *
	 */
	private void mergeCommand() {
		/* "please wait.." message */
		ogt.showHourglass(true);
		
	    /* resets the current stored rule (to allow garbage collection) */
	    ogt.resetRule();
	    
	    /* invokes the domain operation */
	    invokeCommandWorker(new MergeCommandWorker());            
	     		
		/* task completed */
		ogt.showHourglass(false);
		
		displayRuleQuestion("MergeRule");
		
		
	}
	
	/**
	 * adjusts macrorule name according to some policies
	 * @param macrorule
	 * @return macrorule name
	 */
	private String adjustMacroRuleName(String macrorule) {
		return macrorule.toLowerCase();
		
	}
	
	/**
	 * verifies if the name is correct according to some policies. 
	 * Returns an error message, null if correct.
	 * @param name
	 * @return if the name is ok or not
	 */
	private String isNameOk(String name) {
		if(name.length()<1)
			return "Macrorule name must be at least 2 characters long";
		else if(Character.isDigit((name.charAt(0))))
			return "Macrorule name cannot start with a digit";
		else if(name.length()>16)
			return "Macrorule name cannot be longer than 16 digits";
		else
			return null;	
				
	}
	
	/**
	 * performs a save command
	 *
	 */
	private void saveCommand() {

		String macroRuleName = adjustMacroRuleName(ogt.getMacroRuleName());
		String errorMessage = isNameOk(macroRuleName);
		
		/* name check */
		if(errorMessage!=null) {
			ogt.showMessage("Macrorule name error",errorMessage, JOptionPane.ERROR_MESSAGE);
			return;	
		/* macrorule existance check */
		} else if(!ogt.isMacroRuleSet()) {
			ogt.showMessage("Macrorule not generated","You must generate a Macrorule before saving it", JOptionPane.ERROR_MESSAGE);
			return;
		/* all right */
		}	else {
			
			/* "please wait.." message */
			ogt.showHourglass(true);
			
			MacroSaver ms = new MacroSaver(ogt.getMacroRule());
			ms.save(macroRuleName);
			
			/* "please wait end" message */
			ogt.showHourglass(false);
			
			ogt.showMessage("Macrorule saved","Macrorule successfully saved", JOptionPane.INFORMATION_MESSAGE);
			
		}
		
		
	}

}
