package org.modelgen.jpf.opgen.common;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * represents a non-instanciated variant of a MsmConstruct
 * characterized by a set of references
 * @author luigi, francesca
 *
 */
public class KVariant implements Comparable<KVariant> {
	
	private String name;
	/* the corresponding carset */
	private MsmCarset carset; 
	/* the construct to which it belongs */
	private MsmConstruct construct;
	/* if this KVariant is first level */
	private boolean isFirstLevel;
	/* variant constructs */
	private List<VariantConstruct> variantConstructs;
 	
	
	public KVariant(MsmCarset carset, String name, MsmConstruct construct) {
		this.name = name;
		this.carset = carset;
		this.construct = construct;
	}
	
	public void setIsFirstLevel(boolean isFirstLevel) {
		this.isFirstLevel = isFirstLevel;
	}
	
	public boolean isFirstLevel() {
		return this.isFirstLevel;
	}
	
	public MsmCarset getMsmCarset() {
		return this.carset;
	}
	
	public void setMsmConstruct(MsmConstruct construct) {
		this.construct = construct;
		
	}
	
	public MsmConstruct getMsmConstruct() {
		return this.construct;
	}
	
	/**
	 * returns variant constructs corresponding to this kvariant.
	 * This method is recursive beacuse it calls MsmReference.getVariantConstructs()
	 * which recalls this.getVariantConstruct. The firstLevel is the base case.
	 * @return variant construct corresponding to this kvariant
	 */
	public List<VariantConstruct> getVariantConstruct() {
		
		if(this.variantConstructs==null) { // caching of variant constructs
		
			List<List<VariantConstruct>> constructionSet = new LinkedList<List<VariantConstruct>>();
			this.variantConstructs = new LinkedList<VariantConstruct>();
		
			/* if this has no references, returns an empty variant construct */
			/* corresponding to this KVariant */
			if(isFirstLevel) // base case 			
				/* first level variant construct inherit the kvariant name */
				this.variantConstructs.add(new VariantConstruct(this,name)); 
			else { // not first level (recursive case)
				Set<MsmReference> refs = this.carset.getMsmReferenceCollection();
				for(MsmReference ref: refs) {
					/* collects all variant constructs for each included reference */
					constructionSet.add(ref.getVariantConstructs());
				}
			
				/* compute the cartesian product of the construction set to calculate */
				/* all the possible ways to instanciate the references */
				SetOperator setOperator = new SetOperator();
				List<List<VariantConstruct>> constructionSet2 = setOperator.prod(constructionSet);
			
				/* to qualify the variant constructs name */
				int counter = 1;
				
				/* each cvList represents a possible instanciation for the references of this KVariant, so */
				/* for each cvList we create the corresponding VariantConstruct passing the */
				/* reference list and cvList */
				for(List<VariantConstruct> cvList : constructionSet2) {
					this.variantConstructs.add(new VariantConstruct(this,this.name+counter,new LinkedList<MsmReference>(refs),cvList));
					counter++; // change variant construct name
				}
			
			}	
		}
		
		return this.variantConstructs;
		
	}
	
	/**
	 * returns the name of the costruct associated with this kvariant
	 * @return the name of the costruct associated with this kvariant
	 */
	public String getConstructName() {
		
		return this.construct.getName();
	}
	
	/**
	 * returns the names of the properties of the MSM construct 
	 * associated with this kvariant
	 * @return properties names of the MSM construct associated with this kvariant
	 */
	public List<String> getPropertiesNames() {
		
		return this.construct.getPropertiesNames();
		
		
	}
	
	/**
	 * this := kv
	 * kv < kv2 := kv follows kv2
	 * 
	 * 1- kv < kv2 if there is a reference from kv to kv2. (+1)
	 * 2- kv > kv2 if there is a reference from kv2 to kv (-1)
	 * (references loop are not allowed in the supermodel)
	 * 3- if kv is first level and kv2 is not kv > kv2 (-1)
	 * 4- if kv2 is first level and kv is not kv2 > kv (+1)
	 * 5- if kv and kv2 are both first level random (0)
	 * 
	 * (NB this method has to be invoked after getVariantConstruct,
	 * otherwise the sorting returned is random beacuse it is
	 * established according to references stored in VariantConstructs)
	 * 
	 */
	public int compareTo(KVariant kv2) {
		
		int result = 1;
		
		/* 3 */
		if(this.isFirstLevel && !kv2.isFirstLevel())
			result = -1;
		
		/* 4 */
		else if(!this.isFirstLevel && kv2.isFirstLevel())
			result = 1;
		
		/* 5 */
		else if(this.isFirstLevel && kv2.isFirstLevel())
			result =  0;
		
		/* 3,4,5 must be tested before 1,2 to avoid NullPointerException caused by */
		/* possibile null value for carset in firstLevel kvariants */
		
		else {
			
			/* 1 */
			Set<MsmReference> refs = this.getMsmCarset().getMsmReferenceCollection();
			for(MsmReference ref: refs)
				if(ref.getTargetconstructoid().equals(kv2.getMsmConstruct())) 
					result = 1;
		
			/* 2 */
			Set<MsmReference> refs2 = kv2.getMsmCarset().getMsmReferenceCollection();
			for(MsmReference ref: refs2)
				if(ref.getTargetconstructoid().equals(kv2.getMsmConstruct())) 		
					result = -1;
		    
		}
		
		
		return result;
		
		
		
		
	}
	
	
	public String toString() {
		return "KVariant : " + this.name;
	}

}
