package org.modelgen.jpf.opgen.common;

import java.util.*;

/**
 * utilities for kvariant collections
 * @author luigi, francesca
 *
 */
public class KVariantCollections {
	
	/**
	 * returns the list of kvariant sorted topologically (reversed). It is useful
	 * to sort the list according to references dependencies. 
	 * Partial order relationship: 
	 * kv < kv2 := kv follows kv2
	 * 
	 * 1- kv < kv2 if there is a reference from kv to kv2. (+1)
	 * 2- kv > kv2 if there is a reference from kv2 to kv (-1)
	 * (references loop are not allowed in the supermodel)
	 * 3- if kv is first level and kv2 is not kv > kv2 (-1)
	 * 4- if kv2 is first level and kv is not kv2 > kv (+1)
	 * 5- if kv and kv2 are both first level random (0)
	 * @param kvList
	 * @return the list of kvariant sorted topologically (reversed)
	 */
	public List<KVariant> topologicalSorting(List<KVariant> kvList) {
		
		List<KVariant> result = new LinkedList<KVariant>();
		
		/* computes KVariants with inDegree = 0 */
		List<KVariant> nullInDegree = new LinkedList<KVariant>();
		List<KVariant> removed = new LinkedList<KVariant>();
		
		for(KVariant kv: kvList)
			if(kv.getMsmConstruct().getMsmReferenceCollection2().size()==0)
				nullInDegree.add(kv);
		
		/* nodes not toBe considered anymore */
			removed.addAll(nullInDegree);
		
		
		/* first nodes */
		result.addAll(nullInDegree);
		
		/* until there is a not removed kvariant */
		do {
			for(KVariant kv: kvList)
				/* if a kvariant has no in-edges */
				if(!removed.contains(kv) && this.computeReferenceSizes(kv, removed) == 0) {
					/* add it to the result */
					result.add(kv);
					/* and removed it from the input */
					removed.add(kv);
				}		
		} while(result.size() != kvList.size());
		
		Collections.reverse(result);
		
		return result;
		
	}
	
	/**
	 * calculates the number of in-edges for a kv, considering the removed variants.
	 * @param kv
	 * @param removed
	 * @return the number of in-edges for a kv, considering the removed variants.
	 */
	private int computeReferenceSizes(KVariant kv, List<KVariant> removed) {
		
		/* in-edges */
		Collection<MsmReference> refs = kv.getMsmConstruct().getMsmReferenceCollection2();
		
		int size = 0;
		
		/* a kvariant has a in-edge if there is a construct having a kvariant */
		/* wich refers to it wich has not been removed yet. */ 
		for(MsmReference ref: refs) {
			if(!removed.containsAll(ref.getOwnerconstructoid().generateKVariants()))
				size++;
		}
		
		return size;
	}

}
