package org.modelgen.jpf.opgen.common;
import java.util.LinkedList;
import java.util.List;

import org.modelgen.regole.CostanteDatalog;
import org.modelgen.regole.Letterale;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.regole.SkolemFunction;
import org.modelgen.regole.SkolemFunctor;
import org.modelgen.regole.VariabileDatalog;

/**
 *   
 * represents the fully instanciated variant of a MsmConstruct
 * @author luigi, francesca
 *
 */
public class VariantConstruct {
	
	private static final String COPY_FUNCTOR_POSTFIX = "OID_0";
	private static final String IS_STUB_ATTRIBUTE = "isStub";
	private static final String OID = "OID";
	private static final String SM = "SM_";
	private static final String SOURCE_1 = "[SOURCE_1]";
	private static final String SOURCE_2 = "[SOURCE_2]";
	private static final String SOURCE_3 = "[SOURCE_3]";
	private static final String DEST_1 = "[DEST_1]";
	private static final String DEST_2 = "[DEST_2]";	
	private static final String TMP_EQUAL = "TMP_EQUAL_";
	
	private String name;
	/* the kvariant to which it belongs */
	private KVariant kvariant;
	/* the instanciated references */
	private List<InstanciatedReference> instanciatedReferences;
	
	/* used by this construct comes from a first level kvariant */
	public VariantConstruct(KVariant kvariant, String name) {
		instanciatedReferences = new LinkedList<InstanciatedReference>();
		this.kvariant = kvariant;
		this.name = name;
		
	}
	
	/**
	 * builds a Variant Construct from its name, its variant and two lists: references list and
	 * VariantConstructs list that instanciate references.
	 * With the lists the constructor builds the InstanciatedReference which links references and
	 * their instanciations.
	 * @param kvariant
	 * @param name
	 * @param vcList
	 * @param refs
	 */
	public VariantConstruct(KVariant kvariant, String name,  List<MsmReference> refs, List<VariantConstruct> vcList) {
		this.kvariant = kvariant;
		this.name = name;
		instanciatedReferences = new LinkedList<InstanciatedReference>();
		/* builds the instanciated reference for each reference-variantConstruct pair */
		for(int i=0;i<refs.size();i++) 
			this.instanciatedReferences.add(new InstanciatedReference(
											refs.get(i),vcList.get(i))
											);
	}
	
	public String getName() {
		return this.name;
	}
	
	public KVariant getKVariant() {
		return this.kvariant;
	}
	
	/**
	 * generates TMP equivalence rule for this variant construct
	 * @param scheme1 source scheme 1
	 * @param scheme2 source scheme 2
	 * @param destScheme dest scheme 
	 * @param stubEquivalence if it must consider the stubness of the second constructs
	 * @param tmpSuffix suffix to add to the tmp literal name
	 * @return TMP equivalence rule for this variant construct
	 */
	public RegolaDatalog getRegolaTMP(String scheme1, String scheme2, String destScheme, boolean stubEquivalence, String tmpSuffix) {
		
		RegolaDatalog regola = new RegolaDatalog();
		String nomeCostrutto = this.kvariant.getConstructName();
		/* creates the literals for the rule */
		Letterale let1 = new Letterale(SM + nomeCostrutto, scheme1);
		Letterale let2 = new Letterale(SM + nomeCostrutto, scheme2);
		Letterale testa = new Letterale(TMP_EQUAL + this.getName() + tmpSuffix, destScheme);
		
		/* inserts literals into the rule */
		testa.setIsInTesta(true);
		regola.addElementoCorpo(let1);
		regola.addElementoCorpo(let2);
		regola.setTesta(testa);
		
		/* creates OID variables for the literals */
		VariabileDatalog varOid1 = regola.getVariabile("oid1", let1, false);
		VariabileDatalog varOid2 = regola.getVariabile("oid2", let2, false);
		let1.inserisciArgomentoEtichettato(OID, varOid1);
		let2.inserisciArgomentoEtichettato(OID, varOid2);
		
		VariabileDatalog varOid1testa = regola.getVariabile("oid1", testa, false);
		VariabileDatalog varOid2testa = regola.getVariabile("oid2", testa, false);
		testa.inserisciArgomentoEtichettato("OID1", varOid1testa);
		testa.inserisciArgomentoEtichettato("OID2", varOid2testa);
		
		
		namesInTwoLiterals(regola, let1, let2);
		
		/* eventually adds stub property */
		if(stubEquivalence) {
			VariabileDatalog isStubVar = regola.getVariabile("isStub", let1, false);
			VariabileDatalog isStubTestaVar = regola.getVariabile("isStub", testa, false);
			let2.inserisciArgomentoEtichettato(VariantConstruct.IS_STUB_ATTRIBUTE, isStubVar);
			testa.inserisciArgomentoEtichettato(VariantConstruct.IS_STUB_ATTRIBUTE, isStubTestaVar);
			
		}
		
		propertiesInTwoLiterals(regola, let1, let2);
		
		compareReferencesforTMP(regola, let1, let2,tmpSuffix,destScheme);
		
		return regola;
	}
	
	public List<InstanciatedReference> getInstanciatedReferences() {
		return this.instanciatedReferences;
	}
	
	/**
	 * inserts name attribute in two literals
	 * @param regola
	 * @param let1
	 * @param let2
	 */
	private void namesInTwoLiterals(RegolaDatalog regola, Letterale let1, Letterale let2) {
		VariabileDatalog name = regola.getVariabile("name", let1, false);
		VariabileDatalog name2 = regola.getVariabile("name", let2, false);
		
		let1.inserisciArgomentoEtichettato("Name", name);
		let2.inserisciArgomentoEtichettato("Name", name2);
	}

	/**
	 * generates the comparison TMP for the references for the TMP rule generator.
	 * tmpSuffix represents the suffix to use to recall TMP rules
	 * @param regola
	 * @param let1
	 * @param let2
	 * @param tmpSuffix
	 */
	private void compareReferencesforTMP(RegolaDatalog regola, Letterale let1, Letterale let2, String tmpSuffix, String destScheme) {
		/* the names of the references (for the attribute) */
		String referenceName = null;
		
		/* the names of the constructs instanciating the references */
		/* for the TMP literal */		 
		String referenceConstructName = null;
		
		/* for each reference generates the comparison TMP literal */
		Letterale letteraleTMP = null;
		
		VariabileDatalog varReference1;
		VariabileDatalog varReference2;
		
		VariabileDatalog varReference1TMP;
		VariabileDatalog varReference2TMP;
		
		int counter = 1; // to qualify the variable names for the references
		
		for (InstanciatedReference ir: this.instanciatedReferences) {
			referenceName = ir.getReferenceName();
			referenceConstructName = ir.getConstructName();
			letteraleTMP = new Letterale(TMP_EQUAL + referenceConstructName + tmpSuffix, destScheme);
			
			regola.addElementoCorpo(letteraleTMP);
			
			/* generates the oids for each pair of references */
			varReference1 = regola.getVariabile("oidX"+counter, let1, false);
			varReference2 = regola.getVariabile("oidY"+counter, let2, false);
			varReference1TMP = regola.getVariabile("oidX"+counter, letteraleTMP, false);
			varReference2TMP = regola.getVariabile("oidY"+counter, letteraleTMP, false);
			
			/* inserts the variables in the literals and in the TMP */
			let1.inserisciArgomentoEtichettato(referenceName, varReference1);
			let2.inserisciArgomentoEtichettato(referenceName, varReference2);
			letteraleTMP.inserisciArgomentoEtichettato("OID1", varReference1TMP);
			letteraleTMP.inserisciArgomentoEtichettato("OID2", varReference2TMP);
			
			counter++;
			
		}
	}
	
	/**
	 * inserts the properties argomentoEtichettato in the two literals 
	 * passed as parameters
	 * @param regola
	 * @param let1
	 * @param let2
	 */
	private void propertiesInTwoLiterals(RegolaDatalog regola, Letterale let1, Letterale let2) {
		/* creates the properties for the body literals */
		List<String> properties = this.kvariant.getPropertiesNames();
		
		VariabileDatalog varProperty1 = null;
		VariabileDatalog varProperty2 = null;
		
		for(String pname : properties) {
			/* the 'isStub' property is separately managed */
			if(!pname.equals(VariantConstruct.IS_STUB_ATTRIBUTE)) {
				varProperty1 = regola.getVariabile(pname, let1, false);
				varProperty2 = regola.getVariabile(pname, let2, false);
				let1.inserisciArgomentoEtichettato(pname, varProperty1);
				let2.inserisciArgomentoEtichettato(pname, varProperty2);
			}
			
		}
	}
	
	/**
	 * returns the datalog copy rule for this variant construct 
	 * @param isStub true if we need the stub copy rule
	 * @param isNegative true if we need the negative part of the difference
	 * @return the datalog copy rule for this variant construct
	 */
	public RegolaDatalog generateCopyRule(boolean isStub, boolean isNegative) {
		
		RegolaDatalog regola = new RegolaDatalog();
		
		/* creates literals */
		
		String constructName = this.kvariant.getConstructName();
		/* if the difference is negative we copy from the former to the latter and viceversa */
		Letterale let1 = new Letterale(SM + constructName, isNegative ? SOURCE_1 : SOURCE_2);
		regola.addElementoCorpo(let1);
		
		Letterale testa = new Letterale(SM + constructName, isNegative ? DEST_1 : DEST_2);
		testa.setIsInTesta(true);
		testa.setHasSkolem(true);
		regola.setTesta(testa);
		
		Letterale letTMP = new Letterale(TMP_EQUAL + this.name,DEST_1);
		letTMP.setIsNegato(!isStub); // if not stub then negative literal
		regola.addElementoCorpo(letTMP);
		
		VariabileDatalog varOid1 = regola.getVariabile("oid", let1, false);
		VariabileDatalog varOidTesta = regola.getVariabile("oid", testa, true);
		VariabileDatalog varOidTMP = regola.getVariabile("oid", letTMP, false);
		
		let1.inserisciArgomentoEtichettato(OID, varOid1);
		/* if the difference is positive we check OID2, otherwise OID1 */
		letTMP.inserisciArgomentoEtichettato(isNegative ? "OID1" : "OID2", varOidTMP);
		

		SkolemFunctor sk = getCopySkolemFunctor(constructName);
		sk.addArgomento(varOidTesta);
		
		testa.inserisciArgomentoEtichettato(OID, sk);
		
		namesInTwoLiterals(regola, let1, testa);
		
		propertiesInTwoLiterals(regola, let1, testa);
		
		/* if isStub we store it in the db */
		/* note that it is necessary to write the 'false' value */
		/* otherwise the field will contain the null value and */
		/* we cannot predict the outcomes of the merge join. */		
		CostanteDatalog costanteStub = new CostanteDatalog(String.valueOf(isStub));
		testa.inserisciArgomentoEtichettato(IS_STUB_ATTRIBUTE, costanteStub);
		
		
		
		copyReferencesForCopyRule(regola, let1, testa);
		
		/* fully instantiates the constructs */
		List<Letterale> sequence = this.getInstantiationSequence(regola, isNegative ? SOURCE_1 : SOURCE_2, "",false);
		for(Letterale let: sequence)
			regola.addElementoCorpo(let);
		
		
		
		return regola;
	}
	
	/**
	 * returns a list of literals which fully identifies this variant construct by following
	 * the parents chain. The prefix is useful to vary the variables names.
	 * SecondRecursion means that starting from the second recursion we also must join the second level
	 * literal.
	 * @param regola
	 * @param scheme
	 * @param prefix a prefix for the OID variable
	 * @param secondRecursion
	 * @return a list of literals which fully identifies this variant construct
	 */
	private List<Letterale> getInstantiationSequence(RegolaDatalog regola, String scheme, String prefix, boolean secondRecursion ) {
		
		List<Letterale> letterali = new LinkedList<Letterale>();
		VariabileDatalog vd = null;
		
		List<InstanciatedReference> refs = this.instanciatedReferences;
		int counter = 1;
		
		/* starting from the second recursion we must specify the literal */
		/* which joins the reference */
		if(secondRecursion) {
			Letterale letter = new Letterale("SM_" + this.getKVariant().getConstructName(),scheme);
			vd = regola.getVariabile("oid" + prefix, letter, false);
			letter.inserisciArgomentoEtichettato(VariantConstruct.OID, vd);
			letterali.add(letter);
			
			for(InstanciatedReference ref: refs) {
				vd = regola.getVariabile("oid" + prefix + counter, letter, false);
				letter.inserisciArgomentoEtichettato(ref.getReferenceName(), vd);
				
				counter++;
			}
		}
		
		Letterale let1 = null;
		
		List<InstanciatedReference> refs1 = null;
		counter = 1;
		int counter1 = 1;
	
		/* for each reference */
		for(InstanciatedReference ref: refs) {
			/* create a literal for that reference */
			let1 = new Letterale("SM_" + ref.getVariantConstruct().getKVariant().getConstructName(), scheme);
			letterali.add(let1);
			/* connected to the reference name */
			vd = regola.getVariabile("oid"+prefix+counter, let1, false);
			let1.inserisciArgomentoEtichettato(VariantConstruct.OID, vd);
			refs1 = ref.getVariantConstruct().getInstanciatedReferences();
			/* for each reference at this level do the same recursivly */
			/* using counter1 to differentiate the names of variables */
			for(InstanciatedReference ref1: refs1) {
				vd = regola.getVariabile("oidNewVar" + counter1, let1, false);
				let1.inserisciArgomentoEtichettato(ref1.getReferenceName(), vd);
				letterali.addAll(ref1.getVariantConstruct().getInstantiationSequence(regola,scheme,"NewVar" + String.valueOf(counter1),true));
				counter1++;
			}
			
		
			counter++;
		}
		
		return letterali;
		
		
	}
	

	/**
	 * returns the copy skolem functor for this variant construct
	 * @param constructName
	 * @return the copy skolem functor for this variant construct
	 */
	private SkolemFunctor getCopySkolemFunctor(String constructName) {
		/* extracts the copy skolem funcion */
		SkolemFunction skfn = this.getKVariant().getMsmConstruct().getSkolemFunctionByName(constructName + COPY_FUNCTOR_POSTFIX).getSkolemFunction();
		SkolemFunctor sk = new SkolemFunctor(skfn, true); // it is target
		return sk;
	}

	/**
	 * copies the references OID for the copy rule generator
	 * @param regola
	 * @param let1
	 * @param testa
	 */
	private void copyReferencesForCopyRule(RegolaDatalog regola, Letterale let1, Letterale testa) {
	
		int counter = 1; // to qualify the reference variable
		
		/* for each reference the copy rule with the not target functor */
		for(InstanciatedReference ir: this.instanciatedReferences) {			
			copySingleReference(regola, let1, testa, counter, counter, ir);
			counter++;
		}
	}
	
	/**
	 * generates attribute that copiey a reference from let1 to testa using the copy skolem functor and generating
	 * a variable with a suffix given by counter value.
	 * counter1 is the suffix for source literal variable
	 * counter2 is the suffix for head literal variable. The presence of
	 * two counters allows to specify different variable names for the source and the destination to
	 * define an eventual join with another literal.
	 * @param regola
	 * @param let1
	 * @param testa
	 * @param counter
	 * @param ir
	 */
	private void copySingleReference(RegolaDatalog regola, Letterale let1, Letterale testa, int counter1, int counter2, InstanciatedReference ir) {
	
		String constructName;
		SkolemFunctor sk;
		
		String referenceName;
		VariabileDatalog referenceVar1;
		VariabileDatalog referenceVar2;
		referenceName = ir.getReferenceName();
		referenceVar1 = regola.getVariabile("oid" + counter1, let1, false);
		referenceVar2 = regola.getVariabile("oid" + counter2, testa, true);
		
		/* gets the skolem function for the construct corresponding */
		/* to the current reference */ 
		constructName = ir.getVariantConstruct().getKVariant().getMsmConstruct().getName();
		
		sk = ir.getVariantConstruct().getCopySkolemFunctor(constructName);			
		sk.addArgomento(referenceVar2);
		sk.setIsTargetSkolem(false);
		
		let1.inserisciArgomentoEtichettato(referenceName, referenceVar1);
		testa.inserisciArgomentoEtichettato(referenceName, sk);
	}
	
	
	/**
	 * returns the rule which copies this construct from S1 to S2 in the case
	 * specified by the enum crc
	 * @param crc
	 * @return the rule which copies this construct from S1 to S2 
	 */
	public RegolaDatalog getMergeCopyRule(CopyRuleCase crc) {
		
		RegolaDatalog regola = new RegolaDatalog();
		
		String constructName = this.kvariant.getConstructName();
		
		/* creates literals */
		Letterale let1 = new Letterale("SM_" + constructName,VariantConstruct.SOURCE_1);
		Letterale testa = new Letterale("SM_" + constructName,VariantConstruct.DEST_1);
		testa.setIsInTesta(true);
		testa.setHasSkolem(true);
		/* to compare constructs from S1, S3'- */
		Letterale tmp1 = new Letterale(VariantConstruct.TMP_EQUAL + this.name + "_1",VariantConstruct.DEST_1);
		/* negative only in right cases */
		tmp1.setIsNegato(crc.equals(CopyRuleCase.NotPresentNotPresent) || crc.equals(CopyRuleCase.NotPresentStub));
		
		/* to compare literals constructs from S1, S3'+ */
		Letterale tmp2 = new Letterale(VariantConstruct.TMP_EQUAL + this.name + "_2",VariantConstruct.DEST_1);
		/* negative only in right cases */
		tmp2.setIsNegato(crc.equals(CopyRuleCase.NotPresentNotPresent) || crc.equals(CopyRuleCase.StubNotPresent));
		
		/* stores literals in the rule */
		regola.addElementoCorpo(let1);
		regola.addElementoCorpo(tmp1);
		regola.addElementoCorpo(tmp2);
		regola.setTesta(testa);
		
		VariabileDatalog oidVar1 = regola.getVariabile("oid", let1, false);
		VariabileDatalog oidVarTmp1 = regola.getVariabile("oid", tmp1, false);
		VariabileDatalog oidVarTmp2 = regola.getVariabile("oid", tmp2, false);
		
		let1.inserisciArgomentoEtichettato(VariantConstruct.OID, oidVar1);
		tmp1.inserisciArgomentoEtichettato(VariantConstruct.OID+"1", oidVarTmp1);
		tmp2.inserisciArgomentoEtichettato(VariantConstruct.OID+"1", oidVarTmp2);
		
		/* functor in the head literal */
		VariabileDatalog oidVarTesta = regola.getVariabile("oid", testa, true);
		SkolemFunctor skfn = this.getCopySkolemFunctor(constructName);
		skfn.addArgomento(oidVarTesta);
		testa.inserisciArgomentoEtichettato(VariantConstruct.OID, skfn);
		
		/* names properties and references */
		namesInTwoLiterals(regola, let1, testa);
		propertiesInTwoLiterals(regola, let1, testa);
		copyReferencesForCopyRule(regola, let1, testa);
		
		/* inserts the isStub ArgomentoEtichettato in right cases */
		if(!crc.equals(CopyRuleCase.NotPresentNotPresent)) {
			CostanteDatalog cd = new CostanteDatalog("true");
			if(crc.equals(CopyRuleCase.StubNotPresent) || crc.equals(CopyRuleCase.StubStub))
				tmp1.inserisciArgomentoEtichettato(VariantConstruct.IS_STUB_ATTRIBUTE, cd);
			if(crc.equals(CopyRuleCase.NotPresentStub) || crc.equals(CopyRuleCase.StubStub))
				tmp2.inserisciArgomentoEtichettato(VariantConstruct.IS_STUB_ATTRIBUTE, cd);
			
		}
		
		/* fully instantiates the constructs */
		List<Letterale> sequence = this.getInstantiationSequence(regola,SOURCE_1, "",false);
		for(Letterale let: sequence)
			regola.addElementoCorpo(let);
		
		
		return regola;
		
	}
	
	/**
	 * returns the list of the generated copy rule for the second copy rule of
	 * the merge operator. These rules copy from S3'+ to S2.
	 * @return the list of the generated copy rule for the second copy rule
	 */
	public List<RegolaDatalog> getCopyRules2List() {
		
		List<RegolaDatalog> regole = new LinkedList<RegolaDatalog>();
		
		/* gets possible stubness instantiations for references */
		List<List<Boolean>> ris = getReferencesStubnessInstantiations();
		
		String constructName = this.kvariant.getConstructName();
		RegolaDatalog rd = null;
		Letterale let1 = null;
		Letterale let2 = null;
		Letterale testa = null;
		Letterale tmp = null;
		InstanciatedReference ref = null;
		boolean isStub = false;
		VariabileDatalog oidVar1 = null, oidVarT = null, oidRefVar = null,
						 oidVarTmp1 =null, oidVarTmp2 = null;
		
		SkolemFunctor copyFunctor = null;
		int counter = 0;
		
		
		/* for each-possible instantiations of references plus 1 */
		/* "plus 1" allows to cycle once for constructs without references */
		/* that otherwise would be omitted. */
		
		
		List<Boolean> inst = null;
		
		for(int countTMP = 0; countTMP <= ris.size(); countTMP++) {
			
			/* to avoid NullPointerException in the "plus one" iteration */
			if(countTMP<ris.size())
				inst = ris.get(countTMP);
			
			/* create a new datalog rule */
			rd = new RegolaDatalog();
			regole.add(rd);
			
			/* create literals for this rule */
			let1 = new Letterale(VariantConstruct.SM+constructName,VariantConstruct.SOURCE_3);
			rd.addElementoCorpo(let1);
			testa = new Letterale(VariantConstruct.SM+constructName,VariantConstruct.DEST_1);
			testa.setIsInTesta(true);
			testa.setHasSkolem(true);
			rd.setTesta(testa);
			
			/* oids */
			oidVar1 = rd.getVariabile("oid", let1, false);
			oidVarT = rd.getVariabile("oid", testa, true);
			
			copyFunctor = this.getCopySkolemFunctor(constructName);
			copyFunctor.addArgomento(oidVarT);
			
			testa.inserisciArgomentoEtichettato(VariantConstruct.OID, copyFunctor);
			let1.inserisciArgomentoEtichettato(VariantConstruct.OID, oidVar1);
			
			/* names and properties */
			namesInTwoLiterals(rd, let1, testa);
			propertiesInTwoLiterals(rd, let1, testa);
			
			/* the construct in S3+ must be not stub */
			CostanteDatalog costanteStub = new CostanteDatalog("false");
			let1.inserisciArgomentoEtichettato(IS_STUB_ATTRIBUTE, costanteStub);
			
			
			int references = instanciatedReferences.size();
			
			/* creates a new literal for each reference to verify */
			/* if it is stub or not */
			/* if the reference isn't stub then a simple copy with functor */
			/* if the reference is stub then we must retrieve the original oid with the tmp rules */
			/* used to store equivalence. */
			counter = 0;
			for(int i=0;i<references;i++) {
				if(inst!=null)
					isStub = inst.get(i); // if this reference is stub
				ref = instanciatedReferences.get(i);
				let2 = new Letterale(VariantConstruct.SM+ref.getVariantConstruct().getKVariant().getConstructName(),VariantConstruct.SOURCE_3);
				rd.addElementoCorpo(let2);
				oidRefVar = rd.getVariabile("oid"+ counter, let2, false);
				let2.inserisciArgomentoEtichettato(VariantConstruct.OID, oidRefVar);
				/* let's add a different constant which expressess the fact that in this */
				/* instantiation the reference points to a stub construct or not */
				let2.inserisciArgomentoEtichettato(VariantConstruct.IS_STUB_ATTRIBUTE, new CostanteDatalog(String.valueOf(isStub)));
				
				/* if not stub just copy the reference with copy functor */
				if(!isStub)
					this.copySingleReference(rd, let1, testa, counter, counter, ref);
				else { // if stub
					tmp = new Letterale(VariantConstruct.TMP_EQUAL+ref.getVariantConstruct().getName()+"_2",VariantConstruct.DEST_1);
					rd.addElementoCorpo(tmp);
					/* variable for the tmp literal and for the reference in let1 (must be different) */
					oidVarTmp1 = rd.getVariabile("oid"+(counter+1), tmp, false); 
					oidVarTmp2 = rd.getVariabile("oid"+counter, tmp, false);
					tmp.inserisciArgomentoEtichettato(VariantConstruct.OID+1, oidVarTmp1);
					tmp.inserisciArgomentoEtichettato(VariantConstruct.OID+2, oidVarTmp2);					
					/* copy reference with joining the tmp literal */
					copySingleReference(rd, let1, testa, counter, counter+1, ref);
					counter++; 
				}
			
				counter++;
			}
			
			
		}
		
		/* fully instantiates the constructs */
		List<Letterale> sequence = this.getInstantiationSequence(rd,SOURCE_3, "",false);
		for(Letterale let: sequence)
			if(rd!=null)
				rd.addElementoCorpo(let);
		
		return regole;
		
	}

	/**
	 * returns a list of boolean list. Each internal list contains a sequence of booleans.
	 * The boolean in i position is true iff the reference in position i refers to a stub construct.
	 * The boolean in i position is false iff the reference in position i refers to a non-stub construct.
	 * @return a list of boolean list. Each internal list contains a sequence of booleans
	 */
	private List<List<Boolean>> getReferencesStubnessInstantiations() {
		
		int references = this.instanciatedReferences.size();
		
		List<List<Boolean>> argomenti = new LinkedList<List<Boolean>>();
		List<Boolean> bList = null;
		for(int i=0;i<references;i++) {
			bList = new LinkedList<Boolean>();
			bList.add(true);
			bList.add(false);
			argomenti.add(bList);
			
		}
		
		
		if(argomenti.size()==0) // not to invoke cartesian product with empty set
			return argomenti;
		else
			return (new SetOperator()).prod(argomenti);
	}
	
	
	public String toString() {
		return "VariantConstruct :" + name;
	}

}
