package org.modelgen.jpf.signatureApplication.facade;

import org.modelgen.commons.beans.scheme.Scheme;
import org.modelgen.signature.DiffContainer;
import org.modelgen.signature.MacroRuleSignature;
import org.modelgen.signature.ModelSignature;
import org.modelgen.signature.application.macroruleapplication.SignatureApplicationUtility;
import org.modelgen.signature.application.ruleapplication.ApplicateRuleSignature;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Facciata progettata per un'applicazione interattiva in cui l'utente sceglie
 * passo passo le macroregole da applicare al suo schema, per raggiungere lo
 * schema finale
 * @author Fabrizio Celli
 * @since 05-06-2007
 *
 */
public class SignatureApplicationFacade {

	//modello obiettivo e sorgente, inizializzati dal metodo viewUsefulMacro
	private ModelSignature target = null;
	private ModelSignature source = null;

	//macroregole applicate
	private List<MacroRuleSignature> scelte;

	//singleton
	private static SignatureApplicationFacade instance = null;

	//metodi Getter
	public ModelSignature getTarget() {
		return target;
	}

	public ModelSignature getSource() {
		return source;
	}
	
	public List<MacroRuleSignature> getScelte(){
		return this.scelte;
	}
	
	//costruttore reso privato
	private SignatureApplicationFacade(){
		
	}

	//singleton
	public static SignatureApplicationFacade getInstance(){
		if(instance==null)
			instance = new SignatureApplicationFacade();
		return instance;
	}

	/**
	 * Metodo che prende come input uno schema sorgente e la ModelSignature dello schema target.
	 * Restituisce una mappa, le cui chiavi sono le macroregole applicabili (o meglio, utili): ad
	 * ognuna di queste  associato un oggetto DiffContainer, che contiene le informazioni sullo
	 * schema che risulta dall'applicazione della macroregola allo schema sorgente, sui costrutti
	 * aggiunti, rimossi o rimasti invariati.
	 * @param schema sorgente
	 * @param target ModelSignature dello schema target
	 * @return mappa che ha per chiavi le MacroRuleSignature applicabili e per valori oggetti DiffContainer
	 */
	public Map<MacroRuleSignature, DiffContainer> viewUsefulMacro(Scheme schema, ModelSignature target) {
		//costruisco la ModelSignature dello schema da cui parto
		ModelSignature schemaSig = schema.getModelSignature();
		return this.viewUsefulMacro(schemaSig);
	}

	public Map<MacroRuleSignature, DiffContainer> viewUsefulMacro(ModelSignature schema, ModelSignature target) {
		//inizializzo il target
		this.target = target;
		return this.viewUsefulMacro(schema);
	}

	public Map<MacroRuleSignature, DiffContainer> viewUsefulMacro(ModelSignature schema) {
		//inizializzo il source
		this.source = schema;
		Map<MacroRuleSignature, DiffContainer> result = new HashMap<MacroRuleSignature, DiffContainer>();
		List<MacroRuleSignature> usefulMacro = SignatureApplicationUtility.getInstance().usefulMacroRules(schema);
		//per l'applicazione delle macroregole
		ApplicateRuleSignature applicator = new ApplicateRuleSignature();
		//scandisco la lista delle macroregole applicabili
		for(MacroRuleSignature macro: usefulMacro){
			//risultato dell'applicazione della macroregola
			ModelSignature schemaRes = applicator.createModelSignature(macro, schema);
			if(schemaRes!=null){
				DiffContainer container = new DiffContainer(schema, schemaRes, macro);
				result.put(macro, container);
			}
			//se tutto funziona, non  possibile arrivare qui, perch abbiamo gi selezionato
			//le sole macroregole applicabili, che dunque danno risultato non nullo
			else 
				result.put(macro, null);
		}
		return result;
	}

	/**
	 * Permette di verificare se un modello  compatibile col target.
	 * Se il modello target non  ancora stato inizializzato, ritorna false.
	 */
	public boolean isCompatibile(ModelSignature modello){
		if(modello!=null){
			return SignatureApplicationUtility.getInstance().isCompatibile(this.target, modello);
		}
		return false;
	}

	/**
	 * Permette di verificare se un modello source  compatibile col target.
	 */
	public boolean isCompatibile(ModelSignature source, ModelSignature target){
		if(source!=null && target!=null){
			return SignatureApplicationUtility.getInstance().isCompatibile(target, source);
		}
		return false;
	}

	/**
	 * Verifica se i modelli source e target memeorizzati in questo oggetto sono compatibili
	 */
	public boolean compatibili() {
		return this.isCompatibile(this.source, this.target);
	}

	/**
	 * Restituisce la macroregola, dato il nome del file
	 */
	public MacroRuleSignature getMacroRegola(String macro) {
		return SignatureApplicationUtility.getInstance().getMapMacroRegole().get(macro);
	}

	/*
	 * Aggiunge una nuova macroregola scelta
	 */
	public void addMacroScelta(MacroRuleSignature macro){
		if(this.scelte==null)
			scelte = new LinkedList<MacroRuleSignature>();
		if(macro!=null)
			scelte.add(macro);
	}

	/*
	 * Svuota la lista delle macroregole scelte 
	 */
	public void resetScelte(){
		this.scelte = null;
	}
}
