package org.modelgen.jpf.visualizer.debugger;

import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.scheme.Scheme;
import org.modelgen.facade.GeneralManagementFacade;
import org.modelgen.jpf.visualizer.commons.ToolConfigurationManager;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.AbstractModelGenBuilderFacade;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.AbstractConstruct;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.MultiDirectionAbstractStructure;
import org.modelgen.jpf.visualizer.util.Pair;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Class <code>RuleDebugger</code> defines
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class RuleDebugger {

    private List<RuleDebuggerProgressListener> listeners = new LinkedList<RuleDebuggerProgressListener>();
    private MultiDirectionAbstractStructure schemeToTranslate;
    private Map<String, String> steps;
    private DatabaseClone clone;
    private int lastOID;

    public RuleDebugger(MultiDirectionAbstractStructure schemeToTranslate, String ruleToApply, File macroRuleDir) throws IOException {
    	//System.out.println("Scheme to translate: "+schemeToTranslate);
    	//System.out.println("Rule to apply: "+ruleToApply);
	this.schemeToTranslate = schemeToTranslate;
	// split the chosen macrorule with N simple rules
	// in N macrorules with progressive simple rule
	steps = splitRule(ruleToApply, macroRuleDir);
    }

    /**
     * Avvia il debug della regola
     * @return gli schemi creati gia' caricati
     * @throws Exception errore nell'applicazione della regola
     * o errore nella clonazione del server
     */
    public Collection<MultiDirectionAbstractStructure> startDebug() throws Exception {
	Collection<MultiDirectionAbstractStructure> schemesCreated = new LinkedList<MultiDirectionAbstractStructure>();

	// calcola l'ultimo oid per stabilire 
	// quali saranno gli schemi step creati
	MultiDirectionAbstractStructure sm = schemeToTranslate.getParent();
	this.lastOID = 1;
	for (MultiDirectionAbstractStructure child : sm.getChildrens()) {
	    if (MultiDirectionAbstractStructure.isScheme(child)) { // se smScheme
		this.lastOID = Math.max(this.lastOID, child.getOID());
	    }
	}

	this.clone = null;
	try {
	    fireStepStarting("Creating a clone server");
	    
	    // Changes the working database: becomes the clone temporary database
	    clone = new DatabaseClone();
	    
	    // load scheme and model
	    Scheme workingSchemeLazy = (Scheme) schemeToTranslate.getRealSubject();
	    Model workingModelLazy = workingSchemeLazy.getMetaInfo();
	    
	    // apply rules in clone server
	    int workingOID = workingSchemeLazy.getOID();

	    for (String step : steps.keySet()) {
		String description = steps.get(step);
		fireStepStarting(description);
		GeneralManagementFacade.getInstance().applySchemeRules(new String[]{step}, description, workingOID, workingModelLazy.getOID());
	    }

	    // return debug data for viewing
	    fireStepStarting("Caching created schemes");
	    // ricarica il supermodello con le nuove info
	    sm = AbstractModelGenBuilderFacade.reload().getSuperModel();
	    
	    for (MultiDirectionAbstractStructure child : sm.getChildrens()) {
		if (MultiDirectionAbstractStructure.isScheme(child) && child.getOID() > this.lastOID) {
		    child.getAbstractConstructs(); // carica in cache (attiva proxy)
		    schemesCreated.add(child);
		}
	    }
	    
	    // analyze to obtain debug data
	    fireStepStarting("Linking constructs");
	    Collection<AbstractConstruct> sourceConstructs = schemeToTranslate.getAbstractConstructs();
	    for (MultiDirectionAbstractStructure targetScheme : schemesCreated) {
		for (AbstractConstruct targetConstruct : targetScheme.getAbstractConstructs()) {
		    Pair<Collection<AbstractConstruct>, AbstractConstruct> links = SkolemHelper.calculateLinks(targetConstruct, sourceConstructs);
		    Collection<AbstractConstruct> weakLinks = links.getFirst();
		    AbstractConstruct hardLink = links.getSecond();
		    targetConstruct.setWeakLinks(weakLinks);
		    targetConstruct.setHardLink(hardLink);
		}
	    }

	    // l'evento fine deve essere notificato
	    // per poter eventualmente essere trattato
	    fireStepStarting("Finished");
	} catch (Error e) { // cattura qualunque errore grave
	    throw new Exception(e);
	} catch (NullPointerException e) { // cattura qualunque unchecked
	    throw new Exception(e);
	} catch (RuntimeException e) { // cattura qualunque unchecked
	    throw new Exception(e);
	} catch (Exception e) {
	    throw e;
	} finally {
		/* Comunque vada, ripristina il database originale */
	    if (clone != null) {
		clone.resetOriginalDatabase();
	    }
	}
	return schemesCreated;
    }

    /**
     * Divide il file se una riga non e' vuota o non e' un commento
     * @param ruleToApply il file da dividere
     * @param macroRuleDir la cartella di lavoro
     * @return associazioni [fileDiviso, descrizione]
     * @throws IOException se errore I/O
     */
    private Map<String,String> splitRule(String ruleToApply, File macroRuleDir) throws IOException {
	// hashmap preserva l'ordine
	Map<String, String> steps = new LinkedHashMap<String,String>();
	LineNumberReader in = null;
	PrintWriter out = null;
	try {
	    String sourcePath = macroRuleDir.getAbsolutePath();
	    String targetPath = calculateTargetPath().getAbsolutePath();
	    sourcePath += File.separator + ruleToApply;
	    targetPath += File.separator + ruleToApply;
	    in = new LineNumberReader(new FileReader(sourcePath));
	    String currentOverallText = "";
	    String currentLine;
	    while((currentLine = in.readLine()) != null) {
		currentOverallText += currentLine + "\n";
		boolean validLine = ! currentLine.isEmpty() &&
		Character.isLetterOrDigit(currentLine.charAt(0));
		if (validLine) {
		    String stepPath = targetPath + "_" + in.getLineNumber();
		    File stepFilePath = new File(stepPath);
		    //stepFile.createNewFile();
		    //stepFile.deleteOnExit();
		    out = new PrintWriter(new FileWriter(stepFilePath), true);
		    out.print(currentOverallText);
		    out.flush();
		    out.close();
		    String stepFileName = stepFilePath.getName();
		    
		    // anteponi il path che costituisce il cambiamento
		    // della directory macrorule originale
		    File myMacrorulePath = this.calculateTargetPath();
		    String macrorulePathModify = ".." + File.separator + ".." + File.separator + ".." + File.separator + myMacrorulePath + File.separator;
		    //String macrorulePathModify = myMacrorulePath.toString() + File.separator;
		    stepFileName = macrorulePathModify + stepFileName;

		    // ottieni il nome della regola semplice
		    // gli ultimi 7 caratteri saltano: sono '.dtr();'
		    int endPosition = currentLine.length() - 7;
		    int slashPosition = 0;

		    for(int c = endPosition; c>=0; c--) {
			// al primo slash salva la posizione e esci
			char currChar = currentLine.charAt(c);
			if( currChar == '/' || currChar == '\\' ) {
			    slashPosition = c;
			    break;
			}
		    }
		    String description = currentLine.substring(slashPosition+1, endPosition);
		    steps.put(stepFileName, description);
		}
	    }
	} catch (IOException e) {
	    throw e; // propaga
	} finally {
	    if (in != null) in.close();
	    if (out != null) out.close();
	}
	return steps;
    }

    private File calculateTargetPath() {
    	File result = ToolConfigurationManager.sharedInstance.getRulesPath();
    	if(result == null) {
	    	result = new File(".");
	    }
	    return result;
	}

	public void addListener(RuleDebuggerProgressListener bar) {
	this.listeners.add(bar);
    }

    private void fireStepStarting(String message) {
	for (RuleDebuggerProgressListener listener : listeners) {
	    listener.stepStarting(message);
	}
    }

    public int getStepsNumber() {
	// numero di trasformazioni eseguite
	// piu' "Creating a clone server"
	// piu' "Caching created schemes"
	// piu' "Linking constructs"
	// piu' "Finished"
	return steps.keySet().size() + 4;
    }

}
