package org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters;

import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp.AbstractConstructFlyweightFactory;

import java.io.Serializable;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Class <code>BaseInfo</code> defines
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public abstract class BaseInfo<T extends BaseInfo<T>> implements Serializable, Comparable<BaseInfo<T>> {
	public static final String EMPTY_NAME = "(blankValue)";
    private static final long serialVersionUID = 1L;
    protected transient T parent = null;
    protected transient SortedSet<T> childrens = new TreeSet<T>();
    protected transient Object realSubject = null;
    private transient int oid;
    private String sid; // string id 
    protected transient String name;
    
    public abstract TypeID getTypeId();
    public abstract String getTypeString();
    
    private static BaseInfoDatabase database = new BaseInfoDatabase();
    private static boolean saveDatabaseEnabled = true;
    
    public static void setNewDatabase() {
    	BaseInfo.database = null;
    	System.gc(); // svuota memoria
	BaseInfo.database = new BaseInfoDatabase();
	BaseInfo.saveDatabaseEnabled = true;
	AbstractConstructFlyweightFactory.createNewWorkingInstance();
    }
    
    public static BaseInfoDatabase getDatabase() {
	return BaseInfo.database;
    }
    
    public static void setSaveDatabaseEnabled(boolean saveDatabaseEnabled) {
	BaseInfo.saveDatabaseEnabled = saveDatabaseEnabled;
    }
    
    public static BaseInfoDatabase getDatabaseImage() {
	try {
	    return (BaseInfoDatabase) BaseInfo.database.clone();
	} catch (CloneNotSupportedException e) {
	    e.printStackTrace();
	    return null;
	}
    }

    @SuppressWarnings("unchecked")
	public BaseInfo(T parent, Object real, int oid, String name) {
	assert ! (real instanceof BaseInfo); // altrimenti loop
	this.oid = oid;
	this.name = name;
	if( this.name != null && this.name.isEmpty() ) {
		this.name = EMPTY_NAME;
	}

	this.parent = parent;
	realSubject = real;
	this.sid = BaseInfo.createSid(this);
	if (parent != null) {
		this.parent.childrens.add((T) this);
	}
	if (BaseInfo.saveDatabaseEnabled) {
	    BaseInfo.database.add(this); // salva
	}
    }
    
    public BaseInfo(T parent, int oid, String name) {
	this(parent, null, oid, name);
    }

    /**
     * Identifica completamente il baseinfo secondo la sintassi:
     * ROOTID[TYPEID]_CHILDID[TYPEID]_CHILDCHILDID[TYPEID]_...
     * Utile anche in caso di salvataggio su disco del layout
     * @return una stringa che identifica completamente il baseinfo
     */
    private static String createSid(BaseInfo<?> info) {
	String result = "";
	if(info.parent != null) {
	    result += info.parent.sid + "_";
	}
	result += info.oid + "["+info.getTypeId().ordinal()+"]";
	return result;
    }
    
    @Override
    public boolean equals(Object obj) {
	boolean result;
	if ( obj == null || ! (obj instanceof BaseInfo) ) {
	    result = false;
	} else {
	    BaseInfo<?> otherInfo = (BaseInfo<?>) obj;
	    result = this.sid.equals(otherInfo.sid);
	}
	return result;
    }
    
    @Override
    public int hashCode() {
	return sid.hashCode();
    }

    public T getParent() {
	return parent;
    }

    public final int getOID() {
	return oid;
    }
    public final String getName() {
	return name;
    }
    
    public final String getDebugName() {
	String name = this.name;
	if (this.getTypeString() == "Scheme") {
	    name = "Schema " + this.getOID();
	} else if (this.getTypeString() == "Instance") {
	    name = "Istanza " + this.getOID();
	}
	return name;
    }

    public synchronized SortedSet<T> getChildrens() throws Exception {
	return childrens;
    }
    
    @Override
    public String toString() {
	return this.getName();
    }

    public String getDescription() {
	return this.getName() + " (" + this.getOID() + ")";
    }

    public BaseInfo<T> getRootBaseInfo() {
	return getRootBaseInfo(this); // ricorsione
    }

    private BaseInfo<T> getRootBaseInfo(BaseInfo<T> c) {
    	BaseInfo<T> parent = c.getParent();
	if(parent == null) {
	    return c;
	} else {
	    return getRootBaseInfo(parent);
	}
    }

    public String getSid() {
        return sid;
    }
    
    public Object getRealSubject() {
        return realSubject;
    }
    public int compareTo(BaseInfo<T> b) {
    	if(this==b) return 0;
    	
    	int resultCompare;
    	// ordina per tipo e per oid
    	boolean areDifferentTypes = this.getTypeId() != b.getTypeId();
    	boolean maybeAreDifferentTypes = this.sid.equals(b.sid);
    	if (areDifferentTypes || maybeAreDifferentTypes) {
    	    resultCompare = 1; // ammetti inserimenti multipli
    	} else {	    
    	    resultCompare = this.getOID() - b.getOID();
    	    if(resultCompare == 0) {
    		resultCompare = this.sid.compareTo(b.sid);
    	    }
    	}
    	return resultCompare;
    }
    
	public abstract MultiDirectionAbstractStructure getStructure();

}
