package org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters;

import org.modelgen.commons.beans.AbstractStructure;
import org.modelgen.commons.persistence.PersistenceException;
import org.modelgen.commons.persistence.PersistenceHandler;
import org.modelgen.commons.persistence.PersistenceHandlerFactory;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp.SimpleStructure;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp.StructureModelAdapter;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp.StructureSchemeAdapter;
import org.modelgen.jpf.visualizer.util.CollectionFunctions;

import java.util.SortedSet;

/**
 * Class <code>StructureAdapter</code> defines
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public abstract class MultiDirectionAbstractStructure extends BaseInfo<MultiDirectionAbstractStructure> {
    protected transient SortedSet<AbstractConstruct> abstractConstructs = null;
    private transient MultiDirectionAbstractStructure structureLoaded = null;
    private transient boolean errorLoading = false;
    private transient AbstractStructure realSubject = null;
    public static enum LoadedState { NOT_LOADED, LOADED, ERROR_LOADED }

    public MultiDirectionAbstractStructure(MultiDirectionAbstractStructure parent, AbstractStructure realSubject, int oid, String name) {
	super(parent, realSubject, oid, name);
	this.realSubject = realSubject;
    }

    public MultiDirectionAbstractStructure(MultiDirectionAbstractStructure parent, int oid, String name) {
	this(parent, null, oid, name);
    }

    protected abstract MultiDirectionAbstractStructure doLoad(PersistenceHandler handler) throws PersistenceException, ClassNotFoundException, Exception;
    protected abstract SortedSet<AbstractConstruct> doCreateAbstractConstructs(MultiDirectionAbstractStructure proxyStructure);

    /**
     * Restituisce tutti gli AbstractConstruct appartenenti al modello.
     * Il metodo si comporta come un virtual proxy: se necessario
     * carica completamente la struttura da database, memorizzandola
     * internamente.
     * 
     * @return tutti gli AbstractConstruct appartenenti al modello
     * @throws Exception 
     * @throws ClassNotFoundException 
     */
    public final SortedSet<AbstractConstruct> getAbstractConstructs() throws ClassNotFoundException, Exception {
	    this.load(); // carica se necessario
	return this.abstractConstructs;
    }

    /**
     * Se non e' gia' stato eseguito, carica completamente la struttura.
     * @throws Exception 
     * @throws ClassNotFoundException 
     */
    protected final void load() throws ClassNotFoundException, Exception {
	try {
	    if (this.structureLoaded == null) {
		if(parent != null) {
		    parent.load();
		}
		PersistenceHandler handler = PersistenceHandlerFactory.getPersistenceHandler();;
		this.structureLoaded = doLoad(handler);

		/* Forza il caricamento iniziale di tutti gli abstractConstruct */
		this.abstractConstructs = this.structureLoaded.doCreateAbstractConstructs(this);

		/* Forza il caricamento iniziale di tutti gli abstractProperty e abstractReference */
		for (AbstractConstruct c : abstractConstructs) {
		    c.createPropertiesAndReferences();
		}
		
		this.abstractConstructs = CollectionFunctions.reverse(this.abstractConstructs);

		//System.out.println("*** " +this.getName() + " (" + this.getClass().getSimpleName() + ") loaded ***");
		errorLoading = false;
	    }
	} catch (ClassCastException e) {
	    errorLoading = true;
	    throw e; // propaga
	} catch (Exception e) {
	    errorLoading = true;
	    throw e; // propaga
	}
    }

	/**
     * Risolve il problema di sincronizzare le iterazioni sugli
     * elementi della collezione: se si permettesse ad un'utente
     * di caricare i children, verrebbe sollevata una
     * ConcurrentModificationException a causa della strategia
     * troppo conservativa degli iteratori fail-fast.
     * 
     * Si sarebbe potuto utilizzare in BaseInfo un
     * ConcurrentSkipListSet per impedire eccezioni 
     * ConcurrentModificationException, ma tale collezione
     * avrebbe ordinato secondo l'ordinamento naturale imposto da
     * Comparable, che non vogliamo impostare, perche' vogliamo
     * garantire l'ordinamento di inserimento.
     * 
     * Se pero' copiare tutte le collezioni in array separati
     * crea problemi tangibili di prestazioni, considerare l'ipotesi
     * di utilizzare un ConcurrentSkipListSet
     * @throws Exception le eccezioni vengono raggruppate (continua il caricamento)
     */
    protected final void loadAllChildrens() throws Exception {
	MultiDirectionAbstractStructure[] childrensArray = new MultiDirectionAbstractStructure[this.getChildrens().size()];
	childrensArray = this.getChildrens().toArray(childrensArray);
	    String listFailed = "";
	for(int i = 0; i < childrensArray.length; i++) {
	    MultiDirectionAbstractStructure children = childrensArray[i];
	    try {
		children.load();
	    } catch (Exception childEx) {
		listFailed += "      " +children.getDescription() + "\n";
	    }
	}
	if ( !listFailed.isEmpty() ) {
	    throw new Exception("Failed loading:\n" + listFailed);
	}
    }

    public final LoadedState getLoadedState() {
	LoadedState state;
	if (this.structureLoaded != null || (this.realSubject != null && this.realSubject.isFullyLoaded())) {
	    state = LoadedState.LOADED;
	} else {
	    if (this.errorLoading) {
		state = LoadedState.ERROR_LOADED;
	    } else {
		state = LoadedState.NOT_LOADED;
	    }
	}
	return state;
    }

    public void setAbstractConstructs(
    		SortedSet<AbstractConstruct> abstractConstructs) {
	this.abstractConstructs = abstractConstructs;
    }

    public static boolean isSuperModel(MultiDirectionAbstractStructure s) {
	return s instanceof StructureModelAdapter && s.parent == null;
    }

    public static boolean isModel(MultiDirectionAbstractStructure s) {
	return s instanceof StructureModelAdapter && s.parent != null;
    }

    public static boolean isScheme(MultiDirectionAbstractStructure s) {
	return s instanceof StructureSchemeAdapter;
    }

    public static boolean isInstance(MultiDirectionAbstractStructure s) {
	return s instanceof SimpleStructure;
    }

    @Override
    public TypeID getTypeId() {
	return TypeID.AbstractStructure;
    }

	@Override
	public MultiDirectionAbstractStructure getStructure() {
		return this;
	}
    
    
    
    

}
