package org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp;

import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Property;
import org.modelgen.commons.beans.model.Reference;
import org.modelgen.commons.beans.scheme.SchemeObject;
import org.modelgen.commons.beans.scheme.SchemeObjectProperty;
import org.modelgen.commons.beans.scheme.SchemeObjectReference;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.AbstractConstruct;
import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.MultiDirectionAbstractStructure;
import org.modelgen.jpf.visualizer.util.Pair;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * La classe <code>AbstractConstructFactory</code> potrebbe essere
 * vista come un flyweight factory: si assicura che gli AbstractConstruct
 * siano condivisi in modo appropriato.
 * 
 * Quando un client richiede un <code>AbstractConstruct</code>,
 * la classe <code>AbstractConstructFactory</code> restituisce un'istanza
 * esistente, oppure, se non esiste alcuna istanza, prima la crea e poi la
 * restituisce.
 * 
 * In tutte le operazioni e' presente un OID speciale che indica il supermodello,
 * rappresentato dalla costante {@link #SUPERMODELVALUE}
 * 
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class AbstractConstructFlyweightFactory {
    /* Le mappe consentono di memorizzare l'associazione tra
     * construct originale di ModelGen e construct adattato.
     * 
     * Attenzione: l'associazione si baserebbe sul costrutto equals e hashCode
     * di Construct, Reference, Property, SchemeObject, SchemeObjectProperty,
     * SchemeObjectReference che percio' verificano il riferimento.
     * 
     * FIXME: i riferimenti sono molteplici: esistono diversi metacostrutti "Abstract"
     * per il modello "ObjectOriented" o per il modello "ER", quindi il controllo
     * sui riferimenti salta.
     * 
     * Per realizzarlo, si utilizza come chiave:
     * (StructureOID, ConstructOID)
     */
    
    private Map<ConstructKey,ConstructAdapter> constructKeyMap = new LinkedHashMap<ConstructKey,ConstructAdapter>();
    private Map<ConstructKey,ConstructReference> referenceMap = new LinkedHashMap<ConstructKey,ConstructReference>();
    private Map<ConstructKey,ConstructProperty> propertyMap = new LinkedHashMap<ConstructKey,ConstructProperty>();
    private Map<SchemeKey,SchemeObjectAdapter> schemeObjectKeyMap = new LinkedHashMap<SchemeKey,SchemeObjectAdapter>();
    private Map<SchemeKey,SchemeObjectAdapter> smSchemeObjectKeyMap = new LinkedHashMap<SchemeKey,SchemeObjectAdapter>();
    private Map<SchemeKey,SchemeReference> schemeObjectReferenceMap = new LinkedHashMap<SchemeKey,SchemeReference>();
    private Map<SchemeKey,SchemeProperty> schemeObjectPropertyMap = new LinkedHashMap<SchemeKey,SchemeProperty>();
    private Map<SchemeKey,SchemeReference> smSchemeObjectReferenceMap = new LinkedHashMap<SchemeKey,SchemeReference>();
    private Map<SchemeKey,SchemeProperty> smSchemeObjectPropertyMap = new LinkedHashMap<SchemeKey,SchemeProperty>();
    private Map<Construct, ConstructAdapter> constructMap = new LinkedHashMap<Construct, ConstructAdapter>();
    
    public static int SUPERMODELVALUE = -1;
    
    public static AbstractConstructFlyweightFactory sharedInstance = new AbstractConstructFlyweightFactory();
    
    public static void createNewWorkingInstance() {
	AbstractConstructFlyweightFactory.sharedInstance = new AbstractConstructFlyweightFactory();
    }
    
    /* Si realizza una classe interna che rappresenta la chiave.
     * L'utilizzo della classe interna salvaguarda l'information hiding
     * per l'utente di AbstractConstructFlyweightFactory  */
    private class ConstructKey extends Pair<Integer, Integer> {
		private static final long serialVersionUID = 1L;

	public ConstructKey(int structureOID, int constructOID) {
	    super(new Integer(structureOID), new Integer(constructOID));
	}
    }
    private class SchemeKey extends Pair<Integer, Pair<Integer, Integer>> {
		private static final long serialVersionUID = 1L;

	public SchemeKey(int structureOID, int schemeObjectOID, int constructOID) {
	    super(new Integer(structureOID), new Pair<Integer, Integer>(new Integer(schemeObjectOID), new Integer(constructOID)));
	}
    }
    
    public synchronized ConstructAdapter getConstructAdapter(Construct construct, MultiDirectionAbstractStructure proxy) {
	ConstructKey k;
	StructureModelAdapter proxyStructure = (StructureModelAdapter) proxy;
	if ( ! proxyStructure.isSuperModel() ) {
	    k = new ConstructKey(proxyStructure.getOID(), construct.getOID());
	} else {
	    k = new ConstructKey(SUPERMODELVALUE, construct.getOID());
	}
	
	if ( ! constructKeyMap.containsKey(k) ) {
	    ConstructAdapter aParent = null;
	    Construct parent = null;
	    if ( ! proxyStructure.isSuperModel() ) {
		parent = construct.getSuperModelConstructs().iterator().next();
		aParent = getConstructAdapter(parent, proxyStructure.getParent());
	    }
	    ConstructAdapter aConstruct = new ConstructAdapter(proxyStructure, aParent, construct);
	    constructKeyMap.put(k, aConstruct);
	    constructMap.put(construct, aConstruct);
	}
	return constructKeyMap.get(k);
    }

    public synchronized ConstructProperty getConstructProperty(Property p, ConstructAdapter construct) {
	ConstructKey k;
	if ( ! construct.isSuperModelConstruct() ) {
	    k = new ConstructKey(construct.getOID(), p.getOID());
	} else {
	    k = new ConstructKey(SUPERMODELVALUE, p.getOID());
	}
	
	if ( ! propertyMap.containsKey(k) ) {
	    ConstructProperty aParent = null;
	    if ( ! construct.isSuperModelConstruct() ) {
		Property parent = p.getMetaInfo();
		aParent = getConstructProperty(parent, (ConstructAdapter) construct.getParent());
	    }
	    propertyMap.put(k, new ConstructProperty(construct, aParent, p));
	}
	return propertyMap.get(k);
    }

    public synchronized ConstructReference getConstructReference(Reference r, ConstructAdapter owner, ConstructAdapter target) {
	ConstructKey k;
	if ( ! owner.isSuperModelConstruct() ) {
	    k = new ConstructKey(owner.getOID(), r.getOID());
	} else {
	    k = new ConstructKey(SUPERMODELVALUE, r.getOID());
	}
	
	if ( ! referenceMap.containsKey(k) ) {
	    ConstructReference aParent = null;
	    if ( ! owner.isSuperModelConstruct() ) {
		Reference parent = r.getMetaInfo();
		aParent = getConstructReference(parent, (ConstructAdapter) owner.getParent(), (ConstructAdapter) target.getParent());
	    }
	    referenceMap.put(k, new ConstructReference(aParent, r, owner, target));
	}
	return referenceMap.get(k);
    }

    public synchronized SchemeObjectAdapter getSchemeObject(SchemeObject schemeObject, MultiDirectionAbstractStructure proxy) {
	StructureSchemeAdapter proxyStructure = (StructureSchemeAdapter) proxy;
	SchemeKey k = new SchemeKey(proxyStructure.getOID(), schemeObject.getOID(), schemeObject.getMetaInfo().getOID());
	Map<SchemeKey,SchemeObjectAdapter> workingMap;
	if (proxyStructure.isSuperScheme()) {
	    workingMap = smSchemeObjectKeyMap;
	} else {
	    workingMap = schemeObjectKeyMap;
	}
	
	if ( ! workingMap.containsKey(k) ) {
	    Construct parent = schemeObject.getMetaInfo();
	    StructureModelAdapter proxyParent = (StructureModelAdapter) proxyStructure.getParent();
	    
	    ConstructAdapter aParent = getConstructAdapter(parent, proxyParent);
	    workingMap.put(k, new SchemeObjectAdapter(proxyStructure, aParent, schemeObject));
	}
	return workingMap.get(k);
    }
    
    public synchronized SchemeProperty getSchemeProperty(SchemeObjectProperty p, SchemeObjectAdapter schemeObject) {
	SchemeKey k = new SchemeKey(schemeObject.getStructure().getOID(), schemeObject.getOID(), p.getMetaInfo().getOID());
	Map<SchemeKey,SchemeProperty> workingMap;
	if (schemeObject.isSuperScheme()) {
	    workingMap = smSchemeObjectPropertyMap;
	} else {
	    workingMap = schemeObjectPropertyMap;
	}
	
	if ( ! workingMap.containsKey(k) ) {
	    Property parent = p.getMetaInfo();
	    ConstructProperty aParent = getConstructProperty(parent, (ConstructAdapter) schemeObject.getParent());
	    workingMap.put(k, new SchemeProperty(schemeObject, aParent, p));
	}
	return workingMap.get(k);
    }

    public synchronized SchemeReference getSchemeReference(SchemeObjectReference schemeReference, SchemeObjectAdapter owner, SchemeObjectAdapter value) {
	SchemeKey k = new SchemeKey(owner.getStructure().getOID(), owner.getOID(), schemeReference.getMetaInfo().getOID());
	Map<SchemeKey,SchemeReference> workingMap;
	if (owner.isSuperScheme()) {
	    workingMap = smSchemeObjectReferenceMap;
	} else {
	    workingMap = schemeObjectReferenceMap;
	}
	
	if ( ! workingMap.containsKey(k) ) {
	    Reference parent = schemeReference.getMetaInfo();
	    ConstructReference aParent = null;
	    if (value != null) {
		aParent = getConstructReference(parent, (ConstructAdapter) owner.getParent(), (ConstructAdapter) value.getParent());
	    }
	    workingMap.put(k, new SchemeReference(aParent, schemeReference, owner, value));
	}
	return workingMap.get(k);
    }
    
    public void debugConstructs() throws Exception {
	System.out.println("------------------");
	System.out.println("(Super) Model Constructs: ");
	for (ConstructKey k : constructKeyMap.keySet()) {
	    System.out.println("\t" +k + " = " +constructKeyMap.get(k).getName());
	    for (AbstractConstruct c : constructKeyMap.get(k).getChildrens()) {
		System.out.println("\t\t -> " +c.getName());
	    }
	}
	System.out.println("Super Model Scheme Objects: ");
	for (SchemeKey k : smSchemeObjectKeyMap.keySet()) {
	    System.out.println("\t" + k + " = " +smSchemeObjectKeyMap.get(k).getName());
	    for (AbstractConstruct c : smSchemeObjectKeyMap.get(k).getChildrens()) {
		System.out.println("\t\t -> " +c.getName());
	    }
	}
	System.out.println("Model Scheme Objects: ");
	for (SchemeKey k : schemeObjectKeyMap.keySet()) {
	    System.out.println("\t" + k + " = " +schemeObjectKeyMap.get(k).getName());
	    for (AbstractConstruct c : schemeObjectKeyMap.get(k).getChildrens()) {
		System.out.println("\t\t -> " +c.getName());
	    }
	}
	System.out.println("------------------");
    }



}
