package org.modelgen.jpf.visualizer.modelgenview.mvc.model;

import org.modelgen.jpf.visualizer.modelgenview.modelgen.adapters.imp.AbstractConstructFlyweightFactory;
import org.modelgen.jpf.visualizer.util.HtmlString;
import org.modelgen.jpf.visualizer.util.MathFunctions;
import org.modelgen.jpf.visualizer.util.Pair;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.swing.JComponent;

/**
 * Classe base di tutti i componenti.
 * 
 * E' l'analogo delle JComponent
 * 
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class ModelComponent implements MvcModel, Serializable, Cloneable, Comparable<ModelComponent> {
	private static final long serialVersionUID = 1L;

	private boolean firstGroupInserted = false;

	private LinkedList<ModelComponent> childrens = new LinkedList<ModelComponent>();

	private ModelComponent parent;

	protected String name;

	protected HtmlString tooltip = null; // niente tooltip all'inizio

	protected HtmlString htmlName = new HtmlString("");

	protected List<String> extraInfo = new LinkedList<String>();

	protected Serializable info;

	protected JComponent view;

	private boolean structural; // se visualizzarlo nel grafo

	private transient ShapeInfo shapeInfo = null; // info grafiche

	private Map<ModelComponent,Pair<Integer,Integer>> indexByColumnTypeFactory = new LinkedHashMap<ModelComponent, Pair<Integer,Integer>>();

	// a partire dall'istanza corrente (INDIRIZZO) ottiene in quale riga:
	// Via pescaglia (1), 00146 (1), 00178 (2), Via Luigi (2)
	private Map<ModelComponent,Pair<Integer,Integer>> cellPositions = new LinkedHashMap<ModelComponent, Pair<Integer,Integer>>();

	/*
	 * Associa il valore dell'istanza con la riga. La riga a sua volta
	 * coincide con un children della radice
	 */
	protected Map<ModelComponent, List<ModelComponent>> rowPositions = new LinkedHashMap<ModelComponent, List<ModelComponent>>();

	private Map<ModelComponent, ModelComponent> rowOwners = new LinkedHashMap<ModelComponent, ModelComponent>();

	private Vector<Vector<List<ModelComponent>>> instances;

	public ModelComponent(Serializable info) {
		this(info, info != null? info.toString() : null, true);
	}

	public ModelComponent(Serializable info, String name) {
		this(info, name, true);
	}

	public ModelComponent(Serializable info, boolean structural) {
		this(info, info.toString(), structural);
	}

	public ModelComponent(Serializable info, String name, boolean structural) {
		this.info = info;
		this.name = name;
		this.structural = structural;
		/* Inserisci il tooltip solo se ho una vera stringa */
		if (this.name != null) {
			this.htmlName = new HtmlString(name);
			if (this.name.equals("")) {
				this.tooltip = new HtmlString("empty value");
			} else {
				this.tooltip = new HtmlString(this.name);
			}
		} else {
			this.tooltip = new HtmlString("null value");
		}
	}

	protected Vector<ModelComponent> calculateColumnPositions() {
		Vector<ModelComponent> rowPositions = new Vector<ModelComponent>();

		calculateColumnPositions(rowPositions); // ricorsione

		return rowPositions;
	}

	protected void calculateColumnPositions(Vector<ModelComponent> rowPositions) {
		for (ModelComponent rootChild : this.rowPositions.keySet()) {
			/* Se la riga ancora non esiste, creala */
			int index = rowPositions.indexOf(rootChild);
			if (index == -1) {
				rowPositions.add(rootChild);
			}
		}
		for (ModelComponent child : this.childrens) {
			child.calculateColumnPositions(rowPositions); // ricorsione
		}
	}

	public Vector<Vector<List<ModelComponent>>> getInstances() {
		if(this.instances == null) {
			this.instances = calculateInstances();
		}
		return this.instances;
	}
	
	public Vector<Vector<List<ModelComponent>>> calculateInstances() {

		Vector<Vector<List<ModelComponent>>> result = new Vector<Vector<List<ModelComponent>>>();
		
		Vector<ModelComponent> columnPositions = calculateColumnPositions();
		calculateInstances(result, columnPositions); // ricorsione

		/* Inverti le colonne con le righe */
		if (!result.isEmpty()) {
			result = MathFunctions.transposeMatrix(result);
		}

		return result;
	}

	/**
	 * Calcola nella forma: name: Mario, Anna, Giacomo, ... anni: 18, 25,
	 * 30, ... successivamente si esegue la trasposta della matrice
	 * @param columnPositions 
	 *                su quale colonna inserire le istanze
	 */
	protected void calculateInstances(
			Vector<Vector<List<ModelComponent>>> result, Vector<ModelComponent> columnPositions) {
		if( this.isLeaf() ) { // caso base
			/* Crea e inizializza la riga */
			Vector<List<ModelComponent>> row = new Vector<List<ModelComponent>>(
					columnPositions.size());
			for (int i = 0; i < row.capacity(); i++) {
				row.add(null); // inizializza
			}

			/*
			 * Per tutte le istanze correnti, inserisci il valore alla riga
			 * corrispondente, in modo da mantenere la consistenza tra le varie
			 * righe
			 */
			for (ModelComponent rootChild : this.rowPositions.keySet()) {
				// ottieni il size totale
				int size = 1;
				for(Pair<Integer,Integer> cellBounds : rootChild.cellPositions.values()) {
					int cellSize = cellBounds.getFirst() + cellBounds.getSecond();
					size = Math.max(size, cellSize);
				}
				List<ModelComponent> multivalue = this.rowPositions.get(rootChild);
				multivalue = moveCellsToTheOwnerRow(multivalue, size);

				int index = columnPositions.indexOf(rootChild);
				row.set(index, multivalue);
			}

			result.add(row);
		} else { // group

			for (ModelComponent child : this.childrens) {
				child.calculateInstances(result, columnPositions); // ricorsione
			}
		}
	}

	private List<ModelComponent> moveCellsToTheOwnerRow(List<ModelComponent> multivalue, int size) {
		if(multivalue == null || multivalue.isEmpty()) return null;

		List<ModelComponent> result = new LinkedList<ModelComponent>();
		boolean isMultiValue = multivalue.size() > 1;
		
		for(int i = 0; i < size; i++) {
			result.add(new ModelComponent(""));
		}

		for(ModelComponent cell : multivalue) {
			int index = 0;
			if (isMultiValue) {
				index = calculateAbsoluteCellValue(cell);
			}
			result.set(index, cell);
		}

		return result;
	}

	/**
	 * Gli indici delle celle sono nella forma: padre: 1, figlio: 2, figliodifiglio: 2.
	 * Trasformo in assoluto: padre: 1, figlio: 3, figliodifiglio: 5 
	 */
	private int calculateAbsoluteCellValue(ModelComponent cell) {
		int index = 0;
		ModelComponent p = this.getLogicalParent(cell);
		if (p != null &&
				p.cellPositions.containsKey(cell)) {
			index = p.cellPositions.get(cell).getFirst();
			index += calculateAbsoluteCellValue(p);
		}
		return index;
	}

	/**
	 * Aggiungo un'istanza nella riga specificata
	 * 
	 * @param aChildOfRoot
	 *                l'istanza radice a cui si vuole aggiungere il valore
	 * @param value
	 *                il valore di istanza
	 */
	public void addInstance(ModelComponent aChildOfRoot, ModelComponent value, ModelComponent valueInfo) {
		// associa info ma non notificare il parent
		/*if(value != null && value.getName() != null && value.getName().startsWith("t11 v1_2a")) {
			boolean breakpoint = true;
		}*/
		// modifica il padre senza notificare
		//value.parent = valueInfo;
		
		// forza il ricalcolo del vettore delle istanze su richiesta
		this.instances = null;
		
		// salva riga di appartenenza
		this.rowOwners.put(value, valueInfo);
		
		List<ModelComponent> values;
		/* E' una riga con multivalori */
		if (this.rowPositions.containsKey(aChildOfRoot)) {
			values = this.rowPositions.get(aChildOfRoot);
		} else { // primo valore aggiunto
			values = new LinkedList<ModelComponent>();
			this.rowPositions.put(aChildOfRoot, values);
		}
		values.add(value);

		if(valueInfo == null) return;

		if (!valueInfo.cellPositions.containsKey(value)) {
			// crea un nuovo indice in base al tipo (VIA) e al valore padre
			// di riferimento (INDIRIZZO (2026))
			//System.out.println("-------\n Nuovo valore entrato: " + value + "\n-------");
			this.setNewIndex(valueInfo, value);
		}

	}

	/**
	 * Crea un nuovo indice in base al tipo (VIA) e al valore padre
	 * di riferimento (INDIRIZZO (2026))
	 */
	private void setNewIndex(ModelComponent valueInfo, ModelComponent value) {
		if(valueInfo == null) return;

		/*if(value != null && value.getName() != null && value.getName().startsWith("N/A (2481)")) {
			boolean breakpoint = true;
		}*/

		int lastIndex = 0;
		int advanceCached = 0;
		if(this.indexByColumnTypeFactory.containsKey(valueInfo)) {
			Pair<Integer,Integer> columnTypePosition = this.indexByColumnTypeFactory.get(valueInfo);
			lastIndex = columnTypePosition.getFirst();
			advanceCached = columnTypePosition.getSecond();
		}
		
		int newIndex = lastIndex + advanceCached;
		
		Pair<Integer,Integer> newColumnTypePosition = new Pair<Integer,Integer>(0, newIndex+1); 
		this.indexByColumnTypeFactory.put(valueInfo, newColumnTypePosition);
		
		Pair<Integer,Integer> newCellPosition = new Pair<Integer,Integer>(newIndex, 1); 
		valueInfo.cellPositions.put(value, newCellPosition);
		
		//System.out.println("Inizializzato " +value+" a posizione: "+newCellPosition);
		
		// aggiorna il size
		int size = newIndex + 1;
		if(this.parent != null) {
			ModelComponent valueInfoParent = getLogicalParent(valueInfo);
			ModelComponent valueParent = getLogicalParent(value);
			this.parent.notifySize(valueInfoParent, valueParent, size);
		}
	}
	
	// restituisce il padre o se non esiste la riga di appartenenza
	private ModelComponent getLogicalParent(ModelComponent m) {
		ModelComponent logicalParent = m.parent;
		if(logicalParent == null ) {
			logicalParent = this.rowOwners.get(m);
		}
		return logicalParent;
	}

	private void notifySize(ModelComponent valueInfo, ModelComponent value, int size) {
		if(valueInfo == null) return;
		
		//System.out.println("Notifica size di "+ size + " verso " + value);
		
		if(!valueInfo.cellPositions.containsKey(value)) {  // crealo
			this.setNewIndex(valueInfo, value);
			return;
		}
		
		Pair<Integer,Integer> oldCellPosition = valueInfo.cellPositions.get(value);
		int oldIndex = oldCellPosition.getFirst();
		int oldSize = oldCellPosition.getSecond();
		int indexToAdvance = Math.max(size - oldSize, 0);
		int newSize = oldSize + indexToAdvance;
		
		// notifica il padre
		if(this.indexByColumnTypeFactory.containsKey(valueInfo)) {
			Pair<Integer,Integer> columnTypePosition = this.indexByColumnTypeFactory.get(valueInfo);
			int oldTypeSize = columnTypePosition.getSecond();
			newSize += oldIndex;
			
			boolean isUpdated = newSize > oldTypeSize;
			if(isUpdated) {
				Pair<Integer,Integer> newColumnTypePosition = new Pair<Integer,Integer>(0, newSize); 
				this.indexByColumnTypeFactory.put(valueInfo, newColumnTypePosition);
			}
		}
		
		// notifica le celle
		newSize -= oldIndex;
		int newIndex = oldIndex;
		if(indexToAdvance > 0) {
			Pair<Integer,Integer> cellPosition = new Pair<Integer,Integer>(oldIndex, newSize);
			valueInfo.cellPositions.put(value, cellPosition);
			//System.out.println("Aggiornato " +value+" da "+oldCellPosition + " a " + cellPosition);
			Pair<Integer, Integer> lastPosition = updateBrothers(valueInfo, value, indexToAdvance);
			if(lastPosition != null) {
				newIndex =  lastPosition.getFirst();
				newSize = lastPosition.getSecond();
			}
		}
		
		if(this.parent != null) {
			ModelComponent valueInfoParent = getLogicalParent(valueInfo);
			ModelComponent valueParent = getLogicalParent(value);
			this.parent.notifySize(valueInfoParent, valueParent, newSize + newIndex);
		}
	}

	/**
	 * Aggiorna fratelli in modo da traslarli se necessario
	 * @return l'ultimo indice traslato e l'ultimo size di tale indice
	 */
	private Pair<Integer, Integer> updateBrothers(ModelComponent valueInfo, ModelComponent value, int indexToAdvance) {
		if(valueInfo == null || indexToAdvance <= 0) return null;
		
		Pair<Integer, Integer> lastPosition = null;

		// salva il risultato in una copia
		Map<ModelComponent,Pair<Integer,Integer>> cellPositionsTranslated = 
			new LinkedHashMap<ModelComponent, Pair<Integer,Integer>>();

		boolean startUpdate = false;
		
		for(ModelComponent child : valueInfo.cellPositions.keySet()) {
			Pair<Integer,Integer> childPosition = valueInfo.cellPositions.get(child);
			int childIndex = childPosition.getFirst();
			int childRowSpanned = childPosition.getSecond();
			if(startUpdate) {
				childIndex += indexToAdvance;
			}
			Pair<Integer, Integer> newChildPosition = new Pair<Integer,Integer>(childIndex,childRowSpanned);
			cellPositionsTranslated.put(child, newChildPosition);
			/*if(startUpdate) {
				System.out.println("Traslato " +child+" da "+childPosition+" a " + newChildPosition);
			}*/
			lastPosition = newChildPosition;
			if(child == value) { // da component in poi trasla
				startUpdate = true;
			}
		}
		valueInfo.cellPositions = cellPositionsTranslated;
		
		return lastPosition;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getTooltipText() {
		if (tooltip != null) {
			return tooltip.getHtmlText();
		} else {
			return null;
		}
	}

	public HtmlString getTooltip() {
		return tooltip;
	}

	public void setTooltip(HtmlString tooltip) {
		this.tooltip = tooltip;
	}

	public void setView(JComponent view) {
		this.view = view;

	}

	public JComponent getView() {
		return view;
	}

	/**
	 * Arrivati a questo punto, si considerano due elementi grafici uguali
	 * solo se i riferimenti BaseInfo sono identici.
	 * 
	 * @see AbstractConstructFlyweightFactory
	 */
	@Override
	public boolean equals(Object obj) {
		ModelComponent viewObj = (ModelComponent) obj;
		if(this.info == null) {
			return viewObj.info == null;
		} else {
			return this.info.equals(viewObj.info);
		}
	}

	@Override
	public int hashCode() {
		if (this.info != null) {
			return this.info.hashCode();
		} else {
			return super.hashCode();
		}
	}

	@Override
	public String toString() {
		return name;
	}

	public Serializable getInfo() {
		return info;
	}

	/**
	 * Trova il ViewNode della gerarchia che possiede il valore value
	 * 
	 * @param value
	 *                il valore a cui corriponde un determinato ViewNode
	 * @return il ViewNode trovato, null altrimenti
	 */
	public ModelComponent findOwner(ModelComponent value) {
		ModelComponent result = findOwner(this, value); // ricorsione

		return result;
	}

	private ModelComponent findOwner(ModelComponent searchBase,
			ModelComponent value) {
		ModelComponent result = null;

		if (searchBase.equals(value)) {
			result = searchBase;
		} else {
			for (ModelComponent child : searchBase.childrens) {
				result = findOwner(child, value);
				if (result != null)
					break;
			}
		}

		return result;

	}

	public void addChildren(ModelComponent child, boolean addLast) {
		if (addLast) {
			childrens.addLast(child); // ultima colonna
		} else {
			childrens.addFirst(child); // prima colonna
		}
		child.parent = this;
	}

	// inserisce alla penultima posizione
	public void addChildrenConcatenateLast(ModelComponent child) {
		childrens.add(childrens.size()-1,child); // penultima colonna
		child.parent = this;
	}

	public void addChildren(ModelComponent child) {
		this.addChildren(child, true);
	}

	public boolean isRoot() {
		return this.parent == null;
	}

	/**
	 * Calcola se e' una foglia strutturale
	 * cioe' un elemento che non ha figli visibili
	 * nella struttura
	 * @return se ha figli visibili nella struttura
	 */
	public boolean hasStructuralLeaf() {
		if (isLeaf()) {
			return true;
		} else {
			for(ModelComponent child : childrens) {
				if(child.isStructural()) return false;
			}
		}
		return true; 
	}

	public boolean isLeaf() {
		return this.childrens.isEmpty();
	}

	public List<ModelComponent> getChildrens() {
		return childrens;
	}

	/**
	 * Restituisce le foglie che si incontrano solo al
	 * primo livello di profondita', non procedendo quindi
	 * ricorsivamente a cercarne altre
	 * @returnle foglie che si incontrano solo al
	 * primo livello di profondita'
	 */
	public List<ModelComponent> getLeafsAtFirstDepth() {
		List<ModelComponent> leafs = new LinkedList<ModelComponent>();
		for(ModelComponent child : this.childrens) {
			if(child.isLeaf()) {
				leafs.add(child);
			}
		}
		return leafs;
	}

	public ModelComponent calculateRoot() {
		if (this.isRoot()) {
			return this;
		} else {
			return parent.calculateRoot();
		}
	}

	public boolean isStructural() {
		return structural;
	}

	public List<String> getExtraInfo() {
		return extraInfo;
	}

	public void setExtraInfo(List<String> extraInfo) {
		this.extraInfo = extraInfo;
	}

	public void addExtraInfo(String extraInfo) {
		this.extraInfo.add(extraInfo);
	}

	public void addExtraInfo(String... extraInfos) {
		for(String extraInfo : extraInfos) {
			this.extraInfo.add(extraInfo);
		}
	}

	public HtmlString getHtmlName() {
		return htmlName;
	}

	public void setHtmlName(HtmlString htmlName) {
		this.htmlName = htmlName;
	}

	public ModelComponent getParent() {
		return parent;
	}

	public void setStructural(boolean structural) {
		this.structural = structural;
	}

	@Override
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

	public void setParent(ModelComponent parent) {
		this.parent = parent;
	}

	public ShapeInfo getShapeInfo() {
		return shapeInfo;
	}

	public void setShapeInfo(ShapeInfo shapeInfo) {
		this.shapeInfo = shapeInfo;
	}

	public boolean isFirstGroupInserted() {
		return firstGroupInserted;
	}

	public void setFirstGroupInserted(boolean firstGroupInserted) {
		this.firstGroupInserted = firstGroupInserted;
	}

	
	public int compareTo(ModelComponent other) {
		@SuppressWarnings("unchecked")
		Comparable<Object> c = (Comparable<Object>) this.info;
		return c.compareTo(other.info);
	}



}
