package org.modelgen.jpf.visualizer.modelgenview.mvc.model;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Consente di definire un semplice grafo.
 *  
 * Si noti che e' possibile definire completamente un grafo
 * anche solo con il metodo add edge  
 * ma in tal caso deve essere un grafo che non contenga nodi isolati.
 * 
 * In caso di nodi isolati e' possibile comunque aggiungerli
 * esplicitamente tramite add vertex
 * 
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class ModelGraph implements Iterable<ModelComponent>, MvcModel {
    /* I client sono inconsapevoli della presenza della classe
     * ModelVertex. Rappresenta esclusivamente un dettaglio interno */
    private class ModelVertex {
	List<ModelComponent> outVertices = new LinkedList<ModelComponent>();
	ModelComponent model;
	public ModelVertex(ModelComponent model) {
	    this.model = model;
	}
	@Override
	public String toString() {
	    return model.toString()+"V->"+outVertices;
	}
    }

    private Map<ModelComponent, ModelVertex> components = new LinkedHashMap<ModelComponent, ModelVertex>();
    private Set<ModelComponent> vertices = new LinkedHashSet<ModelComponent>();
    private String id;
    private String name;
    
    /**
     * Creates a new <code>ModelGraph</code>
     * @param id l'identificativo per salvataggio su disco di info
     */
    public ModelGraph(String name, String id) {
	this.name = name;
	this.id = id;
    }
    
    private ModelVertex getComponent(ModelComponent model, boolean saveVertex) {
	ModelVertex vertex;
	if (! this.components.containsKey(model) ) {
	    vertex = new ModelVertex(model);
	    this.components.put(model, vertex);
	    if (saveVertex) {
		this.vertices.add(model);
	    }
	} else {
	    vertex = this.components.get(model);
	}
	return vertex;
    }
    
    private void addEdge(ModelComponent source, ModelComponent target, boolean saveSource, boolean saveTarget) {
	ModelVertex sourceVertex = this.getComponent(source, saveSource);
	sourceVertex.outVertices.add(0,target);
    }
    
    public void addConnection(ModelComponent source, ModelComponent target, ModelComponent edge) {
	assert source != null;
	//assert target != null;
	this.addEdge(source, edge, false, false);
	if (target != null) { // FIXME: necessario?
	    this.addEdge(edge, target, false, false);
	}
    }
    
    public void addConnection(ModelComponent source, ModelComponent target, Serializable edgeInfo) {
	assert !(edgeInfo instanceof ModelComponent);
	ModelComponent edge = new ModelComponent(edgeInfo);
	this.addConnection(source,target,edge);
    }
    
    public void addVertex(ModelComponent model) {
	getComponent(model, true); // crea il vertice se non esiste
    }

    public Set<ModelComponent> getVertices() {
        return vertices;
    }
    
    /**
     * Restituisce i component a cui punta. Se
     * il parametro e' un nodo, restituisce tutti gli archi
     * uscenti. Se il parametro e' un arco, restituisce
     * tutti i nodi a cui punta
     * @param source un nodo sorgente o un arco
     * @return a cosa punta
     */
    public List<ModelComponent> getOutModelComponent(ModelComponent source) {
	ModelVertex sourceV = this.getComponent(source,false);
        return sourceV.outVertices;
    }

    public Iterator<ModelComponent> iterator() {
	return this.getVertices().iterator();
    }

    public void remove(ModelComponent c) {
	/* Rimuovi tutti gli archi che puntano al nodo da eliminare */
	for(ModelVertex v : components.values()) {
	    v.outVertices.remove(c);
	}
	this.vertices.remove(c);
    }

    public String getId() {
        return id;
    }

    @Override
    public String toString() {
	return name;
    }
    
    
    
}
