package org.modelgen.jpf.visualizer.modelgenview.mvc.view.jgraph;

import org.modelgen.jpf.visualizer.modelgenview.mvc.model.ModelComponent;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jgraph.groupable.GroupableJGraph;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jgraph.groupable.JGraphDoubleClickListener;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jgraph.prototypes.DefaultGraphCellFactory;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jgraph.prototypes.GraphCellFactory;
import org.modelgen.jpf.visualizer.pac.top.views.view.ModelDoubleClickListener;
import org.modelgen.jpf.visualizer.util.Pair;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.GraphCell;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.ParentMap;

/**
 * Class <code>JGraphBuilder</code> defines
 * 
 * Attenzione: per il corretto funzionamento degli archi paralleli
 * creare prima i vertici e poi gli archi.
 * FIXME: gli archi paralleli creano comunque problema in questo caso:
 * 2 archi con lo stesso TO ma provenienti da diversi children FROM 
 *    appartenenti allo stesso padre 
 * 
 * @param <V> il tipo del valore. Utile poiche' forza
 * gli utilizzatori del grafo ad essere consistenti con il tipo
 * utilizzato
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class JGraphFacade<V extends ModelComponent> implements Serializable {
    private static final long serialVersionUID = 1L;
    protected transient GroupableJGraph<V> jGraph;
    /* La mappa serve ad associare il vertice passato dall'utente
     * con il vertice creato in jgraph. In questo modo l'utilizzatore
     * di questa classe rimane inconsapevole della classe
     * DefaultGraphCell, che rappresenta solo un dettaglio interno */
    private List<V> verticesList = new LinkedList<V>();
    private transient Map<V,DefaultGraphCell> vertices = new LinkedHashMap<V,DefaultGraphCell>();

    private transient List<ModelDoubleClickListener> clickListeners = new LinkedList<ModelDoubleClickListener>();

    private transient GraphCellFactory cellFactory;
    private transient DefaultGraphModel model;
    private transient GraphLayoutCache view;

    public JGraphFacade(GraphCellFactory cellFactory) {
	this.cellFactory = cellFactory;
	this.model = new DefaultGraphModel();
	this.view = new GraphLayoutCache(this.model, new ModelComponentViewCellFactory(), true);
	this.jGraph = new GroupableJGraph<V>(this.model, this.view);
    }

    public JGraphFacade() {
	this(new DefaultGraphCellFactory());
    }

    /**
     * Se non esiste, crea la struttura che rappresenta
     * un nodo del grafo
     * @param value il valore che rappresenta tale nodo
     * @return il nodo grafico che avvolge il 
     *         valore passato per parametro
     */
    public DefaultGraphCell getVertex(V value, boolean serialize) {
	DefaultGraphCell vertex;
	if (!this.vertices.containsKey(value)) {
	    vertex = cellFactory.createDefaultGraphCell(value);
	    vertex.getAttributes().applyMap(cellFactory.createCellAttributes(jGraph, new Point(0,0),value));
	    this.vertices.put(value, vertex);
	    
	    if (serialize) this.verticesList.add(value);
	} else {
	    vertex = this.vertices.get(value);
	}

	return vertex;
    }
    
    public DefaultGraphCell getVertex(V value) {
	return this.getVertex(value, true);
    }

    public void addVertex(V value) {
	this.addVertex(value, true);
    }

    public void addVertex(V value, boolean isLeaf) {
    	DefaultGraphCell cell;
	if ( isLeaf ) {
		cell = this.getVertex(value); // crea il vertice se non esiste
	} else { // isGroup
		cell = this.groupLast(value, null);
	}
	// Insert the Vertex (including child port and attributes)
    jGraph.getGraphLayoutCache().insert(cell);
    
    }

    public DefaultEdge addEdge(V edgeValue, V from, V to) {
	//addChildrenLast(from, edgeValue, false);
	DefaultGraphCell fromV = this.getVertex(from, false);
	DefaultGraphCell toV = this.getVertex(to);
	DefaultEdge edge = cellFactory.createDefaultEdge(edgeValue, null, null);

	// Create a Map thath holds the attributes for the edge
	edge.getAttributes().applyMap(cellFactory.createEdgeAttributes());

	/*
	 * Assumo che esista per ogni cell un'unico port posizionato
	 * all'inizio dell'array
	 */
	jGraph.getGraphLayoutCache().insertEdge(edge, fromV.getFirstChild(),
		toV.getFirstChild());
	return edge;
    }

    /**
     * Crea un gruppo che contiene le celle specificate
     * @param value il valore che avvolge il gruppo
     * @param cells i figli che il gruppo deve inglobare
     * 	se null si suppone nessun figlio: risulta utile
     *  nel caso in cui si vuole dare al nodo la forma
     *  grafica di un nodo composito, ma non inserire
     *  nulla al suo interno.
     * @return il gruppo creato
     */
    private DefaultGraphCell getGroup(V value, V[] cells) {
	DefaultGraphCell group = this.getVertex(value);
	if (cells != null) {
	    for(V cell : cells) {
		this.addChildrenLast(value, cell);
	    }
	} 

	group.getAttributes().clear();
	group.getAttributes().applyMap(
		cellFactory.createGroupAttributes(group, value));

	return group;
    }


    /**
     * Crea un gruppo con i figli specificati in ordine inverso
     * @param value il gruppo da creare
     * @param cells i figli del gruppo in ordine inverso
     */
    public DefaultGraphCell groupLast(V value, V[] cells) {
	 return this.getGroup(value, cells); // crea il gruppo se non esiste
    }
    
    public void addChildrenLast(V cell, V child) {
	this.addChildrenLast(cell, child, true);
    }

    /**
     * Aggiunge il figlio in coda
     * @param cell il padre
     * @param child il figlio da aggiungere al padre
     */
    public void addChildrenLast(V cell, V child, boolean serialize) {
	//System.out.println("addChildrenLast " +cell+", "+child);
	DefaultGraphCell cellV = this.getVertex(cell);
	DefaultGraphCell childV = this.getVertex(child, serialize);
	ParentMap parentMap;

	/* Nel caso sia un vertice semplice, trasformalo 
	 * in un vertice composite */
	if ( cellV.getChildCount() == 1 ) { // se contiene solo il port
	    //System.out.println("changing " + cellV);
	    cellV.getAttributes().clear();
	    cellV.getAttributes().applyMap(
		    cellFactory.createGroupAttributes(cellV, cell));
	} // else gia' avvenuto un inserimento

	/* Assegno a child il padre cell */
	parentMap = new ParentMap();
	parentMap.addEntry(childV, cellV);
	this.view.edit(null, null, parentMap, null);
    }

    /**
     * Restituisce il grafo composto.
     * Si assume che questo metodo venga richiamato
     * alla fine della composizione del grafo. Per questo
     * motivo ad ogni invocazione viene eseguito il
     * layout
     * @return il grafo costruito
     */
    public GroupableJGraph<V> getJGraph(String layoutFile) {
    this.jGraph.clearSelection();
	this.jGraph.saveViewsInfo(); // salva info sulle cellView
	LayoutManager.load(layoutFile, this).doLayout(this);

	/* Verifica il nodo che viene cliccato. Ad ogni click
	 * su un DefaultGraphCell, inoltra la notifica restituendo
	 * stavolta non il DefaultGraphCell, che rappresenta
	 * un dettaglio interno, ma il piu' astratto ModelComponent */
	jGraph.addDoubleClickListener(new JGraphDoubleClickListener() {
	    public void doubleClickedPerformed(DefaultGraphCell nodeClicked) {
		/* la mappa vertices e' in realta' bidirezionale */
		ModelComponent modelClicked = (ModelComponent) nodeClicked.getUserObject();
		/* Notifica tutti gli ascoltatori del doppio click */
		for (ModelDoubleClickListener clickListener : clickListeners) {
		    clickListener.doubleClickedPerformed(modelClicked);
		}
	    }
	});
	//JGraphDebug.dumpGroupStructure(jGraph.getModel());
	return jGraph;
    }

    public GroupableJGraph<V> getCreatedJGraph() {
	return this.jGraph;
    }

    /**
     * Registra gli ascoltatori al doppio click su un nodo
     * @param clickListener l'ascoltatore da registrare
     */
    public void addDoubleClickListener(ModelDoubleClickListener clickListener) {
	this.clickListeners.add(clickListener);
    }

    /**
     * Verifica l'uguaglianza tra grafi.
     * Per semplicita', si assumono uguali due grafi che hanno
     * gli stessi nodi nel modello. 
     * Gli archi non vengono considerati.
     * Il layout non viene considerato.
     */
    @Override
    public boolean equals(Object obj) {
	boolean result;
	if ( obj == null || ! (obj instanceof JGraphFacade) ) {
	    result = false;
	} else {
	    JGraphFacade<?> otherGraph = (JGraphFacade<?>) obj;
	    result = this.verticesList.equals(otherGraph.verticesList);
	}
	return result;
    }

    public Pair<GraphCell, Map<V,Point2D>> getPositionAttributeMapToSet(V v, double x, double y) {
	Map<V,Point2D> attributesToSet = new HashMap<V,Point2D>();
	DefaultGraphCell cell = this.vertices.get(v);
	assert cell != null;
	GraphConstants.setBounds(attributesToSet, new Rectangle2D.Double(x,y,0D,0D));
	return new Pair<GraphCell, Map<V,Point2D>>(cell, attributesToSet);
    }

    public Point2D getPosition(V v) {
	DefaultGraphCell cell = this.vertices.get(v);
	Rectangle2D bounds = GraphConstants.getBounds(cell.getAttributes());
	return new Point2D.Double(bounds.getMinX(), bounds.getMinY());
    }

    public Map<V,Point2D> getAllPositions() {
	Map<V,Point2D> positions = new LinkedHashMap<V,Point2D>();
	for ( V v : this.vertices.keySet() ) {
	    Point2D position = getPosition(v);
	    positions.put(v, position);
	}
	return positions;
    }

    public void saveLayout(String filename) throws IOException {
	LayoutManager.save(filename, this);
    }
    
    public void expand(V component) {
	this.jGraph.expand(component);
    }

    public Map<V, Boolean> getAllExpandedStates() {
	return this.jGraph.getAllExpandedStates();
    }

	public void click() {
		for (ModelDoubleClickListener clickListener : clickListeners) {
		    clickListener.doubleClickedPerformed(null);
		}
	}

}
