package org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable;

import org.modelgen.jpf.visualizer.modelgenview.mvc.model.ModelComponent;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.columns.groupable.ColumnGroup;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.columns.groupable.GroupableTableColumnModel;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.rows.multivalue.MultiValueTableModel;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.table.ColumnHeaderToolTips;
import org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.table.MultiValueGroupableTable;
import org.modelgen.jpf.visualizer.util.Pair;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.JTable;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

/** 
 * Fornisce un'interfaccia semplificata per la creazione
 * di tabelle multivalore multicolonna
 * 
 * La genericita' viene utilizzata per assicurare
 * la consistenza dei tipi e la chiarezza sull'utilizzo
 * 
 * @param <R> il tipo dei valori di riga
 * @param <C> il tipo dei valori di colonna
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class JTableFacade<C extends ModelComponent,R extends ModelComponent> {
    /* Alcuni typedef */
    public static class ColumnsIndexOutOfBoundsException extends Exception {
		private static final long serialVersionUID = 1L;

	public ColumnsIndexOutOfBoundsException(String message) {
	    super(message);
	}
    };

    private MultiValueTableModel<R> model;
    private Vector<R> rowToInsert;
    private DefaultTableModel nestedModel;
    /* Salva le colonne da inserire e* da quale colonna inserirle e per quante colonne */
    private Map<ColumnGroup,Pair<Integer,Integer>> columnGroupsToAdd = new LinkedHashMap<ColumnGroup,Pair<Integer,Integer>>();
    private Set<ColumnGroup> rootColumnGroups = new LinkedHashSet<ColumnGroup>();
    /* Viene utilizzato per accedere al valore
     * memorizzato dalla colonna */
    private Vector<C> simpleColumnValue = new Vector<C>();
    private Vector<C> groupColumnValue = new Vector<C>();

    public JTableFacade() {
	this.model = new MultiValueTableModel<R>();
    }

    /**
     * Aggiunge la colonna con il valore specificato
     * @param column il valore della colonna
     * @return l'indice della colonna
     */
    public int addColumn(C column) {
	model.addColumn(column);
	simpleColumnValue.add(column);
	return model.getColumnCount()-1;
    }

    public void addColumns(C... columns) {
	for(C column : columns) {
	    this.addColumn(column);
	}
    }

    /**
     * Aggiunge una colonna composita
     * @param columnGroup il valore della colonna composita da inserire
     * @param startLeafIndex il valore della colonna semplice su cui collegarsi
     * @param spanQuantity per quante colonne espandere la selezione a
     * partire da startLeafIndex incluso (un valore di 1 include
     * solo startLeafIndex)
     * @return la colonna composita generata
     */
    public ColumnGroup addColumnGroup(C columnGroup, int startLeafIndex, int spanQuantity) {
	assert spanQuantity >= 0;
	ColumnGroup group = new ColumnGroup( columnGroup.getHtmlName().getHtmlText() );
	rootColumnGroups.add(group);
	groupColumnValue.add(columnGroup);
	columnGroupsToAdd.put(group, new Pair<Integer,Integer>(startLeafIndex, spanQuantity));
	return group;
    }

    public ColumnGroup addColumnGroup(C columnGroup, int leafIndex) {
	return this.addColumnGroup(columnGroup, leafIndex, 1);
    }

    public void addColumnGroup(ColumnGroup parentGroup, ColumnGroup childGroup) {
	rootColumnGroups.remove(childGroup);
	rootColumnGroups.add(parentGroup);
	parentGroup.add(childGroup);
	/* Non e' necessario aggiungere la colonna parentGroup
	 * alla collezione groupColumnValue, poiche' per ottenere
	 * un ColumnGroup e' necessario averla creata con
	 * addColumnGroup, che memorizza la colonna */

    }

    public ColumnGroup addColumnGroup(C columnGroup, ColumnGroup childGroup) {
	ColumnGroup parentGroup = new ColumnGroup( columnGroup.getHtmlName().getHtmlText() );
	this.addColumnGroup(parentGroup, childGroup);
	return parentGroup;
    }

    /**
     * Aggiunge la riga alla tabella corrente
     * @param row la riga da aggiungere
     * @throws ColumnsIndexOutOfBoundsException se la riga sfora
     * rispetto al numero di colonne attualmente presenti
     */
    public void addRow(Vector<R> row) throws ColumnsIndexOutOfBoundsException {
	if (rowToInsert != null) {
	    model.addRow(rowToInsert);
	    rowToInsert = null;
	}
	checkBounds(row);

	/* Non inserire subito la riga: potrebbe essere una riga composita */
	rowToInsert = row; 
    }

    /**
     * Funzione di convenienza che sfrutta i varargs
     * Utile per i test
     * @param cells le celle della riga
     * @throws ColumnsIndexOutOfBoundsException se la riga sfora
     * rispetto al numero di colonne attualmente presenti
     */
    public void addRow(R... cells) throws ColumnsIndexOutOfBoundsException {
	Vector<R> row = createVectorFromVarArgs(cells);
	this.addRow(row);
    }

    /**
     * Verifica se la riga non sfora
     * @param row la riga da analizzare
     * @throws ColumnsIndexOutOfBoundsException se la riga sfora
     * rispetto al numero di colonne attualmente presenti
     */
    private void checkBounds(Vector<R> row) throws ColumnsIndexOutOfBoundsException {
	if(row.size() > model.getColumnCount()) {
	    throw new ColumnsIndexOutOfBoundsException("maximum size of row exceeded: " + row.size() + " > " + model.getColumnCount());
	}
    }

    /**
     * Aggiunge all'ultima riga creata con il metodo
     * {@link #addRow(Vector)} una riga composita
     * @param row la riga composita da aggiungere all'ultima riga 
     * (ordinaria) creata
     * @throws ColumnsIndexOutOfBoundsException se la riga sfora
     * rispetto al numero di colonne attualmente presenti
     */
    @SuppressWarnings("unchecked")
	public void addCompositeRow(Vector<R> row) throws ColumnsIndexOutOfBoundsException {
	if (rowToInsert != null) {
	    nestedModel = model.createMultiValueContainer();
	    Vector<Vector<R>> dataVector = nestedModel.getDataVector();
		model.addCompositeRow(rowToInsert, dataVector);
	    rowToInsert = null; 
	}
	checkBounds(row);
	nestedModel.addRow( row );
    }

    public void addCompositeRow(R... cells) throws ColumnsIndexOutOfBoundsException {
	Vector<R> row = createVectorFromVarArgs(cells);
	this.addCompositeRow(row);
    }

    /**
     * Crea la tabella utilizzando i valori specificati
     * nelle precedenti funzioni 
     * @return la tabella creata
     */
    public JTable createJTable() {
	/* Inserisci l'ultimo valore memorizzato */
	if (rowToInsert != null) {
	    model.addRow(rowToInsert);
	    rowToInsert = null;
	}

	MultiValueGroupableTable table = new MultiValueGroupableTable(model);
	GroupableTableColumnModel cm = (GroupableTableColumnModel)table.getColumnModel();

	/* Aggiungi le colonne composite al modello.
	 * Prima non era possibile poiche' non avevamo il 
	 * GroupableTableColumnModel */
	for (ColumnGroup columnGroup : this.columnGroupsToAdd.keySet()) {
	    Pair<Integer,Integer> childsInfo = columnGroupsToAdd.get(columnGroup);
	    int leafIndex = childsInfo.getFirst();
	    int spanQuantity = childsInfo.getSecond();
	    for(int i = 0; i < spanQuantity; i++) {
		int currentLeafIndex = leafIndex + i;
		TableColumn col = cm.getColumn(currentLeafIndex);
		columnGroup.add(col);
	    }
	}

	/* Aggiungi le colonne composite radice nel
	 * GroupableTableColumnModel */
	for (ColumnGroup columnGroup : this.rootColumnGroups) {
	    cm.addColumnGroup(columnGroup);
	}
	
	/* Assegna il tooltip ad ogni colonna semplice */
	ColumnHeaderToolTips tips = new ColumnHeaderToolTips();
	table.getTableHeader().addMouseMotionListener(tips);
	for (int c=0; c<table.getColumnCount(); c++) {
	    TableColumn col = cm.getColumn(c);
	    ModelComponent columnInfo = simpleColumnValue.get(c);
	    col.setHeaderValue(columnInfo.getHtmlName().getHtmlText());
	    tips.setToolTip(col, columnInfo.getTooltipText());
	}
	
	/* Assegna il tooltip ad ogni colonna composita 
	 * FIXME: anche se assegnato, il focus del mouse
	 * rimane sulla colonna semplice  */
	Iterator<ColumnGroup> groupsIt = table.columnGroupIterator();
	int columnIndex = 0;
	while (groupsIt.hasNext()) {
	    ColumnGroup colGroup = groupsIt.next();
	    ModelComponent columnInfo = groupColumnValue.get(columnIndex);
	    tips.setToolTip(colGroup, columnInfo.getTooltipText());
	    columnIndex++;
	}

	return table;
    }

    private Vector<R> createVectorFromVarArgs(R... values) {
	Vector<R> result = new Vector<R>();
	for (R value : values) {
	    result.add(value);
	}
	return result;
    }

    /**
     * @return un modello inizializzato con le stesse
     *    colonne del MultiValueTableModel corrente
     */
    public DefaultTableModel createMultiValueContainer() {
	return this.model.createMultiValueContainer();
    }

}
