package org.modelgen.jpf.visualizer.modelgenview.mvc.view.jtable.rows.multivalue;

import org.modelgen.jpf.visualizer.modelgenview.mvc.model.ModelComponent;
import org.modelgen.jpf.visualizer.util.Pair;
import org.modelgen.jpf.visualizer.util.ReverseIterator;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;
import java.util.concurrent.ConcurrentSkipListMap;

import javax.swing.table.DefaultTableModel;

/**
 * La <code>MultiValueTableModel</code> consente
 * di espandere e comprimere le righe che contengono
 * valori multipli.
 * 
 * @author Riccardo Pietrucci
 * @version 0.1
 */
public class MultiValueTableModel<CellType extends ModelComponent> extends DefaultTableModel {
    private static final long serialVersionUID = 1L;

    /*
     * Mappa l'indice della riga con:
     * - i valori che sono memorizzati
     * - lo stato della riga (espanso - collassato) (true - false) 
     * 
     * Viene istanziato come ConcurrentSkipListMap per evitare
     * eccezioni ConcurrentModificationException: in questo caso
     * infatti gli iteratori fail-fast di TreeMap 
     * espongono un contratto troppo conservativo
     */
    private Map<Integer,Pair<Vector<Vector<CellType>>,Boolean>> values = new ConcurrentSkipListMap<Integer,Pair<Vector<Vector<CellType>>,Boolean>>();

    /* Memorizza le tuple childrens per consentire una successiva differenziazione */
    private Set<Integer> childrens = new TreeSet<Integer>();

    public Vector<Vector<CellType>> getValues(int row) {
	return values.get(new Integer(row)).getFirst();
    }

    private boolean isExpanded(int row) {
	return values.get(new Integer(row)).getSecond();
    }

    public boolean isCompositeElement(int row) {
	assert row < super.getRowCount();
	return values.containsKey(new Integer(row));
    }

    /** 
     * Viene visto come caso particolare di elemento
     * che restituisce false in {@link #isCompositeElement(int)}
     */
    public boolean isChildren(int row) {
	return this.childrens.contains(new Integer(row));
    }

    private void setExpanded(int row, boolean expanded) {
	Integer key = new Integer(row);
	assert values.containsKey(key) : "la chiave " + row + " non esiste";
	Vector<Vector<CellType>> value = getValues(row);
	values.put(key, new Pair<Vector<Vector<CellType>>,Boolean>(value, expanded));
    }

    public void addCompositeRow(CellType[] rowData, Vector<Vector<CellType>> value) {
	super.addRow(rowData);
	addCompositeElement(value);
    }

    public void addCompositeRow(Vector<CellType> rowData, Vector<Vector<CellType>> value) {
	super.addRow(rowData);
	addCompositeElement(value);
    }


    private void addCompositeElement(Vector<Vector<CellType>> value) {
	/* Esegui il check */
	assert childrensValid(value) : "Il numero delle colonne dei figli deve essere al massimo quello del padre";
	this.values.put(new Integer(super.getRowCount() - 1), new Pair<Vector<Vector<CellType>>,Boolean>(value, false));	
    }

    /* Esegui il check delle righe dei children */
    private boolean childrensValid(Vector<Vector<CellType>> value) {
	for(Vector<CellType> row : value) {
	    if (row.size() > super.getColumnCount()) {
		return false;
	    }
	}
	return true;
    }

    public void collapseOrExpand(int row) {
	if (isCompositeElement(row)) {
	    if (isExpanded(row)) {
		collapse(row);
	    } else {
		expand(row);
	    }
	} // else non fare niente
    }

    private void expand(int row) {
	Vector<Vector<CellType>> rowValues = getValues(row);
	assert rowValues != null && isExpanded(row) == false;

	/* Salva lo stato della riga */
	setExpanded(row, true);
	
	/* Trasla gli indici che servono ad accedere ai valori delle righe */
	moveIndexes(row, + rowValues.size() );
	

	/* Aggiungi le righe */
	for(int rowNumber = 0; rowNumber < rowValues.size(); rowNumber++) {
	    Vector<CellType> aValue = rowValues.elementAt(rowNumber);
	    int rowToAdd = row + rowNumber + 1;
	    super.insertRow(rowToAdd, aValue);
	    /* Memorizza i figli per consentire una successiva differenziazione */
	    this.childrens.add(rowToAdd);
	}

    }

    private void collapse(int row) {
	Vector<Vector<CellType>> rowValues = getValues(row);
	assert rowValues != null && isExpanded(row) == true;

	/* Rimuovi le righe */
	for(int rowNumber = rowValues.size(); rowNumber > 0 ; rowNumber--) {
	    int rowToRemove = row + rowNumber;
	    super.removeRow(rowToRemove);
	    /* Cancella i figli memorizzati */
	    this.childrens.remove(rowToRemove);
	}

	/* Salva lo stato della riga */
	setExpanded(row, false);

	/* Trasla gli indici che servono ad accedere ai valori delle righe */
	moveIndexes(row,  - rowValues.size() );
    }

    /**
     * Trasla gli indici che servono ad accedere ai valori delle righe
     * @param start l'inizio (escluso)
     * @param rowQuantity di quanto traslare
     */ 
    private void moveIndexes(int start, int rowQuantity) {
	Set<Integer> keyIndexes = values.keySet();
	Iterator<Integer> it = signIterator(keyIndexes,rowQuantity);

	/* Utilizzo un'iteratore diretto o inverso per 
	 * evitare di sovrascrivere valori */
	while(it.hasNext()) {
	    int keyIndex = it.next();
	    if (keyIndex > start) { // se e' il momento di iniziare...
		Pair<Vector<Vector<CellType>>,Boolean> value = values.get(new Integer(keyIndex));
		values.remove(new Integer(keyIndex)); // elimina il vecchio valore
		int newRowIndex = keyIndex + rowQuantity; // trasla
		values.put(new Integer(newRowIndex), value); // aggiorna
		//System.out.println("Moving from " + keyIndex + " to " + newRowIndex);

		/* Trasla anche tutti i figli se sono stati espansi */
		Vector<Vector<CellType>> childs = value.getFirst();
		boolean isExpanded = value.getSecond();
		Iterator<Vector<CellType>> itChilds = signIterator(childs,rowQuantity);
		if(isExpanded) {
		    while(itChilds.hasNext()) {
			Vector<CellType> child = itChilds.next();
			int childIndex = childs.indexOf(child) + keyIndex + 1; 
			this.childrens.remove(childIndex);
			childIndex += rowQuantity; // trasla
			this.childrens.add(childIndex);
			// Moving (childIndex - rowQuantity) in (childIndex)
		    }
		}
	    }
	}
    }

    /* A seconda del segno, utilizza un iteratore inverso
     * per impedire di sovrascrivere i valori */
    private <T> Iterator<T> signIterator(Collection<T> collection, int offsetQuantity) {
	Iterator<T> it;
	if (offsetQuantity > 0) {
	    it = new ReverseIterator<T>(collection);
	} else {
	    it = collection.iterator();
	}
	return it;
    }

    @Override
    public boolean isCellEditable(int row, int column) {
	/* Le celle non sono mai modificabili */
	return false;
    }


    @Override
    public Class<?> getColumnClass(int columnIndex) {
	/* Trova la classe della prima tupla non nulla della colonna
	 * specificata */ 
	for(int i = 0; i < super.getRowCount(); i++) {
	    Object value = getValueAt(i, columnIndex);
	    if (value != null) {
		return value.getClass();
	    }
	}
	/* Ho solo valori nulli. Restituisci la classe base */
	return Object.class;
    }

    /**
     * @return un modello inizializzato con le stesse
     *    colonne del MultiValueTableModel corrente
     */
    public DefaultTableModel createMultiValueContainer() {
	DefaultTableModel model = new DefaultTableModel();
	for (int i = 0; i < super.getColumnCount(); i++) {
	    model.addColumn(super.getColumnName(i));
	}
	return model;
    }

}
