/* JFlex sintax file for MacroRule language lexer specification */

package org.modelgen.applyrule.reader;

import org.jconfig.Configuration;
import org.jconfig.ConfigurationManager;
import org.apache.log4j.Category;
import org.modelgen.applyrule.*;

import java.util.Stack;
import java.io.FileReader;

import java.io.IOException;
import java.io.FileNotFoundException;

%%

%class MacroRuleReader
%public


%standalone

%unicode
%ignorecase

/* These will be used to track down errors. */
%line
%column

%init{
	macroRule = new MacroRule();
	stack = new Stack();
%init}

/* Code to be copy/pasted into the class body */
%{
	private static Category log = Category.getInstance(MacroRuleReader.class.getName());
	private static Configuration config = ConfigurationManager.getConfiguration("modelgen");
	private static MacroRule macroRule = null;
	private static boolean finishedReading = false;
	private static Stack stack = null;
	
	public static MacroRule read(String filename) throws IOException, FileNotFoundException, Exception {
		macroRule = new MacroRule();
		MacroRuleReader scanner = null;
		FileReader reader = null;

		reader = new FileReader(filename); //Throws IOE, FNFE
		scanner = new MacroRuleReader(reader);
		while (!scanner.yy_atEOF)
			scanner.yylex(); //Throws Exception
		//Reading has completed.
		reader.close();
		
		//Check stack status.
		if (finishedReading) {
			if (macroRule != null) { //everything went good.
				macroRule.setFilename(filename);
				return macroRule;
			}
			else { //It was a bad input file.
				throw new Exception("Bad input file. Maybe a block wasn't closed properly?");
			}
		}
		else //main() has not been called.
			throw new Exception("Bad call on MacroRuleReader: reader was called without initialization.");
		
		//return macroRule;
	}
	
	/**
	 * Returns the MacroRule tree. If called before the reader has read the file, will fail.
	 * @return MacroRule tree.
	 * @throws Exception, if the method is called before main() completition.
	 */
	 
	public static MacroRule getMacroRule111() throws Exception {
		if (finishedReading) {
			if (macroRule != null) { //everything went good.
				return macroRule;
			}
			else { //It was a bad input file.
				throw new Exception("Bad input file. Maybe a block wasn't closed properly?");
			}
		}
		else //main() has not been called.
			throw new Exception("Bad call on MacroRuleReader: reader was called without initialization.");

	}
	
	private static void addSimpleRule(String filename) throws IOException {
		addSimpleRule(filename, false);
	}
	private static void addSimpleRule(String filename, boolean mustBeIterated) throws IOException {
		String rulePath = config.getProperty("rootnode", "", "paths")
								+ config.getProperty("simplerules", "", "paths");
		log.debug("going to read '" + filename + "' as '" + rulePath + filename + "'.");
		SimpleRule rule = null;
		try {
			rule = SimpleRuleReader.read(rulePath + filename); //Can throw exceptions.
		}
		catch (FileNotFoundException fnfe) { //Read failed, throw this up.
			throw new IOException(fnfe.getLocalizedMessage());
		}
		catch (Exception e) { //Read failed, throw this up.
			throw new IOException(e.getLocalizedMessage());
		}
		//Okay, everything went fine.
		log.debug("successfully read file " + filename );
		rule.setMustBeIterated(mustBeIterated);

		if (stack.empty()) { //Add it to the macro rule.
			macroRule.addSimpleRule(rule);
		}
		else { //Add it to the first block in the stack.
			BlockOfRules block = (BlockOfRules) stack.peek();
			block.addSimpleRule(rule);
		}
	}
	
	/**
	 * Instantiates a new BlockOfRules() object. If the stack is empty, it gets added to the MacroRule and the stack,
	 * otherwise it's added to the first object in the stack and then to the stack itself.
	 */
	private static void addBlockOfRules() {
		BlockOfRules block = new BlockOfRules();
		if (stack.empty()) { //This is the first block we encountered, add to the macrorule.
			macroRule.addBlock(block);
		}
		else { //Add it to the first block in the stack, without removing it.
			BlockOfRules b = (BlockOfRules) stack.peek();
			b.addBlock(block);
		} //Now put it in the stack as new first element.
		stack.push(block);
	};
	
	/**
	 * Removes a block from the stack, if present, throwing an exception otherwise.
	 */
	private static void closeBlockOfRules() throws Exception {
		if (stack.empty())
			throw new Exception("Attempted to close a not-existing block.");
		else
			stack.pop(); //Just remove it, we don't need it, and it's already stored in the result onject.
	};
	
%}

%eof{
	log.debug("Reached end of file.");
	finishedReading = true;
	if (!stack.empty()) { //The stack is not empty, but the input file is over!
		macroRule = null; //Set the result to null.
	}
%eof}

/* Some generic macro definitions */
LineTerminator = \r|\n|\r\n
InputCharacter = [^\r\n]
WhiteSpace     = {LineTerminator} | [ \t\f]

Comment = {TraditionalComment} | {EndOfLineComment} | {DocumentationComment}

TraditionalComment   = "/*" [^*] ~"*/"
EndOfLineComment     = "//" {InputCharacter}* {LineTerminator}
DocumentationComment = "/**" {CommentContent} "*"+ "/"
CommentContent       = ( [^*] | \*+ [^/*] )*

/* MacroRule sintax specific definitions */
SimpleRuleFilename = [:jletterdigit:] [:jletterdigit: ]* (.[:jletterdigit:]*)* [ ]* ";" 
StillPointIterationSimpleRuleFilename = "!!" [ ]* { SimpleRuleFilename }

BeginningOfBlock = "{"
EndOfBlock = "}"

BeginningOfExclamationMarkBlock = "!{" | "/!"
EndOfExclamationMarkBlock = "}!" | "!/"

%state DOUBLEEXCLAMATIONMARKBLOCK

%%

<YYINITIAL, DOUBLEEXCLAMATIONMARKBLOCK> {
	{ Comment }	{ /* Ignore */ }
	{ WhiteSpace }	{ /* Ignore */ }
}

<YYINITIAL> {
	{ BeginningOfBlock } {
		log.debug("beginning of a (normal) block...");
		addBlockOfRules();
	}
	
	{ EndOfBlock } {
		log.debug("end of (normal) block...");
		try {
			closeBlockOfRules();
		} catch (Exception e) {
			throw new IOException(e.getMessage());
		}
	}

	{ BeginningOfExclamationMarkBlock }	{
		log.debug("beginning of (!!) block...");
		yybegin(DOUBLEEXCLAMATIONMARKBLOCK);
	}

	{ SimpleRuleFilename }	{
		String filename = yytext().substring(0, yytext().length() - 1).trim();
		log.debug("found this name of file: " + filename);
		addSimpleRule(filename);
	}
	{ StillPointIterationSimpleRuleFilename } {
		String filename = yytext().substring(2, yytext().length() - 1).trim();
		log.debug("found this name of '!!'-modified file: " + filename);
		addSimpleRule(filename, true);
	}
}

<DOUBLEEXCLAMATIONMARKBLOCK> {
	{ EndOfExclamationMarkBlock } {
		yybegin(YYINITIAL);
		log.debug("end of (!!) block...");
	}
	{ SimpleRuleFilename } {
		String filename = yytext().substring(0, yytext().length() - 1).trim();
		log.debug("found (inside a block) this name of file: " + filename);
		addSimpleRule(filename, true);
	}
	<<EOF>> {
		throw new IOException("Unmatched StillPointIteration block at line " + (yyline + 1) + ", col " + (yycolumn + 1));
	}
}

. {
	throw new Error("Invalid sintax in file. Error at line " + (yyline + 1) + ", column " + (yycolumn + 1) + ".");
}
