/**
 * 
 */
package org.modelgen.commons.beans;

import java.util.LinkedList;
import java.util.List;

/**
 * Represents the commons info that every bean reports.
 * 
 * Every bean contains the following info:
 * <ul>
 * <li>the oid</li>
 * <li>the name</li>
 * <li>the metainfo (optional). A metainfo is a bean that contains description infos about this bean.</li>
 * <li>an oid path (automatically generated), that represents the bean's chain of metainfo. This represents the unique key for comparing two beans, used in {@link #equals(Object)} and in {@link #hashCode()} </li>
 * </ul>
 * 
 * @author Riccardo Pietrucci
 *
 * @param <T> the specific type of metaInfo. If this class doesn't contain any metaInfo, you can use {@link org.modelgen.commons.beans.NullBean } 
 */
public abstract class AbstractBean<T extends AbstractBean<?>> implements Comparable<AbstractBean<T>> {
	/** The special constant representing an unspecified value */
	public static final String MDM_UNSPECIFIED = "(MDM_UNSPECIFIED)";
	protected int oid; //0: not configured, > 0: exists in the DB (it's read from there). <0: reserved. :)
	protected String name;
	private T metaInfo;
	private List<Integer> oidPath;

	/**
	 * Creates a bean with the oid and the name specified.
	 * 
	 * This bean it's the base metainfo for the belonging category
	 * 
	 * @param oid the bean's oid
	 * @param name the bean's name
	 */
	public AbstractBean(int oid, String name) {
		this(oid, name, null);
	}

	/**
	 * Creates a bean with the oid, the name and the
	 * metainfo specified.
	 * 
	 * This bean it's a possible specialization of the metainfo specified.
	 * 
	 * @param oid the bean's oid
	 * @param name the bean's name
	 * @param metaInfo the bean's metainfo
	 */
	public AbstractBean(int oid, String name, T metaInfo) {
		this.setOID(oid);
		this.setName(name);
		this.setMetaInfo(metaInfo);
	}

	/**
	 * Sets the bean that contains description infos about this bean
	 * @param metaInfo the bean that contains description infos about this bean
	 */
	public void setMetaInfo(T metaInfo) {
		this.metaInfo = metaInfo;
		this.buildOidPath();
	}
	
	private void buildOidPath() {
		this.oidPath = new LinkedList<Integer>();
		this.oidPath.add(this.getOID());
		AbstractBean tmp = this.metaInfo;
		if(tmp == null) return;
		while( ! tmp.isRoot() ) {
			this.oidPath.add(tmp.getOID());
			tmp = tmp.getMetaInfo();
		}
	}

	/**
	 * @return the bean's name
	 */
	public final String getName() {
		return name;
	}
	
	/**
	 * @param name the bean's name
	 */
	public final void setName(String name) {
		if (name!=null) {
			this.name = name; 
		} else {
			this.name = MDM_UNSPECIFIED;
		}
	}
	
	/**
	 * @return the bean's oid
	 */
	public final int getOID() {
		return this.oid;
	}
	
	/**
	 * @param oid the bean's oid
	 */
	public final void setOID(int oid) {
		this.oid = oid;
		this.buildOidPath();
	}
	
	/**
	 * @return if it doesn't contain any metainfo
	 */
	public boolean isRoot() {
		return this.getMetaInfo() == null;
	}

	/**
	 * Returns the bean that contains description infos about this bean
	 * @return the bean that contains description infos about this bean
	 */
	public final T getMetaInfo() {
		return this.metaInfo;
	}

	/**
	 * @return the oid's name
	 */
	@Override
	public String toString() {
		return this.name;
	}

	/**
	 * @param other the bean to compare
	 * @return the compare value
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public final int compareTo(AbstractBean<T> other) {
		// TODO Auto-generated method stub
		return 1;
	}
	
	/**
	 * Returns if two beans have the same oid path, where
	 * the <i>oid path</i> is the bean's chain of metainfo.
	 * 
	 * Two beans are equals if contain the same oid
	 * and if their oid path contains the same beans.
	 */
	@Override
	public final boolean equals(Object obj) {
		if(this == obj) {
			return true;
		} 
		
		if( obj instanceof AbstractBean ) {
			AbstractBean bean = (AbstractBean) obj;
			return this.oidPath.equals(bean.oidPath) && this.name.equals(bean.name); // prima senza name 
		}
		
		return false;
	}

	/**
	 * Returns the hash code of the oid path.
	 */
	@Override
	public final int hashCode() {
		return this.oidPath.hashCode() + this.name.hashCode(); // prima non c'era il name 
	}

}
