package org.modelgen.commons.beans;

import java.util.LinkedList;

import org.apache.log4j.Logger;

/**
 * Represents a generic structure.
 * 
 * @author Riccardo Pietrucci
 *
 * @param <S> the specific metastructure's type. If this class doesn't contain any metastructure, you can use {@link org.modelgen.commons.beans.NullBean } 
 * @param <C> the specific abstract constructs bean's type
 */
public abstract class AbstractStructure<S extends AbstractStructure<?,?>, C extends AbstractBean<?>> extends AbstractBean<S> {
	private static final Logger log = Logger.getLogger(AbstractStructure.class);
	
	/**
	 * Creates a structure with the oid and the name specified.
	 * 
	 * This structure it's the supermodel.
	 * 
	 * @param oid the structure's oid
	 * @param name the structure's name
	 */
	public AbstractStructure(int oid, String name) {
		super(oid, name);
	}
	
	/**
	 * Creates a structure with the oid, the name and the
	 * metastructure specified.
	 * 
	 * This structure it's a possible specialization of the metastructure specified.
	 * 
	 * @param oid the structure's oid
	 * @param name the structure's name
	 * @param metastructure the structure's metainfo
	 */
	public AbstractStructure(int oid, String name, S metastructure) {
		super(oid, name, metastructure);
	}

	/**
	 * Returns the abstract constructs of this structure.
	 * 
	 * An <i>abstract construct</i> is a generic concept of this structure.
	 * If this structure is a (super)model, we have a (meta)construct.
	 * If this structure is a scheme(instance), we have a scheme(instance) object.
	 * 
	 * If not yet invoked, this method causes the full load of this structure.
	 * 
	 * @return the abstract constructs of this structure
	 */
	public LinkedList<C> getAbstractConstructs() {
		LinkedList<C> abstractConstructs = this.doGetAbstractConstructs();
		if(abstractConstructs == null) {
			try {
				this.doLoadAbstractConstructs();
				log.info("Structure " + this + " loaded!");
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			/* reload: now it must work */
			abstractConstructs = this.doGetAbstractConstructs();
		}
		return abstractConstructs;
	}
	
	/**
	 * @return if this structure is fully loaded or only 
	 */
	public boolean isFullyLoaded() {
		return this.doGetAbstractConstructs() != null;
	}
	
	
	/**
	 * This method works as an adapter to obtain the abstract construct
	 * of the working (concrete) subclass. 
	 * @return the abstract constructs of the working (concrete) subclass
	 */
	protected abstract LinkedList<C> doGetAbstractConstructs();
	
	/**
	 * It loads completely the structure by getting infos from the db and
	 * creating the corresponding abstract constructs.
	 * 
	 * The class implementing this method has to update
	 * the corresponding abstract constructs cached. 
	 * 
	 * @throws ClassNotFoundException
	 * @throws Exception
	 */
	protected abstract void doLoadAbstractConstructs() throws ClassNotFoundException, Exception;
	
}
