/*
 * Created on Jan 25, 2004
 *
 */
package org.modelgen.commons.beans.model;

import org.modelgen.commons.beans.AbstractBean;
import org.modelgen.commons.utils.Validator;

import java.util.Collection;
import java.util.LinkedList;

/**
 * The Construct is a generic concept of a {@link Model}, composed of
 * various {@link Property} and {@link Reference}.<p>
 * 
 * If the construct hasn't a referring construct containing its metainfos we have
 * a <i>MetaConstruct</i> belonging to the SuperModel.<p>
 * 
 * Therefore setting <t>null</t> to {@link #setMetaInfo(Construct)} brings
 * this construct to the MetaConstruct. Vice versa, setting a non null value 
 * to {@link #setMetaInfo(Construct)} sets the referring MetaConstruct to this construct.
 * 
 * @author Luca Santarelli [lucachan@tin.it - hrk@users.sourceforge.net]
 * Jan 25, 2004
 */
public class Construct extends AbstractBean<Construct> {
	private boolean isLexical; // = true;
	private Collection<Property> properties = new LinkedList<Property>(); //The beans are two-way navigable.
	private Collection<Reference> references = new LinkedList<Reference>();
	private Collection<Construct> dependencies = new LinkedList<Construct>();
	
	private Collection<Construct> superModelConstructs = new LinkedList<Construct>();
	
	//signature
	private String signature;
	
	//vincoli da mostrare all'utente
	private String userBonds;

	/**
	 * Creates the MetaConstruct with the oid and the name specified.
	 * 
	 * @param oid the metaconstruct's oid
	 * @param name the metaconstruct's name
	 */
	public Construct(int oid, String name) {
		super(oid, name);
		this.signature = "true";
		this.userBonds = "true";
	}
	
	/**
	 * Creates a Construct with the oid, the name and the
	 * MetaConstruct specified.
	 * 
	 * @param oid the construct's oid
	 * @param name the construct's name
	 * @param metaConstruct the construct's metaConstruct
	 */
	public Construct(int oid, String name, Construct metaConstruct) {
		super(oid, name, metaConstruct);
		this.signature = "true";
		this.userBonds = "true";
	}
	
	/**
	 * Add the metaConstruct to the list {@link #getSuperModelConstructs()}
	 */
	@Override
	public void setMetaInfo(Construct metaConstruct) {
		super.setMetaInfo(metaConstruct);
		if( metaConstruct != null && ! this.superModelConstructs.contains(metaConstruct) ) {
			this.superModelConstructs.add(metaConstruct);
		}
	}
	
	/**
	 * The properties are the SM ones
	 * @return the signature as a String;
	 */
	public String getSignature() {
		return this.signature;
	}

	public void setSignature(String signature) {
		this.signature = signature;
	}

	/**
	 * The properties are the M ones
	 * @return the bonds create by user as String
	 */
	public String getUserBonds() {
		return this.userBonds;
	}

	public void setUserBonds(String userSignature) {
		this.userBonds = userSignature;
	}

	/**
	 * @return the list of properties
	 */
	public Collection<Property> getProperties() {
		return properties;
	}

	/**
	 * @return the list of references
	 */
	public Collection<Reference> getReferences() {
		return references;
	}

	/**
	 * Add a dependency to this construct
	 * @param construct a construct required to effectly use this construct 
	 */
	public void addDependency(Construct construct) {
		dependencies.add(construct);
	}
	
	/**
	 * @return all constructs required to effectly use this construct 
	 */
	public Collection<Construct> getDependencies() {
		return dependencies;
	}

	/**
	 * Adds the given Property to the collection of Properties.
	 * @param p the Property to add
	 */
	public void addProperty(Property p) {
		properties.add(p); // TODO imposta l'owner in p
	}

	/**
	 * Adds the given Reference to the collection of References.
	 * @param r the Reference to add
	 */
	public void addReference(Reference r) {
		references.add(r);
	}

	/**
	 * Returns true if the given name is valid for a Construct.<p>
	 * The check is formal only (i.e.: not null, not empty, ...)
	 * @param constructName the name to check
	 * @return if the given name is valid
	 */
	public static boolean isValidName(String constructName) {
		String[] reservedChars = new String[] { "_" , "-" }; //Add other reserved characters here.
		return Validator.isValidString(constructName, reservedChars);
	}

	/**
	 * @return the MetaConstructs required to build this construct 
	 */
	public Collection<Construct> getSuperModelConstructs() {
		return superModelConstructs;
	}

	/**
	 * @param superModelConstructs the MetaConstructs required to build this construct 
	 */
	public void setSuperModelConstructs(Collection<Construct> superModelConstructs) {
		this.superModelConstructs = superModelConstructs;
		if( ! this.superModelConstructs.isEmpty() ) {
			this.setMetaInfo(superModelConstructs.iterator().next());
		}
	}

	/**
	 * Finds and returns the Property with the specified oid
	 * @param propertyOID the property's oid
	 * @return the property found, <code>null</code> otherwise
	 */
	public Property getProperty(int propertyOID) {
		return this.getProperty(propertyOID, null);
	}
	
	/**
	 * Finds and returns the Property with the specified name
	 * @param propertyName the property's name
	 * @return the property found, <code>null</code> otherwise
	 */
	public Property getProperty(String propertyName) {
		return this.getProperty(-1, propertyName);
	}
	
	private Property getProperty(int propertyOID, String propertyName) {
		for (Property p : properties) {
			if (p.getOID() == propertyOID || p.getName().equalsIgnoreCase(propertyName))
				return p;
		}
		return null;
	}

	/**
	 * Returns the Reference object which has the given referenceOID.
	 * - @see #getReference(int, String)
	 * @param referenceOID the reference's oid
	 * @return the Reference object if found, <code>null</code> otherwise.
	 */	
	public Reference getReference(int referenceOID) {
		return this.getReference(referenceOID, null);
	}
	
	/**
	 * Returns the Reference object which has the given referenceName.
	 * - @see #getReference(int, String)
	 * @param referenceName the reference's name
	 * @return the Reference object if found, <code>null</code> otherwise. 
	 */
	public Reference getReference(String referenceName) {
		return this.getReference(-1, referenceName);
	}
	
	/**
	 * Returns the Reference object which has the given referenceOID or the given referenceName.
	 * @param referenceOID
	 * @param referenceName
	 * @return the Reference object if found, null otherwise.
	 */
	private Reference getReference(int referenceOID, String referenceName) {
		for (Reference r : references) {
			if (r.getOID() == referenceOID || r.getName().equalsIgnoreCase(referenceName))
				return r;
		}
		return null;
	}

	/**
	 * Removes a construct's property
	 * @param property the property to remove
	 */
	public void removeProperty(Property property) {
		for (Property p : properties) {
			if (p.getName().equalsIgnoreCase(property.getName())) {
				properties.remove(property);
				break; //No need to search further.
			}
		}
	}
	/**
	 * @return if it's lexical (if it can contains a real value data)
	 */
	public boolean isLexical() {
		return isLexical;
	}
	/**
	 * @param isLex if it's lexical (if it can contains a real value data)
	 */
	public void setLexical(boolean isLex) {
		this.isLexical = isLex;
	}

	/**
	 * Find if this construct has the specified reference
	 * @param referenceName the reference's name to find
	 * @return if this construct has the specified reference
	 */
	public boolean hasReference(String referenceName) {
		for (Reference r : references) {
			if (r.getName().equalsIgnoreCase(referenceName)) 
				return true;
		}
		return false;
	}

	/**
	 * Find if this construct has the specified property (case insensitive)
	 * @param propertyName the property's name to find (case insensitive)
	 * @return if this construct has the specified reference
	 */
	public boolean hasProperty(String propertyName) {
		propertyName = propertyName.toLowerCase(); // modifica LF
		for (Property p : properties) {
			if (p.getName().equalsIgnoreCase(propertyName)) 
				return true; // modifica LF
		}
		return false;
	}
	
	/**
	 * @return the construct's name and the string rappresentation of
	 * all its properties and references
	 */
	public String toString() {
		StringBuilder constructPrint = new StringBuilder(this.getName());
		constructPrint.append("(\n");
		for (Property prop : properties) {
				constructPrint.append(prop + "\n");
		}
		for (Reference ref : references){
				constructPrint.append(ref + "\n");
		}
		constructPrint.append(")\n");
		return constructPrint.toString();
	}
}
