/*
 * Created on Jan 25, 2004
 *
 */
package org.modelgen.commons.beans.model;

import org.modelgen.commons.beans.AbstractStructure;
import org.modelgen.commons.persistence.PersistenceHandler;
import org.modelgen.commons.persistence.PersistenceHandlerFactory;
import org.modelgen.commons.utils.Validator;
import org.modelgen.signature.ConstructSignature;
import org.modelgen.signature.ModelSignature;

import java.util.Collection;
import java.util.LinkedList;

/**
 * Basic class representing a model.
 * A model is made out of a name and a collection of (scheme) objects.<p>
 * 
 * If the model hasn't a referring model containing its metainfos we have
 * the <i>SuperModel</i>.
 * 
 * Therefore setting <t>null</t> to {@link #setMetaInfo(Model)} brings
 * this model to the SuperModel. Vice versa, setting a non null value 
 * to {@link #setMetaInfo(Model)} sets the referring SuperModel to this model.
 * 
 * @author Luca Santarelli [lucachan@tin.it - hrk@users.sourceforge.net]
 * Feb 5, 2004
 */
public class Model extends AbstractStructure<Model,Construct> {
	private String shortName; //This one is used to build the tables.
	private String renamedName; // the name for the renaming 
	private String renamedShortName;

	private LinkedList<Construct> constructs = null;
	
	//informazioni sulla famiglia del modello
	//anche se il modello � il capofamiglia, l'OID della famiglia � dato dalla
	//variabile "familyOid", mentre quello del modello � il semplice OID.
	private boolean isAncestor = false;
	private int familyOid = 0;
	
	private Model family = null;

	/**
	 * Creates the SuperModel with the oid and the name specified.
	 * 
	 * @param oid the supermodel's oid
	 * @param name the supermodel's name
	 */
	public Model(int oid, String name) {
		super(oid, name);
		renamedName = null;
		renamedShortName = null;
	}
	
	/**
	 * setta un nuovo nome per la rinominazione
	 * @param newName
	 */
	public void setNewNames(String newName, String shortNewName) {
		
		/* se era stato già rinominato, non deve perdere memoria del nome originale */
		if(!isRenamed()) {
			this.renamedName = this.name;
			this.renamedShortName = this.shortName;
		}
		
		this.name = newName;
		this.shortName = shortNewName;
	}
	
	public void clearOldNamesOnSave(){
		this.renamedName = null;
		this.renamedShortName = null;
	}
	/**
	 * torna il nome di prima della ridenominazione
	 */
	public String getOldName() {
		return this.renamedName;
	}
	
	public String getOldShortName() {
		return this.renamedShortName;
	}
	
	/**
	 * se è stato rinominato
	 */
	public boolean isRenamed() {
		return renamedName != null;
	}

	/**
	 * Creates a Model with the oid, the name and the
	 * supermodel specified.
	 * 
	 * @param oid the model's oid
	 * @param name the model's name
	 * @param metastructure the model's metainfo
	 */
	public Model(int oid, String name, Model metastructure) {
		super(oid, name, metastructure);
	}

	/**
	 * Calculates the model signature
	 * @return the model signature calculated
	 */
	public ModelSignature getModelSignature(){		
		ModelSignature signature = null;	
		//if(constructs!=null){
			signature = new ModelSignature(this.name);
			//for(Construct c: constructs){
			for(Construct costrutto: this.getAbstractConstructs()){
				Construct c = costrutto.getMetaInfo();
				ConstructSignature sigConstr = new ConstructSignature(c.getName());
				sigConstr.setFormula(costrutto.getSignature());
				signature.insert(sigConstr);
			}
		//}
		return signature;
	}

	/**
	 * Returns true if the given string is valid as a model name.<br/>
	 * This does NOT check if the name isn't already used in the database, it just performs a basic string checking.<br/>
	 * It does, however, check that the string does not contain "super" and "model" at the same time.
	 * @param modelName the model's name
	 * @return true - if the string can be a valid name.
	 */
	public static boolean isValidName(String modelName) {
		if (!Validator.isValidString(modelName))
			return false;
		String[] reservedChars = { "_", "-" };
		if (!Validator.isValidString(modelName, reservedChars))
			return false;
		if (modelName.toLowerCase().indexOf("super") != -1 && modelName.toLowerCase().indexOf("model") != -1)
			return false;
		return true;
	}
	/**
	 * Returns true if the given string is valid as a short model name.<br/>
	 * This does NOT check if the name isn't already used in the database, it just performs a basic string checking.<br/>
	 * It does, however, check that the given string is not "SM".
	 * @param modelShortName the model's short name
	 * @return true - if the string can be a valid name.
	 */
	public static boolean isValidShortName(String modelShortName) {
		if (!Validator.isValidString(modelShortName))
			return false;
		String[] reservedChars = { "_", "-" };
		if (!Validator.isValidString(modelShortName, reservedChars))
			return false;
		if (modelShortName.equalsIgnoreCase("SM"))
			return false;
		return true;
	}

	/**
	 * Adds the given Construct to the collection of constructs of this model.<br/>
	 * NOTE: No check is performed to know if a duplicate exists.
	 * @param c the construct to add
	 */
	public void addConstruct(Construct c) {
		if (constructs == null) {
			constructs = new LinkedList<Construct>();
		}
		constructs.add(c);
	}

	public void initConstruct() {
		if (constructs == null) {
			constructs = new LinkedList<Construct>();
		}
	}

	/**
	 * Returns the construct object which corresponds to the given name.
	 * @param constructName - the name of the construct.
	 * @return the construct if it's in the model, null otherwise.
	 */
	public Construct getConstruct(String constructName) {
		return this.getConstruct(-1, constructName);
	}
	/**
	 * Returns the construct object which corresponds to the given OID.
	 * @param constructOID - the OID of the construct.
	 * @return the construct if it's in the model, null otherwise.
	 */
	public Construct getConstruct(int constructOID) {
		return this.getConstruct(constructOID, null);
	}
	private Construct getConstruct(int constructOID, String constructName) {
		if (constructs == null)
			return null;

		for (Construct c : constructs) {
			if (c.getOID() == constructOID || c.getName().equalsIgnoreCase(constructName))
				return c;
		}
		return null;
	}
	/**
	 * Sorts the current model based upon the dependencies found in the super model.
	 * @param superModel
	 */
	public void sort(Model superModel) {
		LinkedList<Construct> unsortedConstructs = new LinkedList<Construct>(this.getAbstractConstructs());
		LinkedList<Construct> sortedConstructs = new LinkedList<Construct>();
		while(unsortedConstructs.size() > 0) {
			//System.out.println("Ci sono " + unsortedConstructs.size() + " non ordinati.");
			for (int i = 0; i < unsortedConstructs.size(); i++) {
				Construct unsortedConstruct = unsortedConstructs.get(i);
				//System.out.println("Controllo le dipendenze per " + unsortedConstruct.getName());
				Construct unsortedSMConstruct = unsortedConstruct.getSuperModelConstructs().iterator().next();
				//System.out.println("Mappa a " + unsortedSMConstruct.getName());
				//If it has fulfilled dependencies, we append it to the sorted list.
				boolean fulfilled = true;
				for (Construct dependency : unsortedSMConstruct.getDependencies()) {
					//System.out.println("Ha una dipendenza verso " + dependency.getName());
					//Is this already in the sorted list?
					boolean found = false;
					for (Construct sortedConstruct : sortedConstructs) {
						//System.out.println("Gi� sorted: " + sortedConstruct.getName());
						Construct sortedSMConstruct = sortedConstruct.getSuperModelConstructs().iterator().next();
						if (sortedSMConstruct.getName().equals(dependency.getName())) {
							found = true;
							break;
						}
					}
					if (!found) {
						fulfilled = false;
					}
				}
				if (fulfilled) {
					//System.out.println("Uno lo ho aggiunto.");
					//To the sorted list.
					sortedConstructs.addLast(unsortedConstruct);
					unsortedConstructs.remove(i--);
				}
			}
		}
		//Okay, it's done.
		this.constructs = sortedConstructs;
	}
	/**
	 * Returns a collection of all the Super Model Constructs which can be mapped to Model Constructs.<br/>
	 * If there are no available Constructs (they were all mapped, but this is an unlikely event) the returned
	 * collection is empty, not null.<br/>
	 * A Super Model Construct is available if it has not yet been mapped to a Model Construct and if all its
	 * dependencies have. Example: in a fresh model, the list will contain Abstract but not AttributeOfAbstract.
	 * When the former will be mapped to a Model Construct, the latter will be available too.
	 * @param superModel the Super Model  
	 * @return the collection of available super model constructs.
	 */
	public Collection<Construct> getAvailableConstructs(Model superModel) {
		LinkedList<Construct> availableConstructs = new LinkedList<Construct>();
		for (Construct smc : superModel.getAbstractConstructs()) {
			//Is there a model construct mapped to this construct?
			if (this.getConstructMappedTo(smc) == null) { //There is no construct mapped to smc.
				//If the dependences are all fulfilled, this construct can be considered available.
				boolean allAreFulfilled = true;
				for (Construct dependency : smc.getDependencies()) {
					if (this.getConstructMappedTo(dependency) == null) {
						//This dependency is not mapped yet. This means the dependencies are not fulfilled.
						allAreFulfilled = false;
						break; //No need to keep looping.
					}
				}
				if (allAreFulfilled == true) {
					availableConstructs.addLast(smc);
				}
			}
		}
		return availableConstructs;
	}
	/**
	 * Returns a collection of all the Super Model Properties which can be mapped to
	 * Model Properties for the given Construct <code>construct</code>.<br/>
	 * If there are no available properties (either because the mapped SM constructs
	 * have no properties or because they were already mapped) the returned Collection is empty, not null.  
	 * @param construct
	 * @return a Collection of available Super Model Property objects.
	 */
	public Collection<Property> getAvailableProperties(Construct construct) {	
		LinkedList<Property> availableProperties = new LinkedList<Property>();
		//What are the available properties? All the properties of the sm-constructs mapped to this construct which haven't been mapped already.
		//For every super model construct...
		for (Construct smConstruct : construct.getSuperModelConstructs()) {
			//For every property it has...
			for (Property smProperty : smConstruct.getProperties()) {
				//If it hasn't already been mapped, add it.
				//This means that in the construct owning this property, no other property has a sm-mapping to the sm-property we're adding.
				boolean bAlreadyMapped = false;
				for (Property mExistingProperty : construct.getProperties()) {
					if (mExistingProperty.getMetaInfo().getOID() == smProperty.getOID()) {
						//This particular SM property (smProperty) has the same OID of one already mapped. Too bad.
						bAlreadyMapped = true;
						break; //Don't waste other time on the following existing properties.
					}
				}
				if (bAlreadyMapped == false)
					availableProperties.addLast(smProperty);
			}
		}
		//The list of properties is done.
		return availableProperties;
	}
	/**
	 * Returns a collection of all the Super Model References which can be mapped to
	 * Model References for the given Construct <code>construct</code>.<br/>
	 * If there are no available references (either because the mapped SM constructs
	 * have no references or because they were already mapped) the returned Collection is empty, not null.  
	 * @param construct
	 * @return a Collection of available Super Model Reference objects.
	 * @author paoloc
	 * 20070327
	 */
	public Collection<Reference> getAvailableReferences(Construct construct) {	
		LinkedList<Reference> availableReferences = new LinkedList<Reference>();
		//What are the available references? All the references of the sm-constructs mapped to this construct which haven't been mapped already.
		//For every super model construct...
		for (Construct smConstruct : construct.getSuperModelConstructs()) {
			//For every property it has...
			for (Reference smReference : smConstruct.getReferences()) {
				//If it hasn't already been mapped, add it.
				//This means that in the construct owning this reference, no other reference has a sm-mapping to the sm-reference we're adding.
				boolean bAlreadyMapped = false;
				for (Reference mExistingReference : construct.getReferences()) {
					if (mExistingReference.getMetaInfo().getOID() == smReference.getOID()) {
						//This particular SM reference (smReference) has the same OID of one already mapped. Too bad.
						bAlreadyMapped = true;
						break; //Don't waste other time on the following existing properties.
					}
				}
				if (bAlreadyMapped == false)
					availableReferences.addLast(smReference);
			}
		}
		//The list of properties is done.
		return availableReferences;
	}
	/**
	 * Returns (if it exists) the (model) construct mapped to the given SuperModel construct
	 * @param smConstruct the SuperModel construct to check for a mapping
	 * @return the mapping from the SuperModel construct to this model's available constructs.
	 */
	public Construct getConstructMappedTo(Construct smConstruct) {
		if (constructs == null)
			return null;

		for (Construct c : constructs) {
			for (Construct smc : c.getSuperModelConstructs()) {
				if (smc.getOID() == smConstruct.getOID()) //We've found it!
					return c;
			}
		}
		return null;
	}

	/**
	 * Removes a construct from this model
	 * @param constructName the name of the construct to find and remove (if exists)
	 */
	public void removeConstruct(String constructName) {
		this.removeConstruct(this.getConstruct(constructName));
	}

	/**
	 * Removes a construct from this model
	 * @param construct the construct to remove
	 */
	public void removeConstruct(Construct construct) {
		constructs.remove(construct);
	}
	/**
	 * @return Returns the shortName.
	 */
	public String getShortName() {
		return shortName;
	}
	/**
	 * @param shortName The shortName to set.
	 */
	public void setShortName(String shortName) {
		this.shortName = shortName;
	}

	/**
	 * @return familyOid
	 */
	public int getFamilyOid() {
		return this.familyOid;
	}

	/**
	 * @param familyOid familyOid da impostare
	 */
	public void setFamilyOid(int familyOid) {
		this.familyOid = familyOid;
	}

	/**
	 * @return isAncestor
	 */
	public boolean isAncestor() {
		return this.isAncestor;
	}

	/**
	 * @param isAncestor isAncestor da impostare
	 */
	public void setAncestor(boolean isAncestor) {
		this.isAncestor = isAncestor;
	}

	/**
	 * @return family
	 */
	public Model getFamily() {
		if(this.isAncestor())
			return this;
		else
			return this.family;
	}

	/**
	 * @param family family da impostare
	 */
	public void setFamily(Model family) {
		this.family = family;
	}

	@Override
	protected void doLoadAbstractConstructs() throws ClassNotFoundException, Exception {
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler();
		Model modelLoaded = ph.loadModelInternally(this.getOID(), ph.getSuperModel());
		if(modelLoaded != null) {
			this.constructs = modelLoaded.getAbstractConstructs();
		} else {
			this.constructs = new LinkedList<Construct>();
		}
	}

	@Override
	protected LinkedList<Construct> doGetAbstractConstructs() {
		return this.constructs;
	}
}