/*
 * Created on Jan 25, 2004
 *
 */
package org.modelgen.commons.beans.model;

import org.modelgen.commons.beans.AbstractBean;
import org.modelgen.commons.utils.Validator;
/**
 * The Reference is a generic reference of a {@link Construct}.<p>
 * 
 * If the reference hasn't a referring reference containing its metainfos we have
 * a <i>MetaReference</i> belonging to the SuperModel.<p>
 * 
 * Therefore setting <t>null</t> to {@link #setMetaInfo(Reference)} brings
 * this reference to act as a MetaReference. Vice versa, setting a non null value 
 * to {@link #setMetaInfo(Reference)} sets the referring MetaReference to this reference.
 * 
 * 
 * @author Luca Santarelli [lucachan@tin.it - hrk@users.sourceforge.net]
 * Jan 25, 2004
 */
public class Reference extends AbstractBean<Reference> {
	private Construct ownerConstruct;
	private Construct targetConstruct;
	private ReferenceType type;
	
	/** 
	 * The type of reference
	 * @author Riccardo Pietrucci
	 */
	public static enum ReferenceType { 
		/**
		 * Simple reference from owner to target constructs 
		 */
		RELATION,
		/**
		 * Decorates a description to the target construct
		 * by using the infos available in the owner construct 
		 */
		DESCRIPTION
	}
	
	/**
	 * Creates the MetaReference with the oid and the name specified.
	 * 
	 * @param oid the metaReference's oid
	 * @param name the metaReference's name
	 */
	public Reference(int oid, String name) {
		super(oid, name);
	}
	
	/**
	 * Creates a Reference with the oid, the name and the
	 * MetaReference specified.
	 * 
	 * @param oid the reference's oid
	 * @param name the reference's name
	 * @param metaReference the metaReference
	 */
	public Reference(int oid, String name, Reference metaReference) {
		super(oid, name, metaReference);
	}

	/**
	 * @return the target construct
	 */
	public Construct getTargetConstruct() {
		return targetConstruct;
	}

	/**
	 * @param construct the target construct
	 */
	public void setTargetConstruct(Construct construct) {
		targetConstruct = construct;
	}

	/**
	 * @return the owner construct
	 */
	public Construct getOwnerConstruct() {
		return ownerConstruct;
	}

	/**
	 * @param construct the owner construct
	 */
	public void setOwnerConstruct(Construct construct) {
		ownerConstruct = construct;
	}

	/**
	 * Returns true if the given name is valid for a Reference.<p>
	 * The check is formal only (i.e.: not null, not empty, ...)
	 * @param referenceName the name to check
	 * @return if the given name is valid
	 */
	public static boolean isValidName(String referenceName) {
		return Validator.isValidString(referenceName);
	}
	
	/**
	 * @param type the type of this reference
	 */
	public void setType(ReferenceType type) {
		this.type = type;
	}
	
	/**
	 * @param typeString the string rappresentation of this type (case insensitive)
	 */
	public void setType(String typeString) {
		for(ReferenceType type : ReferenceType.values() ) {
			if (typeString.equalsIgnoreCase(type.toString())) {
				this.type = type;
				return;
			}
		}
		throw new IllegalArgumentException("\"" + type + "\" is an illegal type for a Reference object.");
	}
	
	/**
	 * @return the type of this reference
	 */
	public ReferenceType getType() {
		return this.type;
	}
	
	/**
	 * @return the lower case string rappresentation of this type
	 */
	public String getTypeAsString() {
		return this.type.toString().toLowerCase();
	}
	
	/**
	 * @return the textual rappresentation of this reference's name and type
	 */
	public String toString() {
		return this.getName() + " - " +this.getTypeAsString();
	}
}
