/*
 * Created on 9-apr-2004
 *
 */
package org.modelgen.commons.beans.scheme;

import org.modelgen.commons.beans.AbstractBean;
import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Property;
import org.modelgen.commons.beans.model.Reference;
import org.modelgen.commons.utils.Validator;

import java.util.ArrayList;
import java.util.Collection;

/**
 * The SchemeObject is a structure element of a {@link Scheme} composed of
 * various {@link SchemeObjectProperty} and {@link SchemeObjectReference}.<p>
 * 
 * It contains the <i>values</i> of the properties and references declared of the
 * referring {@link org.modelgen.commons.beans.model.Construct}.<p>
 * 
 * Therefore we can't assign a <code>null</code> value to {@link #setMetaInfo(Construct)} 
 * 
 * @author Luca
 *
 */
public class SchemeObject extends AbstractBean<Construct> {
	private ArrayList<SchemeObjectReference> references; //The number of items is equal to construct.getReferences().size()
	private ArrayList<SchemeObjectProperty> properties;
	private SchemeObjectKey key = null;

	/**
	 * Creates a scheme object with the oid, the name and the
	 * referring construct specified.
	 * 
	 * <p>If it's a super model scheme, the referring model it's the
	 * <i>source model</i>, useful with rule transformations.
	 * 
	 * @param oid the scheme object's oid
	 * @param name the scheme object's name
	 * @param parent the referring construct
	 */
	public SchemeObject(int oid, String name, Construct parent) {
		super(oid, name, parent);
	}

	/**
	 * Assigns the construct specified and creates the
	 * corresponding {@link SchemeObjectProperty} and 
	 * {@link SchemeObjectReference} with empty value
	 * @param construct the new referrring construct
	 */
	@Override
	public void setMetaInfo(Construct construct) {
		super.setMetaInfo(construct);
		//The properties coming from this construct:
		this.properties = new ArrayList<SchemeObjectProperty>(construct.getProperties().size());
		for (Property p : construct.getProperties()) {
			SchemeObjectProperty sop = new SchemeObjectProperty(0, "", p);
			properties.add(sop);
		}
		//The references coming from this construct:
		this.references = new ArrayList<SchemeObjectReference>(construct.getReferences().size());
		for (Reference r : construct.getReferences()) {
			SchemeObjectReference sor = new SchemeObjectReference(0, null, r);
			references.add(sor); 
		}
	}


	/**
	 * Assign the property value
	 * @param name the property's name (corresponding to one of its referring model's available properties with type boolean)
	 * @param value the property's value
	 */
	public void setProperty(String name, boolean value) {
		if (value == true)
			this.setProperty(name, "true");
		else
			this.setProperty(name, "false");
	}

	/**
	 * @return the properties list
	 */
	public Collection<SchemeObjectProperty> getProperties() {
		return properties;
	}

	/**
	 * Returns true if the given name is valid for a SchemeObject.<p>
	 * The check is formal only (i.e.: not null, not empty, ...)
	 * @param objectName the name to check
	 * @return if the given name is valid
	 */
	public static boolean isValidName(String objectName) {
		if (!Validator.isValidString(objectName))
			return false;
		return true;
	}

	/**
	 * Returns the SchemeObjectReference object which has the given referenceName.
	 * @param referenceName the reference's name
	 * @return the SchemeObjectReference object if found, <code>null</code> otherwise.
	 */	
	public SchemeObjectReference getReference(String referenceName) {
		for (SchemeObjectReference sor : this.references) {
			if (sor.getMetaInfo().getName().equals(referenceName)) //Found it!
				return sor;
		}
		return null;
	}

	/**
	 * @return the available references
	 */
	public Collection<SchemeObjectReference> getReferences() {
		return references;
	}


	/**
	 * Assigns the pointing reference value
	 * @param referenceName the working reference's name
	 * @param object the SchemeObject pointed to the specified {@link SchemeObjectReference}.
	 * If <code>null</code>, brings the working reference to the optional value.
	 */
	public void setReference(String referenceName, SchemeObject object) {
		SchemeObjectReference sor = this.getReference(referenceName);
		if (sor != null) { //It was found, let's see if the objects are compatible.
			// paoloc 20060511
			// added handling for null (optional) references. This may break the code somewhere...
			if (object==null) 
				sor.setValue(object);
			else
				if (sor.getMetaInfo().getTargetConstruct().getName().equals(object.getMetaInfo().getName())) {
					//Yes, they are.
					sor.setValue(object);
				}
				else {//The objects are instances of different constructs!
					throw new ClassCastException(
							"The reference \"" + sor.getMetaInfo().getName()
							+ "\" points to instances of constructs of type \"" + sor.getMetaInfo().getTargetConstruct().getName()
							+ "\" but you passed \"" + object.getName() + "\" which is an instance of \"" + object.getMetaInfo().getName()
							+ "\".");
				}
		}
		else { //A reference by that name was not found in the list.
			throw new IllegalArgumentException(
					"The object \"" + this.name + "\", instance of \"" + this.getMetaInfo().getName()
					+ "\" has no reference named \"" + referenceName + "\".");
		}
	}

	/**
	 * Returns the SchemeObjectProperty which has the given propertyName.
	 * @param propertyName the property's name
	 * @return the SchemeObjectProperty object if found, <code>null</code> otherwise.
	 */	
	public SchemeObjectProperty getProperty(String propertyName) {
		for (SchemeObjectProperty sop : this.properties) {
			if (sop.getMetaInfo().getName().equals(propertyName))
				return sop;
		}
		return null;
	}

	/**
	 * Assign the property value
	 * @param propertyName the property's name (corresponding to one of its referring model's available properties)
	 * @param value the property's value. If <code>null</code> sets the property's value to {@link org.modelgen.commons.beans.AbstractBean#MDM_UNSPECIFIED }
	 */
	public void setProperty(String propertyName, String value) {
		if (value == null) {
			// commentata da paolo c il 07-10-2004 per ammettere la presenza di 
			// valori nulli negli argomenti; dovrebbe capitare solo con le string

			// throw new IllegalArgumentException("\"null\" is not allowed as a value.");
			value = AbstractBean.MDM_UNSPECIFIED;
		}
		SchemeObjectProperty sop = this.getProperty(propertyName);
		if (sop != null) { //It was found.
			// 2005-01-29, rimuovo tutto questo controllo perch mi intralcia nel caricamento dei super-schemi.
			// prima o poi bisogner capire se e come reintrodurlo (serve solo a rendere robusta la libreria
			// ma tanto lo strumento che la utilizza fornisce dati validi in partenza)
			sop.setValue(value);
			/*
			String type = sop.getModelProperty().getSuperModelProperty().getType();
			if (TypeTranslator.getJavaType(type).equals(Boolean.class)) {
				if (value.equals("true") || value.equals("false"))
					sop.setValue(value);
				else { //The property is boolean, but the string is not a valid boolean.
					throw new IllegalArgumentException(
						"The property \"" + propertyName + "\" is of type \"" + type
						+ "\" but \"" + value + "\" is not a valid " + type + "\".");
				}
			} //Boolean.class
			else if (TypeTranslator.getJavaType(type).equals(String.class)) {
				//It's always valid.
				sop.setValue(value);
			}
			 */
		}
		else { //A property by this name does not exist (which happens if propertyName is null)
			throw new IllegalArgumentException(
					"The object \"" + this.name + "\", instance of \"" + this.getMetaInfo().getName()
					+ "\" has no property named \"" + propertyName + "\".");
		}
	}
	/**
	 * Returns in a thread-safe way the unique SchemeObjectKey object representing the key of this particular SchemeObject.
	 * @see SchemeObjectKey
	 * @return the unique SchemeObjectKey object representing the key of this particular SchemeObject.
	 */
	public synchronized SchemeObjectKey getKey() {
		if (this.key == null) {
			//In questo modo, la chiave viene generata solo quando serve realmente.
			this.key = new SchemeObjectKey();
		}
		return this.key;
	}
}
