/*
 * Created on 10-mar-2004
 *
 */
package org.modelgen.commons.persistence;

import org.modelgen.commons.beans.instance.Instance;
import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.scheme.Scheme;
import org.modelgen.commons.beans.skolem.SkolemFunction;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Contains all functions useful to obtain ModelGen's data.
 * 
 * @author Luca
 *
 */
public interface PersistenceHandler {
	/**
	 * Returns true if there is already a model with the given name, false otherwise (which means it returns false on a null string too)
	 * @param modelName
	 * @return true, if a model with the name modelName does exist already.
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract boolean doesModelExist(String modelName) throws PersistenceException;
	/**
	 * Returns true if the shortName is already used for a model, false otherwise (which means it returns false on a null string too)
	 * @param shortName the model's short name
	 * @return true, if a model with the name modelName does exist already.
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract boolean doesModelExistWithShortname(String shortName) throws PersistenceException;
	/**
	 * @return a collection of the currently available models
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Collection<Model> getModels() throws PersistenceException;
	/**
	 * Returns a collection of the currently available schemes.
	 * @return a collection of the currently available schemes
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Collection<Scheme> getSchemes() throws PersistenceException;
	
	/**
	 * @return the supermodel
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Model getSuperModel() throws PersistenceException;
	/**
	 * Returns a Model object whose OID in the database is modelOID.
	 * @param modelOID the model's oid
	 * @return the model found
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Model getModel(int modelOID) throws PersistenceException;
	
	/**
	 * Returns a Scheme object whose OID in the database is schemeOID.
	 * @param schemeOID the scheme's oid
	 * @return the scheme found
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Scheme getScheme(int schemeOID) throws PersistenceException;
	
	/**
	 * Returns a SuperModel's Scheme object whose OID 
	 * in the database is schemeOID.
	 * @param schemeOID the scheme's oid
	 * @return the scheme found
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Scheme getSuperModelScheme(int schemeOID) throws PersistenceException;
	/**
	 * Returns a Instance object whose OID in the database is instanceOID.
	 * @param instanceOID the model's oid
	 * @return the model found
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Instance getSuperModelInstance(int instanceOID) throws PersistenceException;

	/**
	 * Completely loads a Model object given its database OID and an OO representation of the super model.
	 * 
	 * <p>Warning: don't use it! Only used internally in {@link Model}.
	 * 
	 * <p>Because of the structure of the class 
	 * {@link org.modelgen.commons.beans.AbstractStructure}
	 * this modify the internal structure of the lazy loaded Model
	 * 
	 * @param modelOID
	 * @param superModel 
	 * @return the model loaded
	 * @throws PersistenceException if a persistence error occurs
	 * 
	 */
	public abstract Model loadModelInternally(int modelOID, Model superModel) throws PersistenceException;
	
	/**
	 * Completely loads a Scheme object given its database OID.
	 * 
	 * <p>Warning: don't use it! Only used internally in {@link Scheme}.
	 * 
	 * <p>Because of the structure of the class 
	 * {@link org.modelgen.commons.beans.AbstractStructure}
	 * this modify the internal structure of the lazy loaded Scheme
	 * 
	 * @param schemeOID the scheme's oid
	 * @return the loaded scheme
	 * @throws PersistenceException if a persistence error occurs
	 * 
	 */
	public abstract Scheme loadSchemeInternally(int schemeOID) throws PersistenceException;
	
	/**
	 * The family ancestor
	 * @param familyOID
	 * @return the Model that is tha family's ancestor
	 * @throws PersistenceException
	 */
	public abstract Model getFamilyAncestor(int familyOID) throws PersistenceException;
	
	/**
	 * Retrieves all families ancestors
	 * @return the list of Model that are family's ancestor
	 * @throws PersistenceException
	 */
	public abstract List<Model> getAncestors() throws PersistenceException;
	
	/**
	 * Saves a complete Model object into the database.
	 * @param model the model to save
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeModel(Model model) throws PersistenceException;
	/**
	 * Saves a complete Scheme object into the supermodel portion of the database.
	 * This method has been introduced to replace the copyrules step.
	 * @param scheme the scheme to save
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeModelSchemeIntoSuperModel(Scheme scheme) throws PersistenceException;
	/**
	 * Saves a complete Scheme object into the supermodel portion of the database.
	 * This method has been introduced to replace the copyrules step.
	 * @param superModelScheme the super model scheme to store
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeSuperModelSchemeIntoModel(Scheme superModelScheme) throws PersistenceException;
	/**
	 * Saves a complete Scheme object into the database.
	 * @param scheme
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeScheme(Scheme scheme) throws PersistenceException;
	/**
	 * Deletes the model which maps to the given modelOID.
	 * @param modelOID
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void deleteModel(int modelOID) throws PersistenceException;
	/**
	 * Deletes the scheme which maps to the given schemeOID.
	 * @param schemeOID
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void deleteScheme(int schemeOID) throws PersistenceException;
	/**
	 * Renames the model with OID modelOID to the name modelName
	 * @param modelOID
	 * @param modelName
	 * @param shortName 
	 * @throws PersistenceException if the model does not exist or if the name is already taken.
	 */
	public abstract void renameModel(int modelOID, String modelName, String shortName) throws PersistenceException;
	/**
	 * Renames the scheme with OID schemeOID to the name schemeName
	 * @param schemeOID
	 * @param schemeName
	 * @throws PersistenceException if the model does not exist or if the name is already taken.
	 */
	public abstract void renameScheme(int schemeOID, String schemeName) throws PersistenceException;
	/**
	 * Returns true if there is already a scheme with the given name, false otherwise (which means it returns false on a null string too)
	 * @param schemeName
	 * @return true, if a scheme with the name schemeName does exist already.
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract boolean doesSchemeExist(String schemeName) throws PersistenceException;
	/**
	 * @return the map of all the available (e.g.: already defined) Skolem functions
	 * @deprecated This function is deprecated because uses a table-based model.
	 * Use {@link #getSkolemFunctions()} instead, which uses a object-bean model.
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Map<String,org.modelgen.regole.SkolemFunction> getMappaFuntoriSkolem() throws PersistenceException;
	/**
	 * @return the map of all the available (e.g.: already defined) Skolem functions
	 * 
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Collection<org.modelgen.commons.beans.skolem.SkolemFunction> getSkolemFunctions() throws PersistenceException;
	/**
	 * @return the map of the user defined Skolem function (i.e: those not created
	 * at initialization time by the tool).
	 * 
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Collection<org.modelgen.commons.beans.skolem.SkolemFunction> getUserDefinedSkolemFunctions() throws PersistenceException;
	/**
	 * Stores the Skolem Functions contained in the given Map <code>map</code>.
	 * @param map the map of all the available (e.g.: already defined) Skolem functions to save
	 * 
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeSkolemFunctions(Map<String,org.modelgen.regole.SkolemFunction> map) throws PersistenceException;
	/**
	 * Find a skolem function
	 * @param superModel the super model
	 * @param construct the skolem function's owner construct
	 * @param skfName the skolem function's name
	 * @return the corresponding skolem function
	 * 
	 * @throws PersistenceException if a persistence error occurs or if not found
	 */
	public abstract SkolemFunction getSkolemFunction(Model superModel, Construct construct, String skfName) throws PersistenceException;
	/**
	 * Save a skolem function
	 * @param superModel the super model
	 * @param skolemFunction the skolem function to save
	 * 
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void storeSkolemFunction(Model superModel, SkolemFunction skolemFunction) throws PersistenceException;
	/**
	 * @return the super model's schemes
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract Collection<Scheme> getSuperModelSchemes() throws PersistenceException;
	
	/**
	 * Returns the next valid OID to be used in an INSERT statement for the table and field passed as arguments.
	 * 
	 * <p>A typical invocation is:
	 * <code>getNewOID("SchemeSM","OID")</code> or
	 * <code>getNewOID("InstanceSM","OID")</code>
	 * 
	 * @param table the working table
	 * @param field the field to check for the next valid oid
	 * @return the next valid OID to be used
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract int getNewOID(String table, String field) throws PersistenceException;

	/**
	 * Completely loads a supermodel's Scheme object given its database OID.
	 * 
	 * <p>Warning: don't use it! Only used internally in {@link Scheme}.
	 * 
	 * <p>Because of the structure of the class 
	 * {@link org.modelgen.commons.beans.AbstractStructure}
	 * this modify the internal structure of the lazy loaded Scheme
	 * 
	 * @param schemeOID the scheme's oid
	 * @return the loaded scheme
	 * @throws PersistenceException if a persistence error occurs
	 * 
	 */
	public abstract Scheme loadSuperModelSchemeInternally(int schemeOID) throws PersistenceException;
	
	/**
	 * Finds and deletes the skolem function found.
	 * @param superModel the super model
	 * @param construct the skolem function's owner construct
	 * @param functionName the skolem function's name
	 * @throws PersistenceException if a persistence error occurs
	 */
	public abstract void deleteSkolemFunction(Model superModel, Construct construct, String functionName) throws PersistenceException;

	/**
	 * 
	 * @author Gio
	 * @param model a model
	 * @return the corresponding oid with the specified model's name
	 * @throws PersistenceException if a persistence error occurs
	 */
	public int getModelOID(Model model) throws PersistenceException;
	
	/**
	 * Test a simple query on db, reporting the result.
	 * 
	 * @return if the db works as expected
	 * @author Gio
	 */
	public boolean isInitialized();
	
	/**
	 * @return the connection with the persistence support
	 */
	public Connection getConnection();
	
	/**
	 * @return the super model's instances
	 * @throws PersistenceException if a persistence error occurs
	 */
	public Collection<Instance> getSuperModelInstances() throws PersistenceException;
	
	/**
	 * Completely loads a Instance object given its database OID.
	 * 
	 * <p>Warning: don't use it! Only used internally in {@link Instance}.
	 * 
	 * <p>Because of the structure of the class 
	 * {@link org.modelgen.commons.beans.AbstractStructure}
	 * this modify the internal structure of the lazy loaded Instance
	 * 
	 * @param instanceOID the instance's oid
	 * @return the loaded instance
	 * @throws PersistenceException if a persistence error occurs
	 * 
	 */
	public abstract Instance loadSuperModelInstanceInternally(int instanceOID) throws PersistenceException;
	
	
	/**
	 * Creates the base infos useful to store a super model scheme
	 * @param schemeOID
	 * @param sourceModelOID
	 * @param newName
	 * @throws SQLException if a persistence error occurs
	 */
	public void createSuperModelSchemeUgly(int schemeOID, int sourceModelOID, String newName) throws SQLException;
	
	/**
	 * Creates the base infos useful to store a super model instance
	 * @param instanceOID
	 * @param schemaOID
	 * @param newName
	 * @throws SQLException if a persistence error occurs
	 */
	public void createSuperModelInstanceUgly(int instanceOID, int schemaOID, String newName) throws SQLException;
	
}