/*
 * Created on 10-mar-2004
 *
 */
package org.modelgen.commons.persistence;

import org.modelgen.commons.persistence.database.CacheProxyDBHandler;
import org.modelgen.commons.persistence.database.DBHandler;
import org.modelgen.commons.persistence.database.RealDBHandler;
import org.modelgen.commons.utils.ModelgenConfigurationManager;

import java.sql.SQLException;
import java.util.LinkedList;
import java.util.List;

/**
 * Factory to obtain the shared {@link PersistenceHandler} istance.
 * 
 * <p>The concrete type of the {@link PersistenceHandler} depends on
 * the configuration file.
 * 
 * @author Luca
 *
 */
public class PersistenceHandlerFactory {
	private static List<PersistenceHandlerFactoryReloadListener> listeners = new LinkedList<PersistenceHandlerFactoryReloadListener>();
	
	/* riccardop: l'istanza e' condivisa */
	private static DBHandler currentDBHandler = null;
	
	/**
	 * Creates if not yet present or get from cache the default persistence handler. 
	 * 
	 * @return the persistence handler to obtain
	 * @throws ClassNotFoundException if the default persistence handler is unknown
	 * @throws Exception if can't create a connection
	 */
	public static final synchronized PersistenceHandler getPersistenceHandler() throws ClassNotFoundException, Exception {
		return getPersistenceHandler(null);
	}
	
	/**
	 * Creates if not yet present or get from cache the specified persistence handler. 
	 * @param preferredHandler the type of handler to obtain
	 * 
	 * @return the persistence handler to obtain
	 * @throws ClassNotFoundException if the default persistence handler is unknown
	 * @throws Exception if can't create a connection
	 */
	public static final synchronized PersistenceHandler getPersistenceHandler(String preferredHandler) throws ClassNotFoundException, Exception {
		if (preferredHandler == null) { //Return the preferred one, as found in the configuration file.
			//DBHandler is set as default, so if jConfig doesn't find the configuration, a valid handler is returned.
			return getPersistenceHandler(ModelgenConfigurationManager.getIstance().getProperty("preferredHandler", "DBHandler", "persistence"));
		} else if (preferredHandler.equals("DBHandler")) {
			/* riccardop: l'istanza e' condivisa */
			if(currentDBHandler != null ) {
				return currentDBHandler;
			} else {
				/* riccardop: ora si crea il cache proxy */
				RealDBHandler realHandler = new RealDBHandler();
				CacheProxyDBHandler proxyHandler = new CacheProxyDBHandler(realHandler);
				currentDBHandler = proxyHandler;
				return proxyHandler;
			}
	    } else {
			throw new ClassNotFoundException();
		}
	}
	
	/**
	 * It forces to use a new {@link PersistenceHandler}.
	 * 
	 * <p>Therefore we now use updated values of db.
	 * 
	 * <p>It notifies all the {@link PersistenceHandlerFactoryReloadListener} registered
	 * through {@link #addReloadListener(PersistenceHandlerFactoryReloadListener)}
	 * 
	 * @author Riccardo Pietrucci
	 */
	public static synchronized final void reload() {
		
		currentDBHandler = null;
		System.gc();
		
		
		/* reload è chiamato (praticamente, vd. reload2() ) solo in avvio da StartDBServerAction */
		/* la notify è sempre senza conseguenze perché si verifica */
		/* un'eccezione dovuta alla assenza di connessione con il db */
		/* il passo successivo di StartDBServerAction porta a termine la connessione */
		/* si mantiene questo reload() solo per gli effetti collaterali precedenti, la notifyListeners() è inutile */
		//notifyListeners();
	}
	
	/**
	 * per mantenere la compatibilità con il debugger
	 * che (?) accede direttamente alla persistenza
	 */
	public static synchronized final void reload2() {
		reload();
		notifyListeners();
	}
	
	private static void notifyListeners() {
		for(PersistenceHandlerFactoryReloadListener listener : listeners) {
			listener.dbReloaded();
		}
	}
	
	/**
	 * Add a reload listener that will be notified in a reload event.
	 * @param listener the reload listener to notify in a reload event
	 * @author Riccardo Pietrucci
	 */
	public static void addReloadListener(PersistenceHandlerFactoryReloadListener listener) {
		listeners.add(listener);
	}
	
	/**
	 * It resets the old {@link DBHandler} after a {@link #reload()} operation.
	 * 
	 * <p>Therefore we now use the old values of db.
	 * 
	 * <p>It notifies all the {@link PersistenceHandlerFactoryReloadListener} registered
	 * through {@link #addReloadListener(PersistenceHandlerFactoryReloadListener)}
	 * 
	 * @author Riccardo Pietrucci
	 * @param oldDBHandler the old {@link DBHandler} to reset
	 * @throws ClassNotFoundException if the sql driver can't be found
	 * @throws SQLException if the connection can't be re-established
	 */
	public static synchronized void resetDBHandler(DBHandler oldDBHandler) throws ClassNotFoundException, SQLException {
		
		currentDBHandler = oldDBHandler;
		if(currentDBHandler instanceof CacheProxyDBHandler) {
			CacheProxyDBHandler cacheProxy = (CacheProxyDBHandler) currentDBHandler;
			/* Riconnetti */
			RealDBHandler realHandler = new RealDBHandler();
			cacheProxy.setTrueDBHandler(realHandler);
			resetCache();
		}
		
		notifyListeners();
	}

	public static void resetCache() {
		/* cancella vecchi modelli e schemi */
		CacheProxyDBHandler cacheProxy = (CacheProxyDBHandler) currentDBHandler;
		cacheProxy.clean("models");
		cacheProxy.clean("schemes");
	}
}
