package org.modelgen.commons.persistence.database;

import org.modelgen.commons.beans.instance.Instance;
import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.scheme.Scheme;
import org.modelgen.commons.persistence.PersistenceException;
import org.modelgen.regole.SkolemFunction;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Holds in cache the results of {@link RealDBHandler} methods to minimize
 * database accesses and to provide consistency in pointed structures.<p>
 * 
 * @author Riccardo Pietrucci
 *
 */
public class CacheProxyDBHandler implements DBHandler {
	
	private RealDBHandler trueDBHandler;
	
	private Model superModel = null;
	private SortedMap<Integer,Model> modelsLoaded = new TreeMap<Integer,Model>(); 
	private SortedMap<Integer,Scheme> schemesLoaded = new TreeMap<Integer,Scheme>(); 
	private SortedMap<Integer,Scheme> superModelSchemesLoaded = new TreeMap<Integer,Scheme>();
	private SortedMap<Integer,Instance> instancesLoaded = new TreeMap<Integer,Instance>();
	
	public CacheProxyDBHandler(RealDBHandler trueDBHandler) {
		this.setTrueDBHandler(trueDBHandler);
	}

	public void deleteModel(int modelOID) throws PersistenceException {
		trueDBHandler.deleteModel(modelOID);
		this.modelsLoaded.remove(modelOID);
	}

	public void deleteScheme(int schemeOID) throws PersistenceException {
		trueDBHandler.deleteScheme(schemeOID);
		this.schemesLoaded.remove(schemeOID);
	}

	public void deleteSkolemFunction(Model model, Construct construct, String functionName) throws PersistenceException {
		trueDBHandler.deleteSkolemFunction(model, construct, functionName);
	}

	public boolean doesModelExist(String modelName) throws PersistenceException {
		return trueDBHandler.doesModelExist(modelName);
	}

	public boolean doesModelExistWithShortname(String shortName) throws PersistenceException {
		return trueDBHandler.doesModelExistWithShortname(shortName);
	}

	public boolean doesSchemeExist(String schemeName) throws PersistenceException {
		return trueDBHandler.doesSchemeExist(schemeName);
	}

	public Connection getConnection() {
		return trueDBHandler.getConnection();
	}

	public Map<String, SkolemFunction> getMappaFuntoriSkolem() throws PersistenceException {
		return trueDBHandler.getMappaFuntoriSkolem();
	}

	public Model getModel(int modelOID) throws PersistenceException {
		Model model;
		if(this.modelsLoaded.containsKey(modelOID)) {
			model = modelsLoaded.get(modelOID);
		} else {
			model = trueDBHandler.getModel(modelOID);
			modelsLoaded.put(modelOID, model);
		}
		return model;
	}

	public int getModelOID(Model model) throws PersistenceException {
		return trueDBHandler.getModelOID(model);
	}

	public Collection<Model> getModels() throws PersistenceException {
		Collection<Model> models = trueDBHandler.getModels();
		for(Model model : models) {
			int modelOID = model.getOID();
			if( ! this.modelsLoaded.containsKey(modelOID) ) {
				this.modelsLoaded.put(modelOID, model);
			}
		}
		/* Restituisce i gia' salvati e aggiunge i nuovi */
		return this.modelsLoaded.values();
	}

	public int getNewOID(String table, String field) throws PersistenceException {
		return trueDBHandler.getNewOID(table, field);
	}
	
	public Model getFamilyAncestor(int familyOID) throws PersistenceException{
		return trueDBHandler.getFamilyAncestor(familyOID);
	}
	
	public List<Model> getAncestors() throws PersistenceException{
		return trueDBHandler.getAncestors();
	}

	public Collection<Scheme> getSchemes() throws PersistenceException {
		Collection<Scheme> schemes = trueDBHandler.getSchemes();
		for(Scheme scheme : schemes) {
			int schemeOID = scheme.getOID();
			if( ! this.schemesLoaded.containsKey(schemeOID) ) {
				this.schemesLoaded.put(schemeOID, scheme);
			}
		}
		/* Restituisce i gia' salvati e aggiunge i nuovi */
		return this.schemesLoaded.values();
	}

	public org.modelgen.commons.beans.skolem.SkolemFunction getSkolemFunction(Model model, Construct construct, String skfName) throws PersistenceException {
		return trueDBHandler.getSkolemFunction(model, construct, skfName);
	}

	public Collection<org.modelgen.commons.beans.skolem.SkolemFunction> getSkolemFunctions() throws PersistenceException {
		return trueDBHandler.getSkolemFunctions();
	}

	public Model getSuperModel() throws PersistenceException {
		if(this.superModel == null) {
			this.superModel = trueDBHandler.getSuperModel();
		}
		return this.superModel;
	}

	public Collection<Instance> getSuperModelInstances() throws PersistenceException {
		Collection<Instance> instances = trueDBHandler.getSuperModelInstances();
		for(Instance instance : instances) {
			int instanceOID = instance.getOID();
			if( ! this.instancesLoaded.containsKey(instanceOID) ) {
				this.instancesLoaded.put(instanceOID, instance);
			}
		}
		/* Restituisce i gia' salvati e aggiunge i nuovi */
		return this.instancesLoaded.values();
	}

	public Collection<Scheme> getSuperModelSchemes() throws PersistenceException {
		Collection<Scheme> schemes = trueDBHandler.getSuperModelSchemes();
		for(Scheme scheme : schemes) {
			int schemeOID = scheme.getOID();
			if( ! this.superModelSchemesLoaded.containsKey(schemeOID) ) {
				this.superModelSchemesLoaded.put(schemeOID, scheme);
			}
		}
		/* Restituisce i gia' salvati e aggiunge i nuovi */
		return this.superModelSchemesLoaded.values();
	}

	public Collection<org.modelgen.commons.beans.skolem.SkolemFunction> getUserDefinedSkolemFunctions() throws PersistenceException {
		return trueDBHandler.getUserDefinedSkolemFunctions();
	}

	public boolean isInitialized() {
		return trueDBHandler.isInitialized();
	}

	public void renameModel(int modelOID, String modelName, String shortName) throws PersistenceException {
		trueDBHandler.renameModel(modelOID, modelName, shortName);
	}

	public void renameScheme(int schemeOID, String schemeName) throws PersistenceException {
		trueDBHandler.renameScheme(schemeOID, schemeName);
	}

	public void storeModel(Model model) throws PersistenceException {
		trueDBHandler.storeModel(model);
		this.modelsLoaded.put(model.getOID(), model);
	}

	public void storeModelSchemeIntoSuperModel(Scheme scheme) throws PersistenceException {
		trueDBHandler.storeModelSchemeIntoSuperModel(scheme);
		this.superModelSchemesLoaded.put(scheme.getOID(), scheme);
	}

	public void storeScheme(Scheme scheme) throws PersistenceException {
		trueDBHandler.storeScheme(scheme);
		this.schemesLoaded.put(scheme.getOID(), scheme);
	}

	public void storeSkolemFunction(Model model, org.modelgen.commons.beans.skolem.SkolemFunction skolemFunction) throws PersistenceException {
		trueDBHandler.storeSkolemFunction(model, skolemFunction);
	}

	public void storeSkolemFunctions(Map<String,SkolemFunction> map) throws PersistenceException {
		trueDBHandler.storeSkolemFunctions(map);
	}

	public void storeSuperModelSchemeIntoModel(Scheme superModelScheme) throws PersistenceException {
		trueDBHandler.storeSuperModelSchemeIntoModel(superModelScheme);
		this.schemesLoaded.put(superModelScheme.getOID(), superModelScheme);
	}

	public void createSuperModelSchemeUgly(int schemeOID, int modelOID, String newName) throws SQLException {
		trueDBHandler.createSuperModelSchemeUgly(schemeOID, modelOID, newName);
	}

	public void createSuperModelInstanceUgly(int instanceOID, int schemaOID, String newName) throws SQLException {
		trueDBHandler.createSuperModelInstanceUgly(instanceOID, schemaOID, newName);
	}

	public void setTrueDBHandler(RealDBHandler trueDBHandler) {
		this.trueDBHandler = trueDBHandler;
		this.trueDBHandler.setCache(this);
	}

	/**
	 * Search from the cache values obtained 
	 * from {@link #getSuperModelInstances()} the instance with
	 * the specified oid. Therefore it doesn't execute any 
	 * specific sql query.
	 * 
	 * @param instanceOID instance's oid to search
	 * @return the corresponding instance, null otherwise
	 * @throws PersistenceException errors in {@link #getSuperModelInstances()}
	 */
	public Instance getSuperModelInstance(int instanceOID) throws PersistenceException {
		/* update cache values */
		this.getSuperModelInstances();
		return this.instancesLoaded.get(instanceOID);
	}
	
	/**
	 * Search from the cache values obtained 
	 * from {@link #getScheme(int)} the scheme with
	 * the specified oid. Therefore it doesn't execute any 
	 * specific sql query.
	 * 
	 * @param schemeOID scheme's oid to search
	 * @return the corresponding scheme, null otherwise
	 * @throws PersistenceException errors in {@link #getSuperModelInstances()}
	 */
	public Scheme getScheme(int schemeOID) throws PersistenceException {
		/* update cache values */
		this.getSchemes();
		return this.schemesLoaded.get(schemeOID);
	}
	
	/**
	 * Search from the cache values obtained 
	 * from {@link #getScheme(int)} the supermodel's scheme with
	 * the specified oid. Therefore it doesn't execute any 
	 * specific sql query.
	 * 
	 * @param schemeOID supermodel scheme's oid to search
	 * @return the corresponding supermodel's scheme, null otherwise
	 * @throws PersistenceException errors in {@link #getSuperModelInstances()}
	 */
	public Scheme getSuperModelScheme(int schemeOID) throws PersistenceException {
		/* update cache values */
		this.getSuperModelSchemes();
		return this.superModelSchemesLoaded.get(schemeOID);
	}

	public Model loadModelInternally(int modelOID, Model superModel) throws PersistenceException {
		return this.trueDBHandler.loadModelInternally(modelOID, superModel);
	}

	public Scheme loadSchemeInternally(int schemeOID) throws PersistenceException {
		return this.trueDBHandler.loadSchemeInternally(schemeOID);
	}

	public Instance loadSuperModelInstanceInternally(int instanceOID) throws PersistenceException {
		return this.trueDBHandler.loadSuperModelInstanceInternally(instanceOID);
	}

	public Scheme loadSuperModelSchemeInternally(int schemeOID) throws PersistenceException {
		return this.trueDBHandler.loadSuperModelSchemeInternally(schemeOID);
	}
	
	/**
	 * Resetta una porzione della cache, forzando il reloading al prossimo utilizzo.
	 * Utile ad esempio per le cancellazione di schemi che possono essere riferiti a modelli
	 * in cache, ma modificati sul db.
	 * @param toBeCleaned una stringa che identifica quale parte della cache deve essere resettata
	 */
	public void clean(String toBeCleaned) {
		/* clean degli schemi memorizzati */
		if(toBeCleaned.equals("schemes"))
			this.schemesLoaded = new TreeMap<Integer,Scheme>();
		/* clean dei modelli memorizzati */
		else if(toBeCleaned.equals("models"))
			this.modelsLoaded = new TreeMap<Integer,Model>();
	}

	
}
