package org.modelgen.commons.persistence.database;

import java.util.LinkedList;
import java.util.TreeMap;

import org.apache.log4j.Logger;

/**
 * Implements the 'move-to-front' strategy, useful
 * in {@link CacheProxyDBHandler} (<i>for now unused</i>).
 * 
 * <p>Whenever a client access a cache entry through
 * {@link #getValue(Object)}, the corresponding value is moved to
 * the front of a frequencyTable of fixed size.
 * 
 * <p>When new entries have to be added through 
 * {@link #put(Object, Object)}, if the frequencyTable if full
 * entries can be deleted from the back of the list.
 * 
 * @author Riccardo Pietrucci
 *
 * @param <K> the key type
 * @param <V> the value type
 */
public class MoveToFrontMap<K, V> extends TreeMap<K, V> {
	private static final long serialVersionUID = 1L;
	/**
	 * The maximum number of values cachable
	 */
	public final static int CACHE_ENTRIES_LIMIT = 100; 
	private final Logger log = Logger.getLogger(MoveToFrontMap.class);
	private LinkedList<K> frequencyTable = new LinkedList<K>();
	private String keyDescription;
	private String valueDescription;

	/**
	 * Creates a new MoveToFrontMap with the key and value description
	 * specified (useful for logging)
	 * @param keyDescription the key description
	 * @param valueDescription the value description
	 */
	public MoveToFrontMap(String keyDescription, String valueDescription) {
		super();
		this.keyDescription = keyDescription;
		this.valueDescription = valueDescription;
	}
	
	/** 
	 * Obtain the corresponding value and
	 * moves to front its key
	 * @param key the key 
	 * @return the corresponding value
	 * @see java.util.TreeMap#get(java.lang.Object)
	 */
	public V getValue(K key) {
		V v = super.get(key);
		this.moveToFront(key);
		return v;
	}

	/**
	 * @deprecated use {@link #getValue(Object)} instead
	 */
	@Override
	public V get(Object key) {
		return super.get(key);
	}



	/**
	 * @see java.util.TreeMap#put(java.lang.Object, java.lang.Object)
	 */
	@Override
	public V put(K key, V newValue) {
		V oldValue = super.put(key, newValue);
		
		/* If a **new** cache entry */
		if(oldValue == null) {
			/* Free space if necessary */
			this.createToFront(key);
		} else { /* If a **modified** cache entry */
			/* Remove old cache entry */
			this.frequencyTable.remove(oldValue);
			this.moveToFront(key);
		}
		
		return oldValue;
	}
	
	private void createToFront(K k) {
		this.freeCache();
		this.frequencyTable.addFirst(k);
		log.debug("New " + valueDescription + "'s "+keyDescription+" ("+k+") ["+this.frequencyTable.size()+"/"+CACHE_ENTRIES_LIMIT+"]");
	}

	/**
	 * If the frequencyTable if full
	 * entries can be deleted from the back of the list
	 */
	private void freeCache() {
		if(this.frequencyTable.size() >= CACHE_ENTRIES_LIMIT) {
			K k = this.frequencyTable.pollLast();
			log.debug("Cleared the " + valueDescription + "'s "+keyDescription+" ("+k+").");
		}
	}

	private void moveToFront(K k) {
		this.frequencyTable.remove(k);
		this.frequencyTable.addFirst(k);
		log.debug("Moved the " + valueDescription + "'s "+keyDescription+" ("+k+") to front");
	}
	
}
