/*
 * Created on 9-gen-2005
 *
 */
package org.modelgen.commons.persistence.xml;

import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Property;
import org.modelgen.commons.beans.model.Reference;
import org.modelgen.commons.beans.scheme.SchemeObject;
import org.modelgen.commons.beans.scheme.SchemeObjectKey;
import org.modelgen.commons.beans.skolem.SkolemFunction;
import org.modelgen.commons.beans.skolem.SkolemFunctionArgument;
import org.modelgen.commons.persistence.PersistenceException;
import org.modelgen.commons.persistence.PersistenceHandler;
import org.modelgen.commons.persistence.PersistenceHandlerFactory;
import org.modelgen.commons.persistence.xml.jaxb.model.Model;
import org.modelgen.commons.persistence.xml.jaxb.model.ModelConstructType;
import org.modelgen.commons.persistence.xml.jaxb.model.ModelPropertyType;
import org.modelgen.commons.persistence.xml.jaxb.model.ModelReferenceType;
import org.modelgen.commons.persistence.xml.jaxb.nestedmodel.ConstructType;
import org.modelgen.commons.persistence.xml.jaxb.scheme.SchemeObjectPropertyType;
import org.modelgen.commons.persistence.xml.jaxb.scheme.SchemeObjectReferenceType;
import org.modelgen.commons.persistence.xml.jaxb.scheme.SchemeObjectType;
import org.modelgen.commons.persistence.xml.jaxb.skolemnew.ArgumentType;
import org.modelgen.commons.persistence.xml.jaxb.skolemnew.FunctionType;
import org.modelgen.commons.persistence.xml.jaxb.skolemnew.Functions;
import org.modelgen.facade.GeneralManagementFacade;

import java.io.File;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;

import org.apache.log4j.Logger;

/**
 * XML import functions for {@link org.modelgen.commons.beans.model.Model},
 * {@link org.modelgen.commons.beans.scheme.Scheme} and
 * {@link org.modelgen.commons.beans.skolem.SkolemFunction}
 * 
 * @author lucasantarelli
 *
 */
public class Importer {

	private static Logger log = Logger.getLogger(Importer.class);

	private static void printUsage() {
		System.out.println("To import a (flat xml) model, use: --model filename.xml.");
		System.out.println("To import a (nested xml) model, use: --nestedmodel filename.xml.");
		System.out.println("To import a scheme, use: --scheme filename.xml.");
		System.out.println("To import a set of skolem functions, use: --skolem filename.xml.");
		return;
	}

	/**
	 * ModelGen Importer.
	 *  
	 * @param args
	 * <ul>
	 * <li><tt>--model &lt;xmlfile&gt;</tt>: import a (flat xml) model</li>
	 * <li><tt>--nestedmodel &lt;xmlfile&gt;</tt>: import a (nested xml) model</li>
	 * <li><tt>--scheme &lt;xmlfile&gt;</tt>: import a scheme</li>
	 * <li><tt>--skolem &lt;xmlfile&gt;</tt>: import a set of skolem functions</li>
	 * </ul>
	 * @throws Exception errors in db connection or in import
	 */
	public static void main(String[] args) throws Exception {

		if (args.length < 1 || args[0].equals("--help")) {
			printUsage();
		}
		else {
			String inputfile = args[1];
			if (args[0].equals("--model")) {
				importModel(inputfile);
			}
			if (args[0].equals("--nestedmodel")) {
				importNestedModel(inputfile);
			}
			else if (args[0].equals("--scheme")) {
				importScheme(inputfile);
			}
			else if (args[0].equals("--skolem")) {
				importSkolemFunctions(inputfile);
			}
			else {
				System.out.println("Wrong usage.");
				printUsage();
			}
		}
	}

	/**
	 * Import a model from a xml flat file and store it in the db.
	 * @param fileName the input xml flat file
	 * @return the model imported (and stored in the db)
	 * @throws Exception errors in db connection or in import routine
	 */
	public static org.modelgen.commons.beans.model.Model importModel(String fileName) throws Exception {
		org.modelgen.commons.persistence.xml.jaxb.model.Model jaxbModel = null;
		//try {

		JAXBContext context = JAXBContext.newInstance("org.modelgen.commons.persistence.xml.jaxb.model", org.modelgen.commons.persistence.xml.jaxb.model.ObjectFactory.class.getClassLoader());
		Unmarshaller unmarshaller = context.createUnmarshaller();
		jaxbModel = (Model) unmarshaller.unmarshal(new File(fileName));
		log.debug("successfully unmarshalled the model from the xml input file.");
		org.modelgen.commons.beans.model.Model ourModel = XML2Model(jaxbModel);
		log.debug("successfully translated the JAXB object model into our own object model.");
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");
		if (ph.doesModelExist(ourModel.getName()) || ph.doesModelExistWithShortname(ourModel.getShortName())) {
			throw new PersistenceException("A model with the same keys (name or shortname) already exists, please ensure you are not importing an already imported model.");
		}
		//inserisco le informazioni sulla famiglia
		ourModel.setAncestor(false);
		ourModel.setFamilyOid(GeneralManagementFacade.getInstance().findFamily(ourModel.getModelSignature()));
		if(ourModel.getFamilyOid()!=0 && !ourModel.isAncestor())
			ourModel.setFamily(GeneralManagementFacade.getInstance().getFamilyAncestor(ourModel.getFamilyOid()));
				
		ph.storeModel(ourModel);
		log.debug("successfully stored the imported model into the database.");
		return ourModel;
	}

	/**
	 * Import a model from a xml nested file and store it in the db.
	 * @param fileName the input xml nested file
	 * @return the model imported (and stored in the db)
	 * @throws Exception errors in db connection or in import routine
	 */
	public static org.modelgen.commons.beans.model.Model importNestedModel(String fileName) throws Exception {
		org.modelgen.commons.persistence.xml.jaxb.nestedmodel.Model jaxbModel = null;
		//try {
		JAXBContext context = JAXBContext.newInstance("org.modelgen.commons.persistence.xml.jaxb.nestedmodel", org.modelgen.commons.persistence.xml.jaxb.nestedmodel.ObjectFactory.class.getClassLoader());
		Unmarshaller unmarshaller = context.createUnmarshaller();
		jaxbModel = (org.modelgen.commons.persistence.xml.jaxb.nestedmodel.Model) unmarshaller.unmarshal(new File(fileName));
		log.debug("successfully unmarshalled the model from the xml input file.");
		org.modelgen.commons.beans.model.Model ourModel = NestedXML2Model(jaxbModel);
		log.debug("successfully translated the JAXB object model into our own object model.");
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");
		if (ph.doesModelExist(ourModel.getName()) || ph.doesModelExistWithShortname(ourModel.getShortName())) {
			throw new PersistenceException("A model with the same keys (name or shortname) already exists, please ensure you are not importing an already imported model.");
		}
		//inserisco le informazioni sulla famiglia
		ourModel.setAncestor(false);
		ourModel.setFamilyOid(GeneralManagementFacade.getInstance().findFamily(ourModel.getModelSignature()));
		if(ourModel.getFamilyOid()!=0 && !ourModel.isAncestor())
			ourModel.setFamily(GeneralManagementFacade.getInstance().getFamilyAncestor(ourModel.getFamilyOid()));
		
		ph.storeModel(ourModel);
		log.debug("successfully stored the imported model into the database.");
		return ourModel;
	}

	/**
	 * Import a scheme from a xml file and store it in the db.
	 * @param fileName the input xml file
	 * @return the scheme imported (and stored in the db)
	 * @throws Exception errors in db connection or in import routine
	 */
	public static org.modelgen.commons.beans.scheme.Scheme importScheme(String fileName) throws Exception {
		org.modelgen.commons.persistence.xml.jaxb.scheme.Scheme jaxbScheme = null;
		//try {
		JAXBContext context = JAXBContext.newInstance("org.modelgen.commons.persistence.xml.jaxb.scheme", org.modelgen.commons.persistence.xml.jaxb.scheme.ObjectFactory.class.getClassLoader());
		Unmarshaller unmarshaller = context.createUnmarshaller();
		jaxbScheme = (org.modelgen.commons.persistence.xml.jaxb.scheme.Scheme) unmarshaller.unmarshal(new File(fileName));
		log.debug("successfully unmarshalled the scheme from the xml input file.");
		org.modelgen.commons.beans.scheme.Scheme ourScheme = XML2Scheme(jaxbScheme);
		log.debug("successfully translated the JAXB object scheme into our own object scheme.");
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");
		ph.storeScheme(ourScheme);
		log.debug("successfully stored the imported scheme into the database.");
		return ourScheme;
		/*}
		catch (JAXBException je) {
			je.printStackTrace();
			return null;
		}*/

	}

	/**
	 * Import a set of skolem functions from a xml file and store it in the db.
	 * @param fileName the input xml file
	 * @return the set of skolem functions imported (and stored in the db)
	 * @throws Exception errors in db connection or in import routine
	 */
	public static Collection<SkolemFunction> importSkolemFunctions(String fileName) throws Exception {
		Functions jaxbList = null;
		//try {
		JAXBContext context = JAXBContext.newInstance("org.modelgen.commons.persistence.xml.jaxb.skolemnew", org.modelgen.commons.persistence.xml.jaxb.skolemnew.ObjectFactory.class.getClassLoader());
		Unmarshaller unmarshaller = context.createUnmarshaller();
		jaxbList = (Functions) unmarshaller.unmarshal(new File(fileName));
		log.debug("successfully unmarshalled the skolem function set from the xml input file.");
		Collection<SkolemFunction> ourList = XML2SkolemFunctions(jaxbList);
		log.debug("successfully translated the JAXB object list into our own object list.");
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");
		LinkedList<SkolemFunction> storedList = new LinkedList<SkolemFunction>();
		org.modelgen.commons.beans.model.Model superModel = ph.getSuperModel();
		//Okay, let's try to insert one function at a time.
		for (Iterator<SkolemFunction> itFunctions = ourList.iterator(); itFunctions.hasNext();) {
			SkolemFunction skf = itFunctions.next();
			if (ph.getSkolemFunction(superModel, skf.getOwnerConstruct(), skf.getName()) != null) {
				//The function already exists in the database, let's skip it.
				//log.info("skipping function " + skf.getName() + " for construct "
				//		+ skf.getOwnerConstruct().getName() + " because it is already in the database.");
				throw new Exception("The provided skolem function (" + skf.getName() + ") already exists in the database.");
			}
			else { //The skolem function does not exist in the db, we can import it.
				log.debug("storing skolem function " + skf.getName() + " for construct "
						+ skf.getOwnerConstruct().getName() + " in the db.");
				ph.storeSkolemFunction(superModel, skf);
				storedList.add(skf);
			}
		}
		log.debug("successfully stored the imported skolem functions into the database.");
		return storedList;
		/*}
		catch (JAXBException je) {
			je.printStackTrace();
			return null;
		}*/
	}

	/**
	 * Computes the skolem function form from its xml rappresentation 
	 * @param srcFunctions the xml skolem function to transform
	 * @return the rappresentation of the specified skolem function
	 * @throws Exception errors in db connection or in the XML/Java binding
	 */
	public static Collection<SkolemFunction> XML2SkolemFunctions(Functions srcFunctions) throws Exception {
		org.modelgen.commons.beans.model.Model superModel = PersistenceHandlerFactory.getPersistenceHandler("DBHandler").getSuperModel();
		LinkedList<SkolemFunction> functions = new LinkedList<SkolemFunction>();
		for (Iterator itFunctions = srcFunctions.getFunction().iterator(); itFunctions.hasNext();) {
			FunctionType xmlFunction = (FunctionType) itFunctions.next();
			SkolemFunction function = new SkolemFunction();
			//Check if the data is valid. If it is, add to the list.
			if (!SkolemFunction.isValidName(xmlFunction.getName())) {
				throw new Exception("The provided name (" + xmlFunction.getName() + ") is not valid for a skolem function.");
			}
			function.setName(xmlFunction.getName());
			Construct smConstruct = superModel.getConstruct(xmlFunction.getTargetConstruct());
			if (smConstruct == null) {
				throw new Exception("The construct " + xmlFunction.getTargetConstruct() + " does not exist in the supermodel.");
			}
			function.setOwnerConstruct(smConstruct);
			for (Iterator itArguments = xmlFunction.getArguments().getArgument().iterator(); itArguments.hasNext();) {
				ArgumentType xmlArgument = (ArgumentType) itArguments.next();
				Construct sourceConstruct = superModel.getConstruct(xmlArgument.getSourceConstruct());
				if (sourceConstruct == null) {
					throw new Exception("The provided source construct (" + xmlArgument.getSourceConstruct() + ") is not valid for a skolem function argument.");
				}
				SkolemFunctionArgument skfa = new SkolemFunctionArgument(-1, xmlArgument.getName(), xmlArgument.getType(), sourceConstruct);
				if (!SkolemFunctionArgument.isValidName(xmlArgument.getName())) {
					throw new Exception("The provided name (" + xmlArgument.getName() + ") is not valid for a skolem function argument.");
				}
				if (!SkolemFunctionArgument.isValidType(xmlArgument.getType())) {
					throw new Exception("The provided type (" + xmlArgument.getType() + ") is not valid for a skolem function argument.");
				}
				if (function.contains(skfa)) {
					throw new Exception("The skolem function " + function.getName() + " already contains an argument " + xmlArgument.getName() + ".");
				}
				//The fields are valid, the argument isn't present yet, the construct exists. Okay.
				function.addArgument(skfa);
			}
			//All the arguments have been analyzed
			functions.add(function);
		}
		if (functions.isEmpty())
			return null;
		else
			return functions;
	}

	/**
	 * Computes the scheme form from its xml rappresentation 
	 * @param srcScheme the xml scheme to transform
	 * @return the rappresentation of the specified scheme
	 * @throws Exception errors in db connection or in the XML/Java binding
	 */
	public static org.modelgen.commons.beans.scheme.Scheme XML2Scheme(org.modelgen.commons.persistence.xml.jaxb.scheme.Scheme srcScheme) throws Exception {
		org.modelgen.commons.beans.model.Model model = null; //We'll use these later.
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");

		//As for the model, the first validaton happens at the JAXB level.
		log.debug("initiating scheme validation.");
		if (!org.modelgen.commons.beans.scheme.Scheme.isValidName(srcScheme.getName())) {
			throw new Exception("The provided name (" + srcScheme.getName() + ") is not valid for a scheme.");
		}

		//Now to obtain the model, we have to do a bit of work. Alternatively we could add a method in PersistenceHandler to do this.
		//TODO: discuss about moving this into PersistenceHandler itself.
		Collection<org.modelgen.commons.beans.model.Model> models = ph.getModels();
		for (org.modelgen.commons.beans.model.Model aux : models) {
			if (aux.getShortName().equals(srcScheme.getModel())) {
				model = ph.getModel(aux.getOID()); //And this is sure to exist!
				break; //We have what we want.
			}
		}
		if (model == null) {
			throw new Exception("The provided scheme (" + srcScheme.getName() + ") claims to be instance of a model ("
					+ srcScheme.getModel() + ") which does not exist in the database.");
		}

		String name = srcScheme.getName();
		org.modelgen.commons.beans.scheme.Scheme scheme = new org.modelgen.commons.beans.scheme.Scheme(0, name, model, false);

		//Okay, the scheme has a valid name and claims to be instance of an existing model. Let's check the objects now.
		//There can be objects with unfulfilled dependencies, like in the import model routine, so we'll use two auxiliary lists.
		LinkedList<SchemeObjectType> checkable = new LinkedList<SchemeObjectType>(srcScheme.getObjects().getSchemeobject());
		LinkedList<SchemeObjectType> failedDeps = new LinkedList<SchemeObjectType>();
		Hashtable<SchemeObjectKey, String> keyTable = new Hashtable<SchemeObjectKey, String>(); //We need this since the keys in memory will be different from the ones in the file!
		while (true) {
			//Iterate on the list.
			log.debug("iterating on the scheme objects list.");

			for (Iterator<SchemeObjectType> itObjects = checkable.iterator(); itObjects.hasNext();) {
				SchemeObjectType xmlso = itObjects.next();
				log.debug("analyzing XML scheme object: " + xmlso.getName() + " (" + xmlso.getKey() + ").");
				boolean dependenciesAreOk = true; //The default is true because there are objects with no dependencies to fulfill!
				//In opposition to what we did for the model import (creating the valid object at the end) we need to do it as we go on.
				if (!SchemeObject.isValidName(xmlso.getName())) {
					throw new Exception("The provided name (" + xmlso.getName() + ") is not valid for a scheme object.");
				}
				Construct ourConstruct = model.getConstruct(xmlso.getConstruct());
				if (ourConstruct == null) {
					throw new Exception("The provided scheme object (" + xmlso.getName()
							+ ") claims to be instance of a model construct (" + xmlso.getConstruct()
							+ ") which does not exist in the given model (" + model.getShortName() + ").");
				}
				SchemeObject validSchemeObject = new SchemeObject(0, xmlso.getName(), ourConstruct);
				log.debug("the object seems good, moving to its properties.");
				//TODO: do we need a check on the key?
				//Let's see the properties.
				for (Iterator itObjectProperties = xmlso.getProperties().getProperty().iterator(); itObjectProperties.hasNext();) {
					SchemeObjectPropertyType xmlsop = (SchemeObjectPropertyType) itObjectProperties.next();
					log.debug("examining property " + xmlsop.getName() + ".");
					Property ourProperty = ourConstruct.getProperty(xmlsop.getName());
					if (ourProperty == null) {
						throw new Exception("The declared property (" + xmlsop.getName()
								+ ") does not exist in the referenced construct (" + xmlso.getName() + ").");
					}
					if (xmlsop.getValue().length() == 0) {
						throw new Exception("The value for the property " + xmlsop.getName()
								+ " of the object " + xmlso.getName() + " is not valid.");
					}
					//The property is valid.
					validSchemeObject.setProperty(xmlsop.getName(), xmlsop.getValue());
				}
				log.debug("now moving to the references.");
				//And now the references. This is where we may encounter dependency issues.
				//TODO: check the reference type against the super model reference type.
				for (Iterator itObjectReferences = xmlso.getReferences().getReference().iterator(); itObjectReferences.hasNext();) {
					SchemeObjectReferenceType xmlsor = (SchemeObjectReferenceType) itObjectReferences.next();
					log.debug("examining reference " + xmlsor.getName() + ".");
					//Does the reference exist?
					Reference ourReference = ourConstruct.getReference(xmlsor.getName());
					if (ourReference == null) {
						throw new Exception("The declared reference (" + xmlsor.getName()
								+ ") does not exist in the referenced construct (" + xmlso.getName() + ").");
					}
					//Now there can be multiple possibilities. If the target object has been loaded in the scheme,
					//it means this reference is valid and we can load this object too.
					//If the object is not present in the scheme, it's a failed dependency. This can be a temporary
					//failure or a permanent one, so this is where we may put objects in failedDeps (to iterate again).
					SchemeObject targetObject = null;
					for (SchemeObject temp : scheme.getAbstractConstructs()) {
						String oldKey = keyTable.get(temp.getKey());
						log.debug("newkey: " + temp.getKey().toString() + ", oldkey: " + oldKey + ", xmlsor value: " + xmlsor.getValue());
						if (/*temp.getKey().toString()*/oldKey.equals(xmlsor.getValue())) { //xmlsor.getValue() contains the SOK.
							targetObject = temp;
							break;
						}
					}
					// if the reference value is MDM_UNSPECIFIED it is valid and must be set to null.
					if ((targetObject == null)&&(!xmlsor.getValue().equals(SchemeObject.MDM_UNSPECIFIED))) {
						dependenciesAreOk = false;
						log.debug("this reference points to nowhere, this is a (temporarily?) unfulfilled dependency.");
					}
					else { //The dependency is fulfilled!
						validSchemeObject.setReference(xmlsor.getName(), targetObject);
					}
				}
				if (dependenciesAreOk) { //If everything went good, we can add the object to the tree.
					log.debug("the object is fine, adding it to the tree.");
					scheme.addObject(validSchemeObject);
					//Let's store its old key value.
					keyTable.put(validSchemeObject.getKey(), xmlso.getKey());
				}
				else {
					log.debug("the object wasn't fine, adding it to the failed dependency list");
					failedDeps.add(xmlso);
				}
			}
			if (failedDeps.size() == 0) { //All the objects were added! Good!
				break;
			}
			else { //Some dependencies were unfulfilled.
				log.debug("there are unfulfilled dependencies. failedDeps:" + failedDeps.size() + ", checkable: " + checkable.size());
				if (failedDeps.size() == checkable.size()) {
					//This iteration didn't add any object to the tree: we're stuck.
					String exceptionMessage = "The following scheme objects have unfulfilled dependencies:\n";
					for (Iterator<SchemeObjectType> itFailed = failedDeps.iterator(); itFailed.hasNext();) {
						SchemeObjectType failure = itFailed.next();
						exceptionMessage += failure.getName() + " (" + failure.getKey() + ")\n";
					}
					throw new Exception(exceptionMessage);
				}
				else { //There are unfulfilled dependencies, but we can hope to fix them in the next step.
					checkable = failedDeps;
					failedDeps = new LinkedList<SchemeObjectType>();
					log.debug("we're going to iterate again.");
				}
			}
		}
		return scheme;
	}

	/**
	 * Computes the model form from its xml rappresentation 
	 * @param srcModel the xml model to transform
	 * @return the rappresentation of the specified model
	 * @throws Exception errors in db connection or in the XML/Java binding
	 */
	public static org.modelgen.commons.beans.model.Model XML2Model(Model srcModel) throws Exception {
		org.modelgen.commons.beans.model.Model superModel = PersistenceHandlerFactory.getPersistenceHandler("DBHandler").getSuperModel();

		//The first validity check happens at JAXB level, so if we reach this point, the model is well-formed in regards to XSD.
		//We must validate it in terms of valid internal references (D and R references pointing to existing and allowed constructs)
		//and external references (to the supermodel).
		log.debug("The name is: " + srcModel.getName());
		if (!org.modelgen.commons.beans.model.Model.isValidName(srcModel.getName())) {
			throw new Exception("The provided name (" + srcModel.getName() + ") is not valid for a model.");			
		}
		String name = srcModel.getName();
		org.modelgen.commons.beans.model.Model model = new org.modelgen.commons.beans.model.Model(0, name);

		if (!org.modelgen.commons.beans.model.Model.isValidShortName(srcModel.getShortName())) {
			throw new Exception("The provided short name (" + srcModel.getShortName() + ") is not valid for a model.");
		}
		model.setShortName(srcModel.getShortName());

		//The basic data is in, let's move to the constructs.
		//In order to be valid, a given construct must have a new name, must be mapped to a valid super-construct and must fulfill dependencies.
		//Since we don't know the order of the constructs, we must check for dependencies thinking that they can be fulfilled later.
		//Algorithm: we use two lists: a (starts full of wannabe constructs) and b (starts empty).
		//Iterate over a, if an element is good, put it in the model, if it has missing deps, it goes in b.
		//When a is over, if b contains elements, a = b, b = empty and start again.
		//if, at a end of an iteration on a, no elements were added to the model, we're stuck and can throw an exception.
		LinkedList<ModelConstructType> checkable = new LinkedList<ModelConstructType>(srcModel.getConstructs().getConstruct()); //NOTE: new LinkedList() because we don't want to change the real list!
		LinkedList<ModelConstructType> failedDeps = new LinkedList<ModelConstructType>();

		while (true/*checkable.size() != failedDeps.size()*/) { //treue because we exit with a break or thrown Exception
			for (int i = 0; i < checkable.size(); i++) {
				ModelConstructType srcConstruct = checkable.get(i);
				log.debug("checking validity of XML construct: "
						+ srcConstruct.getName() + ", mapping to: "
						+ srcConstruct.getMappingToConstruct().getName());
				//Is it valid?
				if (!Construct.isValidName(srcConstruct.getName())) {
					//ERROR.
					throw new Exception("The provided name (" + srcConstruct.getName() + ") is not valid for a construct.");
				}
				//Okay, the name is good. NOTE: I don't remember if does check if there's another construct by the same name.
				Construct smConstruct = superModel.getConstruct(srcConstruct.getMappingToConstruct().getName());
				if (smConstruct == null) {
					throw new Exception("Invalid mapping, the provided construct ("
							+ srcConstruct.getMappingToConstruct().getName()
							+ ") does not exist in the super model.");
				}
				//The following call checks if the mapping is valid (it checks dependencies! Kawaii)
				if (!model.getAvailableConstructs(superModel).contains(smConstruct)) {
					//Dependencies aren't met.
					failedDeps.add(srcConstruct);
					log.debug("XML construct " + srcConstruct.getName() + " has unfulfilled dependencies, putting in standby list.");
				}
				else {
					//Ok, the name is valid and the dependencies are met: it's a valid construct, let's add it to the model.
					Construct validConstruct = new Construct(0,srcConstruct.getName());
					validConstruct.setSignature(srcConstruct.getSignature());
					validConstruct.setUserBonds(srcConstruct.getBonds());
					LinkedList<Construct> mappings = new LinkedList<Construct>(); //TODO: remove this when 1:N become 1:1 mappings.
					mappings.add(smConstruct);
					validConstruct.setSuperModelConstructs(mappings);

					model.addConstruct(validConstruct);
					log.debug("XML construct " + srcConstruct.getName() + " is valid and has been inserted in the model.");
				}
				//we're ready for the next construct!
			}
			//Okay, we've now examined all the items in "checkable". failedDeps is either empty (good!) or it contains some items.
			if (failedDeps.size() == 0) { //Every construct was added, good!
				log.debug("All the XML constructs have been inserted in the model.");
				break;
			}
			else { //Some dependencies weren't met. If failedDeps contains all the objects in checkable, then we're stuck.
				log.debug("failedDeps size: " + failedDeps.size() + ", checkable size: " + checkable.size());
				if (failedDeps.size() == checkable.size()) {
					String exceptionMsg = "The dependencies cannot be fulfilled, the input model is wrong.\n"
						+ "The following constructs have unfulfilled dependencies:\n";
					for (int i = 0; i < failedDeps.size(); i++) {
						exceptionMsg += failedDeps.get(i).getName() + "\n";
					}
					throw new Exception(exceptionMsg);
				}
				else { //Ok, we're here to give it another try.
					checkable = failedDeps;
					failedDeps = new LinkedList<ModelConstructType>();
				}
			}
		}
		//Wow, and this was for the constructs only! Now we need properties and dependencies!
		//In order for a property to be valid, it must have a valid name, point to a valid construct (which is already
		//in the model by now) and have a super model property which is valid (which means the super model property already
		//exists in the super model and belongs to a construct mapped to the one owning the given (model) property.
		//This time, we don't have dependency issues, since all the constructs have already been loaded!
		log.debug("starting import of property objects.");
		for (int i = 0; i < srcModel.getProperties().getProperty().size(); i++) {
			ModelPropertyType srcProperty = (ModelPropertyType) srcModel.getProperties().getProperty().get(i);
			String propertyName = srcProperty.getName();
			log.debug("Analizing property with name: " + propertyName + ", belonging to: " + srcProperty.getBelongingTo());
			if (!Property.isValidName(propertyName)) {
				throw new Exception("The provided name (" + propertyName + ") is not valid for a property.");
			}
			if (!Property.isValidType(srcProperty.getType())) {
				throw new Exception("The provided type (" + srcProperty.getType() + ") is not valid for a property.");
			}
			Construct ownerConstruct = model.getConstruct(srcProperty.getBelongingTo());
			if (ownerConstruct == null) {
				throw new Exception("The property '" + propertyName
						+ "' claims to belong to a construct which does not exist in the model ("
						+ srcProperty.getBelongingTo() + ").");
			}
			//So, the property has a valid name, valid type and claims to belong to an existing construct.
			//Is the mapping valid (i.e.: can this construct own such a property?)
			Construct smConstruct = superModel.getConstruct(srcProperty.getMappingToProperty().getBelongingTo());
			if (smConstruct == null) {
				throw new Exception("The provided property (" + propertyName
						+ ") claims to belong to a super model property belonging to a not existing construct ("
						+ srcProperty.getMappingToProperty().getBelongingTo() + ").");
			}
			//The super model construct exists, but is it mapped to the ownerConstruct?
			if (!ownerConstruct.getSuperModelConstructs().contains(smConstruct)) {
				throw new Exception("The provided property (" + propertyName
						+ ") claims to be mapped to a property belonging to a super model construct ("
						+ smConstruct.getName() + ") which is not mapped to the construct this property claims to belong to ("
						+ ownerConstruct.getName() + ").");
			}
			//Ok, the super model construct exists, but does it own a super model property with the required name?
			Property smProperty = smConstruct.getProperty(srcProperty.getMappingToProperty().getName());
			if (smProperty == null) {
				throw new Exception ("The provided property (" + propertyName
						+ ") claims to map to a super model property (" + srcProperty.getMappingToProperty().getName()
						+ ") belonging to a construct (" + srcProperty.getMappingToProperty().getBelongingTo()
						+ ") which doesn't contain such a property.");
			}
			//Ok, but... is it already mapped?
			if (!model.getAvailableProperties(ownerConstruct).contains(smProperty)) {
				throw new Exception("The provided property (" + propertyName + ") cannot be added to the construct it claims to belong to.");
			}

			//Ok, I think we can say it's a valid property. :-)
			Property validProperty = new Property(0,propertyName,smProperty);
			validProperty.setOwnerConstruct(ownerConstruct);
			validProperty.setType(srcProperty.getType());

			ownerConstruct.addProperty(validProperty);
		}
		for (Construct element : model.getAbstractConstructs()) {
			log.debug("un costrutto: " + element.getName());

		}
		log.debug("finished importing properties, moving to references.");
		for(int i = 0; i < srcModel.getReferences().getReference().size(); i++) {
			ModelReferenceType srcReference = (ModelReferenceType) srcModel.getReferences().getReference().get(i);
			String referenceName = srcReference.getName();
			log.debug("checking reference with name " + referenceName);
			if (!Reference.isValidName(referenceName)) {
				throw new Exception("The provided name (" + referenceName + " is not valid for a reference.");
			}
			Construct ownerConstruct = model.getConstruct(srcReference.getBelongingTo());
			if (ownerConstruct == null) {
				throw new Exception("The provided reference (" + referenceName
						+ ") claims to belong to a construct (" + srcReference.getBelongingTo()
						+ ") which does not exist in the model.");
			}
			log.debug("reference claim to belong to contruct " + srcReference.getBelongingTo()
					+ " is verified on construct " + ownerConstruct.getName() + " (OID: " + ownerConstruct.getOID() + ")");
			Construct targetConstruct = model.getConstruct(srcReference.getTargeting());
			if (targetConstruct == null) {
				throw new Exception("The provided reference (" + referenceName + ") claims to point to a construct ("
						+ srcReference.getTargeting() + ") which does not exist in the model.");
			}
			log.debug("reference claim to target contruct " + srcReference.getTargeting()
					+ " is verified on construct " + targetConstruct.getName() + " (OID: " + targetConstruct.getOID() + ")");
			//From a "model" point of view, the reference is good. Let's check the super model reference is claims to be mapped to.
			//For this to be valid, there must exist a reference with the same name in the super model construct which maps to the
			//ownerConstruct (and it must not be already declared!). Not only this: the target construct must be mapped to the same
			//construct in the super model reference mapping.
			//The check is: get the onwerconstruct, move to the super model construct. Does it have a reference with the declared
			//name? Ok, is the super model reference target construct mapped to the reference target to? If yes, OK!
			Construct smOwnerConstruct = ownerConstruct.getSuperModelConstructs().iterator().next();
			Construct smTargetConstruct = targetConstruct.getSuperModelConstructs().iterator().next();
			//Those two were got from the object model, they are good by definition.
			Reference smReference = smOwnerConstruct.getReference(srcReference.getMappingToReference().getName());
			if (smReference == null) {
				//The mapping is wrong.
				throw new Exception("The provided reference (" + referenceName + ") claims to map to a super model reference ("
						+ srcReference.getMappingToReference().getName() + ") which does not exist in the super-construct ("
						+ smOwnerConstruct.getName() + ").");
			}
			if (!srcReference.getMappingToReference().getTargeting().equals(smTargetConstruct.getName())) {
				throw new Exception("The provided reference (" + referenceName + ") claims to target a construct ("
						+ srcReference.getTargeting() + ") which is not mapped to a valid super model construct.");
			}
			//Okay, it looks like it's a valid reference.
			Reference validReference = new Reference(0,referenceName,smReference);
			validReference.setOwnerConstruct(ownerConstruct);
			validReference.setType(srcReference.getType());
			validReference.setTargetConstruct(targetConstruct);

			ownerConstruct.addReference(validReference);
		}
		log.debug("finished checking references.");
		//Wow, it was long and hard, but we're done!
		return model;
	}

	/**
	 * Computes the model form from its xml nested rappresentation 
	 * @param srcModel the xml nested model to transform
	 * @return the rappresentation of the specified model
	 * @throws Exception errors in db connection or in the XML/Java binding
	 */
	public static org.modelgen.commons.beans.model.Model NestedXML2Model(org.modelgen.commons.persistence.xml.jaxb.nestedmodel.Model srcModel) throws Exception {
		//You can check the XML2Model method to get comments. This method is quite similar, only differences will be commented.
		org.modelgen.commons.beans.model.Model superModel = PersistenceHandlerFactory.getPersistenceHandler("DBHandler").getSuperModel();

		log.debug("The name is: " + srcModel.getName());
		if (!org.modelgen.commons.beans.model.Model.isValidName(srcModel.getName())) {
			throw new Exception("The provided name (" + srcModel.getName() + ") is not valid for a model.");			
		}

		String name = srcModel.getName();
		org.modelgen.commons.beans.model.Model model = new org.modelgen.commons.beans.model.Model(0, name);

		if (!org.modelgen.commons.beans.model.Model.isValidShortName(srcModel.getShortName())) {
			throw new Exception("The provided short name (" + srcModel.getShortName() + ") is not valid for a model.");
		}
		model.setShortName(srcModel.getShortName());

		LinkedList<ConstructType> checkable = new LinkedList<ConstructType>(srcModel.getConstructs().getConstruct());
		LinkedList<ConstructType> failedDeps = new LinkedList<ConstructType>();

		while (true/*checkable.size() != failedDeps.size()*/) { 
			for (int i = 0; i < checkable.size(); i++) {
				org.modelgen.commons.persistence.xml.jaxb.nestedmodel.ConstructType srcConstruct =
					checkable.get(i);
				log.debug("checking validity of XML construct: "
						+ srcConstruct.getName() + ", mapping to: "
						+ srcConstruct.getMetaConstruct().getName());
				//Is it valid?
				if (!Construct.isValidName(srcConstruct.getName())) {
					//ERROR.
					throw new Exception("The provided name (" + srcConstruct.getName() + ") is not valid for a construct.");
				}
				Construct smConstruct = superModel.getConstruct(srcConstruct.getMetaConstruct().getName());
				if (smConstruct == null) {
					throw new Exception("Invalid mapping, the provided construct ("
							+ srcConstruct.getMetaConstruct().getName()
							+ ") does not exist in the super model.");
				}
				//The following call checks if the mapping is valid (it checks dependencies! Kawaii)
				if (!model.getAvailableConstructs(superModel).contains(smConstruct)) {
					//Dependencies aren't met.
					failedDeps.add(srcConstruct);
					log.debug("XML construct " + srcConstruct.getName() + " has unfulfilled dependencies, putting in standby list.");
				}
				else {
					//Ok, the name is valid and the dependencies are met: it's a valid construct, let's add it to the model.
					Construct validConstruct = new Construct(0,srcConstruct.getName());
					validConstruct.setSignature(srcConstruct.getSignature());
					validConstruct.setUserBonds(srcConstruct.getBonds());
					LinkedList<Construct> mappings = new LinkedList<Construct>(); //TODO: remove this when 1:N become 1:1 mappings.
					mappings.add(smConstruct);
					validConstruct.setSuperModelConstructs(mappings);

					model.addConstruct(validConstruct);
					log.debug("XML construct " + srcConstruct.getName() + " is valid and has been inserted in the model.");
					//Here comes the real difference between methods: we check properties and references directly here.
					log.debug("starting import of property objects.");
					for (int j = 0; j < srcConstruct.getProperties().getProperty().size(); j++) {
						org.modelgen.commons.persistence.xml.jaxb.nestedmodel.PropertyType srcProperty =
							(org.modelgen.commons.persistence.xml.jaxb.nestedmodel.PropertyType)
							srcConstruct.getProperties().getProperty().get(j);
						String propertyName = srcProperty.getName();
						log.debug("Analizing property with name: " + propertyName + ", belonging to: " + srcConstruct.getName());
						if (!Property.isValidName(propertyName)) {
							throw new Exception("The provided name (" + propertyName + ") is not valid for a property.");
						}
						//The type will be taken from the super model.
						//Does the super model property, with the required name, exist?
						Property smProperty = smConstruct.getProperty(srcProperty.getMetaProperty().getName());
						if (smProperty == null) {
							throw new Exception ("The provided property (" + propertyName
									+ ") claims to map to a super model property (" + srcProperty.getMetaProperty().getName()
									+ ") belonging to a construct (" + smConstruct.getName()
									+ ") which doesn't contain such a property.");
						}
						//Ok, but... is it already mapped?
						if (!model.getAvailableProperties(validConstruct).contains(smProperty)) {
							throw new Exception("The provided property (" + propertyName + ") cannot be added to the construct it claims to belong to.");
						}
						//Ok, I think we can say it's a valid property. :-)
						Property validProperty = new Property(0,propertyName,smProperty);
						validProperty.setOwnerConstruct(validConstruct);
						//The type is taken directly from the supermodel. :-( Sayonara customization. :-(
						validProperty.setType(smProperty.getType());

						validConstruct.addProperty(validProperty);
					}
					log.debug("finished importing properties, moving to references.");
					for(int k = 0; k < srcConstruct.getReferences().getReference().size(); k++) {
						org.modelgen.commons.persistence.xml.jaxb.nestedmodel.ReferenceType srcReference =
							(org.modelgen.commons.persistence.xml.jaxb.nestedmodel.ReferenceType)
							srcConstruct.getReferences().getReference().get(k);
						String referenceName = srcReference.getName();
						log.debug("checking reference with name " + referenceName);
						if (!Reference.isValidName(referenceName)) {
							throw new Exception("The provided name (" + referenceName + " is not valid for a reference.");
						}
						Construct targetConstruct = model.getConstruct(srcReference.getTo());
						if (targetConstruct == null) {
							throw new Exception("The provided reference (" + referenceName + ") claims to point to a construct ("
									+ srcReference.getTo() + ") which does not exist in the model.");
						}
						//Those two were got from the object model, they are good by definition.
						Reference smReference = smConstruct.getReference(srcReference.getMetaReference().getName());
						if (smReference == null) {
							//The mapping is wrong.
							throw new Exception("The provided reference (" + referenceName + ") claims to map to a super model reference ("
									+ srcReference.getMetaReference().getName() + ") which does not exist in the super-construct ("
									+ smConstruct.getName() + ").");
						}
						//Okay, it looks like it's a valid reference.
						Reference validReference = new Reference(0,referenceName,smReference);
						validReference.setOwnerConstruct(validConstruct);
						validReference.setType(smReference.getType()); //From the supermodel
						validReference.setTargetConstruct(targetConstruct);

						validConstruct.addReference(validReference);
					}
					log.debug("finished checking references.");
				}
				//we're ready for the next construct!
			}
			//Okay, we've now examined all the items in "checkable". failedDeps is either empty (good!) or it contains some items.
			if (failedDeps.size() == 0) { //Every construct was added, good!
				log.debug("All the XML constructs have been inserted in the model.");
				break;
			}
			else { //Some dependencies weren't met. If failedDeps contains all the objects in checkable, then we're stuck.
				log.debug("failedDeps size: " + failedDeps.size() + ", checkable size: " + checkable.size());
				if (failedDeps.size() == checkable.size()) {
					String exceptionMsg = "The dependencies cannot be fulfilled, the input model is wrong.\n"
						+ "The following constructs have unfulfilled dependencies:\n";
					for (int i = 0; i < failedDeps.size(); i++) {
						exceptionMsg += ((ModelConstructType) failedDeps.get(i)).getName() + "\n";
					}
					throw new Exception(exceptionMsg);
				}
				else { //Ok, we're here to give it another try.
					checkable = failedDeps;
					failedDeps = new LinkedList<ConstructType>();
				}
			}
		}
		return model;
	}

}
