package org.modelgen.facade;

import org.modelgen.jpf.core.CorePlugin;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import javax.swing.ImageIcon;

import org.java.plugin.PluginManager;

/**
 * <p>
 * Contains <i>the only</i> functions you can use to create 
 * and open files for your plugin.
 * </p>
 * 
 * <p>
 * For <i>read-only</i> files (icons), see {@link #openFile(Class, String)}
 * and {@link #openImageIcon(Class, String)}.
 * </p>
 * 
 * <p>
 * For <i>read-write</i> files, see {@link #createFileIfNotExist(Class, String, String)}
 * and {@link #createDirIfNotExist(Class, String, String)}
 * </p>
 * 
 * @author Riccardo Pietrucci
 *
 */
public class ResourceFunctions {
	
	/**
	 * Opens the read-only icon specified, looking for 
	 * library type "resources" declared in <tt>plugin.xml</tt>.
	 * 
	 * <p>
	 * For example, if the <tt>plugin.xml</tt> file contains:
	 * <pre>&lt;runtime&gt;
	 &lt;library type="resources" path="icons/" id="icons"/&gt;
	 &lt;library type="resources" path="config/" id="config"/&gt;
	    &lt;/runtime&gt;
	 </pre>
	 * this function looks for the icon in <tt>icons/</tt> and in <tt>config/</tt>.
	 * </p>
	 * 
	 * @param c the context class (a class belonging to the context plugin)
	 * @param image the path of the image to open (relative to "resources")
	 * @return the image opened, <tt>null</tt> otherwise
	 */
	public static ImageIcon openImageIcon(Class<?> c, String image) {
		URL url = c.getClassLoader().getResource(image);
		return url != null ? new ImageIcon(url) : null;
	}
	
	/**
	 * Opens the read-only file specified, looking for 
	 * library type "resources" declared in <tt>plugin.xml</tt>.
	 * 
	 * <p>
	 * For example, if the <tt>plugin.xml</tt> file contains:
	 * <pre>&lt;runtime&gt;
   &lt;library type="resources" path="icons/" id="icons"/&gt;
   &lt;library type="resources" path="config/" id="config"/&gt;
&lt;/runtime&gt;
	 </pre>
	 * this function looks for the icon in <tt>icons/</tt> and in <tt>config/</tt>.
	 * </p>
	 * 
	 * @param c the context class (a class belonging to the context plugin)
	 * @param filename the path of the file to open (relative to "resources")
	 * @return the file opened, <tt>null</tt> otherwise
	 */
	public static InputStream openFile(Class<?> c, String filename) {
		InputStream is = c.getClassLoader().getResourceAsStream(filename);
		return is;
	}
	
	/**
	 * <p>
	 * Opens the read-write file specified, looking for 
	 * library type "resources" declared in <tt>plugin.xml</tt>.
	 * </p>
	 * 
	 * <p>
	 * For example, if the <tt>plugin.xml</tt> file contains:
	 * <pre>&lt;runtime&gt;
   &lt;library type="resources" path="icons/" id="icons"/&gt;
   &lt;library type="resources" path="config/" id="config"/&gt;
&lt;/runtime&gt;
	 </pre>
	 * this function looks for the file in <tt>icons/</tt> and in <tt>config/</tt>.
	 * </p>
	 * 
	 * The behaviour is:
	 * <ul>
	 * <li>
	 * if the file <i>is found</i> in "resources":
	 * it <i>copies</i> the file in the plugin's data folder if not yet exists
	 * in the plugin's data folder
	 * </li> 
	 * <li>
	 * if the file is <i>not</i> found in "resources": 
	 * it creates <i>a new empty file</i> with the specified name 
	 * in the plugin's data folder if not yet exists.
	 * </li>
	 * </ul>
	 * 
	 * @param c the context class (a class belonging to the context plugin)
	 * @param filename the path of the file to open (relative to "resources")
	 * @param pluginID the plugin ID
	 * @return the file opened (or created/copied)
	 */
	public static File createFileIfNotExist(Class<?> c, String filename, String pluginID) {
		File dataFolder = getDataFolder(c, pluginID);
		String filePath = dataFolder+File.separator+filename;
		File fileCreated = new File(filePath);
		
		if (fileCreated.exists()) {
			return fileCreated;
		}
		
		/* Se non esiste crealo nel dataFolder */
		fileCreated.getParentFile().mkdirs();
		try {
			fileCreated.createNewFile();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		/* Apri, se esiste, il contenuto di default del file */
		InputStream is = openFile(c, filename);
		
		/* Se esiste, copia il contenuto iniziale */
		if(is != null) {
			copyFile(is,fileCreated);
		}
		
		return fileCreated;
	}
	
	/**
	 * Creates or opens the read-write directory 
	 * specified in the plugin's data folder.
	 * 
	 * @param c the context class (a class belonging to the context plugin)
	 * @param dirname the relative path of the dir to create
	 * @param pluginID the plugin ID
	 * @return the dir opened (or created)
	 */
	public static File createDirIfNotExist(Class<?> c, String dirname, String pluginID) {
		File dataFolder = getDataFolder(c, pluginID);
		String filePath = dataFolder+File.separator+dirname;
		File dir = new File(filePath);
		
		if (dir.exists()) {
			return dir;
		}
		
		/* Se non esiste crealo nel dataFolder */
		dir.mkdirs();
		
		return dir;
	}
	
    private static void copyFile(InputStream io, File file) {
    	FileOutputStream fos = null;
    	try {
			fos = new FileOutputStream(file);
			byte[] buf = new byte[256];
			int read = 0;
			while ((read = io.read(buf)) > 0) {
				fos.write(buf, 0, read);
			}
    	} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if(fos != null) {
				try {
					fos.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
    }
	
    /**
     * Returns the plugin's data folder path
     * @param c the context class (a class belonging to the context plugin)
	 * @param pluginID the plugin ID
	 * @return the plugin's data folder path
     */
	public static File getDataFolder(Class<?> c, String pluginID) {
    	PluginManager pm = PluginManager.lookup(c);
    	File dataFolder = null;
        try {
			dataFolder= ((CorePlugin) pm.getPlugin(CorePlugin.PLUGIN_ID))
			        .getDataFolder(pm.getRegistry().getPluginDescriptor(pluginID));
		} catch (Exception e) {
			e.printStackTrace();
		}
		return dataFolder;
	}
}
