/*
 * Created on 10-mar-2004
 *
 */
package org.modelgen.makecopyrules;

import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.model.Property;
import org.modelgen.commons.beans.model.Reference;
import org.modelgen.commons.persistence.PersistenceException;
import org.modelgen.commons.persistence.PersistenceHandler;
import org.modelgen.commons.persistence.PersistenceHandlerFactory;
import org.modelgen.commons.utils.ModelgenConfigurationManager;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;

import org.apache.log4j.Logger;
/**
 * Rules Builder
 * 
 * @author Luca
 *
 */
public class CopyRulesMaker {
	private static Logger log = Logger.getLogger(CopyRulesMaker.class);

	/**
	 * Rules Builder
	 * 
	 * @param args
	 * <ul>
	 * <li><tt>--list-models</tt>: lists all the available models</li>
	 * <li><tt>--model &lt;model OID&gt;</tt>: tells the application to build the rules from this model to the super model and viceversa</li>
	 * </ul>
	 */
	public static void main(String[] args) {
		if (args.length >= 1) {
			if (args[0].equals("--list-models")) {
				listModels();
			}
			else if (args[0].equals("--model")) {
				int modelOID = Integer.parseInt(args[1]);
				try {
					makeCopyRules(modelOID);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			else
				printUsage();
		}
		else
			printUsage();
	}
	
	/**
	 * Builds the rules from this model to the super model and viceversa
	 * @param modelOID the model's oid
	 * @throws Exception if a persistence error occurs
	 */
	public static void makeCopyRules(int modelOID) throws Exception {
		PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler();
		Model model = ph.getModel(modelOID);
		String ruleM_to_SM, ruleSM_to_M;
		ruleM_to_SM = makeCopyRules(model, false); //Prepare the rule.
		storeCopyRules(model, false, ruleM_to_SM); //Save it.
		ruleSM_to_M = makeCopyRules(model, true);
		storeCopyRules(model, true, ruleSM_to_M);		
	}
	/**
	 * @param model the model to work on.
	 * @param toModel true if the rule is from SuperModel to Model, false otherwise.
	 * @param rule the string to save.
	 */
	private static void storeCopyRules(Model model, boolean toModel, String copyRuleText) {
		//We're here to store the rule into a simple rule file, then make a macro rule which will call it.
		String rootnode = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths");
		String copyRulePath = ModelgenConfigurationManager.getIstance().getProperty("copyrules", "", "paths");
		String simpleRuleFullPath = rootnode + ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths") + copyRulePath;
		String macroRuleFullPath = rootnode + ModelgenConfigurationManager.getIstance().getProperty("macrorules", "", "paths") + copyRulePath;
		//The above two paths must have a trailing path separator character in order to work.
		//Prepare the filename: if you don't understand what's happening, know that I am extravagant.
		String filename =  "copyrule_";
		if (toModel == true) filename += "SM_to_";
		filename += model.getShortName();
		if (toModel == false) filename += "_to_SM";
		filename += ".txt";
		//Which means...
		simpleRuleFullPath += filename;
		macroRuleFullPath += filename;
		//And now we can write them.
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(simpleRuleFullPath));
			bw.write(copyRuleText);
			bw.close();
			String macroRuleText = "/* This file was generated automatically, do not edit or you'll lose the changes. */\n"
				+ copyRulePath + filename + ";";
			bw = new BufferedWriter(new FileWriter(macroRuleFullPath));
			bw.write(macroRuleText);
			bw.close();
		}
		catch (IOException e) {
			log.fatal(e);
		}
	}

	/**
	 * 
	 * @param model the model object to work on.
	 * @param toModel true if the requested rule is from the SuperModel to the given Model.
	 */
	private static String makeCopyRules(Model model, boolean toModel) {
		/*
		 * A model is made of constructs, each one currently having a 1:1 mapping to a construct of the super model.
		 * Each construct is translated into a table. (e.g. ER2_Entity, Rel_Column, ..., SM_abstract, SM_AttributeOfAggregationOfLexicals, ...)
		 * The copy rules translate the tuples of, say, ER2_Entity into tuples of, say, SM_Abstract.
		 * 
		 * For each construct (table) in the model, insert its data in the corresponding construct in the supermodel.
		 *  
		 */
		String rule = new String("");
		final String tab = "\t", newline = "\n"; //These are used to improve readability of the SQL construction code.
		//Auxiliary objects
		Construct smConstruct = null;

		String sourceTable = null, skolemTable = null, targetTable = null;
		
//		By first, we must add a tuple to either Scheme or SchemeSM.
		if (toModel == true) {
			rule += "INSERT INTO Scheme(OID, name, modelOID) " + newline;
			rule += tab + "SELECT ?1, SchemeSM.name, SchemeSM.modelOID " + newline;
			rule += tab + "FROM SchemeSM " + newline;
			rule += tab + "WHERE SchemeSM.OID = ?2;";
		}
		else {
			rule += "INSERT INTO SchemeSM(OID, name, modelOID, sourceSchemeOID) " + newline;
			rule += tab + "SELECT ?1, Scheme.name, Scheme.modelOID, null " + newline;
			rule += tab + "FROM Scheme " + newline;
			rule += tab + "WHERE Scheme.OID = ?2;";
		}
		
		rule += newline + newline;
		
		//For each construct, we'll need to make a "standard" tablet and a skolem one.
		for (Construct construct : model.getAbstractConstructs()) {
			//Since there's currently a 1:1 mapping between model and super model, the following call is right. 
			smConstruct = construct.getSuperModelConstructs().iterator().next();
			//The queries to translate to the super model and from the super model are dual, we just switch the tables.
			if (toModel == true) { //The source is the super model, the target is the model.
				sourceTable = "SM_" + smConstruct.getName(); 
				targetTable = model.getShortName() + "_" + construct.getName();
			}
			else { //The source is the model, the target is the super model.
				sourceTable = model.getShortName() + "_" + construct.getName();
				targetTable = "SM_" + smConstruct.getName();
			}
			skolemTable = targetTable + "_SK";
			//Queries: first insert into the skolem table. This one looks nice, don't get used to it.
			/*
			rule += "INSERT INTO " + skolemTable + " (OID, schemeOID, OID_copy) " + newline;
			rule += tab + "SELECT max(Skolem.OID), ?1, Source.OID " + newline;
			rule += tab + "FROM " + skolemTable + " AS Skolem, " + sourceTable + " AS Source " + newline;
			rule += tab + "WHERE Source.schemeOID = ?2" + newline;
			rule += tab + "AND NOT EXISTS (" + newline;
			rule += tab + tab + "SELECT *" + newline; 
			rule += tab + tab + "FROM " + skolemTable + " TargetSK" + newline; 
			rule += tab + tab + "WHERE TargetSK.OID_copy = Source.OID" + newline; 
			rule += tab + tab + "AND TargetSK.schemeOID = ?1" + newline; 
			rule += tab + ");" + newline;
			rule += newline;
			*/
			rule += "INSERT INTO " + skolemTable + " (schemeOID, OID_copy) " + newline;
			rule += tab + "SELECT ?1, Source.OID " + newline;
			rule += tab + "FROM " + sourceTable + " AS Source " + newline;
			rule += tab + "WHERE Source.schemeOID = ?2" + newline;
			rule += tab + "AND NOT EXISTS (" + newline;
			rule += tab + tab + "SELECT *" + newline; 
			rule += tab + tab + "FROM " + skolemTable + " TargetSK" + newline; 
			rule += tab + tab + "WHERE TargetSK.OID_copy = Source.OID" + newline; 
			rule += tab + tab + "AND TargetSK.schemeOID = ?1" + newline; 
			rule += tab + ");" + newline;
			rule += newline;
			//Second query: insert into the real table. This is harder to read.
			//Some auxiliary data first: an alias for the tables referenced by the references.
			Hashtable<String,String> aliases = new Hashtable<String,String>();
			int aliasCounter = 0;
			//Let's begin:
			rule += "INSERT INTO " + targetTable + "(OID, schemeOID, name";
			//The properties:
			for (Property property : construct.getProperties()) {
				if (toModel == true) //from the model...
					rule += ", " + property.getName();
				else //from the super model...
					rule += ", " + property.getMetaInfo().getName();
			}
			//And the references:
			for (Reference reference : construct.getReferences()) {
				//We take the chance to populate the alias counter. This can be done in any place where we iterate on the references, so this is a good place. :)
				if (toModel == true) {
					rule += ", " + reference.getName();
					// invertite, paolo c 20040610_15-20
					aliases.put(reference.getName(), "Ref" + (aliasCounter++));
				}
				else {
					rule += ", " + reference.getMetaInfo().getName();
					//invertite, paolo c 20040610_15-20
					aliases.put(reference.getMetaInfo().getName(), "Ref" + (aliasCounter++));
				}
			}
			rule += ")" + newline;
			//The "INSERT" part is over, now what values do we put there? Let's begin with the implicit ones:
			rule += tab + "SELECT" + newline;
			rule += tab + tab + "TargetSK.OID, TargetSK.schemeOID, Source.name";
			//Now the ones from the properties:
			for (Property property : construct.getProperties()) {
				if (toModel == true)
					rule += ", Source." + property.getMetaInfo().getName();
				else
					rule += ", Source." + property.getName();
			}
			//And the references: (and we use the alias, since we have it)
			for (Reference reference : construct.getReferences()) {
				if (toModel == true)
				    // invertite, paolo c 20040610_15-20
					rule += ", " + aliases.get(reference.getName()) + ".OID";
				else
				    // invertite, paolo c 20040610_15-20
					rule += ", " + aliases.get(reference.getMetaInfo().getName()) + ".OID";				
			}
			rule += newline;
			//The FROM clausole needs the source table, the skolem for the ID and another table for each reference!
			rule += tab + "FROM " + newline;
			rule += tab + tab + skolemTable + " AS TargetSK, " + newline;
			rule += tab + tab + sourceTable + " AS Source";
			for (Reference reference : construct.getReferences()) {
				if (toModel == true)
					rule += "," + newline + tab + tab + model.getShortName() + "_" + reference.getTargetConstruct().getName() + "_SK AS " + aliases.get(reference.getName());
				else
					rule += "," + newline + tab + tab + "SM_" + reference.getMetaInfo().getTargetConstruct().getName() + "_SK AS " + aliases.get(reference.getMetaInfo().getName()); 
			}
			rule += newline;
			//WHERE: a join + the scheme + a join (as a where) for each reference.
			rule += tab + "WHERE Source.OID = TargetSK.OID_copy" + newline; //Join as WHERE clausole.
			rule += tab + tab + "AND TargetSK.schemeOID = ?1" + newline;
			for (Reference reference : construct.getReferences()) {
				if (toModel == true) {
					rule += tab + tab + "AND Source." + reference.getMetaInfo().getName() + " = " + aliases.get(reference.getName()) + ".OID_copy" + newline;
					rule += tab + tab + "AND " + aliases.get(reference.getName()) + ".schemeOID = ?1" + newline;
				}
				else {
					rule += tab + tab + "AND Source." + reference.getName() + " = " + aliases.get(reference.getMetaInfo().getName()) + ".OID_copy" + newline;
					rule += tab + tab + "AND " + aliases.get(reference.getMetaInfo().getName()) + ".schemeOID = ?1" + newline;
				}
			}
			
			rule += tab + ";" + newline;
			rule += newline;
		}
		
		log.fatal(rule); //Some console output is always nice (and can be disabled via log4j.properties file)
		return rule;
	}

	private static void printUsage() {
		System.out.println/*log.fatal*/("Usage: java -jar makecopyrules.jar [--list-models | --model <model OID>]");
		System.out.println/*log.fatal*/("--list-models: lists all the available models.");
		System.out.println/*log.fatal*/("--model <model OID>: tells the application to build the rules from this model to the super model and viceversa. ");
	}
	
	private static void listModels() {
		try {
			PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler("DBHandler");
			System.out.println/*log.fatal*/("Available models:");
			for (Model model : ph.getModels()) {
				System.out.println/*log.fatal*/("[" + model.getOID() + "] " + model.getName() + " (" + model.getShortName() + ")");
			}
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (PersistenceException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
