package org.modelgen.polyhandler;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * This class is used to generate combinations of elements via appopriate mathematical algorithms
 * 
 * @author Daniele
 * 
 */
public class CombinationGenerator {
	private int[] a;
	private int n;
	private int r;
	private BigInteger numLeft;
	private BigInteger total;

	/**
	 * Constructor for this class
	 * @param n - the number of elements to be combined
	 * @param r - the size of the combination
	 */
	public CombinationGenerator(int n, int r) {
		if (r > n) {
			throw new IllegalArgumentException();
		}
		if (n < 1) {
			throw new IllegalArgumentException();
		}
		this.n = n;
		this.r = r;
		a = new int[r];
		BigInteger nFact = getFactorial(n);
		BigInteger rFact = getFactorial(r);
		BigInteger nminusrFact = getFactorial(n - r);
		total = nFact.divide(rFact.multiply(nminusrFact));
		reset();
	}

	/**
	 * Reset
	 *
	 */
	public void reset() {
		for (int i=0; i< a.length; i++) {
			a[i] = i;
		}
		numLeft = new BigInteger(total.toString());
	}

	/**
	 * This method returns the number of combinations not yet generated
	 * 
	 * @return a BigInteger representing the number of comibations not yet generated
	 */
	public BigInteger getNumLeft() {
		return numLeft;
	}

	/**
	 * Are there more combinations?
	 * 
	 * @return true or false
	 */
	public boolean hasMore() {
		return numLeft.compareTo(BigInteger.ZERO)==1;
	}

	/**
	 * This method returns the total number of combinations
	 * 
	 * @return a BigInteger representing the total number of combinations
	 */
	public BigInteger getTotal() {
		return total;
	}

	
	/**
	 * This method computes the factorial
	 * 
	 * @param n - the input number
	 * @return the computed factorial
	 */
	private static BigInteger getFactorial(int n) {
		BigInteger fact = BigInteger.ONE;
		for (int i=n; i>1; i--) {
			fact = fact.multiply(new BigInteger(Integer.toString(i)));
		}
		return fact;
	}

	
	/**
	 * This method implements the algorithm used to generate the next combination
	 * 
	 * @return a combination in array form
	 */
	public int[] getNext() {
		if (numLeft.equals(total)) {
			numLeft = numLeft.subtract(BigInteger.ONE);
			return a;
		}

		int i = r - 1;
		while (a[i] == n - r + i) {
			i--;
		}
		a[i] = a[i] + 1;
		for (int j = i + 1; j < r; j++) {
			a[j] = a[i] + j - i;
		}

		numLeft = numLeft.subtract(BigInteger.ONE);
		return a;
	}
	
	/**
	 * This method generates the various combinations of k elements taken n at a time,
	 * including combinations featuring repetitions of the same elements or elements in alternate
	 * order. E.g.: {a,b,c} taken 2 at a time and using "_" as a delimiter will result in: {a_b,a_c,a_a,b_a,b_c,c_b,b_b,c_a,c_c}. 
	 * 
	 * @param elements - the array of elements to be combined
	 * @param n - the size of each combination
	 * @param delimiter - the delimiter used within each combination to separate its single elements
	 * @return a Set of String representing the resulting combinations
	 */
	public static Set<String> generateCombinations(String[] elements, int n, String delimiter) {
		Set<String> set = new HashSet<String>();
		
		int[] indices;
		CombinationGenerator x = new CombinationGenerator(elements.length, n);
		StringBuffer combination;
		while (x.hasMore()) {
			combination = new StringBuffer();
			
			indices = x.getNext();
			for(int i=0; i<indices.length; i++) {
				//System.out.println("index: "+indices[i]);
				if(combination.length()==0) {
					combination.append(elements[indices[i]]);
				} else {
					combination.append(delimiter+elements[indices[i]]);
				}
				
				//System.out.println("temp: " +combination);
			}
			//System.out.println("Combination: "+combination.toString());
			String current = combination.toString();
			boolean found = false;
			for(String s: set) {
				if(current.equalsIgnoreCase(s)) {
					found = true;
				}
			}
			if(found==false) {
				set.add(current);
			}	
		}//end of while
		return set;
	}
	
	/**
	 * A method which splits a combination string like "a_b_c" into the single strings it is composed of,
	 * returning them as a List
	 * 
	 * @param s - the String to be split
	 * @param delimiter - the delimiters used to separate the various combinations within the string
	 * @return a List containing the single strings
	 */
	public static List<String> splitCombination(String s, String delimiter) {
		List<String> split = new LinkedList<String>();
		//String delimiters = "_";
	    StringTokenizer st = new StringTokenizer(s, delimiter, false);
	    while (st.hasMoreTokens()) {
	        String t = st.nextToken();
	        split.add(t);
	    }
		/*
	    int i=1;
	    for (String w: split) {
	    	System.out.println("#"+i+" Member of combination: "+w);
	    	i++;
	    }*/
	    
		return split;
	}
	
	/**
	 * Test method
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		String[] elements = {"a", "b", "c", "a", "b", "c"};
		Set<String> set = generateCombinations(elements,2,"_");
		int j=1;
		for(String t: set) {
			System.out.println("***"+j+" result: "+t);
			List<String> l = splitCombination(t,"_");
			int i=1;
			for (String w: l) {
		    	System.out.println("#"+i+" Member of combination: "+w);
		    	i++;
		    }
			
			j++;
		}
	}
}