package org.modelgen.polyhandler;

import java.io.File;

import org.jconfig.Configuration;
import org.jconfig.ConfigurationManager;
import org.jconfig.ConfigurationManagerException;
import org.jconfig.handler.XMLFileHandler;
/**
 * ConfigurationManager used for handling the generalization repository for the polymorphic
 * Datalog rules.
 * 
 * @author Daniele
 *
 */
public class PolyHandlerConfigurationManager {
		
	private static PolyHandlerConfigurationManager instance; //instance of this class
	/**
	 * Default configuration file path
	 */
	public static final String POLYHANDLER_CONFIGURATION_FILE = "generalizations_config.xml";
	private XMLFileHandler handler;
	private ConfigurationManager cm;
	private Configuration configuration;

	/**
	 * Private constructor for this class
	 */
	private PolyHandlerConfigurationManager() {
		handler = new XMLFileHandler();
		cm = ConfigurationManager.getInstance();
		configuration = loadConfiguration();	
	}
	    
	public Configuration getConfiguration() {
	  	return configuration;
	}
	    
	/**
	 * Obtain the one and only instance of this class
	 * @return the modelgen configuration instance
	 */
	public static synchronized PolyHandlerConfigurationManager getInstance(){
	   	if(instance==null)
	   		instance = new PolyHandlerConfigurationManager();
	   	return instance;
	}
	    
	/**
	 * @param key the name of the property
	 * @return the value as String if it is found or null
	 */
	public String getProperty(String key) {
		return configuration.getProperty(key);
	}
	    
	/**
	 * It will return the value of the property with the specific name. 
	 * First of all, it checks if the name of the category exists. 
	 * If not, then it will use the name of the default category. 
	 * The next step is that it will look for the property. 
	 * If it is not found in the category, it will look inside the 
	 * default category (inheritance). 
	 * If it still cannot find the property, it will return the defaultValue
	 *  
	 * @param key the name of the property
	 * @param defaultValue the default value
	 * @param categoryName the name of the category
	 * @return the value as String
	 */
	public String getProperty(String key, String defaultValue, String categoryName) {
		return configuration.getProperty(key, defaultValue, categoryName);
	}
	    
	/**
	 * This method sets the value for a property for the given category.
	 * If the category is null then it uses the default category.
	 * 
	 * @param name the name of the property
	 * @param value the value as String
	 * @param categoryName the name of the category
	 */
	public void setProperty(String name, String value, String categoryName) {
		configuration.setProperty(name, value, categoryName);
	}
	    
	/**
	 * Save the configuration in {@link #POLYHANDLER_CONFIGURATION_FILE}
	 */
	public void saveConfiguration(){
	   	if (configuration==null || handler==null)
	   		return;
	   	try {
			cm.save(handler,configuration);
		} catch (ConfigurationManagerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
	}

	/**
	 * This method loads the correct configuration file
	 * 
	 * @return the Configuration needed
	 */
	private Configuration loadConfiguration(){
	   	String configPATH = POLYHANDLER_CONFIGURATION_FILE;
    	File file = new File(configPATH);
    	handler.setFile(file);  	
   	
		try {
			cm.load(handler,"generalizations");
		} catch (ConfigurationManagerException e) {
			e.printStackTrace();
		}
		return ConfigurationManager.getConfiguration("generalizations");
	}

}