package org.modelgen.polyhandler;

import org.modelgen.commons.utils.ModelgenConfigurationManager;
import org.modelgen.regole.Argomento;
import org.modelgen.regole.ArgomentoEtichettato;
import org.modelgen.regole.Letterale;
import org.modelgen.regole.MacroRegola;
import org.modelgen.regole.RegolaDatalog;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.apache.log4j.Logger;
import org.jconfig.Configuration;

/**
 * This is the core class of the polyhandler package. It is meant to implement a pre-processing for
 * Datalog rules which possess a polymorphic structure (and therefore cannot be directly handed over
 * to the rule interpreter.
 *  
 * @author Daniele
 *
 */
public class PolyHandlerMain {
	private static final String RESERVED_KEYWORD_RESULT = "constructResult";
	private static final String RESERVED_KEYWORD_ORIGIN = "constructOrigin";
	private static final String RESERVED_SKOLEM = "skolemForConstruct";
	private static final String RESERVED_OID = "cOID";
	private static final String RESERVED_SUFFIX_FOR_ARGUMENT_RENAMING = "ZZZ";
	private static JFrame frame;
	private static List<RegolaDatalog> rules;
	private static List<RegolaDatalog> scannedRules;
	private static int ruleNameCounter = 0;
	private static int counterTotal;
	private static int letterCounter = 0;
	private static Configuration gen;
	private static String model;
	private static boolean hasAlreadyBeenAddedAsItsInstances = false;
	private static String macroruleInstance ="";
	private static String macroruleFileName;
	//private static PolyHandlerMain instance;
	private static Logger log = Logger.getLogger(PolyHandlerMain.class.getName());	
	//private static ErrorHandler errorHandler = ErrorHandler.instance;
		
	/**
	 * Public constructor for this class
	 */
	public PolyHandlerMain(String m) {
		model = m;
	}
	
	/*
	/**
	 * This method returns the one and only instance of this class
	 * 
	 * @param m - the model to be specified
	 * @return the instance of this class
	 */
	/*public static synchronized PolyHandlerMain getInstance(String m){
		if(instance==null) {
			instance = new PolyHandlerMain(m);
		}
		return instance;
	}*/

	/**
	 * Main method of this class to be externally invoked, it wraps up the core functionalities
	 * provided by the polymorphism handling.
	 * It accepts a Macrorule filename and performs the appropriate polymorphism transformations for
	 * the polymorphic simplerules found within it, consequently generating the correct instances of the latter rules.
	 *  
	 * @param macroRule - the filename of the Macrorule to be processed
	 */
	public void polyHandlerMainRoutine(String macroRule) {
				
		ruleNameCounter = 0;
		
		/* We tell the program to look in the macrorules directory to search for the specified macrorule */ 
		String macroRulePath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
		+ ModelgenConfigurationManager.getIstance().getProperty("macrorules", "", "paths");				        // macrorules/
		String macroregola = macroRulePath + macroRule;
		
		macroruleFileName = macroRule.substring(5,macroRule.length()-4);
		
		/* We use the appropriate ConfigurationManager to get the specs for the generalizations used in the current logical Supermodel */
		PolyHandlerConfigurationManager manager = PolyHandlerConfigurationManager.getInstance();
		gen = manager.getConfiguration();
				
		/* We load the MacroRegola from the string composed by path+filename */
		MacroRegola mrdp = PolyReader.read(macroregola);
				
		/* We are interested in the sub-rules which compose the MacroRegola: */
		rules = mrdp.getRegole();
		scannedRules = new LinkedList<RegolaDatalog>();
		counterTotal = 0;
				
		for(RegolaDatalog regola: rules) {
			JOptionPane.showMessageDialog(frame,regola.toString(),"Regola Datalog "+counterTotal,JOptionPane.INFORMATION_MESSAGE);
			handlePolyRule(regola);
			/* if a polyrule containing a construct which is recursively involved in generalizations (e.g. StructOfAttributes, it' s a construct
			 * possessing mutually exclusive references and therefore is a parent of a generalization, but it is also a child within other generalizations),
			 * then we have already added the single instances of such a rule within the "generateSingleRules(..) method"
			 */
			 if(hasAlreadyBeenAddedAsItsInstances==false) {
				scannedRules.add(regola);
				counterTotal++;
				
				
			} else {
				hasAlreadyBeenAddedAsItsInstances = false;
			}
		}
		
		fileWrite(macroruleInstance.toString(),macroRulePath,"instance_"+macroRule);
		
	}
	
	/**
	 * This method returns all the children belonging to all the defined generalizations
	 * 
	 * @return all the children of generalizations as a list of strings
	 */
	private static List<String> getAllChildrenOfGeneralizations() {
		List<String> lex = new LinkedList<String>();
		String[] temp = gen.getCategoryNames();
		for(int i=0; i<temp.length; i++) {
			try {
				if(!temp[i].equalsIgnoreCase("general")) {
					String[] array = pruneAllowedLexicals(temp[i]);
					for(int j=0; j<array.length; j++) {
						lex.add(array[j]);
						System.out.println("Child: " + array[j]);
					}
				}	
			} catch(GeneralizationNotFoundException ex) {
				JOptionPane.showMessageDialog(frame,ex.getMessage(),"Exception",JOptionPane.ERROR_MESSAGE);
			}
		}
		return lex;
	}
	
	/**
	 * This method prunes the array containing the various mutually exclusive references for a given construct
	 * according to the ones allowed in a specific model (established when the singleton for this class is instantiated).
	 * 
	 * @param letterale_name - the construct name
	 * @return an array featuring only the constructs allowed in the specified model
	 */
	private static String[] pruneAllowedLexicals(String letterale_name) throws GeneralizationNotFoundException {
		
		String[] temp = gen.getPropertyNames(letterale_name);
		if(temp.length==0) {
			throw new GeneralizationNotFoundException("There are no mutually exclusive reference for "+letterale_name+"! Check the correctness of the construct name or the configuration file!");
		}
		
		List<String> mec = new LinkedList<String>();
		//String[] mutuallyExclusiveConstructs = new String[temp.length];
		int i = 0;
		while(i<temp.length) {
			String value = gen.getProperty(temp[i],"default",letterale_name);
			System.out.println("Property: "+temp[i]+": "+value);
			String[] values = value.split(",");
			for (int j=0; j<values.length; j++) {
				System.out.println("VALUE TOKEN: "+values[j]);
				if(values[j].equalsIgnoreCase(model)) {
					System.out.println("TOKEN "+temp[i]+" WILL BE ADDED");
					mec.add(temp[i]);
					//mutuallyExclusiveConstructs[i] = temp[i];
					//System.out.println(value);
				}
				
			}
			i++;
		}
		String[] mutuallyExclusiveConstructs = new String[mec.size()];
		for(int j=0; j<mec.size(); j++) {
			mutuallyExclusiveConstructs[j] = mec.get(j);
		}
		
		return mutuallyExclusiveConstructs;
	}
	
	/**
	 * This method returns a suitable counter number used for file-sorting purposes
	 * 
	 * @param c - the number to be turned into a more suitable form
	 * @return - a counter in the form: 0NUMBER where NUMBER = 1,2,..,9; NUMBER where NUMBER = 10,11,12...
	 */
	private static String getCounter(int c) {
		String counter ="";
		if(c<10) {
			counter = "0"+c; 
		} else {
			counter = ""+c+"";
		}
		return counter;	
	}
	
	/**
	 * The core method of this class: it accepts a Datalog rule as input, checks whether it is polymorphic or not, and in the former case
	 * handles the whole operation of turning it into single, instantiatable rules (by invoking appropriate sub-methods where necessary);
	 * in the latter, it just increments the scannedRules collection
	 *  
	 * @param regola - the currently scanned Datalog rule
	 */
	private static void handlePolyRule(RegolaDatalog regola) {
		boolean isPolyHead = false;
		String[] mutuallyExclusiveConstructs;
		String letterale_name;
		
		/* We need the Letterale making up the rule head, as well as the other lexicals within the rule body */
		Letterale testa = regola.getTesta();
		
		String SM_letterale_name = "";; //it could either be from head or body, depending on the check performed below
				
		/* Let' s check whether the current rule is a polymorphic one by examining the Letterale in its head */
		isPolyHead = isPolymorphic(testa);
		/* Let' s also check whether a Letterale in the rule body has polymorphic references */
		boolean isPolyBody = false;
		Letterale polyBodyLett = null;;
		List<Letterale> body = regola.getBody();
		for(Letterale let: body) {
			if(isPolymorphic(let)==true) {
				isPolyBody = true;
				polyBodyLett = let;
			}
		}		
		/* *******************
		 * Let' s check! :-)
		 * *******************
		 */
		if(isPolyHead==true) { //ok it is a polymorphic rule, its head contains a polymorphic reference
			if((testa.getPredicateName()).equals(polyBodyLett.getPredicateName())) {
				/* Same polymorphic Letterale in head and body, we can consider either one of them */
				SM_letterale_name = testa.getPredicateName();
			} else {
				/* Different Letterale in the body, we' ll consider this one then */
				SM_letterale_name = polyBodyLett.getPredicateName();
			}	
		} else { //isPolyHead=false, but a lexical in the rule body might have polymorphic references nevertheless, let' s check it out
			if(isPolyBody==false) { // this is not a polymorphic rule
				//JOptionPane.showMessageDialog(frame,"No polymorphic references in this rule, nothing to be done at all!","No poly references",JOptionPane.INFORMATION_MESSAGE);
				
				/* Write the rule as is */
				String simpleRulesPath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
				+ ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths");			
				ruleNameCounter++;
				String polyRulesPath = simpleRulesPath + "polyrules/";
				String ruleCounter = getCounter(ruleNameCounter);
				String ruleFileName = "polyrule_"+ruleCounter+"_NOT_POLY"+".dtr";
				fileWrite(regola.toString(),polyRulesPath+macroruleFileName+"/","polyrule_"+ruleCounter+"_NOT_POLY"+".dtr");		
				
				boolean mustIterate = regola.getMustIterate();
				String iterate = "";
				if(mustIterate==true) {
					iterate ="!!";
				}
				macroruleInstance = macroruleInstance + iterate + "polyrules/"+macroruleFileName+"/"+ruleFileName+"();\n";
				
				return; //let' s stop here
			} else { //isPolyHead=false but isPolyBody=true
				/* The polymorphic references are found only within a Letterale in the rule body */
				SM_letterale_name = polyBodyLett.getPredicateName();
			}	
		}
		/* *******************
		 * Check done! :-)))
		 * *******************
		 */
		
		JOptionPane.showMessageDialog(frame,"REGOLA POLIMORFA!!!", "REGOLA POLIMORFA!!!",JOptionPane.INFORMATION_MESSAGE);
						
		letterale_name = SM_letterale_name.substring(3,SM_letterale_name.length()); //Letterale's name without the "SM_", e.g.: Lexical
		JOptionPane.showMessageDialog(frame,"Letterale polimorfo: "+letterale_name, "REGOLA POLIMORFA!!!",JOptionPane.INFORMATION_MESSAGE);
			
		/* ******************************************************
		 * A polymorphic rule usually has the following structure:
		 * 
		 * SM_GENCONSTRUCT(
		 * 		OID: #skolemForGenConstruct(oid),
		 * 		Name: name,
		 * 		...
		 * 		constructResultOID: #skolemForConstruct(cOID)
		 * )<-
		 * SM_GENCONSTRUCT(
		 * 		OID: oid,
		 * 		Name: name,
		 * 		...
		 * 		constructOriginOID: cOID
		 * ),
		 * ...
		 * ;
		 * 
		 * where "constructOrigin", "constructResult" and "skolemForConstruct" are the reserved keywords to indicate polymorphic references and
		 * their corresponding Skolem function.
		 * Thus, we consider the lexical featuring polymorphic references (there could be three possible cases: 
		 * 		1) polymorphic references in the head and in a body lexical;
		 * 		2) polymorphic references in a body lexical only; 
		 * 		3) no polymorphic references at all;)
		 * and we look into the generalizations repository in order to find the corresponding mutually exclusive references such a construct must SURELY have.
		 * If the construct name is not found anywhere in the repository, EXCEPTION!!!
		 *  
		 **********************************************************************************************************************************
		 */
			
		/* Let' s grab the info about generalizations etc. from the repository */
		//mutuallyExclusiveConstructs = gen.getPropertyNames(letterale_name);
		mutuallyExclusiveConstructs = null;
		try {
			mutuallyExclusiveConstructs = pruneAllowedLexicals(letterale_name);
		} catch(GeneralizationNotFoundException exc) {
			/* Fatal error: no generalizations could be found for the specified Letterale in the current model */
			String msg = exc.getMessage();
			JOptionPane.showMessageDialog(frame,msg+" The application will now abort","ERROR",JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		PolyResolved pr = new PolyResolved();
		/* The structure which will store the association: ConstructOrigin -> {ConstructResult1,ConstructResult2...}
		 * E.g.: Abstract -> {Aggregation}
		 * 	 or: Abstract -> {Aggregation,Abstract} (it means that some Abstract have been translated into Aggregation and some other into Abstract)
		 */
		Map<String,List<String>> resolved = new HashMap<String,List<String>>();
		
		/* The association ConstructOrigin -> {Skolem1,Skolem2...} as above
		 * 
		 */
		Map<String,List<String>> skolem = new HashMap<String,List<String>>();
			
		/* The association ConstructOrigin -> {ruleBody1,ruleBody2...}
		 * in order to specify the various translations when a ConstructOrigin has been translated
		 * into two or more different constructs within a Macrorule
		 * The "ruleBodyK" string will be appended (if necessary) to the instantiated rules derived from the processing
		 * of the polymorphic ones
		 */
		Map<String,List<String>> append = new HashMap<String,List<String>>();
		Map<String,List<String>> specs = new HashMap<String,List<String>>();
		String appendString = "APPEND PROVA";
		//boolean hasNoRefs = false;
		boolean isMain = false;
		
		//boolean hasBeenTranslatedAlready = false;
		/* 
		 * For each mutually exclusive construct... E.g.: Abstract, Aggregation, StructOfAttributes
		 */
		for(int j=0; j<mutuallyExclusiveConstructs.length; j++) {
			boolean hasBeenTranslatedEventually = false;
			
			List<String> resolvedList = new LinkedList<String>();
			List<String> skolemList = new LinkedList<String>();
			List<String> appendList = new LinkedList<String>();
			List<String> specsList = new LinkedList<String>();
			
			/*
			 * In the event of having a polymorphic Letterale which allows for NO REFERENCES
			 */
			if(mutuallyExclusiveConstructs[j].equalsIgnoreCase("none")) {
				generateRuleWithoutRefs(regola,letterale_name,mutuallyExclusiveConstructs);
			} else {
			
			//JOptionPane.showMessageDialog(frame,mutuallyExclusiveConstructs[j],"Reference in mutua esclusione per il costrutto "+letterale_name,JOptionPane.INFORMATION_MESSAGE);
			
			/* 
			 * ...and for each Datalog rule we have previously scanned... 
			 */
			for(RegolaDatalog rule: scannedRules) {
				isMain = isMainConstruct(rule,mutuallyExclusiveConstructs[j]);
				/* if isMain is true, it means that the construct considered has been found and is the main 
				 * construct of the current rule -> therefore, the lexical in the latter' s head is the one we
				 * were looking for. Such a construct will then replace the polymorphic reference in the
				 * polymorphic rule we are currently processing!
				 * 
				 * E.g.: we were scanning the Abstract construct, and we found out that, in a previous rule, it
				 * had been turned into an Aggregation: therefore, in the polymorphic rule, the "constructOrigin" and
				 * "constructResult" references will be replaced respectively with Abstract and Aggregation.
				 * We want to keep track of such an association along with the Skolem function used in the aforementioned
				 * transformation!
				 * 
				 */					
				if(isMain==true) { 
					hasBeenTranslatedEventually = true;
					Letterale head = rule.getTesta(); //the head of the current rule among the scannedRules which we will extract the needed info from
					String SM_headName = head.getPredicateName();
					String headName = SM_headName.substring(3,SM_headName.length());
					String headSkolem = getSkolemWithoutArgsForOIDFromLetterale(head); //e.g.: AbstractOID_1
					//appendString = getBodyAsString(rule);
					appendString = preProcessBodyNew(rule,mutuallyExclusiveConstructs[j],RESERVED_OID);
					String hasFurtherSpecs = checkSpecs(rule);
					//letterCounter++;
					JOptionPane.showMessageDialog(frame,"<html>"+mutuallyExclusiveConstructs[j]+ " wil become " + headName + " in the polymorphic rule!<br>"+headSkolem+" will be used as the Skolem function</html>","ASSOCIATION stage",JOptionPane.WARNING_MESSAGE);
					
					/*
					 * Check whether such a construct has already been listed as a "main" in a previous scanning. If so, just add the new ConstructResult to the proper list etc.
					 */
					if(!resolved.containsKey(mutuallyExclusiveConstructs[j])) {
						//System.out.println(mutuallyExclusiveConstructs[j] +" is not present: we' ll add it");
						resolvedList.add(headName);
						skolemList.add(headSkolem);
						appendList.add(appendString);
						specsList.add(hasFurtherSpecs);
						resolved.put(mutuallyExclusiveConstructs[j],resolvedList);
						skolem.put(mutuallyExclusiveConstructs[j],skolemList);
						append.put(mutuallyExclusiveConstructs[j],appendList);
						specs.put(mutuallyExclusiveConstructs[j],specsList);
					} else { // it wasn' t the first time we had encountered this construct
						//System.out.println(mutuallyExclusiveConstructs[j] +" is PRESENT: we' ll just update its lists");
						resolvedList.add(headName);
						skolemList.add(headSkolem);
						appendList.add(appendString);
						specsList.add(hasFurtherSpecs);
					}
				}//end (isMain==true)
			}// end of for(RegolaDatalog rule: scannedRules)
			
			/*
			 * We should need to include the possibility for a mutually exclusive construct not to be included (and translated) in any previous rules:
			 * that means either that such a construct does not exist in our current model (perhaps we' re dealing with a simplified version of it, or the order which
			 * the rules are executed by makes impossible to find it within the previous rules)
			 * or that there is something wrong with the Datalog rules belonging to the considered Macrorule.
			 */
			if(hasBeenTranslatedEventually==false) {//we haven' t been able to identify such a construct as a main within any of the previously scanned rules	
				JOptionPane.showMessageDialog(frame,"Construct: "+mutuallyExclusiveConstructs[j]+" hasn' t been translated in any previous rules, there might be something wrong with the current set of rules","Construct Not Translated",JOptionPane.WARNING_MESSAGE);
			}
			}//end of else (if m.e.construct=="none"...)
		}//end of for(int j=0; j<mutuallyExclusiveConstructs.length; j++)
		
		/*
		 * Just to check things
		 */
		Set<String> set = resolved.keySet();
		System.out.println("Printing references association for polymorphic "+letterale_name);
		for(String s: set) {
			List<String> l = resolved.get(s);
			List<String> k = skolem.get(s);
			List<String> a = append.get(s);
			int cn = 0;
			System.out.println("Construct "+s+" ,list size: "+l.size());
			for(cn = 0;cn<l.size(); cn++) {
				String t = l.get(cn);
				String sk = k.get(cn);
				String ap = a.get(cn);
				System.out.println(s + " -> " + t +"; Skolem: "+sk+"; Append: "+ap);
			}
			//String t = (String)resolved.get(s);
			//String sk = (String)skolem.get(s);
			//System.out.println(s + " -> " + t +"; Skolem: "+sk);
		}
						
		/*
		 * Now let' s hand over all the PRICELESS associations to the final method!
		 */
		pr.setAllButSpecs(resolved,skolem,append,specs);
		/*
		 * Only if the considered mutually exclusive construct is not "none",
		 * otherwise we have already dealt with it
		 */
		if(resolved.size()>0) {
			generateSingleRules(regola,resolved,skolem,append,specs);
		}	
	}	
	
	/**
	 * Returns the number of elements found in a Map featuring lists as values, for a certain key.
	 * That means that if a certain key K is associated to a certain list V which contains N elements,
	 * N is returned.
	 * 
	 * @param map - the map to be considered
	 * @param key - the seleceted key
	 */
	private static int getNumberOfElementsForKey(Map<String,List<String>> map, String key) {
		int number = 0;
		List<String> l = map.get(key);
		number = l.size();
		return number;
	}
	
	/**
	 * Returns the TOTAL number of elements found in a Map featuring lists as values.
	 * This method returns the number of ALL the elements in ALL the lists.
	 * 
	 * @param map - the map to be considered
	 */
	protected static int getNumberOfElements(Map<String,List<String>> map) {
		int number = 0;
		Set<String> set = map.keySet();
		for(String s: set) {
			List<String> l = map.get(s);
			number += l.size();
		}
		return number;
	}
	
	/*
	private static String getBodyAsString(RegolaDatalog regola) {
		String bodyString =",";
		
		List<Letterale> l = regola.getBody();
		for(Letterale lett: l) {
			bodyString = bodyString + "\n"+lett.toString()+",";
		}
		String sub = bodyString.substring(0,bodyString.length()-1);
		bodyString = sub+";";
		
		return bodyString;
	}
	*/
	
	/**
	 * This method is used to check whether a construct featured as a mutually exclusive reference,
	 * despite having just ONE translation within a Macrorule (i.e., it is the "main" construct just once)
	 * possesses some further specification when its translation occur (e.g. in the rule body there is some
	 * TMP construct or some negated construct used to "select" only a certain subset of the given construct)
	 * 
	 * @param regola - the Datalog rule we need to check for further specs
	 * @return a String which encapsulates a boolean value (used for the application purposes)
	 */
	private static String checkSpecs(RegolaDatalog regola) {
		String hasFurtherSpecs = "false";
		List<Letterale> body = regola.getBody();
		//letterCounter++;
		for(Letterale lett: body) {
			//String SM_lett_name = lett.getPredicateName();
			//String role = lett.getRuolo();
			if(lett.getIsNegato()==true || lett.getIsTMP()==true) {
				/*
				 * Let' s see if we can work out a solution for a couple of Macrorules like 31,32,33
				 */
				hasFurtherSpecs = "true";
			} else {
				// nothing to be done at the moment		
			}
		}
		return hasFurtherSpecs;
	}	
	
	/**
	 * This method accepts as input a Datalog rule, a construct (signalled as "main",i.e. the construct this rule is translating),
	 * and a replacement for its OID according to the RESERVED_OID keyword. It performs such a replacement in the rule body,
	 * as well as modifying other argument values of body constructs in order not to have incongruences when appending such a rule body
	 * to the instance of a polymorphic rule.
	 * 
	 * @param regola - the Datalog rule
	 * @param mainConstruct - the construct identified as "main"
	 * @param replacementOID - the replacement of the main construct's oid value
	 * @return a String containing the rule body modified accordingly
	 */
	private static String preProcessBodyNew(RegolaDatalog regola, String mainConstruct, String replacementOID) {
		String result ="";
		String bodyString = ",";
		
		List<Letterale> body = regola.getBody();
		String OID = null;
		letterCounter++;
		for(Letterale lett: body) {
			String SM_lett_name = lett.getPredicateName();
			String role = lett.getRuolo();
			if(lett.getIsNegato()==false) {
				bodyString = bodyString + "\n" + SM_lett_name + " " + role + " (";
			} else {
				bodyString = bodyString + "\n" + "!" + SM_lett_name + " " + role + " (";
				/*
				 * Let' s see if we can work out a solution for a couple of Macrorules like 31,32,33
				 */
				//hasFurtherSpecs = true;
			}
				
			String lett_name = SM_lett_name.substring(3,SM_lett_name.length());
			
			if(mainConstruct.equals(lett_name) && !role.equalsIgnoreCase("[DEST]")) {
				OID = getOIDvalueFromLetterale(lett);
			}	
			//name = getNameValueFromLetterale(lett);
			System.out.println("OID of construct"+lett_name+" in appendString: "+OID);
			//bodyString = replaceWord(bodyString,OID,replacementOID);
			//result2 = replaceWord(result1,name,name+"Z");
			List<String> attributi = lett.getListaAttributi();
			List<ArgomentoEtichettato> argomenti = lett.getListaArgomentiEtichettati();
			int count = 0;
			//String difference = "ZZZ";
			
			for (count = 0; count<attributi.size(); count++) { 
				//String att = attributi.get(count);
				ArgomentoEtichettato arg = argomenti.get(count);
				Argomento a = arg.getArgomento();
				String at = arg.getAttributo();
				/*
				if(OID!=null && at.equalsIgnoreCase("OID")) {
					bodyString = bodyString + "\n    " + at + ": " + replacementOID+ ",";
				} else {*/
					if(!(arg.toString()).contains("\"" )) {
						bodyString = bodyString + "\n    " + at + ": "+ a.getNome()+RESERVED_SUFFIX_FOR_ARGUMENT_RENAMING+",";
					//difference = difference + 1;
					} else {
						bodyString = bodyString + "\n    " + at + ": "+"\""+ a.getNome()+"\""+",";
					}
					
			} 
			bodyString = bodyString.substring(0,bodyString.length()-1); 
			bodyString = bodyString +"\n" + "),";
		
			if(OID!=null) {
				bodyString = replaceWord(bodyString,OID+RESERVED_SUFFIX_FOR_ARGUMENT_RENAMING,replacementOID);
			}
			//name = getNameValueFromLetterale(lett);
			//bodyString = replaceWord(bodyString,name,name+"ZZZ");
				
		}
		bodyString = bodyString.substring(0,bodyString.length()-1);
		result = bodyString + ";";
		System.out.println("Result: "+result+"--------------");		
		return result;
	}
	
	/**
	 * This method is invoked by the handlePolyRule method when a certain polymorphic Letterale has
	 * "none" as a mutually exclusive reference...i.e., a child of its generalizations has no references at all.
	 * Therefore, when such a case applies, this method is used instead of generateSingleRules.
	 * 
	 * @param regola - the polymorphic Datalog rule
	 * @param construct - the Letterale parent of a generalization which includes a child with no references
	 * @param mutuallyExclusiveConstructs - all the children belonging to the generalizaion whose parent is the considered construct (used for negations)
	 */
	private static void generateRuleWithoutRefs(RegolaDatalog regola, String construct, String[] mutuallyExclusiveConstructs) {
		String ruleInstance = "";
		Letterale lex = regola.getTesta();
		String lex_name = lex.getPredicateName();
		ruleInstance = ruleInstance + lex_name +" (";
		List<ArgomentoEtichettato> args = lex.getListaArgomentiEtichettati();
		for(ArgomentoEtichettato ae: args) {
			String attributo = ae.getAttributo(); // e.g.: OID, Name, etc.
			if(!attributo.equals(RESERVED_KEYWORD_RESULT+"OID")) {
				String argomento = ae.getArgomento().toString();
				ruleInstance = ruleInstance + "\n    "+attributo + ": " + argomento +",";
			}
		}
		ruleInstance = ruleInstance.substring(0,ruleInstance.length()-1);
		ruleInstance = ruleInstance + "\n)\n<-";
		
		List<Letterale> body = regola.getBody();
		for(Letterale bodyLett: body) {
			String lett_name = bodyLett.getPredicateName();
			String lett_name_without_SM = lett_name.substring(3,lett_name.length());
			if(!lett_name_without_SM.equalsIgnoreCase(RESERVED_KEYWORD_ORIGIN)) {
				ruleInstance = ruleInstance + "\n" + lett_name + " (";
				List<ArgomentoEtichettato> arguments = bodyLett.getListaArgomentiEtichettati();
				for(ArgomentoEtichettato ar: arguments) {
					String attributo = ar.getAttributo(); // e.g.: OID, Name, etc.
					if(!attributo.equals(RESERVED_KEYWORD_ORIGIN+"OID")) {
						String argomento = ar.getArgomento().toString();
						ruleInstance = ruleInstance + "\n    "+attributo + ": " + argomento +",";
					}
				}
			
				if(lett_name_without_SM.equals(construct)) {
					int c = 1;
					for(int i=0; i<mutuallyExclusiveConstructs.length; i++) {
						if(!mutuallyExclusiveConstructs[i].equalsIgnoreCase("none")) {
							ruleInstance = ruleInstance + "\n    "+mutuallyExclusiveConstructs[i]+"OID: rOID"+c+",";
							c++;
						}	
					}
				}
			
				ruleInstance = ruleInstance.substring(0,ruleInstance.length()-1);
				ruleInstance = ruleInstance + "\n),";
			}
		}
		int c = 1;
		for(int j=0; j<mutuallyExclusiveConstructs.length; j++) {
			if(!mutuallyExclusiveConstructs[j].equalsIgnoreCase("none")) {
				ruleInstance = ruleInstance + "\n!SM_" + mutuallyExclusiveConstructs[j] + " (";
				ruleInstance = ruleInstance + "\n    " + "OID: rOID"+c+"";
				ruleInstance = ruleInstance + "\n),";
				c++;
			}	
		}
		ruleInstance = ruleInstance.substring(0,ruleInstance.length()-1);
		ruleInstance = ruleInstance + ";";
		
		JOptionPane.showMessageDialog(frame,ruleInstance,"Regola risultato NO REFS",JOptionPane.WARNING_MESSAGE);
		List<String> allChildren = getAllChildrenOfGeneralizations();
		for(String child: allChildren) {
			if(child.equals(construct)) {
				hasAlreadyBeenAddedAsItsInstances = true;
				//JOptionPane.showMessageDialog(frame,letteraleOriginal+"' s polyrule is going to be added as its instances","WARNING",JOptionPane.WARNING_MESSAGE);
			}
		}
		if(hasAlreadyBeenAddedAsItsInstances==true) {
			RegolaDatalog regolaIst = PolyDatalogRulesReader.readString(ruleInstance);
			scannedRules.add(regolaIst);
			counterTotal++;
			JOptionPane.showMessageDialog(frame,construct+"' s polyrule HAS BEEN added as its instances","WARNING",JOptionPane.WARNING_MESSAGE);
		}
		
//		 WRITING TO DISK FOR TESTING PURPOSES
		//String name = regola.getTesta().getPredicateName();
		String simpleRulesPath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
		+ ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths");			
			
		ruleNameCounter++;
		String ruleCounter = getCounter(ruleNameCounter);
		String ruleFileName = "polyrule_"+ruleCounter+".dtr";
		String polyRulesPath = simpleRulesPath +"polyrules/";
		fileWrite(ruleInstance,polyRulesPath+macroruleFileName+"/","polyrule_"+ruleCounter+".dtr");		
		
		boolean mustIterate = regola.getMustIterate();
		String iterate = "";
		if(mustIterate==true) {
			iterate ="!!";
		}
		macroruleInstance = macroruleInstance + iterate + "polyrules/"+macroruleFileName+"/"+ruleFileName +"();\n";
		
	}
		
	/**
	 * This method is used to generate instances of a polymorphic rule according to the various polymorphic references
	 * found within the featured constructs. It is responsibile for processing the data received by the handlePolyRule method
	 * and performing the appropriate replacements in the given rule, thus bringing about the creation of a series of instances
	 * for such a rule, which will be later processed by the rule interpreter.
	 * 
	 * When the basic case cannot be applied (i.e. when there is more than one polymorphic reference belonging to a construct
	 * of the considered rule), it relies on the CombinationGenerator class in order to produce the various combinations of the data
	 * received by the handlePolyRule method (=the constructs' associations obtained by scanning the previous rules), 
	 * taken K at a time (where K is the number of polymorphic references for a construct). 
	 * 
	 * @param regola - the polymorphic Datalog rule to be transformed into single, specific rules
	 * @param resolved - a Map for keeping trace of the construct translations performed in the previous rules related to the mutually exclusive references belonging to the polymorphic construct
	 * @param skolem - a Map for keeping trace of the Skolem functions used in the aforementioned translations
	 * @param append - a Map for keeping trace of the bodies belonging to the rule where the above associations have been detected, used to "specialize" the instantiated rules when necessary
	 * @param specs
	 */
	private static void generateSingleRules(RegolaDatalog regola, Map<String,List<String>> resolved, Map<String,List<String>> skolem, Map <String,List<String>> append, Map<String,List<String>> specs) {
		/*
		 * First of all, let' s get all the polymorphic references belonging to the construct within the given rule
		 */
		String letteraleOriginal = "";
		String regolaPolimorfica = regola.toString();
		List<String> collHead = new LinkedList<String>(); //list of the polymorphic references found in the head construct
		List<String> collBody = new LinkedList<String>(); //list of the polymorphic references found in the body construct (whose name is the same as the head's)
		Letterale testa = regola.getTesta();
		List<ArgomentoEtichettato> argomenti= testa.getListaArgomentiEtichettati();
		for (ArgomentoEtichettato ae: argomenti) {
			String att = ae.getAttributo(); // ex. OID, Name, etc.
			if(att.contains(RESERVED_KEYWORD_RESULT)) {
				JOptionPane.showMessageDialog(frame,"Attributo polimorfo testa: "+att,"Attributi",JOptionPane.INFORMATION_MESSAGE);
				collHead.add(att);
			}
		}
		List<Letterale> body = regola.getBody();
		for(Letterale lett: body) {
			//if(lett.getPredicateName().equals(testa.getPredicateName())) {
				List<ArgomentoEtichettato> args = lett.getListaArgomentiEtichettati();
				for(ArgomentoEtichettato ae: args) {
					String att = ae.getAttributo();
					if(att.contains(RESERVED_KEYWORD_ORIGIN)) {
						String letteraleOriginal_SM = lett.getPredicateName();
						letteraleOriginal = letteraleOriginal_SM.substring(3,letteraleOriginal_SM.length());
						JOptionPane.showMessageDialog(frame,"Attributo polimorfo corpo: "+att,"Attributi",JOptionPane.INFORMATION_MESSAGE);
						collBody.add(att);
					}
				}
			//}
		}
		/* Ok, we got them! :-) */
		
		/* *********************************************************************************************
		 * Complex case: the construct we are considering might have more than one polymorphic reference
		 * *********************************************************************************************
		 */
		int combinationSize = 0;
		if(collHead.size()>=1) {
			combinationSize = collHead.size();
		} else {
			combinationSize = collBody.size();
		}
		int assocSize = getNumberOfElements(resolved);
		String[] assocValues = new String[assocSize];
		Set<String> set = resolved.keySet(); //let' s grab the keys, i.e. the original constructs
		int i = 0;
		for(String s: set) {
			//String t = (String)resolved.get(s);
			List<String> l = resolved.get(s);
			int num = 0;
			for (int j = 0; j<l.size();j++) {  // String ls: l) {
				assocValues[i] = s+"-"+num; //the original construct with a specific number as prefix, indicating its possible different translations! E.g.: Abstract-0, Abstract-1 (if Abstract has been translated into two different constructs)
				num++;
				/*if(num>=10) {
					JOptionPane.showMessageDialog(frame,"TOO MANY TRANSLATIONS FOR A SINGLE CONSTRUCT (>=10), there must be something wrong with the Datalog rules","ERROR",JOptionPane.ERROR_MESSAGE);
				/* We should throw an Exception here... or maybe not */
				//}
				System.out.println("assocValue "+i+"= "+assocValues[i]);
				i++;
			}	
		}
			
		/*
		 * We need a proper array to be handed over to the CombinationGenerator later
		 */
		String[] assocValuesDoubled = new String[assocValues.length*2];
		int l=0;
		for(l=0; l<assocValues.length; l++) {
			assocValuesDoubled[l] = assocValues[l];
		}
		int u=0;
		for(int h=l; h<assocValuesDoubled.length; h++){
			assocValuesDoubled[h] = assocValues[u];
			u++;
		}
			
		/*
		 * Ok, now we have the various constructs used as "origin" in a comfortable array form.
		 * Let' s hand them over to the CombinationGenerator!
		 */
		Set<String> combinations = CombinationGenerator.generateCombinations(assocValuesDoubled,combinationSize,"_");
		//e.g.: {Abstract-0_Aggregation-0, Abstract-0_Abstract-0, Aggregation-0_Abstract-0, Aggregation-0_Aggregation-0}
		int k=1;
		for(String s: combinations) { //for each combination...
			System.out.println(k+" Combination: "+s);
			List<String> list = CombinationGenerator.splitCombination(s,"_");
			int counter = 0;
			String regolaTemp0 = regolaPolimorfica;
			String regolaIstanziata= "";
			//String constructOriginalCopy ="";
			for(counter=0; counter<list.size(); counter++) { //for each element of the combination, ex. Aggregation-0
			//for(String constructOriginal: list) { 
				String constructOriginalPlusID = list.get(counter); //e.g.: Aggregation-0
				System.out.println("Executing cycle: "+counter+" out of "+list.size()+" for "+constructOriginalPlusID);
				int position = constructOriginalPlusID.indexOf("-");
				String identifierString = constructOriginalPlusID.substring(position+1,constructOriginalPlusID.length());
				int identifier = Integer.parseInt(identifierString);
				System.out.println("Identifier for construct "+constructOriginalPlusID+": "+identifier);
				String constructOriginal = constructOriginalPlusID.substring(0,position); //e.g.: Aggregation
				//constructOriginalCopy = constructOriginal;
				
				List<String> skolemList = skolem.get(constructOriginal);
				String skolemForConstruct = skolemList.get(identifier);
				System.out.println("constructOriginal for combination #"+k+": "+constructOriginal);
				System.out.println("skolem for "+constructOriginal+": "+skolemForConstruct);
				String substringHead ="";
				if(collHead.size()>=1) {
					substringHead = collHead.get(counter).substring(RESERVED_KEYWORD_RESULT.length(),collHead.get(counter).length()); //e.g.: "FromOID"
					System.out.println("substringHead: "+substringHead);
				}
				String substringBody = collBody.get(counter).substring(RESERVED_KEYWORD_ORIGIN.length(),collBody.get(counter).length());
				
				System.out.println("substringBody: "+substringBody);
				List<String> resolvedList = resolved.get(constructOriginal);
				String constructNameForHead = resolvedList.get(identifier); 
				System.out.println(constructOriginal+" is associated with construct: "+constructNameForHead);
				
				List<String> appendList = append.get(constructOriginal);
				String appendString = appendList.get(identifier);
				System.out.println(constructOriginal+" has the corresponding appended string: "+appendString);
				
				List<String> specsList = specs.get(constructOriginal);
				String hasFurtherSpecs = specsList.get(identifier);
				System.out.println(constructOriginal+" has specs: "+hasFurtherSpecs);
				/*
				 * Now the replacements begin!
				 */
				int c = counter+1;
				String regolaTemp1 = replaceWord(regolaTemp0,RESERVED_KEYWORD_RESULT+substringHead,constructNameForHead+substringHead); //it might do nothing if no polymorphic keyword has been found in the head
				String regolaTemp2 = replaceWord(regolaTemp1,RESERVED_KEYWORD_ORIGIN+substringBody,constructOriginal+substringBody);
				String regolaTemp3 = replaceWord(regolaTemp2,"SM_"+RESERVED_KEYWORD_ORIGIN+c,"SM_"+constructOriginal); //it might do nothing if there is only one "SM_GenericConstruct" and therefore it is not numbered; the instruction below will do in its stead
				String regolaTemp4 = replaceWord(regolaTemp3,"SM_"+RESERVED_KEYWORD_ORIGIN,"SM_"+constructOriginal); //it might do nothing if there are at least two "SM_GenericConstruct" and therefore they are numbered; the instruction above will do in its stead
				String regolaTemp5 = replaceWord(regolaTemp4,RESERVED_SKOLEM+c,skolemForConstruct); //...same story here with the Skolem Functions...		
				String regolaTemp6 = replaceWord(regolaTemp5,RESERVED_SKOLEM,skolemForConstruct); // ...
				/* it would have been useless to place checks for the instructions above, for if a certain RESERVED_KEYWORD is not found in the rule,
				 * the replaceWord method will simply do nothing
				 */
					
				/*
				 * Check whether we need to add the "append" string
				 */
				if(getNumberOfElementsForKey(resolved,constructOriginal)>1 || hasFurtherSpecs.equalsIgnoreCase("true")) {
					String regolaTemp7 = regolaTemp6.substring(0,regolaTemp6.length()-1); //remove ";"
					String regolaTemp8="";
					if(list.size()>1)  {//in case we have multiple polymorphic references within a single Letterale
						appendString = replaceWord(appendString,RESERVED_OID,RESERVED_OID+c);
											
						appendString = appendString.replaceAll(RESERVED_SUFFIX_FOR_ARGUMENT_RENAMING,RESERVED_SUFFIX_FOR_ARGUMENT_RENAMING+c);
						
						regolaTemp8 = regolaTemp7 + appendString;
					} else {
						regolaTemp8 = regolaTemp7 + appendString;
					}
					regolaTemp0 = regolaTemp8;
					
				
				} else { //no need to append anything
					regolaTemp0 = regolaTemp6;
				}
				//counter++;
					
			}//end of for(each element of the combination...)
			regolaIstanziata = regolaTemp0;
			//JOptionPane.showMessageDialog(frame,regola.toString(),"Regola iniziale",JOptionPane.WARNING_MESSAGE);
			JOptionPane.showMessageDialog(frame,regolaIstanziata,"Regola RISULTATO #"+k,JOptionPane.WARNING_MESSAGE);
			log.info(regolaIstanziata);
			
			List<String> allChildren = getAllChildrenOfGeneralizations();
			for(String child: allChildren) {
				if(child.equals(letteraleOriginal)) {
					hasAlreadyBeenAddedAsItsInstances = true;
					//JOptionPane.showMessageDialog(frame,letteraleOriginal+"' s polyrule is going to be added as its instances","WARNING",JOptionPane.WARNING_MESSAGE);
				}
			}
			if(hasAlreadyBeenAddedAsItsInstances==true) {
				RegolaDatalog regolaIst = PolyDatalogRulesReader.readString(regolaIstanziata);
				scannedRules.add(regolaIst);
				counterTotal++;
				JOptionPane.showMessageDialog(frame,letteraleOriginal+"' s polyrule HAS BEEN added as its instances","WARNING",JOptionPane.WARNING_MESSAGE);
				
			}
			
			
			k++;
				
			// WRITING TO DISK FOR TESTING PURPOSES
			//String name = regola.getTesta().getPredicateName();
			String simpleRulesPath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
			+ ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths");			
				
			ruleNameCounter++;
			String ruleCounter = getCounter(ruleNameCounter);
			String ruleFileName = "polyrule_"+ruleCounter+".dtr";
			String polyRulesPath = simpleRulesPath +"polyrules/";
			fileWrite(regolaIstanziata,polyRulesPath+macroruleFileName+"/","polyrule_"+ruleCounter+".dtr");		
			
			boolean mustIterate = regola.getMustIterate();
			String iterate = "";
			if(mustIterate==true) {
				iterate ="!!";
			}
			macroruleInstance = macroruleInstance + iterate + "polyrules/"+macroruleFileName+"/"+ruleFileName +"();\n";
			
			
		}//end of iteration over combinations
	}
			
	/**
	 * A method which replaces all the occurrences of a certain word within a String
	 * 
	 * @param original - the input String
	 * @param find - the word to be replaced
	 * @param replacement - the word to be used as replacement
	 * @return the resulting String where all occurrences of the specified word have been replaced
	 */
	private static String replaceWord(String original, String find, String replacement) {
	    String result = "";
	    String delimiters = "+-*/(),.:[]# \n";
	    StringTokenizer st = new StringTokenizer(original, delimiters, true);
	    while (st.hasMoreTokens()) {
	        String w = st.nextToken();
	        if (w.equals(find)) {
	            result = result + replacement;
	        } else {
	            result = result + w;
	            //System.out.println(w);
	        }
	    }
	    return result;
	}
	
	/**
	 * A method used for checking whether a Letterale included in a Datalog rule contains a polymorphic reference
	 * (i.e. makes the Datalog rule "polymorphic")
	 * 
	 * @param letterale - the Letterale to be checked
	 * @return true or false depending on the result of the check
	 */
	public static boolean isPolymorphic(Letterale letterale) {
		
		boolean isPolymorphic = false;
		//String lett = letterale.getPredicateName();
		//JOptionPane.showMessageDialog(frame,lett,"Nome del letterale",JOptionPane.INFORMATION_MESSAGE);
		
		for(ArgomentoEtichettato argomentoEtichettato : letterale.getListaArgomentiEtichettati()) {
			//Argomento argomento = argomentoEtichettato.getArgomento();
			/* The column names are the attribute names, so we need them */
			String attributo = argomentoEtichettato.getAttributo();        //ex. OID, Name, isOptional...
			//String nome_argomento = argomento.getNome();	//ex. 018, pincopallino, etc.
			
			/* Check whether the attribute name is a polymorphic reference */
			if(attributo.contains(RESERVED_KEYWORD_RESULT) || attributo.contains(RESERVED_KEYWORD_ORIGIN)) {
				isPolymorphic = true;
				//JOptionPane.showMessageDialog(frame,attributo+": POLIMORFICO!!!","Ricerca argomenti polimorfi",JOptionPane.INFORMATION_MESSAGE);
			}
			//JOptionPane.showMessageDialog(frame,attributo +": "+nome_argomento,"Scansione attributi-argomenti",JOptionPane.INFORMATION_MESSAGE);
		}
		return isPolymorphic;
	}
	
	/**
	 * This method, given a Datalog rule and a construct name (the latter as a string), checks whether such a construct is
	 * included or not within the body of the specified rule AND is the translated construct within that rule (i.e. not a supporting
	 * construct used for further specifications in the translation process). We call this the "main" construct of such a rule.  
	 * 
	 * @param rule - the Datalog rule to be scanned
	 * @param letterale_name - the name of the construct whose membership we desire to check within
	 * the rule, along with its status as the "main" construct of that rule (the translated one)
	 * @return true or false, depending on the result of the check
	 */
	public static boolean isMainConstruct(RegolaDatalog rule, String letterale_name) {
		boolean isMain = false; // we initialize the returning value
		/* We initialize the boolean values that will eventually make up the returning one */
		boolean isIncluded = false; 
		boolean isTranslated = false;
		boolean isFoundWithinReferences = false;
		boolean hasTMPhead = false;
		
		/* Let' s first find the Skolem function used in the OID creation of the head construct, which
		 * we will use later
		 */
		Letterale testa = rule.getTesta();
		if (testa.getIsTMP()==true) {
			hasTMPhead = true;
			return false;
		}
		
		String skolemFunctor = getSkolemForOIDFromLetterale(testa); // e.g. #AggregationOID_1*(oid)
				
		/* Let' s iterate over the constructs found within the given rule body 
		 * in order to carry out our checks */
		String body_SM_lett_name;
		String body_lett_name;
		String OIDvalue;
		for(Letterale body_lett: rule.getBody()) {
			body_SM_lett_name = body_lett.getPredicateName();
			String role = body_lett.getRuolo();
			body_lett_name = body_SM_lett_name.substring(3,body_SM_lett_name.length()); // the construct name without "SM_"
			if(body_lett_name.equals(letterale_name)) {
				//JOptionPane.showMessageDialog(frame,body_lett_name +" = " + letterale_name,"Inclusion check",JOptionPane.INFORMATION_MESSAGE);
				isIncluded = true;
				
				/* Ok, we have made sure the construct was included in the specified rule.
				 * Now let' s see whether such a construct is the one that is being translated in the first place:
				 * we do so by considering the OID value for the construct in the body and checking
				 * whether such a value is included or not in the Skolem Function used to generate the OID of
				 * the Letterale in the rule head, while also making sure that such construct is not a [DEST] construct
				 */
				OIDvalue = getOIDvalueFromLetterale(body_lett); //e.g. "oid", "pincoPallinoOID" ...
				System.out.println("SKOLEM FUNCTOR: "+skolemFunctor);
				List<String> args = getSkolemArguments(skolemFunctor);
				for(String a: args) {
					
					if(a.equals(OIDvalue) && !role.equalsIgnoreCase("[DEST]")) {
					//JOptionPane.showMessageDialog(frame,OIDvalue + " found within "+skolemFunctor,"Main check",JOptionPane.INFORMATION_MESSAGE);
						isTranslated = true;
					
						/* Finally let' s ascertain the fact that such an OID, while being within the Skolem functor for the head OID, is also NOT
						 * found within any other Skolem functor used for the head reference(s) (if it' s the case, the construct we are considering
						 * will surely not be a "main" one).
						 */
						List<String> skolemsForReferences = getSkolemsForReferences(testa);
						for(String s: skolemsForReferences) { //e.g.: AbstractOID_1(oid)
							List<String> ar = getSkolemArguments(s);
							for(String st: ar) {
								if(st.equals(OIDvalue)) {
									isFoundWithinReferences = true;
								}
							}	
						}
					}
				}	
			} //end of if(body_lett_name.equals...)
		}
		System.out.println(isIncluded + " " + isTranslated + " " + !isFoundWithinReferences + " " + !hasTMPhead+" in regola: "+rule.toString()+" for letterale: "+letterale_name+"...\n");
		isMain = isIncluded && isTranslated && !isFoundWithinReferences && !hasTMPhead;
		return isMain;
	}
	
	/**
	 * This method returns a list of the arguments (as strings) of a Skolem function (also as a string).
	 * 
	 * @param skolemFunctor - the string representing the Skolem function - e.g. AbstractOID_0*(oid)
	 * @return a list of strings representing the skolem arguments
	 */
	public static List<String> getSkolemArguments(String skolemFunctor) {
		List<String> arguments = new LinkedList<String>();
		//StringTokenizer st = new StringTokenizer(skolemFunctor,",");
		String[] args = skolemFunctor.split("\\(");
		String[] argum = null;
		for (int i=0; i<args.length; i++) {
			if(args[i].contains(",")) { // oid,oid1,ciaoid,strunz)
				args[i] = args[i].substring(0,args[i].length()-1);
				argum = args[i].split(","); 
				//System.out.println(args[i]);
			}
			else {
				argum = new String[1];
				argum[0] = args[i].substring(0,args[i].length()-1);
			}
		}
		for(int j=0; j<argum.length; j++) {
			arguments.add(argum[j]);
			//System.out.println("Token: "+ argum[j]);
		}
		return arguments;
	}
	
	/**
	 * This method returns the SkolemFunctor found as arguments of a head construct's references
	 * 
	 * @param lett - the head Letterale to be considered
	 * @return a list of Strings featuring the Skolem functions (e.g.: AbstractOID_0(absOID))
	 */
	public static List<String> getSkolemsForReferences(Letterale lett) {
		List<String> skolems = new LinkedList<String>();
		//String skolemFunctor ="";
		List<ArgomentoEtichettato> l = lett.getListaArgomentiEtichettati();
		for(ArgomentoEtichettato a: l) {	
			String s = a.getAttributo();
			if(!s.equalsIgnoreCase("OID")) {
				String arg = a.getArgomento().toString();
				if(arg.startsWith("#")) {
					skolems.add(arg);
					System.out.println("Construct "+lett.getPredicateName()+"; Skolem function for reference: "+arg);
				}
			}
		}	
		return skolems;
	}
		
	/**
	 * A simple supporting method that returns the SkolemFunctor with its arguments (as a string) used
	 * for creating the OID of the given Letterale
	 * 
	 * @param lett - the Letterale to be considered
	 * @return skolemFunctor - the String representing the Skolem functor with its arguments
	 */
	public static String getSkolemForOIDFromLetterale(Letterale lett) {
		String skolemFunctor ="";
		List<ArgomentoEtichettato> l = lett.getListaArgomentiEtichettati();
		for(ArgomentoEtichettato a: l) {	
			String s = a.getAttributo();
			if(s.equalsIgnoreCase("OID")) {
				String nome = a.getArgomento().getNome();
				skolemFunctor = a.getArgomento().toString();
				System.out.println(s +": " + nome +"; valore: " + skolemFunctor);
			}	
		}
		return skolemFunctor;
	}
	
	/**
	 * A simple supporting method that returns the SkolemFunctor w/o arguments (as a string) used
	 * for creating the OID of the given head Letterale
	 * 
	 * @param lett - the head Letterale to be considered
	 * @return skolemFunctor - the String representing the Skolem functor without its arguments
	 */
	public static String getSkolemWithoutArgsForOIDFromLetterale(Letterale lett) {
		String skolemFunctor ="";
		List<ArgomentoEtichettato> l = lett.getListaArgomentiEtichettati();
		for(ArgomentoEtichettato a: l) {	
			String s = a.getAttributo();
			if(s.equalsIgnoreCase("OID")) {
				skolemFunctor = a.getArgomento().getNome();
				//skolemFunctor = a.getArgomento().toString();
				//System.out.println(s +": " + nome +"; valore: " + skolemFunctor);
			}	
		}
		return skolemFunctor;
	}
		
	/**
	 * A simple supporting method which returns the OID value of a Letterale
	 * 
	 * @param lett - the Letterale considered
	 * @return OIDvalue - the value for the OID field of the given Letterale
	 */
	public static String getOIDvalueFromLetterale(Letterale lett) {
		String OIDvalue ="";
		List<ArgomentoEtichettato> list = lett.getListaArgomentiEtichettati();
		for (ArgomentoEtichettato arg: list) {
			String attributeName = arg.getAttributo();
			//System.out.println(attributeName);
			if(attributeName.equalsIgnoreCase("OID")) {
				OIDvalue = arg.getArgomento().getNome(); //the value for the OID field which we are going to use
				//System.out.println(OIDvalue);
			}
		}
		return OIDvalue;
	}
	
	/**
	 * A simple supporting method which returns the "Name" value of a Letterale
	 * 
	 * @param lett - the Letterale considered
	 * @return nameValue - the value for the "Name" field of the given Letterale
	 */
	public static String getNameValueFromLetterale(Letterale lett) {
		String nameValue ="";
		List<ArgomentoEtichettato> list = lett.getListaArgomentiEtichettati();
		for (ArgomentoEtichettato arg: list) {
			String attributeName = arg.getAttributo();
			//System.out.println(attributeName);
			if(attributeName.equalsIgnoreCase("name")) {
				nameValue = arg.getArgomento().getNome(); //the value for the OID field which we are going to use
				//System.out.println(OIDvalue);
			}
			
		}
		return nameValue;
	}
	
	/** 
	 * Just a method to write a .txt file from a String
	 * 
	 * @param string    - the content of the text file to be written
	 * @param path      - the path where the file is to be written at
	 * @param filename  - the name of the file
	 */
	public static void fileWrite(String string, String path, String filename) {
		File file = new File(path + filename);
		try {
			FileWriter writer = new FileWriter(file,true);
			writer.write(string);
			writer.close();
		}
		catch(IOException e) {
			JOptionPane.showMessageDialog(frame,e.getMessage(),"ERROR IN WRITING THE FILE",JOptionPane.ERROR_MESSAGE);
		}
	}	
		
	/**
	 * Test method for this class
	 * @param args
	 */
	public static void main(String[] args) {
		//PolyHandlerMain poly = 
			new PolyHandlerMain("Object-Relational");
		//PolyHandlerMain poly = getInstance("Object-Relational");
		//poly.polyHandlerMainRoutine("polyrule.txt");
		String macroRulePath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
		+ ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths") +"poly2/";			
		RegolaDatalog regola = null;
		try {
			regola = PolyDatalogRulesReader.read(macroRulePath+"binAggOfAbs4FakeNaryOne2ManyAggOfAbs.dtr");
		} catch(Exception e) {
			System.out.println(e.getMessage());
		}
		//System.out.println(getBodyAsString(regola));
		//System.out.println(preProcessBodyNew(regola,"BinaryAggregationOfAbstracts","cassiusOID"));
		//System.out.println("Abstract is Main? "+isMainConstruct(regola,"Abstract"));
		//System.out.println("BinaryAggregationOfAbstracts is Main? "+isMainConstruct(regola,"BinaryAggregationOfAbstracts"));
		//System.out.println("AggregationOfAbstracts is Main? "+isMainConstruct(regola,"AggregationOfAbstracts"));
		//getSkolemArguments("#AbstractOID_1*(oid)");
		//PolyHandlerConfigurationManager manager = PolyHandlerConfigurationManager.getInstance();
		//gen = manager.getConfiguration();
		//getAllChildrenOfGeneralizations();
		String s = regola.toString();
		JOptionPane.showMessageDialog(frame,s,"Regola in stringa",JOptionPane.INFORMATION_MESSAGE);
		try {
			RegolaDatalog r =  PolyDatalogRulesReader.readString(s);
			System.out.println(r.toString());
		}
		catch(Exception e) {
			System.out.println(e.getMessage());
		}
	}	
}