package org.modelgen.polyhandler;

import org.modelgen.regole.Argomento;
import org.modelgen.regole.Condizione;
import org.modelgen.regole.CondizioneSelect;
import org.modelgen.regole.Letterale;
import org.modelgen.regole.Manipolabile;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.regole.SkolemFunction;
import org.modelgen.regole.Tabella;
import org.modelgen.regole.Exception.FunctorVariableNotFoundException;
import org.modelgen.regole.Exception.VarNotFoundException;
import org.modelgen.regole.utilities.CollectionFormatter;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * Rappresenta un funtore di Skolem in una RegolaDatalog
 * @author luigi, francesca
 * @author modified by Daniele for Polymorphic rules
 * 
 */
public class PolySkolemFunctor implements Argomento {
	private SkolemFunction funzione; // funzione su cui � definito
	private boolean isTargetSkolem;  // termina o meno per *
	private List<Manipolabile> argomenti; // lista di argomenti manipolabili
	private Letterale letterale; // per tradurre il ruolo a partire da quello del letterale di appartenenza
	
	
	public PolySkolemFunctor(SkolemFunction fn, boolean isTargetSkolem) {
		argomenti = new LinkedList<Manipolabile>();
		funzione = fn;
		this.isTargetSkolem = isTargetSkolem;		
	}
	
	public String getNome() {
		return funzione.getSkolemFunctionName();
	}
	
	public void setLetterale(Letterale letterale) {
		this.letterale = letterale;
	}
	
	public String getTargetOID(){
		return funzione.getTargetOID();
	}
	
	public String getTargetTableSK() {
		return funzione.getTagetTable();
	}
	
	/**
	 * ritorna il letterale di appartenenza
	 */
	public Letterale getLetterale() {
		return this.letterale;
	}
	
	public List<String> getArgomentiFunzione() {
		return funzione.getArgomenti();
	}
		
	public void setFunzione(SkolemFunction funzione) {
		this.funzione = funzione;
	}
	
	public SkolemFunction getFunzione() {
		return this.funzione;
	}
	public String getTargetTable() {
		return this.funzione.getTagetTable();
	}
	
	public void setIsTargetSkolem(boolean b) {
		this.isTargetSkolem = b;
	}
	
	public boolean getIsTargetSkolem() {
		return this.isTargetSkolem;
	}
	
	public void addArgomento(Manipolabile vd) {
		this.argomenti.add(vd);
	}
	
	public String getTargetAttribute(){
		return funzione.getTargetOID();
	}
	
	/**
	 * Ritorna se questo funtore � di istanza o meno
	 * @return se questo funtore � di istanza o meno
	 */
	public boolean isInstance() {
		return this.getFunzione().isInstance();
	}

		
	public Set<PolyCondizioneSkolem> generaCondizioni(String var, Map<String, Tabella> tabelle, String attributo) throws FunctorVariableNotFoundException{
		PolyCondizioneSkolem cs = new PolyCondizioneSkolem(attributo, var, this);
		Iterator<Manipolabile> it = argomenti.iterator();
		
		Manipolabile m = null;
		while (it.hasNext()){
			m = it.next();
			try {
				cs.addArgomento(m.generaSemicondizione(tabelle));
			} catch(VarNotFoundException ex) {
				throw new FunctorVariableNotFoundException(getNome(),ex);				
			}
		}
		Set<PolyCondizioneSkolem> condizioni = new HashSet<PolyCondizioneSkolem>();
		condizioni.add(cs);
		return condizioni;
	}
	
	public CondizioneSelect getSelect(Map<String, Tabella> tabelle) throws FunctorVariableNotFoundException {
		PolyCondizioneSelectSkolem ssc = new PolyCondizioneSelectSkolem(this);
		Iterator<Manipolabile> it = argomenti.iterator();
		Manipolabile m = null;
		List<String> semiCondizioni = new LinkedList<String>();
		while(it.hasNext()) {
			m = it.next();
			try {
				semiCondizioni.add(m.generaSemicondizione(tabelle));
			} catch(VarNotFoundException ex) {
				throw new FunctorVariableNotFoundException(getNome(),ex);
			}
			
			
		}
		
		ssc.setArgomenti(semiCondizioni);
		return ssc;
		
	}	
	
	
	/**
	 * Null in quanto verrebbe invocato solo quando non ci sono funtori di 
	 * Skolem nella testa, ma non essendoci non ha possibilit� di essere invocato.
	 */
	public Condizione getTMP(String var, Map<String, Tabella> tabelle, String attributo) {
		return null;
	}
	
	public Argomento getArgomentoPerIstanza(RegolaDatalog nuova, Letterale newLetterale) {
		/* crea il funtore basato sulla tabella di Skolem per istanza */
		PolySkolemFunctor newSkolemFunctor = new PolySkolemFunctor(this.getFunzione().getFunzionePerIstanza(),this.getIsTargetSkolem());
		Iterator<Manipolabile> it = argomenti.iterator();
		Manipolabile argomento = null;
		Manipolabile newArgomento = null;
		while(it.hasNext()) {
			argomento = it.next();
			/* si genera l'argomento per istanza corrispondente, true perch� se � una variabile segnale che */
			/* questa si trova in un funtore */
			newArgomento = argomento.getManipolabilePerIstanza(nuova,newLetterale,true);
			newSkolemFunctor.addArgomento(newArgomento);
			
		}
		
		return newSkolemFunctor;
		
	}
	
	public Argomento duplicate(RegolaDatalog nuova, Letterale newLetterale, boolean isSpecialContext) {
		/* la copia in un contesto speciale non genera mai funtori target */
		PolySkolemFunctor skf = new PolySkolemFunctor(this.getFunzione(),isSpecialContext ? false : this.getIsTargetSkolem());
		Iterator<Manipolabile> it = argomenti.iterator();
		Manipolabile argomento = null;
		Manipolabile newArgomento = null;
		while(it.hasNext()) {
			argomento = it.next();
			newArgomento = argomento.getCopia(nuova,newLetterale,true);
			skf.addArgomento(newArgomento);
		}
		
		return skf;
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("#"+funzione.getSkolemFunctionName()+ (isTargetSkolem ? "*" : ""));
		buf.append("(");
		buf.append(CollectionFormatter.getInstance().getSequenzaComponenti(argomenti));
		buf.append(")");
		return buf.toString();
	}

	/**
	 * @return the argomenti
	 */
	public List<Manipolabile> getArgomentiFuntore() {
		return this.argomenti;
	}
	
	/*
	 * (non Javadoc)
	 * Ritorna null, non avendo alcun valore associato utile per scrivere
	 * la segnatura di una regola
	 * @see org.modelgen.regole.Argomento#getValore()
	 */
	public String getValore() {
		return null;
	}
	

}
