/*
 * Created on 15-giu-2005
 *
 */
package org.modelgen.regole;

import org.modelgen.applyrule.MacroRule;
import org.modelgen.applyrule.SimpleRule;
import org.modelgen.applyrule.Statement;
import org.modelgen.regole.Exception.InstanceFromTempException;
import org.modelgen.regole.Exception.InstanceTranslationException;
import org.modelgen.regole.utilities.CollectionFormatter;
import org.modelgen.regole.utilities.TranslatorWorker;
import org.modelgen.signature.MacroRuleSignature;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;


/**
 * Rappresenta una macroregola, ossia un insieme di regole Datalog.
 * @author luigi, francesca
 *
 */
public class MacroRegola {
	
	private List<RegolaDatalog> regole; // le regole Datalog che la compongono
	private MacroRuleSignature signature;	//segnatura della macroregola
	
	public MacroRegola() {
		regole = new LinkedList<RegolaDatalog>();
	}
	
	public void addRegola(RegolaDatalog rd) {
		regole.add(rd);
	}
	
	/**
	 * Restituisce la segnatura della macroregola
	 */
	public MacroRuleSignature getSignature(){
		if(signature==null)
			signature = new MacroRuleSignature(this);
		return signature;
	}
	
	public MacroRuleSignature getSignature(String fileName){
		if(signature==null)
			signature = new MacroRuleSignature(fileName, this);
		return signature;
	}
	
	
	/**
	 * returns the number of different source schemes in the macrorule 
	 */
	public int getSourceSchemesNumber() {
		
		Set<String> source = new HashSet<String>();		
		for(RegolaDatalog rd : this.regole)
			source.addAll(rd.getSourceSchemesNames());
		
		return source.size();
		
	}
	
	/**
	 * returns the number of different dest schemes in the macrorule
	 */
	public int getDestSchemesNumber() {
		Set<String> dest = new HashSet<String>();		
		for(RegolaDatalog rd : this.regole)
			dest.addAll(rd.getDestSchemesNames());
		
		return dest.size();
		
	}
	
	/**
	 * setta il numero di schemi sorgente e destinazione nelle
	 * regole datalog
	 *
	 */
	private void setSourceDestNumbers() {
		
		int sourceNumber = this.getSourceSchemesNumber();
		int destNumber = this.getDestSchemesNumber();
		
		for(RegolaDatalog rd: regole) {
			rd.setSourceNumber(sourceNumber);
			rd.setDestNumber(destNumber);
		}
			
		
	}
	
	/**
	 * Istanzia una MacroRule a partire dalla traduzione SQLB
	 * di tutte le sue regole Datalog.
	 * @param nomeFile il nome del file della MacroRule
	 * @return la MacroRule da applicare
	 */
	public MacroRule istanziaMacroRuleSQL(String nomeFile) {
		
		setSourceDestNumbers();
		
		MacroRule mr = new MacroRule();
		mr.setFilename(nomeFile);
		Iterator<RegolaDatalog> it = regole.iterator();
		RegolaDatalog rd = null;
		RegolaB rb = null;
		SimpleRule sr = null;
		
		/* traduce tutte le regole in multithreading */
		TranslatorWorker tw = new TranslatorWorker();
		List<Regola> l = tw.traduci(regole, false);
		int contatore = 0; // per accede alla giusta traduzione
		
		while(it.hasNext()) {
			rd = it.next();		
			rb = (RegolaB)l.get(contatore); // downcasting sicuro, invocazione di traduci con false
			contatore++;
			sr = new SimpleRule();
			sr.setMustBeIterated(rd.getMustIterate());
			for(String st : rb.getStatements()) {
				sr.addStatement(new Statement(st,false));
			}
			mr.addRule(sr);	
		}
		
		return mr;
	}
	
	public String toString() {
		return CollectionFormatter.getInstance().getSequenzaComponenti(regole,"\n*************************\n");
		
	}
	
	/**
	 * @param tradottoA se si deve generare la traduzione SQLA vale true altrimenti false
	 * @return tutte le regole Datalog contenute tradotte
	 */
	private List<Regola> traduciTutte(boolean tradottoA) {
		
		setSourceDestNumbers();
		
		List<Regola> l = new LinkedList<Regola>();
		TranslatorWorker tw = new TranslatorWorker();
		
		
		/* traduce le regole in multithreading */
		try {
			l = tw.traduci(regole,tradottoA);
		} catch(Exception ex) { }
		
		
		return l;
	}
	
	/**
	 * Data una macroregola sugli schemi, genera la corrispondente per istanza
	 * @return la corrispondente regola per istanza
	 * @throws InstanceTranslationException 
	 */
	public MacroRegola getMacroRegolaPerIstanza() throws InstanceTranslationException {
		
		MacroRegola mr = new MacroRegola();
		
		/* aggiunge alla nuova macroregola la traduzione per istanza di ogni regola Datalog */
		/* della macroregola precedente */
		for(RegolaDatalog rd : regole) { 
			try {
				mr.addRegola(rd.traduciPerIstanza());
			} catch (InstanceFromTempException e) {
//				do nothing: TMP literals do not produce a Datalog rule for 
//				the instance level
				System.out.println("TMP: I don't produce nothing!" + rd);
			} 
//			catch (InstanceTranslationException e) {
//				do nothing
//				System.out.println("ERROR: " + rd);
//			}
		}
		
		return mr;
		
	}
	 
	
	/**
	 * 
	 * @param tradottoA se si deve generare la traduzione SQLA vale true altrimenti false
	 * @return la stringa rappresentante la macroregola tradotta
	 */
	public String toStringTradotto(boolean tradottoA) {
		return CollectionFormatter.getInstance().getSequenzaComponenti(traduciTutte(tradottoA),"\n*************************\n");
	}
	
	public List<RegolaDatalog> getRegole() {
		return regole;
	}
	

}
