/*
 * Created on 14-mag-2005
 *
 */
package org.modelgen.regole;

import org.modelgen.regole.Exception.ATranslationException;
import org.modelgen.regole.Exception.ConditionCreationException;
import org.modelgen.regole.Exception.InstanceFromTempException;
import org.modelgen.regole.Exception.InstanceTranslationException;
import org.modelgen.regole.Exception.LiteralTranslationException;
import org.modelgen.regole.utilities.CollectionFormatter;
import org.modelgen.regole.utilities.VarGenerator;
import org.modelgen.signature.RuleSignature;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * Rappresenta una regola Datalog
 * @author luigi, francesca
 * 
 *
 */
public class RegolaDatalog {
	
	public final static String SOURCE = "[SOURCE]";
	public final static String DEST = "[DEST]";
	private Letterale testa;
	private List<Letterale> corpo;     // di tipo Letterale
	private List<CondizioneDatalog> condizioni;           // di tipo CondizioneDatalog
	private Map<String,VariabileDatalog> variabili;             // le variabili contenute nella regola datalog
	private VarGenerator varGenerator; // per generare gli ID dei letterali
	private boolean isApplyDistinct;   // se nella traduzione verso SQL la select � con distinct
	private boolean mustIterate;       // se in una MacroRegola deve essere iterata
	
	public RegolaDatalog() {
		corpo = new LinkedList<Letterale>();
		condizioni = new LinkedList<CondizioneDatalog>();
		variabili = new HashMap<String,VariabileDatalog>();
		varGenerator = new VarGenerator();
		mustIterate = false;
	}
	
	/**
	 * setta il numero di schemi sorgente dell'intera macroregola nei letterali 
	 * @param sourceNumber
	 */
	public void setSourceNumber(int sourceNumber) {
		
		for(Letterale l: this.corpo)
			l.setSourceNumber(sourceNumber);
		testa.setSourceNumber(sourceNumber);
		
	}
	
	/**
	 * setta il numero di schemi destinazione dell'intera macroregola nei letterali
	 * @param destNumber
	 */
	public void setDestNumber(int destNumber) {
		
		for(Letterale l: this.corpo)
			l.setDestNumber(destNumber);
		testa.setDestNumber(destNumber);
		
	}
	
	/**
	 * returns a set with the different source schemes in the rule (body)
	 */
	public Set<String> getSourceSchemesNames() {
		
		Set<String> sources = new HashSet<String>(); // no duplicates in a set
		for(Letterale l: corpo)
			if(l.getRuolo().startsWith(Letterale.SOURCE.substring(0,Letterale.SOURCE.length()-1)))
				sources.add(l.getRuolo());
		
		return sources;
	}
	
	/**
	 * returns a set with the different dest schemes in the rule (body + head)
	 */
	public Set<String> getDestSchemesNames() {
		
		Set <String> dests = new HashSet<String>();
		for(Letterale l: corpo)
			if(l.getRuolo().startsWith(Letterale.DEST.substring(0,Letterale.DEST.length()-1)))
				dests.add(l.getRuolo());
		
		if(testa.getRuolo().startsWith(Letterale.DEST.substring(0,Letterale.DEST.length()-1)))
			dests.add(testa.getRuolo());
		
		return dests;
		
	}
	
	//restituisce la segnatura
	public RuleSignature getRuleSignature(){
		return new RuleSignature(this);
	}
	
	public void setTesta(Letterale l) {
		this.testa = l;
	}
	
	public Letterale getTesta() {
		return this.testa;
	}
	
	public List<Letterale> getBody() {
		return corpo;
	}

	
	public void addElementoCorpo(Letterale l) {
		l.setID(varGenerator.getID());
		corpo.add(l);
	}
	
	public void setIsApplyDistinct(boolean b) {
		isApplyDistinct = b;
	}
	
	public boolean getIsApplyDistinct() {
		return isApplyDistinct;
	}
	
	public void setMustIterate(boolean b) {
		this.mustIterate = b;
	}
	
	public boolean getMustIterate() {
		return this.mustIterate;
	}
	
	public boolean haCondizioni() {
		return !condizioni.isEmpty();
	}
	
	/**
	 * 
	 * @param nome il nome della variabile
	 * @param l il letterale a cui deve essere associata la variabile
	 * @param isSkolem se la variabile compare nel letterale in un funtore di skolem
	 * @return la variabile associata a quel nome. Se non c'� viene creata e aggiunta alla mappa
	 */
	public VariabileDatalog getVariabile(String nome,Letterale l, boolean isSkolem){
		VariabileDatalog var = this.variabili.get(nome);
		if (var == null) {
			var = new VariabileDatalog(nome);
			variabili.put(nome,var);
		}
		
		/* se viene aggiunta una variabile nell'ambito di un funtore */
		/* la variabile ne � a conoscenza */
		if(!isSkolem)
			var.addLetterale(l);
		else
			var.addLetteraleFuntore(l);
		
		return var;		
	}
	
	
	/**
	 * 
	 * @param op1 il primo operando
	 * @param op2 il secondo operando
	 * @param operatore l'operatore
	 */
	public void aggiungiCondizione(String op1, String op2, String operatore) throws ConditionCreationException {
		Manipolabile man1 = this.variabili.get(op1);
		Manipolabile man2 = this.variabili.get(op2);
		
		
		/* Gestione delle costanti: */
		/* dal parser arriva sempre una stringa, se questa non esiste come variabile Datalog */
		/* e inizia con un apice allora si genera la corrispondente costante. Se non inizia */
		/* con l'apice significa che si � usata per la condizione una variabile */
		/* non presente nei letterali del corpo e si genera un'eccezione. */
		if (man1 == null && (((op1.startsWith("'") && op1.endsWith("'")) || (op1.toLowerCase().equals("true") ||op1.toLowerCase().equals("false"))))) 
			man1 = new CostanteDatalog(op1);
		else if (man1 == null)
			throw new ConditionCreationException(op1);
		if (man2 == null && (((op2.startsWith("'") && op2.endsWith("'")) || (op2.toLowerCase().equals("true") ||op2.toLowerCase().equals("false"))))) 
			man2 = new CostanteDatalog(op2);
		else if (man2 == null)
			throw new ConditionCreationException(op2);
			
		
		CondizioneDatalog condizione = new CondizioneDatalog(man1, man2, operatore);
		this.condizioni.add(condizione);		
	}
	/**
	 * Traduce questa regola Datalog in una regola di SQL alto
	 * @return una regola SQL
	 */
	public Regola traduciInSQLAlto() throws ATranslationException {
		Regola regola = new Regola(isApplyDistinct);
		try {
			regola.traduciCorpo(corpo);
			regola.traduciCondizioni(condizioni);		
			regola.traduciTesta(testa);
		} catch(LiteralTranslationException ex) {
			throw new ATranslationException(ex);
		}
		
		return regola;
	}
	
	
	
	/**
	 * Crea la corrispondente regola Datalog per le istanze
	 * @return una regola Datalog per le istanze
	 * @throws InstanceTranslationException 
	 */
	public RegolaDatalog traduciPerIstanza() throws InstanceTranslationException {
		
//		System.out.println("**************");
//		System.out.println("REGOLA: " + this);
//		System.out.println("**************");
//		tentative: put schema selection in instance translation 
//		if (this.getTesta().getIsTMP()) return this;
			
		/* se è senza corpo eccezione */
		if(this.corpo.isEmpty())
			throw new InstanceTranslationException("Non si può tradurre per istanza una regola senza corpo"); // da dove verrebbero i dati?
		
		
		RegolaDatalog regolaPerIstanza = new RegolaDatalog();
		regolaPerIstanza.setIsApplyDistinct(this.isApplyDistinct);
		regolaPerIstanza.setCorpoPerIstanza(this.corpo);
		regolaPerIstanza.setCondizioniPerIstanza(this.condizioni);
//		paoloc 20070523
//		let's see if iteration work for instances too
		regolaPerIstanza.setMustIterate(this.getMustIterate());
		try {
			regolaPerIstanza.setTestaPerIstanza(this.testa);
		} catch(InstanceFromTempException ex) {
			throw ex;
//				throw new InstanceTranslationException(ex);
		}
		
		
		
		
		return regolaPerIstanza;
		
	}
	
	/**
	 * Data la testa della regola Datalog costruisce la testa della
	 * regola Datalog per istanza
	 * @param testa la testa della regola Datalog originale
	 */
	public void setTestaPerIstanza(Letterale testa) throws InstanceFromTempException, InstanceTranslationException {
		
		/* viene creato il nuovo letterale di testa, si specifica */
		/* come parametro la regola per poter costruire variabili */
		
//		if(testa.getIsTMP()) {
//			InstanceFromTempException ex = new InstanceFromTempException(testa.getPredicateName());
//			throw ex;
//		}
		
		Letterale newTesta = testa.getLetteralePerIstanza(this); 
		this.testa = newTesta;
	}
	
	/**
	 * Dato il corpo della regola Datalog, costruisce il corpo
	 * della regola Datalog per istanza
	 * @param corpo la lista dei letterali del corpo
	 * @throws InstanceTranslationException 
	 */
	public void setCorpoPerIstanza(List<Letterale> corpo) throws InstanceTranslationException {
		Iterator<Letterale> it = corpo.iterator();
		Letterale letterale = null;
		Letterale letteraleIstanza = null;		
		while(it.hasNext()) {
			/* per ogni letterale inserisce nel corpo una copia del letterale stesso */
			/* e una sua traduzione per istanza */
			letterale = it.next();
			/* se il letterale non � temporaneo o negato lo traduce per istanza e poi lo incolla */
			/* altrimenti lo copia solamente. */
			if(!letterale.getIsTMP() && !letterale.getIsNegato()) {				
				letteraleIstanza = letterale.getLetteralePerIstanza(this);
				this.addElementoCorpo(letteraleIstanza);				
			}
			if (letterale.getIsTMP() && !letterale.getIsNegato()) {
				letteraleIstanza = letterale.getLetteralePerIstanza(this);
				this.addElementoCorpo(letteraleIstanza);
			}
			letterale = letterale.duplicate(this);
			this.addElementoCorpo(letterale);
		}
	}
	
	/**
	 * Aggiunge le condizioni alla regola per istanza a partire dalle 
	 * condizioni della regola di origine
	 * @param condizioni le condizioni della regola di origine
	 */
	public void setCondizioniPerIstanza(List<CondizioneDatalog> condizioni) {
		Iterator<CondizioneDatalog> it = condizioni.iterator();
		CondizioneDatalog cd = null;
		Manipolabile m1 = null;
		Manipolabile m2 = null;
		String operatore = null;
		while(it.hasNext()) {
			cd = it.next();
			m1 = cd.getOperando1();
			m2 = cd.getOperando2();
			operatore = cd.getOperatore();
			
			/* la regola sa aggiungere condizioni a se stessa */
			/* a partire dall'espressione della condizione */
			try {
				this.aggiungiCondizione(m1.toString(),m2.toString(),operatore);
			} catch(ConditionCreationException ex) {
				/* la regola tradotta � stata gi� caricata */
				/* di conseguenza � corretta, quest'eccezione */
				/* non pu� essere mai sollevata */
			}
		}
		
	}
	
	/**
	 * Ritorna la lista delle variabili value
	 * @return la lista delle variabili value
	 */
	public List<VariabileDatalog> getVariabiliValue() {
		Set<String> chiavi = variabili.keySet();
		Iterator<String> it = chiavi.iterator();
		List<VariabileDatalog> value = new LinkedList<VariabileDatalog>();
		String chiave= null;
		while(it.hasNext()) {
			chiave = it.next();
			if(chiave.toLowerCase().startsWith("value"))
				value.add(variabili.get(chiave));
		}
		
		return value;
	}
	
	
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(testa.toString());
		buf.append("\n");
		buf.append(isApplyDistinct ? "<*-" : "<-" + "\n");
		buf.append(CollectionFormatter.getInstance().getSequenzaComponenti(corpo,",\n"));
		if(haCondizioni()) {
			buf.append(",");
			buf.append("\n" + CollectionFormatter.getInstance().getSequenzaComponenti(condizioni,",\n"));
		}
		buf.append(";");
		return buf.toString();
	}

}
