package org.modelgen.regole.reader;

import org.modelgen.commons.utils.ModelgenConfigurationManager;
import org.modelgen.regole.MacroRegola;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.regole.Exception.InstanceTranslationException;
import org.modelgen.regole.utilities.InstanceRulesDuplicateGenerationException;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * Generate the repository for the instance rules.
 * For each macrorule, the class generates: a folder with the same
 * name of the macrorule and a list of files with the simplerules 
 * for the instance, and a macrorule for the instance.
 * 
 * @author paolo
 *
 */
public class GenerateInstanceRuleRepository {
	
	/**
	 * 
	 */
	private static final String FILE_SUFFIX_IN_MACRO = ".dtr();";
	private static final String FILE_EXTENSION = ".dtr";
	private String rootnode = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths");
	private String macroRulesPath = ModelgenConfigurationManager.getIstance().getProperty("macrorules", "", "paths");
	private String simpleRulesPath = ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths");
//	private String instanceRuleRoot = config.getProperty("instanceRulesRoot", "", "paths");
	private String instanceMacroRulesPath = ModelgenConfigurationManager.getIstance().getProperty("instanceMacroRules", "", "paths");
	private String instanceSimlpeRulesPath = ModelgenConfigurationManager.getIstance().getProperty("instanceSimlpeRules", "", "paths");
	
	/**
	 * Translate a schema macrorule file to an instance rule, creating the corresponding file.
	 * @param sourceMacroRuleFileName the schema macrorule file
	 * @param overwrite if overwrite existing target file
	 * @throws IOException errors in file writing
	 * @throws InstanceRulesDuplicateGenerationException if (overwrite==false) and macrorule or simplerule for instance data already exists
	 * @throws InstanceTranslationException errors in translation from scheme rule to instance rule 
	 */
	public void createInstanceMacroRuleFile(String sourceMacroRuleFileName, 
			boolean overwrite) throws IOException, InstanceRulesDuplicateGenerationException, InstanceTranslationException {
//		String instanceRootFolderName = rootnode + instanceRuleRoot;
		String instanceSimpleRulesFolderName = rootnode + simpleRulesPath + instanceSimlpeRulesPath + sourceMacroRuleFileName;
		String instanceMacroRulesFolderName = rootnode + macroRulesPath + instanceMacroRulesPath;
		System.out.println("instanceMacroRulesFolderName: " + instanceMacroRulesFolderName);
		System.out.println("instanceSimpleRulesFolderName: " + instanceSimpleRulesFolderName);
//		System.out.println("Ceeeeeecccccioooooooooo! " + overwrite);
		File instanceSimpleRulesFolder = new File(instanceSimpleRulesFolderName);
		File instanceMacroRuleFile = new File(instanceMacroRulesFolderName + 
				File.separator + sourceMacroRuleFileName);
//		check if dataMacroFile or dataSimpleRulesFolder exists: 
//		if yes, then throws an exception to avoid overwrinting.
		if (!overwrite) {
			if (instanceSimpleRulesFolder.exists() || 
					instanceMacroRuleFile.exists())
				throw new InstanceRulesDuplicateGenerationException("MacroRule or " +
						"simplerules for instance data already exists for" +
						"the specific macrorule for schema. Schema macrorule:" +
						sourceMacroRuleFileName);
		}
//		if this point is reached then we are allowed to, eventually, overwrite.
		
//		first, check if the instance rules root folder exists
//		System.out.println(instanceRootFolderName);
//		File instanceRootFolder = new File(instanceRootFolderName);
//		if (!instanceRootFolder.exists()) instanceRootFolder.mkdir();
		
//		create (replacing the existing one) the instance simple rules folder
		if (instanceSimpleRulesFolder.exists()) instanceSimpleRulesFolder.delete();
		instanceSimpleRulesFolder.mkdirs();
		
//		do the same for the instance macro rule folder
		File instanceMacroRulesFolder = new File(instanceMacroRulesFolderName);
		if (!instanceMacroRulesFolder.exists()) {
			instanceMacroRulesFolder.mkdir();
		}
		
//		Directories are OK.
//		Now read the macrorule
		String sourceMacroRulePath = rootnode +  macroRulesPath + sourceMacroRuleFileName;
		MacroRegola mr = ParametricDatalogRulesReader.read(sourceMacroRulePath);
		MacroRegola mrIstanza = mr.getMacroRegolaPerIstanza();
		
//		Write the generated MacroRegola to file
		
//		create the macro file
		String macroRuleInstancePath = instanceMacroRulesFolderName + sourceMacroRuleFileName;
		System.out.println("macroRuleInstancePath: " + macroRuleInstancePath);
		File macroFile = new File(macroRuleInstancePath);
		try {
			if (!macroFile.createNewFile()) {
				macroFile.delete();
				macroFile.createNewFile();
			}
		} catch (SecurityException e) {
			System.out.print("Security exception in: " + this.getClass().getName());
			e.printStackTrace();
		}
		
		PrintWriter macroFileWriter = new PrintWriter(macroFile);
		
		int ruleCounter = 1;
//		System.out.println("Size: " + mrIstanza.getRegole().size())
		for (RegolaDatalog regolaDatalogIstanza : mrIstanza.getRegole()) {
			String regolaDatalogIstanzaFileName = instanceSimpleRulesFolderName + File.separator + ruleCounter + FILE_EXTENSION;
//			System.out.println(regolaDatalogIstanza);
			PrintWriter regolaDatalogIstanzaFile = new PrintWriter(regolaDatalogIstanzaFileName);
			regolaDatalogIstanzaFile.println(regolaDatalogIstanza.toString());
			regolaDatalogIstanzaFile.close();
			if (regolaDatalogIstanza.getMustIterate()) macroFileWriter.print("!!");
			macroFileWriter.println(instanceSimlpeRulesPath
					+ sourceMacroRuleFileName
//					the path separator is forced to /
//					+ File.separator + ruleCounter
					+ "/" + ruleCounter 
					+ FILE_SUFFIX_IN_MACRO);
			ruleCounter++;
		}
		macroFileWriter.close();
	}

	/**
	 * @param args
	 * @throws IOException 
	 * @throws InstanceRulesDuplicateGenerationException 
	 */
//	public static void main(String[] args) throws IOException, InstanceRulesDuplicateGenerationException {
//		GenerateInstanceRuleRepository test = new GenerateInstanceRuleRepository();
//		String fileName = "XSDToERBinary.txt";
//		test.createInstanceMacroRuleFile(fileName, false);
//
//	}

}
