package org.modelgen.signature;

import org.modelgen.signature.application.ruleapplication.ServicesApplication;
import org.modelgen.signature.util.ToolTableaux;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Oggetto che prende le segnature di due modelli, uno di partenza e uno di
 * arrivo, e ne fa la differenza sui costrutti, cio tiene le informazioni
 * su costrutti aggiunti o rimossi. 
 * Inoltre c' una mappa che contiene i costrutti ignorati: per ogni costrutto
 * c' una formula che indica se qualche propriet del costrutto  stata ignorata.
 * Se la formula vale false, il costrutto non  stato ignorato dall'aplicazione
 * della regola.
 * @author Fabrizio Celli
 * @date 05-06-07
 */
public class DiffContainer {

	private ModelSignature sorgente;
	private ModelSignature arrivo;

	//liste di costrutti modificati rispetto al modello sorgente
	private List<String> aggiunti;
	private List<String> rimossi;

	//mappa delle propriet ignorate per i costrutti dello schema sorgente
	private Map<String, String> ignorate;
	
	//macroregola che ha portato da sorgente ad arrivo
	private MacroRuleSignature macro;

	/*
	 * Costruttore
	 */
	public DiffContainer(ModelSignature sorgente, ModelSignature arrivo, MacroRuleSignature macro){
		this.sorgente = sorgente; 
		this.arrivo = arrivo;
		this.macro = macro;
		this.aggiunti = new LinkedList<String>();
		this.rimossi = new LinkedList<String>();
		this.evalConstructs();
		this.calculateIgnoredProperty();
	}

	/*
	 * metodi Getter
	 */
	public List<String> getAggiunti() {
		return aggiunti;
	}

	public ModelSignature getArrivo() {
		return arrivo;
	}

	public List<String> getRimossi() {
		return rimossi;
	}

	public ModelSignature getSorgente() {
		return sorgente;
	}

	public Map<String, String> getIgnorate() {
		return ignorate;
	}

	/*
	 * Esegue le differenze sui costrutti dei due modelli
	 */
	private void evalConstructs(){
		//costruisco l'insieme dei nomi dei costrutti del modello sorgente e di arrivo
		Set<String> nomiConsSourceSet = this.sorgente.getCostrutti().keySet();
		Set<String> nomiConsTargetSet = this.arrivo.getCostrutti().keySet();
		//costruisco le liste dei nomi dei costrutti del modello sorgente e di arrivo
		List<String> nomiConsTarget = new LinkedList<String>();
		nomiConsTarget.addAll(nomiConsTargetSet);
		List<String> nomiConsSource = new LinkedList<String>();
		nomiConsSource.addAll(nomiConsSourceSet);
		//i costrutti rimasti nell'insieme nomiConsSourceSet sono quelli eliminati
		//rimuovo i costrutti temporanei
		for(String nomeSource: nomiConsSource){
			if(nomeSource.charAt(0)!='_')
				this.rimossi.add(nomeSource);
		}
		this.rimossi.removeAll(nomiConsTarget);
		
		//quelli rimasti nell'insieme nomiConsTarget sono quelli aggiunti
		for(String nomeTrg: nomiConsTarget){
			if(nomeTrg.charAt(0)!='_')
				this.aggiunti.add(nomeTrg);
		}
		this.aggiunti.removeAll(nomiConsSource);
	}

	/**
	 * Ad ogni costrutto si associa una formula che vale "false" 
	 * se nessuna propriet del costrutto  stata ignorata dall'applicazione 
	 * della macroregola. 
	 */
	private void calculateIgnoredProperty() {
		this.ignorate = new HashMap<String, String>();
		//creo l'applicatore 
		ServicesApplication sap = new ServicesApplication();
		//determino la segnatura della macroregola
		List<RuleSignature> macroSig = macro.getRulesSignatures();
		//prendo i costrutti dello schema sorgente
		Map<String, ConstructSignature> mapSchemaConstructs = sorgente.getCostrutti();
		Collection<String> namesSchemaConstructs = mapSchemaConstructs.keySet();
		//per tutti i costrutti dello schema sorgente
		for(String cm:namesSchemaConstructs){
			//ignoriamo i costrutti temporanei
			if(cm.charAt(0)!='_'){
				//formula sulle propriet ignorate associata
				String formula = mapSchemaConstructs.get(cm).getFormula();
				//per ogni regola della mcroregola
				for(RuleSignature r: macroSig){
					//se la regola  applicabile
					if(sap.applicable(r, sorgente)){
						//prendo il corpo della regola
						List<ConstructSignature> body = r.getBody();
						//per ogni costrutto nel corpo
						for(ConstructSignature csBody: body){
							//se ha lo stesso nome del costrutto del corpo
							if(cm.equals(csBody.getName())){
								//prendo la formula del costrutto nel corpo della regola
								String formulaConstRule = csBody.getFormula();
								//se la formula vale true, tutte le varianti del costrutto corrispondente unificano col corpo della regola
								if(!formulaConstRule.equals("true")){
									//creo il tableau con la formula del costrutto dello schema
									ToolTableaux tablSchemeConstruct = new ToolTableaux(formula);
									formula = tablSchemeConstruct.ignoredLiteral(formulaConstRule);
									//se il risultato fosse stato "false", verrebbe riportato null
									if(formula==null)
										formula = "false";
								}
								else{
									formula = "false";
									break;
								}
							}
						}
						if(formula.equals("false"))
							break;
					}
				}
				this.ignorate.put(cm, formula.replaceAll(cm+".", ""));
			}
		}

	}



}
