package org.modelgen.signature;

import org.modelgen.regole.MacroRegola;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.signature.util.ToolTableaux;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * La segnatura di una macroregola  data dal nome del file che contiene la
 * macroregole e da una lista di segnature delle regole contenute in essa.
 * @author Fabrizio celli
 * @since 1.2.1
 */
public class MacroRuleSignature {

	private String name = null;
	//mappa che contiene la segnatura di un costrutto, con il suo nome
	private List<RuleSignature> regole;

	public MacroRuleSignature(MacroRegola regola){
		//calcolo le segnature delle regole
		List<RegolaDatalog> regoleDatalog = regola.getRegole(); 
		regole = new ArrayList<RuleSignature>();
		for(RegolaDatalog r: regoleDatalog){
			RuleSignature sig = new RuleSignature(r);
			regole.add(sig);
		}
	}

	public MacroRuleSignature(String nameFile, MacroRegola regola){
		//calcolo le segnature delle regole
		this(regola);
		this.name = nameFile;
	}

	/**
	 * Restituisce la lista delle segnature delle regole contenute
	 * nella macroregola
	 * @return segnatura regole della macroregola
	 */
	public List<RuleSignature> getRulesSignatures(){
		return this.regole;
	}

	/**
	 * Restituisce il nome del file in cui  contenuta la macroregola
	 */
	public String getFileName(){
		return this.name;
	}

	public void setFileName(String fileName){
		this.name = fileName;
	}

	/**
	 * Calcola la ModelSignature associata alla macroregola come unione dei
	 * costrutti nei corpi delle regole;
	 * Se pi costrutti si ripetono, le loro formule sono poste in OR;
	 * Non consideriamo i costrutti temporanei.
	 * @return la ModelSignature associata alla macroregola
	 */
	public ModelSignature getModelSignatureFromBodies(){
		String modelName = "GENERATED_FROM_MACRORULE";
		if(this.name!=null)
			modelName = this.name;
		ModelSignature model = new ModelSignature(modelName);

		//per ogni regola
		for(RuleSignature rule: this.regole){
			//prendo il body
			List<ConstructSignature> body = rule.getBody();

			//per ogni costrutto nel body
			for(ConstructSignature construct: body){
				//prendo il nome
				String constructName = construct.getName();
				//se non  temporaneo
				if(!constructName.startsWith("_")){
					//controllo se gi c' un costrutto con stesso nome nel modello
					if(model.getCostrutti().get(constructName)==null)
						model.insert(construct);
					else{
						String newFormula = 
							"("+model.getCostrutti().get(constructName).getFormula()+" OR "+construct.getFormula()+")";
						ToolTableaux tool = new ToolTableaux(newFormula);
						newFormula = tool.formaNormDisgiuntiva();

						model.update(constructName, newFormula);
					}
				}
			}
		}

		return model;
	}

	/**
	 * Restituisce l'unione dei nomi dei costrutti nella testa
	 * delle RuleSignature di questa MacroRuleSignature
	 * @return l'unione dei nomi dei costrutti delle teste
	 */
	public Set<String> getUnionHeads(){
		Set<String> heads = new HashSet<String>();
		//scandisco le regole
		for(RuleSignature regola: this.regole){
			heads.add(regola.getHead().getName());
		}
		return heads;
	}

	/**
	 * Restituisce l'unione dei nomi dei costrutti nel corpo
	 * delle RuleSignature di questa MacroRuleSignature
	 * @return l'unione dei nomi dei costrutti dei corpi
	 */
	public Set<String> getUnionBodies(){
		Set<String> bodies = new HashSet<String>();
		//scandisco le regole
		for(RuleSignature regola: this.regole){
			List<ConstructSignature> body = regola.getBody();
			for(ConstructSignature corpo: body)
				bodies.add(corpo.getName());
		}
		return bodies;
	}

	//toString
	public String toString(){
		String s = this.getFileName()+"\n";
		for(RuleSignature rule: this.getRulesSignatures()){
			s+= (rule.toString()+"\n");
		}
		return s;
	}

}
