
package org.modelgen.signature;

import org.modelgen.regole.Argomento;
import org.modelgen.regole.ArgomentoEtichettato;
import org.modelgen.regole.Letterale;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.signature.util.ToolTableaux;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Rappresenta la segnatura di una regola, ottenuta a partire dalle
 * informazioni presenti nell'oggetto RegolaDatalog
 * Testato con successo il 31-5-2007
 * @author Fabrizio Celli
 * @since 6-06-2007
 * @version 1.3
 *
 */
public class RuleSignature {

	private ConstructSignature head;
	private List<ConstructSignature> body;
	private RegolaDatalog regola;

	private boolean dicopia;
	private boolean flagdicopia = false;

	/*
	 * per il mapping MAP(p')=Nj(p): la chiave  p, il valore  p'
	 * infatti nelle applicazioni non ci interessa Nj
	 * inoltre sappiamo che p compare in un unico Nj
	 */
	private Map<String, String> map;

	public RuleSignature(RegolaDatalog regola){
		this.body = new ArrayList<ConstructSignature>();
		this.map = new HashMap<String, String>();
		this.regola = regola;
	}

	/**
	 * Prende il letterale della testa della regola: il nome del letterale, tagliato
	 * del prefisso SM_,  il nome della segnatura del costrutto. Per ogni argomento
	 * etichettato del letterale, se l'attributo  associato ad una costante booleana,
	 * viene aggiunto nella formula della segnatura con il suo segno, a seconda che 
	 * la costante valga true o false. L'attributo  preceduto da un prefisso composto
	 * dal nome del letterale a cui appartiene pi la stringa ".": questo  utile
	 * per identificare univocamente le propriet con stessi nomi di costrutti
	 * con nomi diversi
	 */
	public ConstructSignature getHead() {

		if(this.head==null){

			Letterale l = regola.getTesta();
			String nomeLetterale = l.getPredicateName();
			//elimino il prefisso SM_
			nomeLetterale = nomeLetterale.substring(3);
			ConstructSignature testa = new ConstructSignature(nomeLetterale);

			List<ArgomentoEtichettato> args = l.getListaArgomentiEtichettati();
			String formula = "";
			for(ArgomentoEtichettato arg: args){
				String attributo = arg.getAttributo();	//nome propriet del costrutto
				//aggiungiamo all'attributo il prefisso "nomeLetterale"+"."
				attributo = nomeLetterale+"."+attributo;
				Argomento argomento = arg.getArgomento();
				//gestione della formula
				String valoreArgomento = argomento.getValore();
				//considero solo le costanti booleane
				if(valoreArgomento!=null){
					if(valoreArgomento.equalsIgnoreCase("true")){
						if(formula.length()==0)
							formula += attributo;
						else 
							formula = "("+formula+" AND "+attributo+")";
					}
					else if(valoreArgomento.equalsIgnoreCase("false")){
						if(formula.length()==0)
							formula += "(NOT "+attributo+")";
						else 
							formula = "("+formula+" AND (NOT "+attributo+")"+")";
					}
				}
			} 

			//se non ci sono argomenti etichettati
			if(formula.equals(""))
				formula = "true";

			testa.setFormula(formula);
			this.head = testa;
		}
		return head;
	}

	public List<ConstructSignature> getBody() {
		if(this.body.size() == 0){
			List<Letterale> corpo = regola.getBody();
			ConstructSignature tmp = null;
			for(Letterale l: corpo){
				String nomeLetterale = l.getPredicateName();
				//elimino il prefisso SM_
				nomeLetterale = nomeLetterale.substring(3);
				tmp = new ConstructSignature(nomeLetterale);
				List<ArgomentoEtichettato> args = l.getListaArgomentiEtichettati();
				String formula = "";
				for(ArgomentoEtichettato arg: args){
					String attributo = arg.getAttributo();	//nome propriet del costrutto
					//aggiungiamo all'attributo il prefisso "nomeLetterale"+"."
					attributo = nomeLetterale+"."+attributo;
					Argomento argomento = arg.getArgomento();
					//gestione della formula
					String valoreArgomento = argomento.getValore();
					if(valoreArgomento!=null){
						if(valoreArgomento.equalsIgnoreCase("true")){
							if(formula.length()==0)
								formula += attributo;
							else 
								formula = "("+formula+" AND "+attributo+")";
						}
						else if(valoreArgomento.equalsIgnoreCase("false")){
							if(formula.length()==0)
								formula += "(NOT "+attributo+")";
							else 
								formula = "("+formula+" AND (NOT "+attributo+")"+")";
						}
					}
				}

				//se non ci sono argomenti etichettati
				if(formula.equals(""))
					formula = "true";
				
				//se il letterale  negato, occorre negare la formula
				if(l.getIsNegato()){
					ToolTableaux t = new ToolTableaux(formula);
					formula = t.negaFormulaToCNF();
				}

				tmp.setFormula(formula);
				this.body.add(tmp);
			}
		}
		return body;
	}


	/**
	 * E' definita per le propriet della testa associate a variabili, in cui
	 * ogni variabile compare una sola volta nel corpo della regola.
	 * Se una variabile appare per la propriet p' nella testa e per p nel
	 * costrutto Nj nel corpo, si ha map(p') = Nj(p).
	 * Il metodo restituisce una mappa di Stringhe, di cui la chiave  la
	 * propriet p del corpo e il valore  la propriet p' della testa
	 * Dato che ci interessano solo le propriet, escludiamo l'attributo
	 * "role1" oppure "name" oppure "role2" oppure "type".
	 */
	public Map<String, String> getMap() {
		if(this.map.size()==0){
			Letterale testa = regola.getTesta();
			String nomeLetterale = testa.getPredicateName();
			//elimino il prefisso SM_
			nomeLetterale = nomeLetterale.substring(3);
			List<Letterale> corpo = regola.getBody(); 

			List<ArgomentoEtichettato> argsTesta = testa.getListaArgomentiEtichettati();
			for(ArgomentoEtichettato arg: argsTesta){
				String attributo = arg.getAttributo();	//nome propriet del costrutto
				Argomento argomento = arg.getArgomento();

				String valoreArgomento = argomento.getValore();
				//se l'argomento  una variabile
				if(!attributo.equalsIgnoreCase("name") &&
						!attributo.equalsIgnoreCase("role1") &&
						!attributo.equalsIgnoreCase("role2") &&
						!attributo.equalsIgnoreCase("type") &&
						valoreArgomento!=null &&
						!valoreArgomento.equalsIgnoreCase("true") &&
						!valoreArgomento.equalsIgnoreCase("false") &&
						!valoreArgomento.equals("")){
					//aggiungiamo all'attributo il prefisso "nomeLetterale."
					attributo = nomeLetterale+"."+attributo;
					//scandisco il corpo della regola per cercare l'attributo con la stessa variabile
					String attributoCorpo = null;
					for(Letterale l: corpo){
						List<ArgomentoEtichettato>  argsCorpo = l.getListaArgomentiEtichettati();
						for(ArgomentoEtichettato argc: argsCorpo){
							String valore = argc.getArgomento().getValore();
							if(valore!=null && valore.equalsIgnoreCase(valoreArgomento)){
								String letterale = l.getPredicateName();
								//elimino il prefisso SM_
								letterale = letterale.substring(3);
								attributoCorpo = argc.getAttributo();
								attributoCorpo = letterale+"."+attributoCorpo;
								break;
							}

						}
						if(attributoCorpo!=null)
							break;
					}

					//a questo punto, sucuramente l'attributo del corpo  diverso da null
					map.put(attributoCorpo, attributo);
				}
			}
		}
		return map;
	}

	/**
	 * Una regola  di copia se nel suo corpo c' un costrutto con la stessa 
	 * segnatura del costrutto della testa pi eventualmente altri costrutti.
	 * Questa definizione non  corretta in ogni contesto, ma  corretta nel
	 * contesto che usiamo: infatti, poich con le signature vengono ignorate 
	 * le reference, sarebbe riconosciuta come regola di copia anche una regola 
	 * che copia i lexical di abstract facendoli diventare lexical di aggregation.
	 * Tuttavia, se in una macroregola  presente tale regola,  garantita anche
	 * la presenza di una regola che trasforma abstract in aggregation, e quindi, 
	 * globalmente, siamo comunque sicuri che la macroregola, nel complesso, 
	 * non  costituita solo da regole di copia.
	 */
	public boolean idDiCopia(){
		if(!this.flagdicopia){
			for(ConstructSignature bodycostr: this.getBody()){
				if(this.getHead().equals(bodycostr)){
					this.dicopia = true;
					break;
				}
			}
			this.flagdicopia = true;
		}
		return this.dicopia;
	}
	

	public String toString(){
		String descr = "Testa: "+this.getHead()+", " +
		"Corpo: ["+this.getBody()+"], " +
		"Mapping: "+this.getMap();
		return descr;
	}

}
