package org.modelgen.signature.application.macroruleapplication;


import org.modelgen.facade.GeneralManagementFacade;
import org.modelgen.signature.ConstructSignature;
import org.modelgen.signature.MacroRuleSignature;
import org.modelgen.signature.ModelSignature;
import org.modelgen.signature.RuleSignature;
import org.modelgen.signature.application.loader.RulesLoader;
import org.modelgen.signature.application.ruleapplication.ServicesApplication;
import org.modelgen.signature.util.ToolTableaux;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class SignatureApplicationUtility {
	
	//singleton
	private static SignatureApplicationUtility instance = null;
	
	public static SignatureApplicationUtility getInstance(){
		if(instance == null)
			instance = new SignatureApplicationUtility();
		return instance;
	}

	/**
	 * Metodo che, data la ModelSignature di un modello obiettivo e quella del
	 * modello (o schema) corrente, determina se il modello corrente  compatibile 
	 * col modello obiettivo. In pratica, il modello obiettivo deve contenere almeno 
	 * tutti i costrutti del modello corrente (cio i costrutti del modello corrente
	 * devono essere un sottoinsieme di quelli del modello obiettivo) e le formule 
	 * nei costrutti del modello corrente devono soddisfare quelle del modello 
	 * obiettivo, cio al pi essere pi restrittive delle corrispondenti formule
	 * nel modello obiettivo.
	 * Ritorna false anche se una delle due segnature vale null.
	 */
	public boolean isCompatibile(ModelSignature trg, ModelSignature curr){
		if(curr!=null && trg!=null){
			boolean result = true;
			Collection<ConstructSignature> costruttiCurr = curr.getCostrutti().values();
			for(ConstructSignature sig: costruttiCurr){
				//verifico che in dest ci sia un costrutto con lo stesso nome di sig
				String sigName = sig.getName();
				ConstructSignature sigTrg = trg.getCostrutti().get(sigName);
				if(sigTrg==null){
					result = false;
					break;
				}
				//verifico la compatibilit delle formule
				String formulaTrg = sigTrg.getFormula();
				String formulaCurr = sig.getFormula();
				//se la formula del target vale true,  sempre soddisfatta
				result = this.isCompatibile(formulaTrg, formulaCurr);
				if(!result)
					break;
			}
			return result;
		}
		return false;
	}
	
	/**
	 * Data una formula target e una corrente, verifica se sono compatibili
	 */
	public boolean isCompatibile(String trg, String curr){
		if(curr!=null && trg!=null){
			boolean result = true;
			//se la formula del target vale true,  sempre soddisfatta
			if(!trg.equalsIgnoreCase("true")){
				if(!curr.equalsIgnoreCase(trg)){
					ToolTableaux tablCurr = new ToolTableaux(curr);
					ToolTableaux tablTrg = new ToolTableaux(trg);
					result = tablTrg.isSoddisfattoDa(tablCurr);
				}
			}
			return result;
		}
		return false;
	}

	/**
	 * Metodo che, data una lista di macroregole e uno schema, determina
	 * quali macroregole siano applicabili allo schema, riportando in
	 * output la lista delle macroregole applicabili.
	 * Una macroregola  applicabile se contiene almeno una regola 
	 * applicabile al modello. In realt se le uniche macroregole applicabili 
	 * sono di copia, la macroregola non  messa nel risultato.
	 * Restituisce la lista vuota se non ci sono macroregole utili.
	 */
	public List<MacroRuleSignature> usefulMacroRules(List<MacroRuleSignature> macroregole, ModelSignature modelSig){
		List<MacroRuleSignature> applicabili = new ArrayList<MacroRuleSignature>();

		//scandisco la lista delle macroregole
		for(MacroRuleSignature macro: macroregole){
			//se la macroregola  applicabile
			if(this.isApplicableMacroRule(macro, modelSig))
				//aggiungo la macroregola alla lista di quelle applicabili
				applicabili.add(macro);
		}
		return applicabili;	
	}

	/*
	 * considera la lista di tutte le macroregole esistenti
	 */
	public List<MacroRuleSignature> usefulMacroRules(ModelSignature modelSig) {
		return this.usefulMacroRules(this.getMacroRegole(), modelSig);	
	}
	
	public boolean isApplicableMacroRule(MacroRuleSignature macro, ModelSignature modelSig){
		boolean result = false;
		//per verificare l'applicabilit di una regola
		ServicesApplication appl = new ServicesApplication();
		//controllo se almeno una regola  applicabile
		for(RuleSignature regolaSig: macro.getRulesSignatures()){
			if(appl.applicable(regolaSig, modelSig)){
				if(!regolaSig.idDiCopia() && !regolaSig.getHead().getName().startsWith("_")){
					result = true;
					break;
				}
			}
		}
		return result;
	}

	/**
	 * Metodo che produce una lista di tutte le macroregole definite
	 * nell'applicazione. Per farlo carica prima la mappa delle macroregole
	 * (le cui chiavi sono i nomi dei files che le contengono) e poi ne
	 * estrae i valori in una lista.
	 * 
	 * TESTATO 06-06-07
	 * 
	 */
	public List<MacroRuleSignature> getMacroRegole() {
		//carico la cache
		NaiveCacheMacroRuleSignature cache = NaiveCacheMacroRuleSignature.getInstance();
		if(cache.getRegole() == null){
			this.loadCache();
		}
		return cache.getRegole();
	}
	
	/**
	 * Carica TUTTE le macroregole nella cache
	 */
	public void loadCache(){
		Collection<MacroRuleSignature> collRegole = this.getMapMacroRegole().values();
		List<MacroRuleSignature> regole = new LinkedList<MacroRuleSignature>();
		regole.addAll(collRegole);
		NaiveCacheMacroRuleSignature.getInstance().setRegole(regole);
		System.out.println("macrorules cache loaded");
	}

	/**
	 * Produce una mappa con tutte le macroregole definite nell'applicazione
	 * (valori della mappa), associate al nome del file della macroregola
	 * @return una mappa delle macroregole col nome del loro file
	 */
	public Map<String, MacroRuleSignature> getMapMacroRegole() {
		//carico la cache
		NaiveCacheMacroRuleSignature cache = NaiveCacheMacroRuleSignature.getInstance();
		if(cache.getMapRegole() == null){
			//creo la lista con i nomi dei files di tutte le macroregoli
			Map<String, MacroRuleSignature> regole = new HashMap<String, MacroRuleSignature>();
			List<String> rulesName = RulesLoader.getInstance().load();
			//per ciascun nome di file, carico la macroregola
			for (int i = 0; i < rulesName.size(); i++) {//rulesName.size(); i++) {
				MacroRuleSignature mrdp = GeneralManagementFacade.getInstance().getMacroRegolaByfileName(rulesName.get(i));
				regole.put(rulesName.get(i), mrdp);
			}
			cache.setMapRegole(regole);
		}
		return cache.getMapRegole();
	}
}
