package org.modelgen.signature.application.ruleapplication;

import org.modelgen.signature.ConstructSignature;
import org.modelgen.signature.MacroRuleSignature;
import org.modelgen.signature.ModelSignature;
import org.modelgen.signature.RuleSignature;
import org.modelgen.signature.util.ToolTableaux;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Fabrizio Celli
 * @date 11-05-2007
 * @version 1.2
 *
 * TESTATA IL 26-05-07
 */
public class ApplicateRuleSignature {

	private ServicesApplication application;

	public ApplicateRuleSignature(){
		this.application = new ServicesApplication();
	}

	/**
	 * Data una macroregola e la segnatura dello schema sorgente, restituisce la 
	 * ModelSignature dello schema risultante. Se il nome dello schema sorgente 
	 * non contiene la stringa "TRANSLATED", questa viene aggiunta.
	 * Se l'applicazione della macroregola allo schema non produce risultati, 
	 * viene restituito null.
	 */
	public ModelSignature createModelSignature(MacroRuleSignature regole, ModelSignature schema){
		String name = schema.getName();
		if(!name.contains("TRANSLATED."))
			name = "TRANSLATED."+name;
		ModelSignature result = new ModelSignature(name);

		List<ConstructSignature> costrutti = this.applicate(regole, schema);

		//se l'applicazione ha prodotto risultato
		if(costrutti!=null){
			for(ConstructSignature sig: costrutti){
				String formula = sig.getFormula();
				if(!formula.equals("true")){
					//la trasformo in FND per semplificarla
					ToolTableaux t = new ToolTableaux(formula);
					formula = t.formaNormDisgiuntiva();
				}
				sig.setFormula(formula);

				result.insert(sig);
			}

			return result;
		}
		//altrimenti ritorna null
		return null;
	}

	/**
	 * Data la ModelSignature model di un modello (o di uno schema) e una macroregola, 
	 * estrae da questa una lista di segnature di regole e la manda in pasto al
	 * metodo privato "applicate(List<RuleSignature>, ModelSignature)"
	 * che calcola la lista delle ConstructSignature che risultano
	 * dall'applicazione della macroregola alla segnatura del modello.
	 * @param regole
	 * @param model
	 * @return la lista delle ConstructSignature che risultano dall'applicazione della 
	 * macroregola alla segnatura del modello
	 */
	private List<ConstructSignature> applicate(MacroRuleSignature regole, ModelSignature model){
		List<RuleSignature> regoleSig = regole.getRulesSignatures();
		return this.applicate(regoleSig, model);
	}

	/**
	 * Data la ModelSignature model di un modello (o di uno schema) e una lista di 
	 * segnature di regole, applica in sequenza ciascuna segnatura di regola a model. 
	 * Ogni regola non applicabile viene ignorata. Il risultato di ciascuna 
	 * applicazione di regola  messo in una lista di ConstructSignature 
	 * nel seguente modo:
	 * se gi  presente un costrutto con lo stesso nome, la sua formula 
	 * viene messa in OR con quella del costrutto che stiamo aggiungendo, 
	 * altrimenti viene aggiunto alla lista il nuovo costrutto.
	 * Se la lista di regole o la segnatura del modello sono null, ritorna null.
	 * Se l'applicazione non fornisce alcun risultato, ritorna null.
	 * @param regole
	 * @param model
	 * @return la lista di ConstructSignature oppure null
	 */
	private List<ConstructSignature> applicate(
			List<RuleSignature> regole, ModelSignature model){
		List<ConstructSignature> result = null;

		if(regole!=null && model!=null){
			result = new ArrayList<ConstructSignature>();
			//int j = 1;
			for(RuleSignature rule: regole){
				//System.out.println(j+": "+rule.getHead());
				//j++;
				ConstructSignature singleApplication = application.applicateOne(rule, model);
				//System.out.println(singleApplication);
				//se la regola era applicabile al modello
				if(singleApplication!=null){
					//se non ho prodotto un costrutto temporaneo
					if(!singleApplication.getName().startsWith("_")){
						//dice se singleApplication  stato aggiunto al risultato
						boolean flag = false;	
						for(int i=0; i<result.size(); i++){
							/*
							 * Se gi c' nel risultato un costrutto con lo stesso nome, metto
							 * in OR le formule
							 */
							if(result.get(i).getName().equals(singleApplication.getName())){
								String formula = result.get(i).getFormula();
								//se la formula vale true, anche l'OR varr true
								if(!formula.equals("true")){
									if(!singleApplication.getFormula().equals("true"))
										formula = "("+formula+" OR "+singleApplication.getFormula()+")";
									else
										formula = "true";
									result.get(i).setFormula(formula);
								}
								flag = true;
								break;
							}
						}
						if(!flag)
							result.add(singleApplication);
					}
					else{
						//aggiungiamo al modello il costrutto temporaneo
						model.insert(singleApplication);
					}
				}
			}
			
			//rimuovo dam modello tutti gli eventuali costrutti temporanei aggiunti
			model.removeTmpConstruct();
			
			//se non  stato prodotto alcun risultato
			if(result.size() == 0)
				result = null;
		}

		return result;

	}

}
