package org.modelgen.signature.application.treeSearcher;

import org.modelgen.signature.DiffContainer;
import org.modelgen.signature.MacroRuleSignature;
import org.modelgen.signature.ModelSignature;
import org.modelgen.signature.application.macroruleapplication.SignatureApplicationUtility;
import org.modelgen.signature.application.ruleapplication.ApplicateRuleSignature;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * E' un nodo dell'albero utilizzato per la selezione delle macroregole
 * applicabili ad uno schema e del percorso che ha portato a quello
 * schema.
 * Rappresenta la ModelSignature di uno schema con associate le seguenti
 * informazioni satellite: 
 * - un insieme di tutte le MacroRuleSignature ancora da applicare,
 * - il percorso che ha prodotto la ModelSignature attuale (percorso inteso come
 *   lista di MacroRuleSignature applicate fin qui)
 * - una lista di MacroRuleSignature applicabili, ottenuta filtrando le 
 *    MacroRuleSignature ancora da applicare
 * - un DiffContainer che specifica come sono stati modificati i costrutti
 *   rispetto alla ModelSignature del padre
 * @author Fabrizio Celli
 * @since 1.2.1
 * @date 15/12/2007
 */
public class Node {
	
	//le macroregole ancora da applicare
	private List<MacroRuleSignature> daApplicare;
	//lo schema rappresentato dal nodo
	private ModelSignature model;
	//il percorso con cui  stato raggiunto lo schema
	private List<MacroRuleSignature> percorso;
	//le macroregole ancora applicabili
	private List<MacroRuleSignature> applicabili = null;
	//i cambiamenti sui costrutti del modello pardre
	private DiffContainer container = null;
	
	public Node(ModelSignature model, List<MacroRuleSignature> daApplicare){
		this.model = model;
		this.daApplicare = daApplicare;
		this.percorso = new LinkedList<MacroRuleSignature>();
		this.setApplicableMacroRuleSignature();
	}
	
	private Node(ModelSignature model, List<MacroRuleSignature> daApplicare, 
			List<MacroRuleSignature> percorso, DiffContainer diff){
		this.model = model;
		this.daApplicare = daApplicare;
		this.percorso = percorso;
		this.container = diff;
		this.setApplicableMacroRuleSignature();
	}

	public List<MacroRuleSignature> getPercorso() {
		return percorso;
	}
	
	public ModelSignature getModel() {
		return model;
	}
	
	//inizializza le macroregole applicabili
	private void setApplicableMacroRuleSignature(){
		SignatureApplicationUtility appl = new SignatureApplicationUtility();
		if(this.applicabili==null)
			applicabili = new LinkedList<MacroRuleSignature>();
		this.applicabili = appl.usefulMacroRules(daApplicare, model);
	}
	
	/**
	 * Applica al nodo corrente tutte le macroregola applicabili, creando la
	 * lista di figli del nodo corrente. Se non ci sono macroregole applicabili,
	 * restituisce la lista vuota.
	 * @return la lista dei figli del nodo corrente
	 */
	public List<Node> applicateAllApplicable(){
		List<Node> figli = new ArrayList<Node>();
		for(MacroRuleSignature rule: this.applicabili)
			figli.add(this.applicate(rule));
		return figli;
	}
	
	/*
	 * Restituisce il figlio del nodo corrente, ottenuto applicando al
	 * modello del nodo corrente la macroregola passata per parametro
	 */
	private Node applicate(MacroRuleSignature rule){
		ApplicateRuleSignature applicator = new ApplicateRuleSignature();	
		//applico la macroregola al modello
		ModelSignature result = applicator.createModelSignature(rule, this.getModel());
		//calcolo le diff sui costrutti
		DiffContainer diff = new DiffContainer(this.getModel(), result, rule);
		//creo la lista delle macroregole da applicare al nuovo nodo
		List<MacroRuleSignature> newDaApplicare = new ArrayList<MacroRuleSignature>();
		newDaApplicare.addAll(this.daApplicare);
		newDaApplicare.remove(rule);
		//aggiorno il percorso
		List<MacroRuleSignature> newPercorso = new ArrayList<MacroRuleSignature>();
		newPercorso.addAll(this.percorso);
		newPercorso.add(rule);
		//restituisco il figlio
		return new Node(result, newDaApplicare, newPercorso, diff);
	}
	
	/**
	 * Verifica se l'applicazione della regola ha causato l'aggiunta di costrutti
	 * @return true se sono stati aggiunti costrutti
	 */
	public boolean hasCreatedConstructs(){
		if(this.container==null)
			return false;
		return (container.getAggiunti().size()!=0);
	}
	
	/**
	 * Verifica se l'applicazione della regola ha ignorato costrutti
	 * @return true se sono stati ignorati costrutti
	 */
	public boolean hasIgnoredConstructs(){
		if(this.container==null){
			return false;
		}
		boolean result = false;
		Map<String, String> ignored = container.getIgnorate();
		//per ogni costrutto, verifico che la formula associata sia false
		for(String construct: ignored.keySet()){
			if(!ignored.get(construct).equalsIgnoreCase("false")){
				result = true;
				break;
			}
		}
		return (result);
	}
	
	//toString
	public String toString(){
		String result = "Applicate:";
		for(MacroRuleSignature macro: this.getPercorso())
			result += macro.getFileName()+"\n";
		result +="Modello: "+this.model;
		return result;
	}
}
