package org.modelgen.tmphandler;

import org.modelgen.commons.beans.model.Construct;
import org.modelgen.commons.beans.model.Model;
import org.modelgen.commons.beans.model.Property;
import org.modelgen.commons.beans.model.Reference;
import org.modelgen.commons.beans.skolem.SkolemFunction;
import org.modelgen.commons.persistence.PersistenceException;
import org.modelgen.commons.persistence.PersistenceHandler;
import org.modelgen.commons.persistence.PersistenceHandlerFactory;
import org.modelgen.commons.utils.ModelgenConfigurationManager;
import org.modelgen.regole.Argomento;
import org.modelgen.regole.ArgomentoEtichettato;
import org.modelgen.regole.Letterale;
import org.modelgen.regole.MacroRegola;
import org.modelgen.regole.RegolaDatalog;
import org.modelgen.regole.reader.ParametricDatalogRulesReader;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

/** 
 * This is the main class of the tmpHandler package. It is meant to accept as input a Macrorule file
 * and process it by splitting it into the single Datalog rules which it is made of, and then produces
 * as output a text file containing the SQL-code associated to the TMP lexicals found within the various
 * rules, in order to automatize the TMP-processing system.
 * 
 * <p>Since its creation, this class has undergone a key modification: the first implementation created the final SQL string
 * by writing row-by-row the various SQL instructions for the table columns, each related to a single attribute of a TMP lexical
 * found and resolved. The rows were then put together into a string for each TMP lexical, and then added to a Collection.
 * The duplicate elimination was poorly performed, for it was based on string length (the TMP lexical with the greater
 * number of attributes was chosen among the TMP lexicals with the same name).
 * 
 * <p>Currently, instead, the final SQL string is originated in a different way: a collection of LexAtt objects (objects
 * representing TMP lexicals with their attributes) is created and updated each time a TMP lexical is found and resolved
 * in a Datalog rule. If a TMP lexical (identified by its name) is already in the collection (i.e.: it was already found
 * and resolved), it is given more attributes according to the new ones found in its repetition within a Macrorule.
 * 
 * <p>For instance, if TMP_Lexical ( OID: oid, Name: name ) has been already resolved,
 * and TMP_Lexical ( OID: oid, isLeaf: isLeaf ) is subsequently found during the scanning of the same Macrorule,
 * the attribute isLeaf is added to the lexical in the collection, resulting in
 * TMP_Lexical ( OID: oid, Name: name, isLeaf: isLeaf ), thus a lexical with a merge of the attributes
 * found in its repetitions.
 * (Incidentally, the LexTemp representation of a TMP_Lexical currently does not include the argument value of the attributes,
 * but instead features their type that is later used in the SQL string -> ex. OID integer).
 * 
 * 
 * @author Daniele
 *  
 */
public class TmpHandler {
	
	
	/**
	 * Temporary tables path
	 */
	public static final String TEMPORARY_TABLES_FILE = "GoodRules/ModelGen_temporaryTablesHandler.txt";
	private static JFrame frame;
	private static List<RegolaDatalog> regole;
	private static int c = 0;
	private static int counter = 0;	 // a counter used for display and test purposes only
	
	// list of the TMP lexicals resolved up to now during the scanning of a Macrorule
	private static HashMap<String,LexTemp> tmpLexicsResolved; 
	
	/** 
     * Main method of the TMP Handler. It accepts a macrorule filename as input, it processes the macrorule
	 * by splitting it into simplerules, and relinquishes control to the handleTmpRule method in order to
	 * process the single rule.
	 * 
	 * This method is given the task of writing the SQL text file in the correct directory.
	 * @param macroRule	- The filename of the macrorule to be read
	 */
	public static void tmpHandlerMainRoutine(String macroRule) {
	
		/* We tell the program to look in the macrorules directory to search for the specified macrorule */ 
		String macroRulePath = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths")      // ./rules/
		+ ModelgenConfigurationManager.getIstance().getProperty("macrorules", "", "paths");				        // macrorules/
		String macroregola = macroRulePath + macroRule;
		//System.out.println("MacroRule: " + macroRule);
		//System.out.println("Regola, file: " + macroregola);
		//File filez = new File(macroregola);
		//System.out.println("path, file: " + filez.getAbsolutePath());
		/* We load the MacroRegola from the string composed by path+filename */
		MacroRegola mrdp = ParametricDatalogRulesReader.read(macroregola);
				
		tmpLexicsResolved = new HashMap<String,LexTemp>();
		
		/* We are interested in the sub-rules which compose the MacroRegola: */
		regole = mrdp.getRegole();
		//SimpleRule rule = null;
		counter = 0;
				
		/* it pass each single rule within the MacroRegola to the proper method that process it: handleTmpRule  */
		 
		for(RegolaDatalog regola : regole) {
			counter++;
			
			handleTmpRule(regola);
		} 
		
			
		// this string contain the SQL code generated 
		String finalNoDuplicates = printFinalString();
		
		//JOptionPane.showMessageDialog(frame,finalNoDuplicates, "FINAL STRING",JOptionPane.INFORMATION_MESSAGE);
		
		/* Ok, everything' s been set, now we finally write the txt file containing the SQL code
		 * meant to generate the correct tables and columns in the database!
		 * But only if the resulting string is not empty! :-)
		 */
		
		/* First of all, let' s check whether a SQL file was already in the specified directory:*/
		String rootnode = ModelgenConfigurationManager.getIstance().getProperty("rootnode", "", "paths");
		String simplerulePath = ModelgenConfigurationManager.getIstance().getProperty("simplerules", "", "paths");
		String filePath = rootnode+simplerulePath + "ModelGen/hsqldb/";

		File f = new File(filePath);
		String[] fs = f.list();
		int count = 0;
		File file = null;
		while(count<fs.length) {
			if(fs[count].equalsIgnoreCase("ddl_CreateTemporaryTablesHandler.sql")) {
				file = new File(filePath+"ddl_CreateTemporaryTablesHandler.sql");
			}
		count++;
			
		}
		
		/* If another file was there, delete it first!*/
		if(file != null) {
			file.delete();
		}
		
		/* If there is actually something to write... */
		String filename = "ddl_CreateTemporaryTablesHandler.sql";
		fileWrite(finalNoDuplicates,filePath,filename);
		
	}	// fine tmpHandlerMainRoutine
	
	
	/** 
   * A method that accepts as input a datalog rule and works on it, determining the TMP lexicals
	 * within it. It then processes them relinquishing control to the sub-method createSQLfromTMP,
	 * which handles the job of turning a TMP lexical into a sequence of SQL instructions
	 * 
	 * This method grants there' ll be no duplicates in the full SQL file created by this class 
	 * 
	 * @param regola - a RegolaDatalog object
	* 
	 */
	
	public static void handleTmpRule(RegolaDatalog regola) {
		
		
		/* We need the Letterale making up the rule head, as well as the other lexicals within the rule body */
		Letterale testa = regola.getTesta();
		// this list includes all the SM lexicals (no TMP!) belonging to the rule,both from head and body
		LinkedList<Letterale> supportList = listSMlexicals(regola);
		
		
		/* Let' s check whether the head is a TMP lexical: */
		if (testa.getIsTMP()) { 
			
			createSQLfromTMP(testa,supportList);
						
		} 
		
		/* Now let' s see if the body contains TMP lexicals: */
		LinkedList corpo = (LinkedList)regola.getBody();
		Iterator iter_lett = corpo.iterator();
		Letterale letterale_corpo = null;
		while(iter_lett.hasNext()) {
			letterale_corpo = (Letterale)iter_lett.next();
						
			if (letterale_corpo.getIsTMP()) { 
	            
				createSQLfromTMP(letterale_corpo,supportList);
			} 
			
		}
	
	}
	
	
	/** 
	 * This method simply accepts a RegolaDatalog as input and returns a list of all the SM Lexicals
	 * that are included in the rule.
	 *  
	 * @param regola - a RegolaDatalog object
	 * @return supportList - a list of all the SM Lexicals found in the rule
	 */
	public static LinkedList<Letterale> listSMlexicals(RegolaDatalog regola) {
		// this list will be used to include all the SM lexicals (no TMP!) belonging to the rule,both from head and body
		LinkedList<Letterale> supportList = new LinkedList<Letterale>();
		Letterale testa = regola.getTesta();
		if(!testa.getIsTMP()) {
			supportList.add(testa);
		}
		
		LinkedList corpo = (LinkedList)regola.getBody();
		Iterator iter_lett = corpo.iterator();
		Letterale letterale_corpo = null;
		while(iter_lett.hasNext()) {
			letterale_corpo = (Letterale)iter_lett.next();
			if(!letterale_corpo.getIsTMP()) {
				supportList.add(letterale_corpo);
			}
		}		
		return supportList;
		
	}
	
	
	
	/** 
	 * The key submethod of this class: receives a TMP Letterale as input and adds the resolved TMP lexical to the 
	 * tmpLexicsResolved collection, which will later provide the SQL code (see printFinalString() method). 
	 * The type control of the lexical attributes is performed here.
	 *  
	 *  
	 * @param letterale - the TMP Letterale to be scanned; supportList - the list of the SM lexicals in the rule
	 */
	
	
	
	private static void createSQLfromTMP(Letterale letterale, LinkedList<Letterale> supportList) {
		
		c++;
		String nome_letterale = letterale.getPredicateName();
		LexTemp tmp = null;
				
		if(tmpLexicsResolved.containsKey(nome_letterale)) {
			tmp = tmpLexicsResolved.get(nome_letterale);
		} else {
			tmp = new LexTemp(nome_letterale);
		}
			
		// insert schemeOID if is not already in tmp
		Collection attributi = tmp.getAttributes();
		Iterator i_attributi = attributi.iterator();
		AttTemp corrente = null;
		boolean trovato = false;
	
		while(i_attributi.hasNext()) {	
			corrente = (AttTemp)i_attributi.next();
			if ((corrente.getAttribute()).equalsIgnoreCase("schemeOID")){
				trovato = true;
			}
		}
	
		if(trovato == false){
			AttTemp attTemp = new AttTemp("schemeOID","integer");
			try {
				tmp.addAttribute(attTemp);
			}
			catch(Exception e) {
				JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
			}
		
		}

		for(ArgomentoEtichettato argomentoEtichettato : letterale.getListaArgomentiEtichettati()) {
			Argomento argomento = argomentoEtichettato.getArgomento();
			/* The column names are the attribute names, so we need them */
			 
			String nome_argomento = argomento.getNome();			        //ex. 018, pincopallino, etc.
			String attributo = argomentoEtichettato.getAttributo();        //ex. OID, Name, isOptional...
			
			/* A TMP rule might have anything as one of its attributes. The key is to determine its type!
			 * If there is a #SkolemFunction as the attribute, type is integer. But if it is not?
			 * We have to check the SM lexicals within the rule and see where the same argument is found:
			 * the corresponding attribute will have its own, already specified, type!
			 * 
			 * There are standard types like OID: integer, Name: varchar
			 * If these standart cases are not applicable, it means it is a Property or a Reference.
			 * A Property has its own type: property.getType();
			 * A Reference is always an integer.
			 *  
			 */
			
					/****************************************************************************************************************** 
					 *    1)          Let' s check whether the argument is a Skolem Function
					 * 
					 ******************************************************************************************************************/
					boolean isSkolem = false;
					try {
						PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler();
						for (SkolemFunction function :  ph.getSkolemFunctions()) {
										
							if (nome_argomento.equalsIgnoreCase(function.getName())) { // it' s a Skolem Function! -> integer type
								isSkolem = true;																
								AttTemp attTemp = new AttTemp(attributo,"integer");
								tmp.addAttribute(attTemp);
							}
						}	
					} // end of try -> loading Skolem Functions
					catch(ClassNotFoundException ce) {
						JOptionPane.showMessageDialog(frame,ce.getMessage() + "ClassNotFound","Error in loading Skolem Functions",JOptionPane.ERROR_MESSAGE);
					}
					catch(PersistenceException pe)	{
						JOptionPane.showMessageDialog(frame,pe.getMessage() + "Persistence","Error in loading Skolem Functions",JOptionPane.ERROR_MESSAGE);
					}
					catch(Exception ex)	{
						JOptionPane.showMessageDialog(frame,ex.getMessage() + "Exception","Error in loading Skolem Functions",JOptionPane.ERROR_MESSAGE);
					}
					
				
					if(isSkolem==false) { // too bad, it wasn' t a Skolem function... what type is it then?
						
						/******************************************************************************************************************* 
						 *    2)          Let' s try and find this attribute within a SM Lexical of the current rule!
						 * 
						 ********************************************************************************************************************/
						Iterator<Letterale> i_SM = supportList.iterator();
						Letterale SM_lett = null;
						boolean found = false;
						while(i_SM.hasNext() && found==false) {
							SM_lett = i_SM.next();							
							String SM_lett_name = SM_lett.getPredicateName();
							String attr = SM_lett.getAttributoPerVariabile(nome_argomento);
							
							if(attr!=null) { // we have found the right SM lexical within the rule which contains the attribute we were looking for!
								found = true;
								
								/* Basic cases: OID*/ 
								if(attr.equalsIgnoreCase("OID")) {
									AttTemp attTemp = new AttTemp(attributo,"integer");
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
									
								
								} else if (attr.equalsIgnoreCase("Name")) {
									AttTemp attTemp = new AttTemp(attributo,"varchar(50)");
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
									
								
								} else {
									
									try {
										PersistenceHandler ph = PersistenceHandlerFactory.getPersistenceHandler();
										Model model = ph.getSuperModel();

										String constructName = SM_lett_name.substring(3,SM_lett_name.length());
										
										Construct c = model.getConstruct(constructName);
										if(c==null) {
											JOptionPane.showMessageDialog(frame,"No construct?", "ERROR",JOptionPane.ERROR_MESSAGE);
											return; // riccardop: inutile continuare
										}							
										Property p = c.getProperty(attr);
										if(p!=null) { // it was a property!
											String propertyType = p.getType();
											String type = typeConverter(propertyType);
											
											AttTemp attTemp = new AttTemp(attributo,type);
											try {
												tmp.addAttribute(attTemp);
											}
											catch(Exception e) {
												JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
											}
											
											
										} else { // it should be a reference, then!
											Reference r = c.getReference(attr);
											if(r!=null) {
												String type ="integer";
											
												AttTemp attTemp = new AttTemp(attributo,type);
											
												try {
													tmp.addAttribute(attTemp);
												}
												catch(Exception e) {
													JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
												}
											} else
												JOptionPane.showMessageDialog(frame,"Impossible?", "ERROR",JOptionPane.ERROR_MESSAGE);
										}				
								
									} // end of try
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage() + " Failed to load the Supermodel","TMP SQL conversion",JOptionPane.INFORMATION_MESSAGE);
								
									}
								} // end of else -> it is not a basic case!	
							}// end of if(attr!=null)
						} // end of while(i_SM.hasNext())
						
						
					
						if (found==false) { // the argument wasn' t found anywhere in the rule!
							
							/******************************************************************************************************************* 
							 *    3)     Let' s check whether we have already encountered this attribute in a previous rule or in a previous
							 *           TMP lexical of the current MacroRule
							 *******************************************************************************************************************/
				
							boolean fn = false;
							if(tmpLexicsResolved.containsKey(nome_letterale)) {
								LexTemp lexTemp = tmpLexicsResolved.get(nome_letterale);
								LinkedList args = (LinkedList)lexTemp.getAttributes(); 
								Iterator iter = args.iterator();
								AttTemp aT = null;
								boolean f = false;
								while(iter.hasNext()) {
									aT = (AttTemp)iter.next();
									String attributeName = aT.getAttribute();
									if(attributo.equalsIgnoreCase(attributeName)) {
										AttTemp attTemp = new AttTemp(attributo,aT.getType());
										try {
											tmp.addAttribute(attTemp);
										}
										catch(Exception e) {
											JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
										}
										f = true;
										fn = true;
									}
									else {
									}
								}
								
								if(f==false) {
									JOptionPane.showMessageDialog(frame,"<html>A TMP lexical with the name " + nome_letterale + " was already resolved,<br>but the attribute: <font color=#0099FF>"+attributo+"</font> was not included<br>-> there must be TMP lexicals with the same name but with different attributes!</html>", "SQL code",JOptionPane.INFORMATION_MESSAGE);
								}
								
								
								
							} //end of if(tmpLexicsResolved.containsKey(nome_letterale))
							
							
							if(fn==false){ // the argument wasn' t even included in the list of the previous resolved attributes...???
								/* It must have been a constant! ...or something else
								 * If it is "True" or "False" -> boolean;
								 * if it is placed between ' ' -> varchar(60);
								 * else...exception
								 */
								
								if(nome_argomento.equalsIgnoreCase("True") || nome_argomento.equalsIgnoreCase("False")) {
									String type ="bit";
									AttTemp attTemp = new AttTemp(attributo,type);
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
								} else if(nome_argomento.startsWith("'")) {
									String type ="varchar(60)";
									AttTemp attTemp = new AttTemp(attributo,type);
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
								} else if(nome_argomento.contains("OID")) {
									String type ="integer";
									AttTemp attTemp = new AttTemp(attributo,type);
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
									JOptionPane.showMessageDialog(frame,"type of " + attributo + ": " + nome_argomento + " might be integer", "SQL code",JOptionPane.INFORMATION_MESSAGE);
								} else if(nome_argomento.startsWith("is") || nome_argomento.startsWith("Is") || nome_argomento.startsWith("IS")) {
									String type ="bit";
									AttTemp attTemp = new AttTemp(attributo,type);
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
									JOptionPane.showMessageDialog(frame," type of " + attributo + ": " + nome_argomento + " might be boolean (bit)", "SQL code",JOptionPane.INFORMATION_MESSAGE);
								} else { // this is the worst case of all... Force the type to be varchar!
									String type ="varchar(70)";
									AttTemp attTemp = new AttTemp(attributo,type);
									try {
										tmp.addAttribute(attTemp);
									}
									catch(Exception e) {
										JOptionPane.showMessageDialog(frame,e.getMessage(),"IMPOSSIBLE!",JOptionPane.ERROR_MESSAGE);
									}
									
									JOptionPane.showMessageDialog(frame,"<html>Using the <font color=FF0000>Magic Sphere</font>: type of " + attributo + ": " + nome_argomento + " has been forced to be varchar(70)</html>", "SQL code",JOptionPane.INFORMATION_MESSAGE);
								}	
							} // end of else -> the argument wasn' t found anywhere
						} // end of if(found==false)
						
					} // end of if(isSkolem==false)	
						
						
		} //END OF THE ARGUMENT SCANNING
		
		if(!tmpLexicsResolved.containsKey(nome_letterale)) {
			tmpLexicsResolved.put(nome_letterale,tmp);
		}	
		
	}
	
	
	/** 
	 * This method provides a string of SQL code, i.e. the main output of the entire class.
	 * Duplicates encountered during the scanning of the Macrorule are handled EXTREMELY WELL
	 * (if there were TMP lexicals with same name but different attributes, the resulting
	 * SQL code will be one table corresponding to that lexical, with a merging of the various
	 * attributes as its own attributes (columns).
	 *  
	 * 
	 * @return finalNoDuplicates - a String containing the final SQL code resulted from the scanning of the TMP lexicals
	 * 							   resolved within the Macrorule ->
	 */
	public static String printFinalString() {
		String finalNoDuplicates ="";
		Collection<LexTemp> tmpL = tmpLexicsResolved.values();
		Iterator<LexTemp> itmp = tmpL.iterator();
		LexTemp lex = null;
		while (itmp.hasNext()) {
			lex = itmp.next();
			finalNoDuplicates = finalNoDuplicates + "NOROLLBACK DROP TABLE " + lex.getPredicateName() + ";\n\n";
			finalNoDuplicates = finalNoDuplicates + "create table " + lex.getPredicateName() + " ( \n";
			LinkedList attrs = (LinkedList)lex.getAttributes();
			Iterator i = attrs.iterator();
			AttTemp attTemp = null;
			while(i.hasNext()) {
				attTemp = (AttTemp)i.next();
				finalNoDuplicates = finalNoDuplicates + "    " + attTemp.getAttribute() + " " + attTemp.getType() + ",\n";
				
			}
			int length = finalNoDuplicates.length();
			finalNoDuplicates = finalNoDuplicates.substring(0,length-2); // we remove the last comma
			finalNoDuplicates = finalNoDuplicates + "\n);\n\n";
			
		}
		

		// add to finalNoDuplicates the TMP instance
		Collection tmpIn = tmpLexicsResolved.values();
		Iterator it = tmpIn.iterator();
		LexTemp lexIn = null;
		while (it.hasNext()) {
			lexIn = (LexTemp)it.next();
			finalNoDuplicates = finalNoDuplicates + "NOROLLBACK DROP TABLE " + "i_" + lexIn.getPredicateName() + ";\n\n";
			finalNoDuplicates = finalNoDuplicates + "create table " + "i_" + lexIn.getPredicateName() + " ( \n";
			LinkedList attrIn = (LinkedList)lexIn.getAttributes();
			Iterator itAttr = attrIn.iterator();
			AttTemp attTempIn = null;
			while(itAttr.hasNext()) {
				attTempIn = (AttTemp)itAttr.next();
				
				System.out.println("\n attributo:  " + attTempIn.getAttribute());
				
				// add the attribute to finalNoDuplicates 
				if((attTempIn.getAttribute()).equalsIgnoreCase("schemeOID") ){
					
					finalNoDuplicates = finalNoDuplicates + "    " + "InstanceOID" + " " + attTempIn.getType() + ",\n";
															
				} else if((attTempIn.getAttribute()).equalsIgnoreCase("OID")){
					
					finalNoDuplicates = finalNoDuplicates + "    " + "OID" + " " + attTempIn.getType() + ",\n";
					
				} else {
					finalNoDuplicates = finalNoDuplicates + "    " + "i_" + attTempIn.getAttribute() + " " + attTempIn.getType() + ",\n";
					
				  }
					
			}
			int length = finalNoDuplicates.length();
			finalNoDuplicates = finalNoDuplicates.substring(0,length-2); // we remove the last comma
			finalNoDuplicates = finalNoDuplicates + "\n);\n\n";
			
		}
		
		return finalNoDuplicates;
	
	}
	
	
	/** 
	 * Just a method to write a .txt file from a String, used in the tmpHandlerMainRoutine
	 * 
	 * @param string    - the content of the text file to be written
	 * @param path      - the path where the file is to be written at
	 * @param filename  - the name of the file
	 */
	public static void fileWrite(String string, String path, String filename) {
		File file = new File(path + filename);
		try {
			FileWriter writer = new FileWriter(file,true);
			writer.write(string);
			writer.close();
		}
		catch(IOException e) {
			JOptionPane.showMessageDialog(frame,e.getMessage(),"ERROR IN WRITING THE FILE",JOptionPane.ERROR_MESSAGE);
		}
	}
	
	 
	
	/** 
	 * A simplest Datalog->SQL type converter method
	 * 
	 * @param type - the Datalog rule type to be converted into SQL
	 * @return result - the string featuring the converted type 
	 */
	public static String typeConverter(String type) {
		String result = type;
		
		if(type.equalsIgnoreCase("integer") || type.equalsIgnoreCase("bit") || type.startsWith("varchar") || type.equalsIgnoreCase("char"))
			result = type;
		else if(type.equalsIgnoreCase("boolean")) {
			result = "bit";
		}
				
		return result;
		
	}
	
	

	/** 
	 * Test method 
	 * 
	 */
	public static void main(String[] args) {
		
		//config.setProperty("tmp","tmp/","paths");
		
		//	     tmpHandlerMainRoutine("ERNaryToERBinary.txt"); //-> OK!!!
		//  	 tmpHandlerMainRoutine("ERBinaryNoGenNoAttOnRelationshipsNoManyToManyToRelational.txt"); // -> PROBLEMS -> SOLVED!!!
		//	 	 tmpHandlerMainRoutine("ERBinaryNoManyToManyNoAttOnRelationshipsToObjectOriented.txt"); -> EMPTY STRING -> NO TEMP RULES
		//		 tmpHandlerMainRoutine("XSDToERBinary.txt"); //-> EMPTY STRING -> NO TEMP RULES
		//		 tmpHandlerMainRoutine("ERBinaryToXSD.txt"); -> INVALID SYNTAX
		//  	 tmpHandlerMainRoutine("ObjectOrientedNoGenToRelational.txt"); // -> PROBLEMS -> SOLVED!!!
		//	   	 tmpHandlerMainRoutine("RelationalToRelationalWithoutNullColumns.txt"); -> EMPTY STRING -> NO TEMP RULES
		//	     tmpHandlerMainRoutine("ERBinaryNoManyToManyToERBinaryNoAttributesOnRelationships.txt"); //-> DIFFERENT ATTRIBUTES ON SAME TMP LEXICAL WITHIN A MACRORULE -> SOLVED!!!
		//	   	 tmpHandlerMainRoutine("ERBinaryToERBinaryNoGen(Both Parent and Children).txt"); -> EMPTY STRING -> NO TEMP RULES
		//		 tmpHandlerMainRoutine("ERBinaryToERBinaryNoGen(Children only).txt"); // BIG TROUBLES -> SOLVED!!!
		//		 tmpHandlerMainRoutine("pippo.txt");	
		//	     tmpHandlerMainRoutine("ERBinaryToERBinaryNoGen(Parent only).txt");// -> OK!!!
		//	     tmpHandlerMainRoutine("ERBinaryToERBinaryNoManyToManyAggregations.txt");// -> OK!!!
		//	  	 tmpHandlerMainRoutine("ObjectOrientedNoIDToObjectOrientedWithID.txt");
		//	   	 tmpHandlerMainRoutine("ObjectOrientedToERBinary.txt");
		//	  	 tmpHandlerMainRoutine("ObjectOrientedToObjectOrientedNoGen(Both Parent and Children).txt");
		//	     tmpHandlerMainRoutine("ObjectOrientedToObjectOrientedNoGen(Children only).txt");
		//	     tmpHandlerMainRoutine("ObjectOrientedToObjectOrientedNoGen(Parent only).txt");
		//     	 tmpHandlerMainRoutine("RelationalToERBinary.txt");
		//	     tmpHandlerMainRoutine("RelationalToObjectOriented.txt");
		//	   	 tmpHandlerMainRoutine("RelationalToRelationalWithoutNullColumns.txt");
		
		//		tmpHandlerMainRoutine("28.00.rel2ER.txt");
		//      tmpHandlerMainRoutine("37.01.absAttFromChild2Father4GenCUSTOM.txt");
		//		tmpHandlerMainRoutine("32.00.elimAbsLeavesOfGen.txt");
		
		
	}
}
