/*
 * $Id: OrthogonalLinkRouter.java,v 1.5 2007/03/19 12:31:46 david Exp $
 * Copyright (c) 2005-2007, David Benson
 *
 * All rights reserved.
 *
 * This file is licensed under the JGraph software license, a copy of which
 * will have been provided to you in the file LICENSE at the root of your
 * installation directory. If you are unable to locate this file please
 * contact JGraph sales for another copy.
 */package com.jgraph.layout.orthogonal;

import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import org.jgraph.graph.EdgeView;
import org.jgraph.graph.GraphConstants;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.JGraphLayout;

/**
 * A cache of the rectangular regions describing an orthogonal layout.
 * This layout only applies to edges
 */
public class OrthogonalLinkRouter implements JGraphLayout {

	/**
	 * The maximum depth region searches will go up to.
	 * Increasing this value improves the result, but performance
	 * is degraded by more than a linear amount
	 */
	private int MAXIMAL_DEPTH = 4;

	/**
	 * The minimum distance from the vertex side before an edge may turn
	 */
	private static int firstLengthDistance = 20;

	private static final int CORNER_LENGTH = 200;

	private static final boolean UPDATE_COLLISION_LISTS = false; // set true for
																// faster
																// computation
																// but not
																// optimal
																// results

	private static final boolean UPDATE_COLLISION_LISTS_REMOVE = true;

	static final boolean IGNORE_LINKS_WITH_ACTUAL_PORTS = false; // cause
																	// links to
																	// try to
																	// merge as
																	// a bus,
																	// multilinks
																	// are
																	// merged
																	// too

	private static final boolean OPTIMALIZE_REGIONS = true;

	Object edge;

	Object sourcePort;

	Object targetPort;

	Point2D sourcePoint;

	Point2D targetPoint;

	int sourceDirection;

	int targetDirection;

	Collection verticalCollisions = new HashSet(); // bounds for vertical
													// segments

	Collection horizontalCollisions = new HashSet(); // bounds for horizontal
														// segments

	private final OrthogonalLinkRouterRegion[] regions = new OrthogonalLinkRouterRegion[MAXIMAL_DEPTH];

	private Point2D[] bestControlPoints;

	private int bestControlPointsLength;

	/**
	 * The API method used to exercise the layout upon the facade description
	 * and produce a separate description of the edge routing changes made. It
	 * runs each stage of the layout that has been created.
	 * 
	 * @param facade
	 *            the facade object that describes and filters the graph to be
	 *            acted upon
	 */
	public void run(JGraphFacade facade) {
		// Determine regions
		Collection edgesColl = facade.getEdges();
		Object[] edges = edgesColl.toArray();
		for (int i = 0; i < edges.length; i++) {
			// Route each edge by intersecting with the fewest number of
			// vertices possible
			routeEdge(edges[i], facade);
		}
	}

	/**
	 * Routing algorithmn for an individual edge
	 * @param edge the edge to be routed
	 * @param facade the facade to act upon
	 */
	public void routeEdge(Object edge, JGraphFacade facade) {
		this.edge = edge;
		Object sourcePort = facade.getSourcePort(edge);
		Object targetPort = facade.getTargetPort(edge);

		if (sourcePort == null || targetPort == null) {
			return;
		}

		// clear required due to collision lists update during routing
		verticalCollisions.clear();
		horizontalCollisions.clear();
		bestControlPoints = null;
		bestControlPointsLength = Integer.MIN_VALUE;

		OrthogonalLinkRouterUtils.calculateCollisionBounds(facade, edge,
				sourcePort, targetPort, verticalCollisions,
				horizontalCollisions);

		List newPoints = new ArrayList();
		Object tmpEdgeView = facade.getCellView(edge);
		if (tmpEdgeView != null && tmpEdgeView instanceof EdgeView) {
			EdgeView edgeView = (EdgeView) tmpEdgeView;
			int n = edgeView.getPointCount();
			Point2D from = edgeView.getPoint(0);
			Point2D to = edgeView.getPoint(n - 1);
			if (from != null && to != null) {

				sourcePoint = new Point2D.Double((int) from.getX(), (int) from
						.getY());

				sourceDirection = OrthogonalLinkRouterUtils.getDirection(
						facade, sourcePort);
				targetDirection = OrthogonalLinkRouterUtils.getDirection(
						facade, targetPort);

				// If the direction is centered we have to pick a direction
				if (sourceDirection == IDirectionable.CENTER
						|| targetDirection == IDirectionable.CENTER) {
					int alterSourceDirection = -1;
					int alterTargetDirection = -1;
					double diffX = to.getX() - from.getX();
					double diffY = to.getY() - from.getY();
					if (Math.abs(diffY) > Math.abs(diffX)) {
						if (diffY > 0) {
							alterSourceDirection = IDirectionable.BOTTOM;
							alterTargetDirection = IDirectionable.TOP;
						} else {
							alterSourceDirection = IDirectionable.TOP;
							alterTargetDirection = IDirectionable.BOTTOM;
						}
					} else {
						if (diffX > 0) {
							alterSourceDirection = IDirectionable.RIGHT;
							alterTargetDirection = IDirectionable.LEFT;
						} else {
							alterSourceDirection = IDirectionable.LEFT;
							alterTargetDirection = IDirectionable.RIGHT;
						}
					}
					if (sourceDirection == IDirectionable.CENTER) {
						sourceDirection = alterSourceDirection;
					}
					if (targetDirection == IDirectionable.CENTER) {
						targetDirection = alterTargetDirection;
					}
				}

				// Work out the targetPoint based on the direction of the to
				// point
				switch (targetDirection) {
				case IDirectionable.LEFT:
					targetPoint = new Point2D.Double(to.getX()
							- firstLengthDistance, (int) to.getY());
					break;
				case IDirectionable.RIGHT:
					targetPoint = new Point2D.Double(to.getX()
							+ firstLengthDistance, (int) to.getY());
					break;
				case IDirectionable.TOP:
					targetPoint = new Point2D.Double(to.getX(), (int) (to
							.getY() - firstLengthDistance));
					break;
				case IDirectionable.BOTTOM:
					targetPoint = new Point2D.Double(to.getX(), (int) (to
							.getY() + firstLengthDistance));
					break;
				default:
				}

				search(facade, new OrthogonalLinkRouterRegion(sourcePoint
						.getX(), sourcePoint.getY(), 0, 0, sourceDirection, 0));

				if (bestControlPoints != null) {
					for (int i = 0; i < bestControlPoints.length; i++) {
						newPoints.add(bestControlPoints[i]);
					}
					facade.setPoints(edge, newPoints);
					facade.disableRouting(edge);
					GraphConstants.setLineStyle(facade.getAttributes(edge),
							GraphConstants.STYLE_ORTHOGONAL);
					;
				}
			}
		}
	}

	/**
	 * Searches for the targetPoint, starting within the given region
	 * 
	 * @param facade
	 * @param region
	 */
	private void search(JGraphFacade facade, OrthogonalLinkRouterRegion region) {
		if (region.width < 0 || region.height < 0)
			return;

		// If the recursion is getting too deep, end it
		int depth = region.getDepth();
		if (depth >= MAXIMAL_DEPTH) {
			return;
		}

		// Extend the curret region by a _lot_ in its current direction
		region.extendToInfinity();
		regions[depth] = region;

		Collection collisions;
		if (region.isHorizontal()) {
			collisions = horizontalCollisions;
		} else {
			collisions = verticalCollisions;
		}

		ArrayList subRegions = region.parseSubRegions(collisions);
		Rectangle updateCollisionsRect = null;

		if (!region.isEmpty()) {
			if (UPDATE_COLLISION_LISTS) {
				updateCollisionsRect = new Rectangle(region);
				horizontalCollisions.add(updateCollisionsRect);
				verticalCollisions.add(updateCollisionsRect);
			}
		}

		// TODO - not true - point could lay on right or bottom edge ! - this is
		// not checked yet
		if (region.containsInsideEdges(targetPoint)) {
			constructControlPoints(facade, depth);
		} else {
			if (region.getLength() > 0) {
				search(facade, region.cloneWithCounterClockwiseEdge());
				search(facade, region.cloneWithClockwiseEdge());
			}

			if (!subRegions.isEmpty()) {
				for (int i = 0; i < subRegions.size(); i++) {
					OrthogonalLinkRouterRegion reg = (OrthogonalLinkRouterRegion) subRegions
							.get(i);
					if (!region.equals(reg)) {
						search(facade, reg);
					}
				}
			}
		}

		if (UPDATE_COLLISION_LISTS && UPDATE_COLLISION_LISTS_REMOVE) {
			horizontalCollisions.remove(updateCollisionsRect);
			verticalCollisions.remove(updateCollisionsRect);
		}
	}

	private void constructControlPoints(JGraphFacade facade, int depth) {
		// List controlPoints = new ArrayList();
		Point2D[] controlPoints = new Point2D.Double[depth + 4];
		Object tmpEdgeView = facade.getCellView(edge);
		if (tmpEdgeView != null && tmpEdgeView instanceof EdgeView) {
			EdgeView edgeView = (EdgeView) tmpEdgeView;
			int n = edgeView.getPointCount();
			Point2D localFrom = edgeView.getPoint(0);
			Point2D localTo = edgeView.getPoint(n - 1);
			controlPoints[0] = edgeView.getAllAttributes().createPoint(
					(int) localFrom.getX(), (int) localFrom.getY());
			// Create Point2D version of targetPoint
			controlPoints[depth + 2] = edgeView.getAllAttributes().createPoint(
					(int) targetPoint.getX(), (int) targetPoint.getY());
			controlPoints[depth + 3] = edgeView.getAllAttributes().createPoint(
					(int) localTo.getX(), (int) localTo.getY());
			double yy, xx;
			for (int a = 0; a <= depth; a++) {
				final OrthogonalLinkRouterRegion region = regions[a];
				final int direction = region.getDirection();
				final int anext = a + 1;
				switch (direction) {
				case IDirectionable.LEFT:
				case IDirectionable.RIGHT: {
					if (a > 0) {
						final int y1 = region.y;
						final int y2 = region.y + region.height;
						if (y1 <= OrthogonalLinkRouterRegion.MIN_INT_REGION
								&& y2 < OrthogonalLinkRouterRegion.MAX_INT_REGION) {
							yy = y2 - OrthogonalLinkRouterUtils.SPACING;
						} else if (y1 > OrthogonalLinkRouterRegion.MIN_INT_REGION
								&& y2 >= OrthogonalLinkRouterRegion.MAX_INT_REGION) {
							yy = y1 + OrthogonalLinkRouterUtils.SPACING;
						} else {
							yy = region.y + region.height / 2;
						}
						if (controlPoints[anext] == null) {
							controlPoints[anext] = new Point2D.Double(0, yy);
						} else {
							controlPoints[anext].setLocation(
									controlPoints[anext].getX(), (double) yy);
						}
					} else {
						yy = localFrom.getY();
					}
					if (a == depth) {
						xx = (int) targetPoint.getX();
					} else
						xx = 0;
				}
					break;
				case IDirectionable.TOP:
				case IDirectionable.BOTTOM: {
					if (a > 0) {
						final int x1 = region.x;
						final int x2 = region.x + region.width;
						if (x1 <= OrthogonalLinkRouterRegion.MIN_INT_REGION
								&& x2 < OrthogonalLinkRouterRegion.MAX_INT_REGION) {
							xx = x2 - OrthogonalLinkRouterUtils.SPACING;
						} else if (x1 > OrthogonalLinkRouterRegion.MIN_INT_REGION
								&& x2 >= OrthogonalLinkRouterRegion.MAX_INT_REGION) {
							xx = x1 + OrthogonalLinkRouterUtils.SPACING;
						} else {
							xx = region.x + region.width / 2;
						}
						controlPoints[anext].setLocation((double) xx,
								controlPoints[anext].getY());
					} else {
						xx = localFrom.getX();
					}
					if (a == depth) {
						yy = targetPoint.getY();
					} else
						yy = 0;
				}
					break;
				default:
					throw new IllegalArgumentException();
				}
				if (a == 0) {
					controlPoints[anext] = edgeView.getAllAttributes()
							.createPoint((int) xx, (int) yy);
				}
				controlPoints[anext + 1] = edgeView.getAllAttributes()
						.createPoint((int) xx, (int) yy);
			}

			if (OPTIMALIZE_REGIONS) {
				for (int a = depth; a > 0; a--) {
					final double newY = controlPoints[a + 3].getY();
					final OrthogonalLinkRouterRegion regionY = regions[a];
					final OrthogonalLinkRouterRegion regionY1 = regions[a - 1];
					if (newY >= regionY.y && newY < regionY.y + regionY.height
							&& newY >= regionY1.y
							&& newY < regionY1.y + regionY1.height) {
						controlPoints[a + 2].setLocation(controlPoints[a + 2]
								.getX(), newY);
						controlPoints[a + 1].setLocation(controlPoints[a + 1]
								.getX(), newY);
					}
					final double newX = controlPoints[a + 3].getX();
					final OrthogonalLinkRouterRegion regionX = regions[a];
					final OrthogonalLinkRouterRegion regionX1 = regions[a - 1];
					if (newX >= regionX.x && newX < regionX.x + regionX.width
							&& newX >= regionX1.x
							&& newX < regionX1.x + regionX1.width) {
						controlPoints[a + 2].setLocation(newX,
								controlPoints[a + 2].getY());
						controlPoints[a + 1].setLocation(newX,
								controlPoints[a + 1].getY());
					}
				}
			}

			int newPointsLength = 0;
			for (int a = 1; a < controlPoints.length - 1; a++) {
				Point2D p0 = controlPoints[newPointsLength];
				Point2D p1 = controlPoints[a];
				Point2D p2 = controlPoints[a + 1];

				if (p0.getX() == p1.getX() && p1.getX() == p2.getX())
					continue;
				if (p0.getY() == p1.getY() && p1.getY() == p2.getY())
					continue;

				newPointsLength++;
				if (newPointsLength != a)
					controlPoints[newPointsLength] = p1;
			}
			newPointsLength++;
			if (newPointsLength < controlPoints.length - 1)
				controlPoints[newPointsLength++] = controlPoints[controlPoints.length - 1];

			int length = 0;
			for (int a = 1; a < newPointsLength; a++) {
				final Point2D p1 = controlPoints[a - 1];
				final Point2D p2 = controlPoints[a];
				length += Math.abs(p2.getY() - p1.getY())
						+ Math.abs(p2.getX() - p1.getX()) + CORNER_LENGTH;
			}
			if (newPointsLength > 0) {
				int average = length / newPointsLength;
				int diff = 0;
				for (int a = 1; a < newPointsLength; a++) {
					final Point2D p1 = controlPoints[a - 1];
					final Point2D p2 = controlPoints[a];
					diff += Math.abs(Math.abs(p2.getY() - p1.getY())
							+ Math.abs(p2.getX() - p1.getX()) - average);
				}
				diff /= newPointsLength;
				length += diff;
			}

			if (bestControlPoints == null || bestControlPointsLength > length) {
				if (newPointsLength < controlPoints.length) {
					bestControlPoints = new Point2D.Double[newPointsLength];
					System.arraycopy(controlPoints, 0, bestControlPoints, 0,
							newPointsLength);
				} else
					bestControlPoints = controlPoints;
				bestControlPointsLength = length;
			}
		}
	}
	
	/**
	 * Returns <code>Orthogonal Edges</code>, the name of this algorithm.
	 */
	public String toString() {
		return "Orthogonal Edges";
	}

	/**
	 * @return Returns the mAXIMAL_DEPTH.
	 */
	public int getMAXIMAL_DEPTH() {
		return MAXIMAL_DEPTH;
	}

	/**
	 * @param maximal_depth The mAXIMAL_DEPTH to set.
	 */
	public void setMAXIMAL_DEPTH(int maximal_depth) {
		MAXIMAL_DEPTH = maximal_depth;
	}

}