/*
 * $Id: OrthogonalLinkRouterRegion.java,v 1.4 2007/03/19 12:32:12 david Exp $
 * Copyright (c) 2005-2007, David Benson
 *
 * All rights reserved.
 *
 * This file is licensed under the JGraph software license, a copy of which
 * will have been provided to you in the file LICENSE at the root of your
 * installation directory. If you are unable to locate this file please
 * contact JGraph sales for another copy.
 */package com.jgraph.layout.orthogonal;

import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

final class OrthogonalLinkRouterRegion extends Rectangle {

	public static final int MIN_INT_REGION = -20000;

	public static final int MAX_INT_REGION = 20000;

	//    private OrthogonalLinkRouterRegion parent;
	private int direction;

	private boolean horizontal;

	private int depth;

	public OrthogonalLinkRouterRegion(double x, double y, int width,
			int height, int direction, int depth) {
		super((int) x, (int) y, width, height);
		//        this.parent = parentRegion;
		this.direction = direction;
		switch (direction) {
		case IDirectionable.LEFT:
		case IDirectionable.RIGHT:
			horizontal = true;
			break;
		case IDirectionable.TOP:
		case IDirectionable.BOTTOM:
			horizontal = false;
			break;
		default:
			throw new IllegalArgumentException();
		}
		this.depth = depth;
	}

	public int getDirection() {
		return direction;
	}

	public boolean isHorizontal() {
		return horizontal;
	}

	public int getDepth() {
		return depth;
	}

	public void extendToInfinity() {
		switch (direction) {
		case IDirectionable.LEFT:
			width = x - MIN_INT_REGION;
			x = MIN_INT_REGION;
			break;
		case IDirectionable.RIGHT:
			width = MAX_INT_REGION - x;
			break;
		case IDirectionable.TOP:
			height = y - MIN_INT_REGION;
			y = MIN_INT_REGION;
			break;
		case IDirectionable.BOTTOM:
			height = MAX_INT_REGION - y;
			break;
		default:
			throw new IllegalArgumentException();
		}
	}

	public double compareImportant(Rectangle2D collisionToCheck,
			Rectangle2D currentMostImportant) {
		// return: <0 if important, 0 if the same, >0 if not important
		boolean intersects = false;
		double checkX = collisionToCheck.getX();
		double checkY = collisionToCheck.getY();
		double checkWidth = collisionToCheck.getWidth();
		double checkHeight = collisionToCheck.getHeight();
		if (isEmpty() || checkWidth <= 0 || checkHeight <= 0) {
			intersects = false;
		} else {
			double x0 = getX();
			double y0 = getY();
			intersects = (checkX + checkWidth > x0 && checkY + checkHeight > y0
					&& checkX < x0 + getWidth() && checkY < y0 + getHeight());
		}
		
		if (!intersects)
			return Integer.MAX_VALUE;
		if (currentMostImportant == null)
			return Integer.MIN_VALUE;
		switch (direction) {
		case IDirectionable.LEFT:
			return (currentMostImportant.getX() + currentMostImportant
					.getWidth())
					- (checkX + checkWidth);
		case IDirectionable.RIGHT:
			return checkX - currentMostImportant.getX();
		case IDirectionable.TOP:
			return (currentMostImportant.getY() + currentMostImportant
					.getHeight())
					- (checkY + checkHeight);
		case IDirectionable.BOTTOM:
			return checkY - currentMostImportant.getY();
		default:
			throw new IllegalArgumentException();
		}
	}

	// TODO - optimize
	private void parseIntervalsBy(ArrayList intervals, Rectangle2D collision) {
		int colStart;
		int colEnd;

		if (horizontal) {
			colStart = (int)collision.getY();
			colEnd = (int)(collision.getY() + collision.getHeight());
		} else {
			colStart = (int)(collision.getX());
			colEnd = (int)(collision.getX() + collision.getWidth());
		}

		for (int i = 0; i < intervals.size();) {
			OrthogonalLinkRouterRegion region = (OrthogonalLinkRouterRegion) intervals
					.get(i);
			int regStart;
			int regEnd;
			if (horizontal) {
				regStart = region.y;
				regEnd = region.y + region.height;
			} else {
				regStart = region.x;
				regEnd = region.x + region.width;
			}

			if (colEnd <= regStart || colStart >= regEnd) {
				// outside
				i++;
				continue;
			}
			if (colStart <= regStart && colEnd >= regEnd) {
				// fully contained
				intervals.remove(i);
				continue;
			}
			if (colStart > regStart && colEnd < regEnd) {
				// split to two
				OrthogonalLinkRouterRegion clonedRegion = region.cloneExactly();
				int len1 = colStart - regStart;
				int len2 = regEnd - colEnd;
				if (horizontal) {
					region.height = len1;
					clonedRegion.y = regEnd - len2;
					clonedRegion.height = len2;
					intervals.add(i + 1, clonedRegion);
				} else {
					region.width = len1;
					clonedRegion.x = regEnd - len2;
					clonedRegion.width = len2;
					intervals.add(i + 1, clonedRegion);
				}
				return;
			}
			if (colStart <= regStart && colEnd > regStart && colEnd < regEnd) {
				// cut region start
				int difference = colEnd - regStart;
				if (horizontal) {
					region.y += difference;
					region.height -= difference;
				} else {
					region.x += difference;
					region.width -= difference;
				}
				i++;
				continue;
			}
			if (colEnd >= regEnd && colStart > regStart && colStart < regEnd) {
				// cut region end
				int difference = regEnd - colStart;
				if (horizontal) {
					region.height -= difference;
				} else {
					region.width -= difference;
				}
				i++;
				continue;
			}
			throw new IllegalStateException();
		}
	}

	public OrthogonalLinkRouterRegion cloneExactly() {
		return new OrthogonalLinkRouterRegion(/*parent, */x, y, width, height,
				direction, depth);
	}

	/**
	 * 
	 * @param collisions
	 * @return list of sub regions
	 */
	public ArrayList parseSubRegions(Collection collisions) {
		ArrayList importantCollisions = new ArrayList();

		Iterator iter = collisions.iterator();
		Rectangle2D importantCollision = null;
		while (iter.hasNext()) {
			Rectangle2D collision = (Rectangle2D) iter.next();
			double howMuch = compareImportant(collision, importantCollision);
			if (howMuch <= 0) {
				if (howMuch < 0) {
					importantCollisions.clear();
					importantCollision = collision;
				}
				importantCollisions.add(collision);
			}
		}

		ArrayList subRegions = new ArrayList();
		if (importantCollisions.size() > 0) {
			cutLengthBy(importantCollision);
			subRegions.add(cloneWithForwardEdge());
			for (int i = 0; i < importantCollisions.size(); i++) {
				Rectangle2D collision = (Rectangle2D) importantCollisions.get(i);
				parseIntervalsBy(subRegions, collision);
			}
		}
		return subRegions;
	}

	public OrthogonalLinkRouterRegion cloneWithForwardEdge() {
		return cloneWithEdge(/*parent, */direction, depth);
	}

	public OrthogonalLinkRouterRegion cloneWithCounterClockwiseEdge() {
		return cloneWithEdge(/*this, */DirectionUtils
				.getCounterClockWiseDirection(direction), depth + 1);
	}

	public OrthogonalLinkRouterRegion cloneWithClockwiseEdge() {
		return cloneWithEdge(/*this, */DirectionUtils
				.getClockWiseDirection(direction), depth + 1);
	}

	private OrthogonalLinkRouterRegion cloneWithEdge(
			/*OrthogonalLinkRouterRegion parent, */int dir, int depth) {
		switch (dir) {
		case IDirectionable.LEFT:
			return new OrthogonalLinkRouterRegion(/*parent, */x, y, 0, height,
					dir, depth);
		case IDirectionable.RIGHT:
			return new OrthogonalLinkRouterRegion(/*parent, */x + width, y, 0,
					height, dir, depth);
		case IDirectionable.TOP:
			return new OrthogonalLinkRouterRegion(/*parent, */x, y, width, 0,
					dir, depth);
		case IDirectionable.BOTTOM:
			return new OrthogonalLinkRouterRegion(/*parent, */x, y + height,
					width, 0, dir, depth);
		default:
			throw new IllegalArgumentException();
		}
	}

	private void cutLengthBy(Rectangle2D collision) {
		switch (direction) {
		case IDirectionable.LEFT: {
			double toBeMin = collision.getX() + collision.getWidth();
			int currently = x;
			if (toBeMin > x + width) {
				x += width;
				width = 0;
			} else if (toBeMin > currently) {
				width = (int)(x + width - toBeMin);
				x = (int)toBeMin;
			}
		}
			break;
		case IDirectionable.RIGHT: {
			double toBeMax = collision.getX();
			int currently = x + width;
			if (toBeMax < x) {
				width = 0;
			} else if (toBeMax < currently) {
				width = (int)(toBeMax - x);
			}
		}
			break;
		case IDirectionable.TOP: {
			int toBeMin = (int)(collision.getY() + collision.getHeight());
			int currently = y;
			if (toBeMin > y + height) {
				y += height;
				height = 0;
			} else if (toBeMin > currently) {
				height = y + height - toBeMin;
				y = toBeMin;
			}
		}
			break;
		case IDirectionable.BOTTOM: {
			int toBeMax = (int)collision.getY();
			int currently = y + height;
			if (toBeMax < y) {
				height = 0;
			} else if (toBeMax < currently) {
				height = toBeMax - y;
			}
		}
			break;
		default:
			throw new IllegalArgumentException();
		}
	}

    public boolean containsInsideEdges (Point2D point) {
        return point.getX() >= x  &&  point.getX() <= x + width  &&  point.getY() >= y  &&  point.getY() <= y + height;
    }

	public int getLength() {
		if (horizontal) {
			return width;
		} else {
			return height;
		}
	}
	
	public String toString() {
		String result = super.toString();
		result += "\ndirection = " + direction + ", horizontal = " + horizontal
				+ ", depth = " + depth;
		return result;
	}
}
